#ifndef INCLUDE_GUARD_CUPY_CUDA_COMMON_H
#define INCLUDE_GUARD_CUPY_CUDA_COMMON_H

typedef char cpy_byte;
typedef unsigned char cpy_ubyte;
typedef short cpy_short;
typedef unsigned short cpy_ushort;
typedef int cpy_int;
typedef unsigned int cpy_uint;
typedef long long cpy_long;
typedef unsigned long long cpy_ulong;
typedef float cpy_float;
typedef double cpy_double;

#endif // INCLUDE_GUARD_CUPY_CUDA_COMMON_H
