# Functions from the following NumPy document
# http://docs.scipy.org/doc/numpy/reference/routines.sort.html

# "NOQA" to suppress flake8 warning
from cupy.sorting import count  # NOQA
from cupy.sorting import search  # NOQA
from cupy.sorting import sort  # NOQA
