/* Generated by Cython 0.26.1 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "_GLIBCXX_USE_CXX11_ABI",
                "0"
            ],
            [
                "_FORCE_INLINES",
                "1"
            ],
            [
                "CUPY_NO_CUDA",
                "1"
            ]
        ],
        "depends": [
            "cupy/cuda/cupy_cusparse.h"
        ],
        "include_dirs": [
            "cupy/cuda",
            "/usr/local/cuda/include"
        ],
        "language": "c++",
        "library_dirs": [
            "/usr/local/cuda/lib64"
        ],
        "name": "cupy.cuda.cusparse",
        "runtime_library_dirs": [
            "/usr/local/cuda/lib64"
        ],
        "sources": [
            "cupy/cuda/cusparse.pyx"
        ]
    },
    "module_name": "cupy.cuda.cusparse"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_26_1"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX < 0x030700A0 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject **args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #ifdef __cplusplus
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough) || (defined(__GNUC__) && defined(__attribute__))
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cupy__cuda__cusparse
#define __PYX_HAVE_API__cupy__cuda__cusparse
#include "cupy_cusparse.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "cupy/cuda/cusparse.pyx",
  "cupy/cuda/cusparse.pxd",
};

/*--- Type declarations ---*/

/* "driver.pxd":20
 *     ctypedef int CUjitInputType 'CUjitInputType'
 * 
 * cpdef enum:             # <<<<<<<<<<<<<<
 *     CU_JIT_INPUT_CUBIN = 0
 *     CU_JIT_INPUT_PTX = 1
 */
enum  {
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_CUBIN = 0,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_PTX = 1,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_FATBINARY = 2,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_OBJECT = 3,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_LIBRARY = 4
};

/* "cupy/cuda/cusparse.pxd":20
 * 
 * 
 * cpdef enum:             # <<<<<<<<<<<<<<
 *     CUSPARSE_POINTER_MODE_HOST = 0
 *     CUSPARSE_POINTER_MODE_DEVICE = 1
 */
enum  {
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_POINTER_MODE_HOST = 0,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_POINTER_MODE_DEVICE = 1,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_ACTION_SYMBOLIC = 0,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_ACTION_NUMERIC = 1,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_INDEX_BASE_ZERO = 0,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_INDEX_BASE_ONE = 1,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_MATRIX_TYPE_GENERAL = 0,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_MATRIX_TYPE_SYMMETRIC = 1,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_MATRIX_TYPE_HERMITIAN = 2,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_MATRIX_TYPE_TRIANGULAR = 3,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_OPERATION_NON_TRANSPOSE = 0,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_OPERATION_TRANSPOSE = 1,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE = 2,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_DIRECTION_ROW = 0,
  __pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_DIRECTION_COLUMN = 1
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#else
    #define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
static int __Pyx_CLineForTraceback(int c_line);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE cusparseIndexBase_t __Pyx_PyInt_As_cusparseIndexBase_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE cusparseMatrixType_t __Pyx_PyInt_As_cusparseMatrixType_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cython' */

/* Module declarations from 'cupy.cuda' */

/* Module declarations from 'cupy.cuda.driver' */

/* Module declarations from 'cupy.cuda.stream' */
static size_t (*__pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr)(void); /*proto*/

/* Module declarations from 'cupy.cuda.cusparse' */
static PyObject *__pyx_v_4cupy_4cuda_8cusparse_STATUS = 0;
static size_t __pyx_f_4cupy_4cuda_8cusparse_create(int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_4cupy_4cuda_8cusparse_destroy(size_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4cupy_4cuda_8cusparse_check_status(int, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_createMatDescr(int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_4cupy_4cuda_8cusparse_destroyMatDescr(size_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_4cupy_4cuda_8cusparse_setMatIndexBase(size_t, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_setMatType(size_t, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_setPointerMode(size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_setStream(size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_sgthr(size_t, int, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dgthr(size_t, int, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrmv(size_t, int, int, int, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrmv(size_t, int, int, int, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrmm(size_t, int, int, int, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrmm(size_t, int, int, int, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrmm2(size_t, int, int, int, int, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrmm2(size_t, int, int, int, int, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcsrgeamNnz(size_t, int, int, size_t, int, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrgeam(size_t, int, int, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrgeam(size_t, int, int, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcsrgemmNnz(size_t, int, int, int, int, int, size_t, int, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrgemm(size_t, int, int, int, int, int, size_t, int const , size_t, size_t, size_t, size_t, int const , size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrgemm(size_t, int, int, int, int, int, size_t, int const , size_t, size_t, size_t, size_t, int const , size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcoo2csr(size_t, size_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsc2dense(size_t, int, int, size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsc2dense(size_t, int, int, size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcsr2coo(size_t, size_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsr2csc(size_t, int, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsr2csc(size_t, int, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsr2dense(size_t, int, int, size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsr2dense(size_t, int, int, size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_snnz_compress(size_t, int, size_t, size_t, size_t, size_t, float, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dnnz_compress(size_t, int, size_t, size_t, size_t, size_t, double, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsr2csr_compress(size_t, int, int, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, float, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsr2csr_compress(size_t, int, int, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, float, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_sdense2csc(size_t, int, int, size_t, size_t, int, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_ddense2csc(size_t, int, int, size_t, size_t, int, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_sdense2csr(size_t, int, int, size_t, size_t, int, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_ddense2csr(size_t, int, int, size_t, size_t, int, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_snnz(size_t, int, int, int, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dnnz(size_t, int, int, int, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_createIdentityPermutation(size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_xcoosort_bufferSizeExt(size_t, int, int, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcoosortByRow(size_t, int, int, int, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_xcsrsort_bufferSizeExt(size_t, int, int, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcsrsort(size_t, int, int, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_xcscsort_bufferSizeExt(size_t, int, int, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcscsort(size_t, int, int, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "cupy.cuda.cusparse"
int __pyx_module_is_main_cupy__cuda__cusparse = 0;

/* Implementation of 'cupy.cuda.cusparse' */
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_super;
static const char __pyx_k_A[] = "A";
static const char __pyx_k_B[] = "B";
static const char __pyx_k_C[] = "C";
static const char __pyx_k_P[] = "P";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_s[] = "%s";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_lda[] = "lda";
static const char __pyx_k_ldb[] = "ldb";
static const char __pyx_k_ldc[] = "ldc";
static const char __pyx_k_nnz[] = "nnz";
static const char __pyx_k_tol[] = "tol";
static const char __pyx_k_typ[] = "typ";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_beta[] = "beta";
static const char __pyx_k_dirA[] = "dirA";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_nnzA[] = "nnzA";
static const char __pyx_k_nnzB[] = "nnzB";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_xInd[] = "xInd";
static const char __pyx_k_xVal[] = "xVal";
static const char __pyx_k_alpha[] = "alpha";
static const char __pyx_k_descr[] = "descr";
static const char __pyx_k_inNnz[] = "inNnz";
static const char __pyx_k_inVal[] = "inVal";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_cscVal[] = "cscVal";
static const char __pyx_k_csrVal[] = "csrVal";
static const char __pyx_k_descrA[] = "descrA";
static const char __pyx_k_descrB[] = "descrB";
static const char __pyx_k_descrC[] = "descrC";
static const char __pyx_k_handle[] = "handle";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_outVal[] = "outVal";
static const char __pyx_k_rowPtr[] = "rowPtr";
static const char __pyx_k_status[] = "status";
static const char __pyx_k_stream[] = "stream";
static const char __pyx_k_transA[] = "transA";
static const char __pyx_k_transB[] = "transB";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_cooCols[] = "cooCols";
static const char __pyx_k_cooRows[] = "cooRows";
static const char __pyx_k_cscValA[] = "cscValA";
static const char __pyx_k_csrValA[] = "csrValA";
static const char __pyx_k_csrValB[] = "csrValB";
static const char __pyx_k_csrValC[] = "csrValC";
static const char __pyx_k_idxBase[] = "idxBase";
static const char __pyx_k_pBuffer[] = "pBuffer";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_inColInd[] = "inColInd";
static const char __pyx_k_inRowPtr[] = "inRowPtr";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_cooRowInd[] = "cooRowInd";
static const char __pyx_k_cscColPtr[] = "cscColPtr";
static const char __pyx_k_cscRowInd[] = "cscRowInd";
static const char __pyx_k_csrColInd[] = "csrColInd";
static const char __pyx_k_csrRowPtr[] = "csrRowPtr";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_nnzPerCol[] = "nnzPerCol";
static const char __pyx_k_nnzPerRow[] = "nnzPerRow";
static const char __pyx_k_outColInd[] = "outColInd";
static const char __pyx_k_outRowPtr[] = "outRowPtr";
static const char __pyx_k_copyValues[] = "copyValues";
static const char __pyx_k_cscColPtrA[] = "cscColPtrA";
static const char __pyx_k_cscRowIndA[] = "cscRowIndA";
static const char __pyx_k_csrColIndA[] = "csrColIndA";
static const char __pyx_k_csrColIndB[] = "csrColIndB";
static const char __pyx_k_csrColIndC[] = "csrColIndC";
static const char __pyx_k_csrRowPtrA[] = "csrRowPtrA";
static const char __pyx_k_csrRowPtrB[] = "csrRowPtrB";
static const char __pyx_k_csrRowPtrC[] = "csrRowPtrC";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_CuSparseError[] = "CuSparseError";
static const char __pyx_k_cscSortedValA[] = "cscSortedValA";
static const char __pyx_k_csrSortedValA[] = "csrSortedValA";
static const char __pyx_k_nnzPerRowColumn[] = "nnzPerRowColumn";
static const char __pyx_k_cscSortedColPtrA[] = "cscSortedColPtrA";
static const char __pyx_k_cscSortedRowIndA[] = "cscSortedRowIndA";
static const char __pyx_k_csrSortedColIndA[] = "csrSortedColIndA";
static const char __pyx_k_csrSortedRowPtrA[] = "csrSortedRowPtrA";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_cupy_cuda_cusparse[] = "cupy.cuda.cusparse";
static const char __pyx_k_nnzTotalDevHostPtr[] = "nnzTotalDevHostPtr";
static const char __pyx_k_CuSparseError___init[] = "CuSparseError.__init__";
static const char __pyx_k_cupy_cuda_cusparse_pyx[] = "cupy/cuda/cusparse.pyx";
static const char __pyx_k_CUSPARSE_STATUS_SUCCESS[] = "CUSPARSE_STATUS_SUCCESS";
static const char __pyx_k_CUSPARSE_STATUS_ZERO_PIVOT[] = "CUSPARSE_STATUS_ZERO_PIVOT";
static const char __pyx_k_CUSPARSE_STATUS_ALLOC_FAILED[] = "CUSPARSE_STATUS_ALLOC_FAILED";
static const char __pyx_k_CUSPARSE_STATUS_ARCH_MISMATCH[] = "CUSPARSE_STATUS_ARCH_MISMATCH";
static const char __pyx_k_CUSPARSE_STATUS_INVALID_VALUE[] = "CUSPARSE_STATUS_INVALID_VALUE";
static const char __pyx_k_CUSPARSE_STATUS_MAPPING_ERROR[] = "CUSPARSE_STATUS_MAPPING_ERROR";
static const char __pyx_k_CUSPARSE_STATUS_INTERNAL_ERROR[] = "CUSPARSE_STATUS_INTERNAL_ERROR";
static const char __pyx_k_CUSPARSE_STATUS_MATRIX_TYPE_NOT[] = "CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED";
static const char __pyx_k_CUSPARSE_STATUS_NOT_INITIALIZED[] = "CUSPARSE_STATUS_NOT_INITIALIZED";
static const char __pyx_k_CUSPARSE_STATUS_EXECUTION_FAILED[] = "CUSPARSE_STATUS_EXECUTION_FAILED";
static PyObject *__pyx_n_s_A;
static PyObject *__pyx_n_s_B;
static PyObject *__pyx_n_s_C;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_ALLOC_FAILED;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_ARCH_MISMATCH;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_EXECUTION_FAILED;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_INTERNAL_ERROR;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_INVALID_VALUE;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_MAPPING_ERROR;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_MATRIX_TYPE_NOT;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_NOT_INITIALIZED;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_SUCCESS;
static PyObject *__pyx_n_s_CUSPARSE_STATUS_ZERO_PIVOT;
static PyObject *__pyx_n_s_CuSparseError;
static PyObject *__pyx_n_s_CuSparseError___init;
static PyObject *__pyx_n_s_P;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_alpha;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_beta;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_cooCols;
static PyObject *__pyx_n_s_cooRowInd;
static PyObject *__pyx_n_s_cooRows;
static PyObject *__pyx_n_s_copyValues;
static PyObject *__pyx_n_s_cscColPtr;
static PyObject *__pyx_n_s_cscColPtrA;
static PyObject *__pyx_n_s_cscRowInd;
static PyObject *__pyx_n_s_cscRowIndA;
static PyObject *__pyx_n_s_cscSortedColPtrA;
static PyObject *__pyx_n_s_cscSortedRowIndA;
static PyObject *__pyx_n_s_cscSortedValA;
static PyObject *__pyx_n_s_cscVal;
static PyObject *__pyx_n_s_cscValA;
static PyObject *__pyx_n_s_csrColInd;
static PyObject *__pyx_n_s_csrColIndA;
static PyObject *__pyx_n_s_csrColIndB;
static PyObject *__pyx_n_s_csrColIndC;
static PyObject *__pyx_n_s_csrRowPtr;
static PyObject *__pyx_n_s_csrRowPtrA;
static PyObject *__pyx_n_s_csrRowPtrB;
static PyObject *__pyx_n_s_csrRowPtrC;
static PyObject *__pyx_n_s_csrSortedColIndA;
static PyObject *__pyx_n_s_csrSortedRowPtrA;
static PyObject *__pyx_n_s_csrSortedValA;
static PyObject *__pyx_n_s_csrVal;
static PyObject *__pyx_n_s_csrValA;
static PyObject *__pyx_n_s_csrValB;
static PyObject *__pyx_n_s_csrValC;
static PyObject *__pyx_n_s_cupy_cuda_cusparse;
static PyObject *__pyx_kp_s_cupy_cuda_cusparse_pyx;
static PyObject *__pyx_n_s_descr;
static PyObject *__pyx_n_s_descrA;
static PyObject *__pyx_n_s_descrB;
static PyObject *__pyx_n_s_descrC;
static PyObject *__pyx_n_s_dirA;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_handle;
static PyObject *__pyx_n_s_idxBase;
static PyObject *__pyx_n_s_inColInd;
static PyObject *__pyx_n_s_inNnz;
static PyObject *__pyx_n_s_inRowPtr;
static PyObject *__pyx_n_s_inVal;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_lda;
static PyObject *__pyx_n_s_ldb;
static PyObject *__pyx_n_s_ldc;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_nnz;
static PyObject *__pyx_n_s_nnzA;
static PyObject *__pyx_n_s_nnzB;
static PyObject *__pyx_n_s_nnzPerCol;
static PyObject *__pyx_n_s_nnzPerRow;
static PyObject *__pyx_n_s_nnzPerRowColumn;
static PyObject *__pyx_n_s_nnzTotalDevHostPtr;
static PyObject *__pyx_n_s_outColInd;
static PyObject *__pyx_n_s_outRowPtr;
static PyObject *__pyx_n_s_outVal;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_pBuffer;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_rowPtr;
static PyObject *__pyx_kp_s_s;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_status;
static PyObject *__pyx_n_s_stream;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_tol;
static PyObject *__pyx_n_s_transA;
static PyObject *__pyx_n_s_transB;
static PyObject *__pyx_n_s_typ;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_xInd;
static PyObject *__pyx_n_s_xVal;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_13CuSparseError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, int __pyx_v_status); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_check_status(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_status); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_2create(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_4createMatDescr(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_6destroy(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_8destroyMatDescr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_descr); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_10setMatIndexBase(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_descr, PyObject *__pyx_v_base); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_12setMatType(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_descr, PyObject *__pyx_v_typ); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_14setPointerMode(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_mode); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_16setStream(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_18sgthr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_nnz, size_t __pyx_v_y, size_t __pyx_v_xVal, size_t __pyx_v_xInd, int __pyx_v_idxBase); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_20dgthr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_nnz, size_t __pyx_v_y, size_t __pyx_v_xVal, size_t __pyx_v_xInd, int __pyx_v_idxBase); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_22scsrmv(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_y); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_24dcsrmv(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_y); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_26scsrmm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_28dcsrmm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_30scsrmm2(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_32dcsrmm2(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_34xcsrgeamNnz(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_nnzTotalDevHostPtr); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_36scsrgeam(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_alpha, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_beta, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_38dcsrgeam(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_alpha, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_beta, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_40xcsrgemmNnz(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_nnzTotalDevHostPtr); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_42scsrgemm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_44dcsrgemm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_46xcoo2csr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_cooRowInd, int __pyx_v_nnz, int __pyx_v_m, size_t __pyx_v_csrRowPtr, int __pyx_v_idxBase); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_48scsc2dense(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_cscSortedValA, size_t __pyx_v_cscSortedRowIndA, size_t __pyx_v_cscSortedColPtrA, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_50dcsc2dense(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_cscSortedValA, size_t __pyx_v_cscSortedRowIndA, size_t __pyx_v_cscSortedColPtrA, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_52xcsr2coo(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_csrRowPtr, int __pyx_v_nnz, int __pyx_v_m, size_t __pyx_v_cooRowInd, int __pyx_v_idxBase); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_54scsr2csc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_cscVal, size_t __pyx_v_cscRowInd, size_t __pyx_v_cscColPtr, int __pyx_v_copyValues, int __pyx_v_idxBase); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_56dcsr2csc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_cscVal, size_t __pyx_v_cscRowInd, size_t __pyx_v_cscColPtr, int __pyx_v_copyValues, int __pyx_v_idxBase); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_58scsr2dense(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_60dcsr2dense(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_62snnz_compress(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, size_t __pyx_v_descr, size_t __pyx_v_values, size_t __pyx_v_rowPtr, size_t __pyx_v_nnzPerRow, float __pyx_v_tol); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_64dnnz_compress(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, size_t __pyx_v_descr, size_t __pyx_v_values, size_t __pyx_v_rowPtr, size_t __pyx_v_nnzPerRow, double __pyx_v_tol); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_66scsr2csr_compress(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_inVal, size_t __pyx_v_inColInd, size_t __pyx_v_inRowPtr, int __pyx_v_inNnz, size_t __pyx_v_nnzPerRow, size_t __pyx_v_outVal, size_t __pyx_v_outColInd, size_t __pyx_v_outRowPtr, float __pyx_v_tol); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_68dcsr2csr_compress(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_inVal, size_t __pyx_v_inColInd, size_t __pyx_v_inRowPtr, int __pyx_v_inNnz, size_t __pyx_v_nnzPerRow, size_t __pyx_v_outVal, size_t __pyx_v_outColInd, size_t __pyx_v_outRowPtr, float __pyx_v_tol); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_70sdense2csc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerCol, size_t __pyx_v_cscValA, size_t __pyx_v_cscRowIndA, size_t __pyx_v_cscColPtrA); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_72ddense2csc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerCol, size_t __pyx_v_cscValA, size_t __pyx_v_cscRowIndA, size_t __pyx_v_cscColPtrA); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_74sdense2csr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRow, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_76ddense2csr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRow, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_78snnz(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_dirA, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRowColumn, size_t __pyx_v_nnzTotalDevHostPtr); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_80dnnz(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_dirA, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRowColumn, size_t __pyx_v_nnzTotalDevHostPtr); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_82createIdentityPermutation(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_p); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_84xcoosort_bufferSizeExt(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cooRows, size_t __pyx_v_cooCols); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_86xcoosortByRow(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cooRows, size_t __pyx_v_cooCols, size_t __pyx_v_P, size_t __pyx_v_pBuffer); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_88xcsrsort_bufferSizeExt(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_90xcsrsort(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_P, size_t __pyx_v_pBuffer); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_92xcscsort_bufferSizeExt(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cscColPtr, size_t __pyx_v_cscRowInd); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_94xcscsort(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_cscColPtr, size_t __pyx_v_cscRowInd, size_t __pyx_v_P, size_t __pyx_v_pBuffer); /* proto */
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_9;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_codeobj__2;

/* "cupy/cuda/cusparse.pyx":251
 * class CuSparseError(RuntimeError):
 * 
 *     def __init__(self, int status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         super(CuSparseError, self).__init__('%s' % (STATUS[status]))
 */

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_13CuSparseError_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_13CuSparseError___init__[] = "CuSparseError.__init__(self, int status)";
static PyMethodDef __pyx_mdef_4cupy_4cuda_8cusparse_13CuSparseError_1__init__ = {"__init__", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_13CuSparseError_1__init__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_13CuSparseError___init__};
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_13CuSparseError_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  int __pyx_v_status;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_status,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_status)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 251, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 251, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_status = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_status == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 251, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 251, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.CuSparseError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_13CuSparseError___init__(__pyx_self, __pyx_v_self, __pyx_v_status);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_13CuSparseError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, int __pyx_v_status) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "cupy/cuda/cusparse.pyx":252
 * 
 *     def __init__(self, int status):
 *         self.status = status             # <<<<<<<<<<<<<<
 *         super(CuSparseError, self).__init__('%s' % (STATUS[status]))
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_status, __pyx_t_1) < 0) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":253
 *     def __init__(self, int status):
 *         self.status = status
 *         super(CuSparseError, self).__init__('%s' % (STATUS[status]))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_CuSparseError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_4cupy_4cuda_8cusparse_STATUS == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 253, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_4cupy_4cuda_8cusparse_STATUS, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_s, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":251
 * class CuSparseError(RuntimeError):
 * 
 *     def __init__(self, int status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         super(CuSparseError, self).__init__('%s' % (STATUS[status]))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cupy.cuda.cusparse.CuSparseError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":257
 * 
 * @cython.profile(False)
 * cpdef inline check_status(int status):             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         raise CuSparseError(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4cupy_4cuda_8cusparse_check_status(int __pyx_v_status, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("check_status", 0);

  /* "cupy/cuda/cusparse.pyx":258
 * @cython.profile(False)
 * cpdef inline check_status(int status):
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         raise CuSparseError(status)
 * 
 */
  __pyx_t_1 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_1) {

    /* "cupy/cuda/cusparse.pyx":259
 * cpdef inline check_status(int status):
 *     if status != 0:
 *         raise CuSparseError(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_CuSparseError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 259, __pyx_L1_error)

    /* "cupy/cuda/cusparse.pyx":258
 * @cython.profile(False)
 * cpdef inline check_status(int status):
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         raise CuSparseError(status)
 * 
 */
  }

  /* "cupy/cuda/cusparse.pyx":257
 * 
 * @cython.profile(False)
 * cpdef inline check_status(int status):             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         raise CuSparseError(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cupy.cuda.cusparse.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_check_status[] = "check_status(int status)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status) {
  int __pyx_v_status;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_status (wrapper)", 0);
  assert(__pyx_arg_status); {
    __pyx_v_status = __Pyx_PyInt_As_int(__pyx_arg_status); if (unlikely((__pyx_v_status == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 257, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_check_status(__pyx_self, ((int)__pyx_v_status));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_check_status(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_status) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("check_status", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":265
 * # cuSPARSE Helper Function
 * 
 * cpdef size_t create() except *:             # <<<<<<<<<<<<<<
 *     cdef Handle handle
 *     status = cusparseCreate(& handle)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_3create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_create(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseHandle_t __pyx_v_handle;
  cusparseStatus_t __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("create", 0);

  /* "cupy/cuda/cusparse.pyx":267
 * cpdef size_t create() except *:
 *     cdef Handle handle
 *     status = cusparseCreate(& handle)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t >handle
 */
  __pyx_v_status = cusparseCreate((&__pyx_v_handle));

  /* "cupy/cuda/cusparse.pyx":268
 *     cdef Handle handle
 *     status = cusparseCreate(& handle)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t >handle
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":269
 *     status = cusparseCreate(& handle)
 *     check_status(status)
 *     return <size_t >handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_handle);
  goto __pyx_L0;

  /* "cupy/cuda/cusparse.pyx":265
 * # cuSPARSE Helper Function
 * 
 * cpdef size_t create() except *:             # <<<<<<<<<<<<<<
 *     cdef Handle handle
 *     status = cusparseCreate(& handle)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_3create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_2create[] = "create() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_3create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_2create(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_2create(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("create", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_create(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 265, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusparse.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":272
 * 
 * 
 * cpdef size_t createMatDescr():             # <<<<<<<<<<<<<<
 *     cdef MatDescr desc
 *     status = cusparseCreateMatDescr(& desc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_5createMatDescr(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_createMatDescr(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseMatDescr_t __pyx_v_desc;
  cusparseStatus_t __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createMatDescr", 0);

  /* "cupy/cuda/cusparse.pyx":274
 * cpdef size_t createMatDescr():
 *     cdef MatDescr desc
 *     status = cusparseCreateMatDescr(& desc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>desc
 */
  __pyx_v_status = cusparseCreateMatDescr((&__pyx_v_desc));

  /* "cupy/cuda/cusparse.pyx":275
 *     cdef MatDescr desc
 *     status = cusparseCreateMatDescr(& desc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>desc
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":276
 *     status = cusparseCreateMatDescr(& desc)
 *     check_status(status)
 *     return <size_t>desc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_desc);
  goto __pyx_L0;

  /* "cupy/cuda/cusparse.pyx":272
 * 
 * 
 * cpdef size_t createMatDescr():             # <<<<<<<<<<<<<<
 *     cdef MatDescr desc
 *     status = cusparseCreateMatDescr(& desc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusparse.createMatDescr", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_5createMatDescr(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_4createMatDescr[] = "createMatDescr() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_5createMatDescr(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createMatDescr (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_4createMatDescr(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_4createMatDescr(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createMatDescr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_f_4cupy_4cuda_8cusparse_createMatDescr(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.createMatDescr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":279
 * 
 * 
 * cpdef void destroy(size_t handle):             # <<<<<<<<<<<<<<
 *     status = cusparseDestroy(<Handle >handle)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_7destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static void __pyx_f_4cupy_4cuda_8cusparse_destroy(size_t __pyx_v_handle, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroy", 0);

  /* "cupy/cuda/cusparse.pyx":280
 * 
 * cpdef void destroy(size_t handle):
 *     status = cusparseDestroy(<Handle >handle)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cusparseDestroy(((cusparseHandle_t)__pyx_v_handle));

  /* "cupy/cuda/cusparse.pyx":281
 * cpdef void destroy(size_t handle):
 *     status = cusparseDestroy(<Handle >handle)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":279
 * 
 * 
 * cpdef void destroy(size_t handle):             # <<<<<<<<<<<<<<
 *     status = cusparseDestroy(<Handle >handle)
 *     check_status(status)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusparse.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_7destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_6destroy[] = "destroy(size_t handle) -> void";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_7destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle) {
  size_t __pyx_v_handle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroy (wrapper)", 0);
  assert(__pyx_arg_handle); {
    __pyx_v_handle = __Pyx_PyInt_As_size_t(__pyx_arg_handle); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 279, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_6destroy(__pyx_self, ((size_t)__pyx_v_handle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_6destroy(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_4cupy_4cuda_8cusparse_destroy(__pyx_v_handle, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":284
 * 
 * 
 * cpdef void destroyMatDescr(size_t descr):             # <<<<<<<<<<<<<<
 *     status = cusparseDestroyMatDescr(<MatDescr>descr)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_9destroyMatDescr(PyObject *__pyx_self, PyObject *__pyx_arg_descr); /*proto*/
static void __pyx_f_4cupy_4cuda_8cusparse_destroyMatDescr(size_t __pyx_v_descr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyMatDescr", 0);

  /* "cupy/cuda/cusparse.pyx":285
 * 
 * cpdef void destroyMatDescr(size_t descr):
 *     status = cusparseDestroyMatDescr(<MatDescr>descr)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cusparseDestroyMatDescr(((cusparseMatDescr_t)__pyx_v_descr));

  /* "cupy/cuda/cusparse.pyx":286
 * cpdef void destroyMatDescr(size_t descr):
 *     status = cusparseDestroyMatDescr(<MatDescr>descr)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":284
 * 
 * 
 * cpdef void destroyMatDescr(size_t descr):             # <<<<<<<<<<<<<<
 *     status = cusparseDestroyMatDescr(<MatDescr>descr)
 *     check_status(status)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusparse.destroyMatDescr", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_9destroyMatDescr(PyObject *__pyx_self, PyObject *__pyx_arg_descr); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_8destroyMatDescr[] = "destroyMatDescr(size_t descr) -> void";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_9destroyMatDescr(PyObject *__pyx_self, PyObject *__pyx_arg_descr) {
  size_t __pyx_v_descr;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyMatDescr (wrapper)", 0);
  assert(__pyx_arg_descr); {
    __pyx_v_descr = __Pyx_PyInt_As_size_t(__pyx_arg_descr); if (unlikely((__pyx_v_descr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 284, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.destroyMatDescr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_8destroyMatDescr(__pyx_self, ((size_t)__pyx_v_descr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_8destroyMatDescr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_descr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyMatDescr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_4cupy_4cuda_8cusparse_destroyMatDescr(__pyx_v_descr, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.destroyMatDescr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":289
 * 
 * 
 * cpdef void setMatIndexBase(size_t descr, base):             # <<<<<<<<<<<<<<
 *     status = cusparseSetMatIndexBase(<MatDescr>descr, base)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_11setMatIndexBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static void __pyx_f_4cupy_4cuda_8cusparse_setMatIndexBase(size_t __pyx_v_descr, PyObject *__pyx_v_base, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  __Pyx_RefNannyDeclarations
  cusparseIndexBase_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("setMatIndexBase", 0);

  /* "cupy/cuda/cusparse.pyx":290
 * 
 * cpdef void setMatIndexBase(size_t descr, base):
 *     status = cusparseSetMatIndexBase(<MatDescr>descr, base)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_cusparseIndexBase_t(__pyx_v_base); if (unlikely((__pyx_t_1 == ((cusparseIndexBase_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 290, __pyx_L1_error)
  __pyx_v_status = cusparseSetMatIndexBase(((cusparseMatDescr_t)__pyx_v_descr), __pyx_t_1);

  /* "cupy/cuda/cusparse.pyx":291
 * cpdef void setMatIndexBase(size_t descr, base):
 *     status = cusparseSetMatIndexBase(<MatDescr>descr, base)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cupy/cuda/cusparse.pyx":289
 * 
 * 
 * cpdef void setMatIndexBase(size_t descr, base):             # <<<<<<<<<<<<<<
 *     status = cusparseSetMatIndexBase(<MatDescr>descr, base)
 *     check_status(status)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("cupy.cuda.cusparse.setMatIndexBase", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_11setMatIndexBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_10setMatIndexBase[] = "setMatIndexBase(size_t descr, base) -> void";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_11setMatIndexBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_descr;
  PyObject *__pyx_v_base = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setMatIndexBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_descr,&__pyx_n_s_base,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_base)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setMatIndexBase", 1, 2, 2, 1); __PYX_ERR(0, 289, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setMatIndexBase") < 0)) __PYX_ERR(0, 289, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_descr = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_descr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 289, __pyx_L3_error)
    __pyx_v_base = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setMatIndexBase", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 289, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.setMatIndexBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_10setMatIndexBase(__pyx_self, __pyx_v_descr, __pyx_v_base);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_10setMatIndexBase(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_descr, PyObject *__pyx_v_base) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setMatIndexBase", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_4cupy_4cuda_8cusparse_setMatIndexBase(__pyx_v_descr, __pyx_v_base, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.setMatIndexBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":294
 * 
 * 
 * cpdef setMatType(size_t descr, typ):             # <<<<<<<<<<<<<<
 *     status = cusparseSetMatType(<MatDescr>descr, typ)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_13setMatType(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_setMatType(size_t __pyx_v_descr, PyObject *__pyx_v_typ, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  cusparseMatrixType_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("setMatType", 0);

  /* "cupy/cuda/cusparse.pyx":295
 * 
 * cpdef setMatType(size_t descr, typ):
 *     status = cusparseSetMatType(<MatDescr>descr, typ)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_cusparseMatrixType_t(__pyx_v_typ); if (unlikely((__pyx_t_1 == ((cusparseMatrixType_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L1_error)
  __pyx_v_status = cusparseSetMatType(((cusparseMatDescr_t)__pyx_v_descr), __pyx_t_1);

  /* "cupy/cuda/cusparse.pyx":296
 * cpdef setMatType(size_t descr, typ):
 *     status = cusparseSetMatType(<MatDescr>descr, typ)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cupy/cuda/cusparse.pyx":294
 * 
 * 
 * cpdef setMatType(size_t descr, typ):             # <<<<<<<<<<<<<<
 *     status = cusparseSetMatType(<MatDescr>descr, typ)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusparse.setMatType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_13setMatType(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_12setMatType[] = "setMatType(size_t descr, typ)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_13setMatType(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_descr;
  PyObject *__pyx_v_typ = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setMatType (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_descr,&__pyx_n_s_typ,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_typ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setMatType", 1, 2, 2, 1); __PYX_ERR(0, 294, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setMatType") < 0)) __PYX_ERR(0, 294, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_descr = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_descr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 294, __pyx_L3_error)
    __pyx_v_typ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setMatType", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 294, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.setMatType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_12setMatType(__pyx_self, __pyx_v_descr, __pyx_v_typ);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_12setMatType(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_descr, PyObject *__pyx_v_typ) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setMatType", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setMatType(__pyx_v_descr, __pyx_v_typ, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.setMatType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":299
 * 
 * 
 * cpdef setPointerMode(size_t handle, int mode):             # <<<<<<<<<<<<<<
 *     status = cusparseSetPointerMode(<Handle>handle, <PointerMode>mode)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_15setPointerMode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_setPointerMode(size_t __pyx_v_handle, int __pyx_v_mode, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setPointerMode", 0);

  /* "cupy/cuda/cusparse.pyx":300
 * 
 * cpdef setPointerMode(size_t handle, int mode):
 *     status = cusparseSetPointerMode(<Handle>handle, <PointerMode>mode)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cusparseSetPointerMode(((cusparseHandle_t)__pyx_v_handle), ((cusparsePointerMode_t)__pyx_v_mode));

  /* "cupy/cuda/cusparse.pyx":301
 * cpdef setPointerMode(size_t handle, int mode):
 *     status = cusparseSetPointerMode(<Handle>handle, <PointerMode>mode)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * ########################################
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":299
 * 
 * 
 * cpdef setPointerMode(size_t handle, int mode):             # <<<<<<<<<<<<<<
 *     status = cusparseSetPointerMode(<Handle>handle, <PointerMode>mode)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.setPointerMode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_15setPointerMode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_14setPointerMode[] = "setPointerMode(size_t handle, int mode)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_15setPointerMode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_mode;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setPointerMode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_mode,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPointerMode", 1, 2, 2, 1); __PYX_ERR(0, 299, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setPointerMode") < 0)) __PYX_ERR(0, 299, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 299, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 299, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setPointerMode", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 299, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.setPointerMode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_14setPointerMode(__pyx_self, __pyx_v_handle, __pyx_v_mode);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_14setPointerMode(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_mode) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setPointerMode", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setPointerMode(__pyx_v_handle, __pyx_v_mode, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.setPointerMode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":306
 * # Stream
 * 
 * cpdef setStream(size_t handle, size_t stream):             # <<<<<<<<<<<<<<
 *     status = cusparseSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_17setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_setStream(size_t __pyx_v_handle, size_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setStream", 0);

  /* "cupy/cuda/cusparse.pyx":307
 * 
 * cpdef setStream(size_t handle, size_t stream):
 *     status = cusparseSetStream(<Handle>handle, <driver.Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cusparseSetStream(((cusparseHandle_t)__pyx_v_handle), ((struct CUstream_st*)__pyx_v_stream));

  /* "cupy/cuda/cusparse.pyx":308
 * cpdef setStream(size_t handle, size_t stream):
 *     status = cusparseSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":306
 * # Stream
 * 
 * cpdef setStream(size_t handle, size_t stream):             # <<<<<<<<<<<<<<
 *     status = cusparseSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_17setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_16setStream[] = "setStream(size_t handle, size_t stream)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_17setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_stream;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setStream (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_stream,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setStream", 1, 2, 2, 1); __PYX_ERR(0, 306, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setStream") < 0)) __PYX_ERR(0, 306, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 306, __pyx_L3_error)
    __pyx_v_stream = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_stream == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 306, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setStream", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 306, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_16setStream(__pyx_self, __pyx_v_handle, __pyx_v_stream);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_16setStream(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setStream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_v_stream, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":322
 * # cuSPARSE Level1 Function
 * 
 * cpdef sgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd,             # <<<<<<<<<<<<<<
 *             int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_19sgthr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_sgthr(size_t __pyx_v_handle, int __pyx_v_nnz, size_t __pyx_v_y, size_t __pyx_v_xVal, size_t __pyx_v_xInd, int __pyx_v_idxBase, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgthr", 0);

  /* "cupy/cuda/cusparse.pyx":324
 * cpdef sgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd,
 *             int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseSgthr(
 *         <Handle>handle, nnz, <const float *>y, <float *>xVal,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":325
 *             int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseSgthr(             # <<<<<<<<<<<<<<
 *         <Handle>handle, nnz, <const float *>y, <float *>xVal,
 *         <const int *>xInd, <IndexBase>idxBase)
 */
  __pyx_v_status = cusparseSgthr(((cusparseHandle_t)__pyx_v_handle), __pyx_v_nnz, ((float const *)__pyx_v_y), ((float *)__pyx_v_xVal), ((int const *)__pyx_v_xInd), ((cusparseIndexBase_t)__pyx_v_idxBase));

  /* "cupy/cuda/cusparse.pyx":328
 *         <Handle>handle, nnz, <const float *>y, <float *>xVal,
 *         <const int *>xInd, <IndexBase>idxBase)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":322
 * # cuSPARSE Level1 Function
 * 
 * cpdef sgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd,             # <<<<<<<<<<<<<<
 *             int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.sgthr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_19sgthr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_18sgthr[] = "sgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd, int idxBase)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_19sgthr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_nnz;
  size_t __pyx_v_y;
  size_t __pyx_v_xVal;
  size_t __pyx_v_xInd;
  int __pyx_v_idxBase;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgthr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_nnz,&__pyx_n_s_y,&__pyx_n_s_xVal,&__pyx_n_s_xInd,&__pyx_n_s_idxBase,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgthr", 1, 6, 6, 1); __PYX_ERR(0, 322, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgthr", 1, 6, 6, 2); __PYX_ERR(0, 322, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgthr", 1, 6, 6, 3); __PYX_ERR(0, 322, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgthr", 1, 6, 6, 4); __PYX_ERR(0, 322, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_idxBase)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgthr", 1, 6, 6, 5); __PYX_ERR(0, 322, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgthr") < 0)) __PYX_ERR(0, 322, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L3_error)
    __pyx_v_xVal = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L3_error)
    __pyx_v_xInd = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_xInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L3_error)
    __pyx_v_idxBase = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_idxBase == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 323, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgthr", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 322, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.sgthr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_18sgthr(__pyx_self, __pyx_v_handle, __pyx_v_nnz, __pyx_v_y, __pyx_v_xVal, __pyx_v_xInd, __pyx_v_idxBase);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_18sgthr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_nnz, size_t __pyx_v_y, size_t __pyx_v_xVal, size_t __pyx_v_xInd, int __pyx_v_idxBase) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgthr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_sgthr(__pyx_v_handle, __pyx_v_nnz, __pyx_v_y, __pyx_v_xVal, __pyx_v_xInd, __pyx_v_idxBase, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.sgthr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":330
 *     check_status(status)
 * 
 * cpdef dgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd,             # <<<<<<<<<<<<<<
 *             int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_21dgthr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dgthr(size_t __pyx_v_handle, int __pyx_v_nnz, size_t __pyx_v_y, size_t __pyx_v_xVal, size_t __pyx_v_xInd, int __pyx_v_idxBase, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgthr", 0);

  /* "cupy/cuda/cusparse.pyx":332
 * cpdef dgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd,
 *             int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseDgthr(
 *         <Handle>handle, nnz, <const double *>y, <double *>xVal,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":333
 *             int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseDgthr(             # <<<<<<<<<<<<<<
 *         <Handle>handle, nnz, <const double *>y, <double *>xVal,
 *         <const int *>xInd, <IndexBase>idxBase)
 */
  __pyx_v_status = cusparseDgthr(((cusparseHandle_t)__pyx_v_handle), __pyx_v_nnz, ((double const *)__pyx_v_y), ((double *)__pyx_v_xVal), ((int const *)__pyx_v_xInd), ((cusparseIndexBase_t)__pyx_v_idxBase));

  /* "cupy/cuda/cusparse.pyx":336
 *         <Handle>handle, nnz, <const double *>y, <double *>xVal,
 *         <const int *>xInd, <IndexBase>idxBase)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":330
 *     check_status(status)
 * 
 * cpdef dgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd,             # <<<<<<<<<<<<<<
 *             int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dgthr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_21dgthr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_20dgthr[] = "dgthr(size_t handle, int nnz, size_t y, size_t xVal, size_t xInd, int idxBase)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_21dgthr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_nnz;
  size_t __pyx_v_y;
  size_t __pyx_v_xVal;
  size_t __pyx_v_xInd;
  int __pyx_v_idxBase;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgthr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_nnz,&__pyx_n_s_y,&__pyx_n_s_xVal,&__pyx_n_s_xInd,&__pyx_n_s_idxBase,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgthr", 1, 6, 6, 1); __PYX_ERR(0, 330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgthr", 1, 6, 6, 2); __PYX_ERR(0, 330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgthr", 1, 6, 6, 3); __PYX_ERR(0, 330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgthr", 1, 6, 6, 4); __PYX_ERR(0, 330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_idxBase)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgthr", 1, 6, 6, 5); __PYX_ERR(0, 330, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgthr") < 0)) __PYX_ERR(0, 330, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L3_error)
    __pyx_v_xVal = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L3_error)
    __pyx_v_xInd = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_xInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L3_error)
    __pyx_v_idxBase = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_idxBase == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 331, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgthr", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 330, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dgthr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_20dgthr(__pyx_self, __pyx_v_handle, __pyx_v_nnz, __pyx_v_y, __pyx_v_xVal, __pyx_v_xInd, __pyx_v_idxBase);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_20dgthr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_nnz, size_t __pyx_v_y, size_t __pyx_v_xVal, size_t __pyx_v_xInd, int __pyx_v_idxBase) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgthr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dgthr(__pyx_v_handle, __pyx_v_nnz, __pyx_v_y, __pyx_v_xVal, __pyx_v_xInd, __pyx_v_idxBase, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dgthr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":342
 * # cuSPARSE Level2 Function
 * 
 * cpdef scsrmv(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int m, int n, int nnz,
 *         size_t alpha, size_t descrA, size_t csrSortedValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_23scsrmv(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrmv(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_y, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrmv", 0);

  /* "cupy/cuda/cusparse.pyx":347
 *         size_t csrSortedRowPtrA, size_t csrSortedColIndA,
 *         size_t x, size_t beta, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseScsrmv(
 *         <Handle>handle, <Operation>transA, m, n, nnz,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":348
 *         size_t x, size_t beta, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseScsrmv(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, m, n, nnz,
 *         <const float *>alpha, <MatDescr>descrA, <const float *>csrSortedValA,
 */
  __pyx_v_status = cusparseScsrmv(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((float const *)__pyx_v_alpha), ((cusparseMatDescr_t)__pyx_v_descrA), ((float const *)__pyx_v_csrSortedValA), ((int const *)__pyx_v_csrSortedRowPtrA), ((int const *)__pyx_v_csrSortedColIndA), ((float const *)__pyx_v_x), ((float const *)__pyx_v_beta), ((float *)__pyx_v_y));

  /* "cupy/cuda/cusparse.pyx":353
 *         <const int *>csrSortedRowPtrA, <const int *>csrSortedColIndA,
 *         <const float *>x, <const float *>beta, <float *>y)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dcsrmv(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":342
 * # cuSPARSE Level2 Function
 * 
 * cpdef scsrmv(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int m, int n, int nnz,
 *         size_t alpha, size_t descrA, size_t csrSortedValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_23scsrmv(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_22scsrmv[] = "scsrmv(size_t handle, int transA, int m, int n, int nnz, size_t alpha, size_t descrA, size_t csrSortedValA, size_t csrSortedRowPtrA, size_t csrSortedColIndA, size_t x, size_t beta, size_t y)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_23scsrmv(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_alpha;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrSortedValA;
  size_t __pyx_v_csrSortedRowPtrA;
  size_t __pyx_v_csrSortedColIndA;
  size_t __pyx_v_x;
  size_t __pyx_v_beta;
  size_t __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsrmv (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_alpha,&__pyx_n_s_descrA,&__pyx_n_s_csrSortedValA,&__pyx_n_s_csrSortedRowPtrA,&__pyx_n_s_csrSortedColIndA,&__pyx_n_s_x,&__pyx_n_s_beta,&__pyx_n_s_y,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 1); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 2); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 3); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 4); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 5); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 6); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 7); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 8); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 9); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 10); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 11); __PYX_ERR(0, 342, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, 12); __PYX_ERR(0, 342, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsrmv") < 0)) __PYX_ERR(0, 342, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 344, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 344, __pyx_L3_error)
    __pyx_v_csrSortedValA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_csrSortedValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 344, __pyx_L3_error)
    __pyx_v_csrSortedRowPtrA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrSortedRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L3_error)
    __pyx_v_csrSortedColIndA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrSortedColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 346, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 346, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 346, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsrmv", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 342, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_22scsrmv(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_x, __pyx_v_beta, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_22scsrmv(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrmv", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsrmv(__pyx_v_handle, __pyx_v_transA, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_x, __pyx_v_beta, __pyx_v_y, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":355
 *     check_status(status)
 * 
 * cpdef dcsrmv(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int m, int n, int nnz,
 *         size_t alpha, size_t descrA, size_t csrSortedValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_25dcsrmv(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrmv(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_y, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrmv", 0);

  /* "cupy/cuda/cusparse.pyx":360
 *         size_t csrSortedRowPtrA, size_t csrSortedColIndA,
 *         size_t x, size_t beta, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseDcsrmv(
 *         <Handle>handle, <Operation>transA, m, n, nnz,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":361
 *         size_t x, size_t beta, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseDcsrmv(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, m, n, nnz,
 *         <const double *>alpha, <MatDescr>descrA, <const double *>csrSortedValA,
 */
  __pyx_v_status = cusparseDcsrmv(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((double const *)__pyx_v_alpha), ((cusparseMatDescr_t)__pyx_v_descrA), ((double const *)__pyx_v_csrSortedValA), ((int const *)__pyx_v_csrSortedRowPtrA), ((int const *)__pyx_v_csrSortedColIndA), ((double const *)__pyx_v_x), ((double const *)__pyx_v_beta), ((double *)__pyx_v_y));

  /* "cupy/cuda/cusparse.pyx":366
 *         <const int *>csrSortedRowPtrA, <const int *>csrSortedColIndA,
 *         <const double *>x, <const double *>beta, <double *>y)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":355
 *     check_status(status)
 * 
 * cpdef dcsrmv(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int m, int n, int nnz,
 *         size_t alpha, size_t descrA, size_t csrSortedValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_25dcsrmv(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_24dcsrmv[] = "dcsrmv(size_t handle, int transA, int m, int n, int nnz, size_t alpha, size_t descrA, size_t csrSortedValA, size_t csrSortedRowPtrA, size_t csrSortedColIndA, size_t x, size_t beta, size_t y)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_25dcsrmv(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_alpha;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrSortedValA;
  size_t __pyx_v_csrSortedRowPtrA;
  size_t __pyx_v_csrSortedColIndA;
  size_t __pyx_v_x;
  size_t __pyx_v_beta;
  size_t __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsrmv (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_alpha,&__pyx_n_s_descrA,&__pyx_n_s_csrSortedValA,&__pyx_n_s_csrSortedRowPtrA,&__pyx_n_s_csrSortedColIndA,&__pyx_n_s_x,&__pyx_n_s_beta,&__pyx_n_s_y,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 1); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 2); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 3); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 4); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 5); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 6); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 7); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 8); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 9); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 10); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 11); __PYX_ERR(0, 355, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, 12); __PYX_ERR(0, 355, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsrmv") < 0)) __PYX_ERR(0, 355, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 356, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L3_error)
    __pyx_v_csrSortedValA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_csrSortedValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L3_error)
    __pyx_v_csrSortedRowPtrA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrSortedRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L3_error)
    __pyx_v_csrSortedColIndA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrSortedColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsrmv", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 355, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_24dcsrmv(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_x, __pyx_v_beta, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_24dcsrmv(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrmv", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsrmv(__pyx_v_handle, __pyx_v_transA, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_x, __pyx_v_beta, __pyx_v_y, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":372
 * # cuSPARSE Level3 Function
 * 
 * cpdef scsrmm(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int m, int n, int k, int nnz,
 *         size_t alpha, size_t descrA, size_t csrSortedValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_27scsrmm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrmm(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrmm", 0);

  /* "cupy/cuda/cusparse.pyx":377
 *         size_t csrSortedRowPtrA, size_t csrSortedColIndA,
 *         size_t B, int ldb, size_t beta, size_t C, int ldc):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseScsrmm(
 *         <Handle>handle, <Operation>transA, m, n, k, nnz,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":378
 *         size_t B, int ldb, size_t beta, size_t C, int ldc):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseScsrmm(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, m, n, k, nnz,
 *         <const float *>alpha, <MatDescr>descrA, <const float *>csrSortedValA,
 */
  __pyx_v_status = cusparseScsrmm(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, ((float const *)__pyx_v_alpha), ((cusparseMatDescr_t)__pyx_v_descrA), ((float const *)__pyx_v_csrSortedValA), ((int const *)__pyx_v_csrSortedRowPtrA), ((int const *)__pyx_v_csrSortedColIndA), ((float const *)__pyx_v_B), __pyx_v_ldb, ((float const *)__pyx_v_beta), ((float *)__pyx_v_C), __pyx_v_ldc);

  /* "cupy/cuda/cusparse.pyx":383
 *         <const int *>csrSortedRowPtrA, <const int *>csrSortedColIndA,
 *         <const float *>B, ldb, <const float *>beta, <float *>C, ldc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dcsrmm(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":372
 * # cuSPARSE Level3 Function
 * 
 * cpdef scsrmm(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int m, int n, int k, int nnz,
 *         size_t alpha, size_t descrA, size_t csrSortedValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_27scsrmm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_26scsrmm[] = "scsrmm(size_t handle, int transA, int m, int n, int k, int nnz, size_t alpha, size_t descrA, size_t csrSortedValA, size_t csrSortedRowPtrA, size_t csrSortedColIndA, size_t B, int ldb, size_t beta, size_t C, int ldc)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_27scsrmm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  int __pyx_v_nnz;
  size_t __pyx_v_alpha;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrSortedValA;
  size_t __pyx_v_csrSortedRowPtrA;
  size_t __pyx_v_csrSortedColIndA;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_beta;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsrmm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_nnz,&__pyx_n_s_alpha,&__pyx_n_s_descrA,&__pyx_n_s_csrSortedValA,&__pyx_n_s_csrSortedRowPtrA,&__pyx_n_s_csrSortedColIndA,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_beta,&__pyx_n_s_C,&__pyx_n_s_ldc,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 1); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 2); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 3); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 4); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 5); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 6); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 7); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 8); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 9); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 10); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 11); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 12); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 13); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 14); __PYX_ERR(0, 372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, 15); __PYX_ERR(0, 372, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsrmm") < 0)) __PYX_ERR(0, 372, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 373, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)
    __pyx_v_csrSortedValA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrSortedValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)
    __pyx_v_csrSortedRowPtrA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrSortedRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 375, __pyx_L3_error)
    __pyx_v_csrSortedColIndA = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_csrSortedColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 375, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 376, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 376, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 376, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 376, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[15]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 376, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsrmm", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 372, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_26scsrmm(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_B, __pyx_v_ldb, __pyx_v_beta, __pyx_v_C, __pyx_v_ldc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_26scsrmm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrmm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsrmm(__pyx_v_handle, __pyx_v_transA, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_B, __pyx_v_ldb, __pyx_v_beta, __pyx_v_C, __pyx_v_ldc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":385
 *     check_status(status)
 * 
 * cpdef dcsrmm(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int m, int n, int k, int nnz,
 *         size_t alpha, size_t descrA, size_t csrSortedValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_29dcsrmm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrmm(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrmm", 0);

  /* "cupy/cuda/cusparse.pyx":390
 *         size_t csrSortedRowPtrA, size_t csrSortedColIndA,
 *         size_t B, int ldb, size_t beta, size_t C, int ldc):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseDcsrmm(
 *         <Handle>handle, <Operation>transA, m, n, k, nnz,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":391
 *         size_t B, int ldb, size_t beta, size_t C, int ldc):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseDcsrmm(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, m, n, k, nnz,
 *         <const double *>alpha, <MatDescr>descrA, <const double *>csrSortedValA,
 */
  __pyx_v_status = cusparseDcsrmm(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, ((double const *)__pyx_v_alpha), ((cusparseMatDescr_t)__pyx_v_descrA), ((double const *)__pyx_v_csrSortedValA), ((int const *)__pyx_v_csrSortedRowPtrA), ((int const *)__pyx_v_csrSortedColIndA), ((double const *)__pyx_v_B), __pyx_v_ldb, ((double const *)__pyx_v_beta), ((double *)__pyx_v_C), __pyx_v_ldc);

  /* "cupy/cuda/cusparse.pyx":396
 *         <const int *>csrSortedRowPtrA, <const int *>csrSortedColIndA,
 *         <const double *>B, ldb, <const double *>beta, <double *>C, ldc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef scsrmm2(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":385
 *     check_status(status)
 * 
 * cpdef dcsrmm(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int m, int n, int k, int nnz,
 *         size_t alpha, size_t descrA, size_t csrSortedValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_29dcsrmm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_28dcsrmm[] = "dcsrmm(size_t handle, int transA, int m, int n, int k, int nnz, size_t alpha, size_t descrA, size_t csrSortedValA, size_t csrSortedRowPtrA, size_t csrSortedColIndA, size_t B, int ldb, size_t beta, size_t C, int ldc)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_29dcsrmm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  int __pyx_v_nnz;
  size_t __pyx_v_alpha;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrSortedValA;
  size_t __pyx_v_csrSortedRowPtrA;
  size_t __pyx_v_csrSortedColIndA;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_beta;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsrmm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_nnz,&__pyx_n_s_alpha,&__pyx_n_s_descrA,&__pyx_n_s_csrSortedValA,&__pyx_n_s_csrSortedRowPtrA,&__pyx_n_s_csrSortedColIndA,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_beta,&__pyx_n_s_C,&__pyx_n_s_ldc,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 1); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 2); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 3); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 4); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 5); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 6); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 7); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 8); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 9); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 10); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 11); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 12); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 13); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 14); __PYX_ERR(0, 385, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, 15); __PYX_ERR(0, 385, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsrmm") < 0)) __PYX_ERR(0, 385, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
    __pyx_v_csrSortedValA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrSortedValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
    __pyx_v_csrSortedRowPtrA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrSortedRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 388, __pyx_L3_error)
    __pyx_v_csrSortedColIndA = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_csrSortedColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 388, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[15]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsrmm", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 385, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_28dcsrmm(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_B, __pyx_v_ldb, __pyx_v_beta, __pyx_v_C, __pyx_v_ldc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_28dcsrmm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrmm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsrmm(__pyx_v_handle, __pyx_v_transA, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_B, __pyx_v_ldb, __pyx_v_beta, __pyx_v_C, __pyx_v_ldc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":398
 *     check_status(status)
 * 
 * cpdef scsrmm2(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k, int nnz,
 *         size_t alpha, size_t descrA, size_t csrValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_31scsrmm2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrmm2(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrmm2", 0);

  /* "cupy/cuda/cusparse.pyx":403
 *         size_t csrRowPtrA, size_t csrColIndA,
 *         size_t B, int ldb, size_t beta, size_t C, int ldc):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseScsrmm2(
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k, nnz,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":404
 *         size_t B, int ldb, size_t beta, size_t C, int ldc):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseScsrmm2(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k, nnz,
 *         <const float *>alpha, <MatDescr>descrA, <const float *>csrValA,
 */
  __pyx_v_status = cusparseScsrmm2(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), ((cusparseOperation_t)__pyx_v_transB), __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, ((float const *)__pyx_v_alpha), ((cusparseMatDescr_t)__pyx_v_descrA), ((float const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((float const *)__pyx_v_B), __pyx_v_ldb, ((float const *)__pyx_v_beta), ((float *)__pyx_v_C), __pyx_v_ldc);

  /* "cupy/cuda/cusparse.pyx":409
 *         <const int *>csrRowPtrA, <const int *>csrColIndA,
 *         <const float *>B, ldb, <const float *>beta, <float *>C, ldc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dcsrmm2(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":398
 *     check_status(status)
 * 
 * cpdef scsrmm2(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k, int nnz,
 *         size_t alpha, size_t descrA, size_t csrValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmm2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_31scsrmm2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_30scsrmm2[] = "scsrmm2(size_t handle, int transA, int transB, int m, int n, int k, int nnz, size_t alpha, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t B, int ldb, size_t beta, size_t C, int ldc)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_31scsrmm2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_transB;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  int __pyx_v_nnz;
  size_t __pyx_v_alpha;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_beta;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsrmm2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_transB,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_nnz,&__pyx_n_s_alpha,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_beta,&__pyx_n_s_C,&__pyx_n_s_ldc,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 1); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 2); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 3); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 4); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 5); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 6); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 7); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 8); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 9); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 10); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 11); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 12); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 13); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 14); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 15); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, 16); __PYX_ERR(0, 398, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsrmm2") < 0)) __PYX_ERR(0, 398, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    __pyx_v_transB = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_transB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 400, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 400, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 400, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 401, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 401, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[16]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsrmm2", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 398, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmm2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_30scsrmm2(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_B, __pyx_v_ldb, __pyx_v_beta, __pyx_v_C, __pyx_v_ldc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_30scsrmm2(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrmm2", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsrmm2(__pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_B, __pyx_v_ldb, __pyx_v_beta, __pyx_v_C, __pyx_v_ldc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrmm2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":411
 *     check_status(status)
 * 
 * cpdef dcsrmm2(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k, int nnz,
 *         size_t alpha, size_t descrA, size_t csrValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_33dcsrmm2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrmm2(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrmm2", 0);

  /* "cupy/cuda/cusparse.pyx":416
 *         size_t csrRowPtrA, size_t csrColIndA,
 *         size_t B, int ldb, size_t beta, size_t C, int ldc):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseDcsrmm2(
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k, nnz,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":417
 *         size_t B, int ldb, size_t beta, size_t C, int ldc):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseDcsrmm2(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k, nnz,
 *         <const double *>alpha, <MatDescr>descrA, <const double *>csrValA,
 */
  __pyx_v_status = cusparseDcsrmm2(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), ((cusparseOperation_t)__pyx_v_transB), __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, ((double const *)__pyx_v_alpha), ((cusparseMatDescr_t)__pyx_v_descrA), ((double const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((double const *)__pyx_v_B), __pyx_v_ldb, ((double const *)__pyx_v_beta), ((double *)__pyx_v_C), __pyx_v_ldc);

  /* "cupy/cuda/cusparse.pyx":422
 *         <const int *>csrRowPtrA, <const int *>csrColIndA,
 *         <const double *>B, ldb, <const double *>beta, <double *>C, ldc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":411
 *     check_status(status)
 * 
 * cpdef dcsrmm2(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k, int nnz,
 *         size_t alpha, size_t descrA, size_t csrValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmm2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_33dcsrmm2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_32dcsrmm2[] = "dcsrmm2(size_t handle, int transA, int transB, int m, int n, int k, int nnz, size_t alpha, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t B, int ldb, size_t beta, size_t C, int ldc)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_33dcsrmm2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_transB;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  int __pyx_v_nnz;
  size_t __pyx_v_alpha;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_beta;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsrmm2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_transB,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_nnz,&__pyx_n_s_alpha,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_beta,&__pyx_n_s_C,&__pyx_n_s_ldc,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 1); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 2); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 3); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 4); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 5); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 6); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 7); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 8); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 9); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 10); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 11); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 12); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 13); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 14); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 15); __PYX_ERR(0, 411, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, 16); __PYX_ERR(0, 411, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsrmm2") < 0)) __PYX_ERR(0, 411, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    __pyx_v_transB = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_transB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 413, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 414, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 414, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 415, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 415, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 415, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 415, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[16]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 415, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsrmm2", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 411, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmm2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_32dcsrmm2(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_B, __pyx_v_ldb, __pyx_v_beta, __pyx_v_C, __pyx_v_ldc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_32dcsrmm2(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, int __pyx_v_nnz, size_t __pyx_v_alpha, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_beta, size_t __pyx_v_C, int __pyx_v_ldc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrmm2", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsrmm2(__pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_nnz, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_B, __pyx_v_ldb, __pyx_v_beta, __pyx_v_C, __pyx_v_ldc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrmm2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":428
 * # cuSPARSE Extra Function
 * 
 * cpdef xcsrgeamNnz(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA, int nnzA,
 *         size_t csrRowPtrA, size_t csrColIndA, size_t descrB,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_35xcsrgeamNnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcsrgeamNnz(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_nnzTotalDevHostPtr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsrgeamNnz", 0);

  /* "cupy/cuda/cusparse.pyx":433
 *         int nnzB, size_t csrRowPtrB, size_t csrColIndB,
 *         size_t descrC, size_t csrRowPtrC, size_t nnzTotalDevHostPtr):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcsrgeamNnz(
 *         <Handle>handle, m, n, <const MatDescr>descrA, nnzA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":434
 *         size_t descrC, size_t csrRowPtrC, size_t nnzTotalDevHostPtr):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcsrgeamNnz(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <const MatDescr>descrA, nnzA,
 *         <const int *>csrRowPtrA, <const int *>csrColIndA,
 */
  __pyx_v_status = cusparseXcsrgeamNnz(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t const )__pyx_v_descrA), __pyx_v_nnzA, ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((cusparseMatDescr_t const )__pyx_v_descrB), __pyx_v_nnzB, ((int const *)__pyx_v_csrRowPtrB), ((int const *)__pyx_v_csrColIndB), ((cusparseMatDescr_t const )__pyx_v_descrC), ((int *)__pyx_v_csrRowPtrC), ((int *)__pyx_v_nnzTotalDevHostPtr));

  /* "cupy/cuda/cusparse.pyx":440
 *         <const int *>csrColIndB, <const MatDescr>descrC, <int *>csrRowPtrC,
 *         <int *>nnzTotalDevHostPtr)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef scsrgeam(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":428
 * # cuSPARSE Extra Function
 * 
 * cpdef xcsrgeamNnz(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA, int nnzA,
 *         size_t csrRowPtrA, size_t csrColIndA, size_t descrB,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrgeamNnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_35xcsrgeamNnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_34xcsrgeamNnz[] = "xcsrgeamNnz(size_t handle, int m, int n, size_t descrA, int nnzA, size_t csrRowPtrA, size_t csrColIndA, size_t descrB, int nnzB, size_t csrRowPtrB, size_t csrColIndB, size_t descrC, size_t csrRowPtrC, size_t nnzTotalDevHostPtr)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_35xcsrgeamNnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  int __pyx_v_nnzA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_descrB;
  int __pyx_v_nnzB;
  size_t __pyx_v_csrRowPtrB;
  size_t __pyx_v_csrColIndB;
  size_t __pyx_v_descrC;
  size_t __pyx_v_csrRowPtrC;
  size_t __pyx_v_nnzTotalDevHostPtr;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcsrgeamNnz (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_nnzA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_descrB,&__pyx_n_s_nnzB,&__pyx_n_s_csrRowPtrB,&__pyx_n_s_csrColIndB,&__pyx_n_s_descrC,&__pyx_n_s_csrRowPtrC,&__pyx_n_s_nnzTotalDevHostPtr,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 1); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 2); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 3); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 4); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 5); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 6); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 7); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 8); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 9); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 10); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 11); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 12); __PYX_ERR(0, 428, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzTotalDevHostPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, 13); __PYX_ERR(0, 428, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcsrgeamNnz") < 0)) __PYX_ERR(0, 428, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 429, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 429, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 429, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 429, __pyx_L3_error)
    __pyx_v_nnzA = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_nnzA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 429, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 430, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 430, __pyx_L3_error)
    __pyx_v_descrB = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_descrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 430, __pyx_L3_error)
    __pyx_v_nnzB = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_nnzB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 431, __pyx_L3_error)
    __pyx_v_csrRowPtrB = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrRowPtrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 431, __pyx_L3_error)
    __pyx_v_csrColIndB = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_csrColIndB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 431, __pyx_L3_error)
    __pyx_v_descrC = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_descrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 432, __pyx_L3_error)
    __pyx_v_csrRowPtrC = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_csrRowPtrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 432, __pyx_L3_error)
    __pyx_v_nnzTotalDevHostPtr = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_nnzTotalDevHostPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 432, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcsrgeamNnz", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 428, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrgeamNnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_34xcsrgeamNnz(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrRowPtrC, __pyx_v_nnzTotalDevHostPtr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_34xcsrgeamNnz(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_nnzTotalDevHostPtr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsrgeamNnz", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_xcsrgeamNnz(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrRowPtrC, __pyx_v_nnzTotalDevHostPtr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrgeamNnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":442
 *     check_status(status)
 * 
 * cpdef scsrgeam(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t alpha, size_t descrA,
 *         int nnzA, size_t csrValA, size_t csrRowPtrA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_37scsrgeam(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrgeam(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_alpha, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_beta, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrgeam", 0);

  /* "cupy/cuda/cusparse.pyx":449
 *         size_t csrColIndB, size_t descrC, size_t csrValC,
 *         size_t csrRowPtrC, size_t csrColIndC):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseScsrgeam(
 *         <Handle>handle, m, n, <const float *>alpha,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":450
 *         size_t csrRowPtrC, size_t csrColIndC):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseScsrgeam(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <const float *>alpha,
 *         <const MatDescr>descrA, nnzA, <const float *>csrValA,
 */
  __pyx_v_status = cusparseScsrgeam(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((float const *)__pyx_v_alpha), ((cusparseMatDescr_t const )__pyx_v_descrA), __pyx_v_nnzA, ((float const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((float const *)__pyx_v_beta), ((cusparseMatDescr_t const )__pyx_v_descrB), __pyx_v_nnzB, ((float const *)__pyx_v_csrValB), ((int const *)__pyx_v_csrRowPtrB), ((int const *)__pyx_v_csrColIndB), ((cusparseMatDescr_t const )__pyx_v_descrC), ((float *)__pyx_v_csrValC), ((int *)__pyx_v_csrRowPtrC), ((int *)__pyx_v_csrColIndC));

  /* "cupy/cuda/cusparse.pyx":458
 *         <const MatDescr>descrC, <float *>csrValC, <int *>csrRowPtrC,
 *         <int *>csrColIndC)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":442
 *     check_status(status)
 * 
 * cpdef scsrgeam(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t alpha, size_t descrA,
 *         int nnzA, size_t csrValA, size_t csrRowPtrA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrgeam", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_37scsrgeam(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_36scsrgeam[] = "scsrgeam(size_t handle, int m, int n, size_t alpha, size_t descrA, int nnzA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t beta, size_t descrB, int nnzB, size_t csrValB, size_t csrRowPtrB, size_t csrColIndB, size_t descrC, size_t csrValC, size_t csrRowPtrC, size_t csrColIndC)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_37scsrgeam(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_alpha;
  size_t __pyx_v_descrA;
  int __pyx_v_nnzA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_beta;
  size_t __pyx_v_descrB;
  int __pyx_v_nnzB;
  size_t __pyx_v_csrValB;
  size_t __pyx_v_csrRowPtrB;
  size_t __pyx_v_csrColIndB;
  size_t __pyx_v_descrC;
  size_t __pyx_v_csrValC;
  size_t __pyx_v_csrRowPtrC;
  size_t __pyx_v_csrColIndC;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsrgeam (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_alpha,&__pyx_n_s_descrA,&__pyx_n_s_nnzA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_beta,&__pyx_n_s_descrB,&__pyx_n_s_nnzB,&__pyx_n_s_csrValB,&__pyx_n_s_csrRowPtrB,&__pyx_n_s_csrColIndB,&__pyx_n_s_descrC,&__pyx_n_s_csrValC,&__pyx_n_s_csrRowPtrC,&__pyx_n_s_csrColIndC,0};
    PyObject* values[19] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 1); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 2); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 3); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 4); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 5); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 6); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 7); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 8); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 9); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 10); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 11); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 12); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 13); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 14); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 15); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 16); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 17); __PYX_ERR(0, 442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, 18); __PYX_ERR(0, 442, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsrgeam") < 0)) __PYX_ERR(0, 442, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 19) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 443, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 443, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 443, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 443, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 443, __pyx_L3_error)
    __pyx_v_nnzA = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_nnzA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 444, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 444, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 444, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L3_error)
    __pyx_v_descrB = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_descrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L3_error)
    __pyx_v_nnzB = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_nnzB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 446, __pyx_L3_error)
    __pyx_v_csrValB = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_csrValB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 446, __pyx_L3_error)
    __pyx_v_csrRowPtrB = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_csrRowPtrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 446, __pyx_L3_error)
    __pyx_v_csrColIndB = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_csrColIndB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 447, __pyx_L3_error)
    __pyx_v_descrC = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_descrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 447, __pyx_L3_error)
    __pyx_v_csrValC = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_csrValC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 447, __pyx_L3_error)
    __pyx_v_csrRowPtrC = __Pyx_PyInt_As_size_t(values[17]); if (unlikely((__pyx_v_csrRowPtrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 448, __pyx_L3_error)
    __pyx_v_csrColIndC = __Pyx_PyInt_As_size_t(values[18]); if (unlikely((__pyx_v_csrColIndC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 448, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsrgeam", 1, 19, 19, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 442, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrgeam", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_36scsrgeam(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_beta, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrValB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrValC, __pyx_v_csrRowPtrC, __pyx_v_csrColIndC);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_36scsrgeam(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_alpha, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_beta, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrgeam", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsrgeam(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_beta, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrValB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrValC, __pyx_v_csrRowPtrC, __pyx_v_csrColIndC, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrgeam", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":461
 * 
 * 
 * cpdef dcsrgeam(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t alpha, size_t descrA,
 *         int nnzA, size_t csrValA, size_t csrRowPtrA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_39dcsrgeam(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrgeam(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_alpha, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_beta, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrgeam", 0);

  /* "cupy/cuda/cusparse.pyx":468
 *         size_t csrColIndB, size_t descrC, size_t csrValC,
 *         size_t csrRowPtrC, size_t csrColIndC):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseDcsrgeam(
 *         <Handle>handle, m, n, <const double *>alpha,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":469
 *         size_t csrRowPtrC, size_t csrColIndC):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseDcsrgeam(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <const double *>alpha,
 *         <const MatDescr>descrA, nnzA, <const double *>csrValA,
 */
  __pyx_v_status = cusparseDcsrgeam(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((double const *)__pyx_v_alpha), ((cusparseMatDescr_t const )__pyx_v_descrA), __pyx_v_nnzA, ((double const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((double const *)__pyx_v_beta), ((cusparseMatDescr_t const )__pyx_v_descrB), __pyx_v_nnzB, ((double const *)__pyx_v_csrValB), ((int const *)__pyx_v_csrRowPtrB), ((int const *)__pyx_v_csrColIndB), ((cusparseMatDescr_t const )__pyx_v_descrC), ((double *)__pyx_v_csrValC), ((int *)__pyx_v_csrRowPtrC), ((int *)__pyx_v_csrColIndC));

  /* "cupy/cuda/cusparse.pyx":477
 *         <const MatDescr>descrC, <double *>csrValC, <int *>csrRowPtrC,
 *         <int *>csrColIndC)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":461
 * 
 * 
 * cpdef dcsrgeam(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t alpha, size_t descrA,
 *         int nnzA, size_t csrValA, size_t csrRowPtrA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrgeam", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_39dcsrgeam(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_38dcsrgeam[] = "dcsrgeam(size_t handle, int m, int n, size_t alpha, size_t descrA, int nnzA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t beta, size_t descrB, int nnzB, size_t csrValB, size_t csrRowPtrB, size_t csrColIndB, size_t descrC, size_t csrValC, size_t csrRowPtrC, size_t csrColIndC)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_39dcsrgeam(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_alpha;
  size_t __pyx_v_descrA;
  int __pyx_v_nnzA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_beta;
  size_t __pyx_v_descrB;
  int __pyx_v_nnzB;
  size_t __pyx_v_csrValB;
  size_t __pyx_v_csrRowPtrB;
  size_t __pyx_v_csrColIndB;
  size_t __pyx_v_descrC;
  size_t __pyx_v_csrValC;
  size_t __pyx_v_csrRowPtrC;
  size_t __pyx_v_csrColIndC;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsrgeam (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_alpha,&__pyx_n_s_descrA,&__pyx_n_s_nnzA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_beta,&__pyx_n_s_descrB,&__pyx_n_s_nnzB,&__pyx_n_s_csrValB,&__pyx_n_s_csrRowPtrB,&__pyx_n_s_csrColIndB,&__pyx_n_s_descrC,&__pyx_n_s_csrValC,&__pyx_n_s_csrRowPtrC,&__pyx_n_s_csrColIndC,0};
    PyObject* values[19] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 1); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 2); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 3); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 4); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 5); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 6); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 7); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 8); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 9); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 10); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 11); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 12); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 13); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 14); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 15); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 16); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 17); __PYX_ERR(0, 461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, 18); __PYX_ERR(0, 461, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsrgeam") < 0)) __PYX_ERR(0, 461, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 19) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L3_error)
    __pyx_v_nnzA = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_nnzA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_descrB = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_descrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_nnzB = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_nnzB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L3_error)
    __pyx_v_csrValB = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_csrValB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L3_error)
    __pyx_v_csrRowPtrB = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_csrRowPtrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L3_error)
    __pyx_v_csrColIndB = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_csrColIndB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 466, __pyx_L3_error)
    __pyx_v_descrC = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_descrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 466, __pyx_L3_error)
    __pyx_v_csrValC = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_csrValC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 466, __pyx_L3_error)
    __pyx_v_csrRowPtrC = __Pyx_PyInt_As_size_t(values[17]); if (unlikely((__pyx_v_csrRowPtrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 467, __pyx_L3_error)
    __pyx_v_csrColIndC = __Pyx_PyInt_As_size_t(values[18]); if (unlikely((__pyx_v_csrColIndC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 467, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsrgeam", 1, 19, 19, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 461, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrgeam", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_38dcsrgeam(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_beta, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrValB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrValC, __pyx_v_csrRowPtrC, __pyx_v_csrColIndC);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_38dcsrgeam(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_alpha, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_beta, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrgeam", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsrgeam(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_alpha, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_beta, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrValB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrValC, __pyx_v_csrRowPtrC, __pyx_v_csrColIndC, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrgeam", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":480
 * 
 * 
 * cpdef xcsrgemmNnz(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k,
 *         size_t descrA, int nnzA, size_t csrRowPtrA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_41xcsrgemmNnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcsrgemmNnz(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_nnzTotalDevHostPtr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsrgemmNnz", 0);

  /* "cupy/cuda/cusparse.pyx":486
 *         size_t csrRowPtrB, size_t csrColIndB,
 *         size_t descrC, size_t csrRowPtrC, size_t nnzTotalDevHostPtr):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcsrgemmNnz(
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":487
 *         size_t descrC, size_t csrRowPtrC, size_t nnzTotalDevHostPtr):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcsrgemmNnz(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k,
 *         <const MatDescr>descrA, nnzA, <const int *>csrRowPtrA,
 */
  __pyx_v_status = cusparseXcsrgemmNnz(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), ((cusparseOperation_t)__pyx_v_transB), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cusparseMatDescr_t const )__pyx_v_descrA), __pyx_v_nnzA, ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((cusparseMatDescr_t const )__pyx_v_descrB), __pyx_v_nnzB, ((int const *)__pyx_v_csrRowPtrB), ((int const *)__pyx_v_csrColIndB), ((cusparseMatDescr_t const )__pyx_v_descrC), ((int *)__pyx_v_csrRowPtrC), ((int *)__pyx_v_nnzTotalDevHostPtr));

  /* "cupy/cuda/cusparse.pyx":493
 *         <const int *>csrRowPtrB, <const int *>csrColIndB,
 *         <const MatDescr>descrC, <int *>csrRowPtrC, <int *>nnzTotalDevHostPtr)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":480
 * 
 * 
 * cpdef xcsrgemmNnz(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k,
 *         size_t descrA, int nnzA, size_t csrRowPtrA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrgemmNnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_41xcsrgemmNnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_40xcsrgemmNnz[] = "xcsrgemmNnz(size_t handle, int transA, int transB, int m, int n, int k, size_t descrA, int nnzA, size_t csrRowPtrA, size_t csrColIndA, size_t descrB, int nnzB, size_t csrRowPtrB, size_t csrColIndB, size_t descrC, size_t csrRowPtrC, size_t nnzTotalDevHostPtr)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_41xcsrgemmNnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_transB;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_descrA;
  int __pyx_v_nnzA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_descrB;
  int __pyx_v_nnzB;
  size_t __pyx_v_csrRowPtrB;
  size_t __pyx_v_csrColIndB;
  size_t __pyx_v_descrC;
  size_t __pyx_v_csrRowPtrC;
  size_t __pyx_v_nnzTotalDevHostPtr;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcsrgemmNnz (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_transB,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_descrA,&__pyx_n_s_nnzA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_descrB,&__pyx_n_s_nnzB,&__pyx_n_s_csrRowPtrB,&__pyx_n_s_csrColIndB,&__pyx_n_s_descrC,&__pyx_n_s_csrRowPtrC,&__pyx_n_s_nnzTotalDevHostPtr,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 1); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 2); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 3); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 4); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 5); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 6); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 7); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 8); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 9); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 10); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 11); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 12); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 13); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 14); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 15); __PYX_ERR(0, 480, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzTotalDevHostPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, 16); __PYX_ERR(0, 480, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcsrgemmNnz") < 0)) __PYX_ERR(0, 480, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L3_error)
    __pyx_v_transB = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_transB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L3_error)
    __pyx_v_nnzA = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_nnzA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 483, __pyx_L3_error)
    __pyx_v_descrB = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_descrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 483, __pyx_L3_error)
    __pyx_v_nnzB = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_nnzB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 483, __pyx_L3_error)
    __pyx_v_csrRowPtrB = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_csrRowPtrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L3_error)
    __pyx_v_csrColIndB = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_csrColIndB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L3_error)
    __pyx_v_descrC = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_descrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L3_error)
    __pyx_v_csrRowPtrC = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_csrRowPtrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L3_error)
    __pyx_v_nnzTotalDevHostPtr = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_nnzTotalDevHostPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcsrgemmNnz", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 480, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrgemmNnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_40xcsrgemmNnz(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrRowPtrC, __pyx_v_nnzTotalDevHostPtr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_40xcsrgemmNnz(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_nnzTotalDevHostPtr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsrgemmNnz", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_xcsrgemmNnz(__pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrRowPtrC, __pyx_v_nnzTotalDevHostPtr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrgemmNnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":496
 * 
 * 
 * cpdef scsrgemm(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k,
 *         size_t descrA, const int nnzA, size_t csrValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_43scsrgemm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsrgemm(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int const __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int const __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrgemm", 0);

  /* "cupy/cuda/cusparse.pyx":503
 *         size_t csrColIndB, size_t descrC, size_t csrValC,
 *         size_t csrRowPtrC, size_t csrColIndC):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseScsrgemm(
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":504
 *         size_t csrRowPtrC, size_t csrColIndC):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseScsrgemm(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k,
 *         <const MatDescr>descrA, nnzA, <const float *>csrValA,
 */
  __pyx_v_status = cusparseScsrgemm(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), ((cusparseOperation_t)__pyx_v_transB), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cusparseMatDescr_t const )__pyx_v_descrA), __pyx_v_nnzA, ((float const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((cusparseMatDescr_t const )__pyx_v_descrB), __pyx_v_nnzB, ((float const *)__pyx_v_csrValB), ((int const *)__pyx_v_csrRowPtrB), ((int const *)__pyx_v_csrColIndB), ((cusparseMatDescr_t const )__pyx_v_descrC), ((float *)__pyx_v_csrValC), ((int const *)__pyx_v_csrRowPtrC), ((int *)__pyx_v_csrColIndC));

  /* "cupy/cuda/cusparse.pyx":512
 *         <const MatDescr>descrC, <float *>csrValC,
 *         <const int *>csrRowPtrC, <int *>csrColIndC)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":496
 * 
 * 
 * cpdef scsrgemm(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k,
 *         size_t descrA, const int nnzA, size_t csrValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrgemm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_43scsrgemm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_42scsrgemm[] = "scsrgemm(size_t handle, int transA, int transB, int m, int n, int k, size_t descrA, int nnzA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t descrB, int nnzB, size_t csrValB, size_t csrRowPtrB, size_t csrColIndB, size_t descrC, size_t csrValC, size_t csrRowPtrC, size_t csrColIndC)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_43scsrgemm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_transB;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_descrA;
  int __pyx_v_nnzA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_descrB;
  int __pyx_v_nnzB;
  size_t __pyx_v_csrValB;
  size_t __pyx_v_csrRowPtrB;
  size_t __pyx_v_csrColIndB;
  size_t __pyx_v_descrC;
  size_t __pyx_v_csrValC;
  size_t __pyx_v_csrRowPtrC;
  size_t __pyx_v_csrColIndC;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsrgemm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_transB,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_descrA,&__pyx_n_s_nnzA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_descrB,&__pyx_n_s_nnzB,&__pyx_n_s_csrValB,&__pyx_n_s_csrRowPtrB,&__pyx_n_s_csrColIndB,&__pyx_n_s_descrC,&__pyx_n_s_csrValC,&__pyx_n_s_csrRowPtrC,&__pyx_n_s_csrColIndC,0};
    PyObject* values[20] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 1); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 2); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 3); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 4); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 5); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 6); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 7); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 8); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 9); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 10); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 11); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 12); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 13); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 14); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 15); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 16); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 17); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 18); __PYX_ERR(0, 496, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, 19); __PYX_ERR(0, 496, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsrgemm") < 0)) __PYX_ERR(0, 496, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 20) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
    __pyx_v_transB = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_transB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 498, __pyx_L3_error)
    __pyx_v_nnzA = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_nnzA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 498, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 498, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L3_error)
    __pyx_v_descrB = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_descrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L3_error)
    __pyx_v_nnzB = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_nnzB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_csrValB = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_csrValB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_csrRowPtrB = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_csrRowPtrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L3_error)
    __pyx_v_csrColIndB = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_csrColIndB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L3_error)
    __pyx_v_descrC = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_descrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L3_error)
    __pyx_v_csrValC = __Pyx_PyInt_As_size_t(values[17]); if (unlikely((__pyx_v_csrValC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L3_error)
    __pyx_v_csrRowPtrC = __Pyx_PyInt_As_size_t(values[18]); if (unlikely((__pyx_v_csrRowPtrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 502, __pyx_L3_error)
    __pyx_v_csrColIndC = __Pyx_PyInt_As_size_t(values[19]); if (unlikely((__pyx_v_csrColIndC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 502, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsrgemm", 1, 20, 20, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 496, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrgemm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_42scsrgemm(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrValB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrValC, __pyx_v_csrRowPtrC, __pyx_v_csrColIndC);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_42scsrgemm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrgemm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsrgemm(__pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrValB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrValC, __pyx_v_csrRowPtrC, __pyx_v_csrColIndC, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsrgemm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":515
 * 
 * 
 * cpdef dcsrgemm(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k,
 *         size_t descrA, const int nnzA, size_t csrValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_45dcsrgemm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsrgemm(size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int const __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int const __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrgemm", 0);

  /* "cupy/cuda/cusparse.pyx":522
 *         size_t csrColIndB, size_t descrC, size_t csrValC,
 *         size_t csrRowPtrC, size_t csrColIndC):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseDcsrgemm(
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":523
 *         size_t csrRowPtrC, size_t csrColIndC):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseDcsrgemm(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Operation>transA, <Operation>transB, m, n, k,
 *         <const MatDescr>descrA, nnzA, <const double *>csrValA,
 */
  __pyx_v_status = cusparseDcsrgemm(((cusparseHandle_t)__pyx_v_handle), ((cusparseOperation_t)__pyx_v_transA), ((cusparseOperation_t)__pyx_v_transB), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cusparseMatDescr_t const )__pyx_v_descrA), __pyx_v_nnzA, ((double const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((cusparseMatDescr_t const )__pyx_v_descrB), __pyx_v_nnzB, ((double const *)__pyx_v_csrValB), ((int const *)__pyx_v_csrRowPtrB), ((int const *)__pyx_v_csrColIndB), ((cusparseMatDescr_t const )__pyx_v_descrC), ((double *)__pyx_v_csrValC), ((int const *)__pyx_v_csrRowPtrC), ((int *)__pyx_v_csrColIndC));

  /* "cupy/cuda/cusparse.pyx":531
 *         <const MatDescr>descrC, <double *>csrValC,
 *         <const int *>csrRowPtrC, <int *>csrColIndC)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":515
 * 
 * 
 * cpdef dcsrgemm(             # <<<<<<<<<<<<<<
 *         size_t handle, int transA, int transB, int m, int n, int k,
 *         size_t descrA, const int nnzA, size_t csrValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrgemm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_45dcsrgemm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_44dcsrgemm[] = "dcsrgemm(size_t handle, int transA, int transB, int m, int n, int k, size_t descrA, int nnzA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t descrB, int nnzB, size_t csrValB, size_t csrRowPtrB, size_t csrColIndB, size_t descrC, size_t csrValC, size_t csrRowPtrC, size_t csrColIndC)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_45dcsrgemm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_transA;
  int __pyx_v_transB;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_descrA;
  int __pyx_v_nnzA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_descrB;
  int __pyx_v_nnzB;
  size_t __pyx_v_csrValB;
  size_t __pyx_v_csrRowPtrB;
  size_t __pyx_v_csrColIndB;
  size_t __pyx_v_descrC;
  size_t __pyx_v_csrValC;
  size_t __pyx_v_csrRowPtrC;
  size_t __pyx_v_csrColIndC;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsrgemm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_transA,&__pyx_n_s_transB,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_descrA,&__pyx_n_s_nnzA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_descrB,&__pyx_n_s_nnzB,&__pyx_n_s_csrValB,&__pyx_n_s_csrRowPtrB,&__pyx_n_s_csrColIndB,&__pyx_n_s_descrC,&__pyx_n_s_csrValC,&__pyx_n_s_csrRowPtrC,&__pyx_n_s_csrColIndC,0};
    PyObject* values[20] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 1); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 2); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 3); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 4); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 5); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 6); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 7); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 8); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 9); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 10); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 11); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 12); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 13); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 14); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndB)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 15); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 16); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 17); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 18); __PYX_ERR(0, 515, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndC)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, 19); __PYX_ERR(0, 515, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsrgemm") < 0)) __PYX_ERR(0, 515, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 20) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    __pyx_v_transA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_transA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    __pyx_v_transB = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_transB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
    __pyx_v_nnzA = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_nnzA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_descrB = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_descrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_nnzB = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_nnzB == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 519, __pyx_L3_error)
    __pyx_v_csrValB = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_csrValB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 519, __pyx_L3_error)
    __pyx_v_csrRowPtrB = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_csrRowPtrB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 519, __pyx_L3_error)
    __pyx_v_csrColIndB = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_csrColIndB == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 520, __pyx_L3_error)
    __pyx_v_descrC = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_descrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 520, __pyx_L3_error)
    __pyx_v_csrValC = __Pyx_PyInt_As_size_t(values[17]); if (unlikely((__pyx_v_csrValC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 520, __pyx_L3_error)
    __pyx_v_csrRowPtrC = __Pyx_PyInt_As_size_t(values[18]); if (unlikely((__pyx_v_csrRowPtrC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 521, __pyx_L3_error)
    __pyx_v_csrColIndC = __Pyx_PyInt_As_size_t(values[19]); if (unlikely((__pyx_v_csrColIndC == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 521, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsrgemm", 1, 20, 20, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 515, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrgemm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_44dcsrgemm(__pyx_self, __pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrValB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrValC, __pyx_v_csrRowPtrC, __pyx_v_csrColIndC);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_44dcsrgemm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_transA, int __pyx_v_transB, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_descrA, int __pyx_v_nnzA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_descrB, int __pyx_v_nnzB, size_t __pyx_v_csrValB, size_t __pyx_v_csrRowPtrB, size_t __pyx_v_csrColIndB, size_t __pyx_v_descrC, size_t __pyx_v_csrValC, size_t __pyx_v_csrRowPtrC, size_t __pyx_v_csrColIndC) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrgemm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsrgemm(__pyx_v_handle, __pyx_v_transA, __pyx_v_transB, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_descrA, __pyx_v_nnzA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_descrB, __pyx_v_nnzB, __pyx_v_csrValB, __pyx_v_csrRowPtrB, __pyx_v_csrColIndB, __pyx_v_descrC, __pyx_v_csrValC, __pyx_v_csrRowPtrC, __pyx_v_csrColIndC, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsrgemm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":537
 * # cuSPARSE Format Convrsion
 * 
 * cpdef xcoo2csr(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t cooRowInd, int nnz, int m, size_t csrRowPtr,
 *         int idxBase):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_47xcoo2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcoo2csr(size_t __pyx_v_handle, size_t __pyx_v_cooRowInd, int __pyx_v_nnz, int __pyx_v_m, size_t __pyx_v_csrRowPtr, int __pyx_v_idxBase, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcoo2csr", 0);

  /* "cupy/cuda/cusparse.pyx":540
 *         size_t handle, size_t cooRowInd, int nnz, int m, size_t csrRowPtr,
 *         int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcoo2csr(
 *         <Handle>handle, <const int *>cooRowInd, nnz, m, <int *>csrRowPtr,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":541
 *         int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcoo2csr(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <const int *>cooRowInd, nnz, m, <int *>csrRowPtr,
 *         <IndexBase>idxBase)
 */
  __pyx_v_status = cusparseXcoo2csr(((cusparseHandle_t)__pyx_v_handle), ((int const *)__pyx_v_cooRowInd), __pyx_v_nnz, __pyx_v_m, ((int *)__pyx_v_csrRowPtr), ((cusparseIndexBase_t)__pyx_v_idxBase));

  /* "cupy/cuda/cusparse.pyx":544
 *         <Handle>handle, <const int *>cooRowInd, nnz, m, <int *>csrRowPtr,
 *         <IndexBase>idxBase)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":537
 * # cuSPARSE Format Convrsion
 * 
 * cpdef xcoo2csr(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t cooRowInd, int nnz, int m, size_t csrRowPtr,
 *         int idxBase):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcoo2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_47xcoo2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_46xcoo2csr[] = "xcoo2csr(size_t handle, size_t cooRowInd, int nnz, int m, size_t csrRowPtr, int idxBase)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_47xcoo2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_cooRowInd;
  int __pyx_v_nnz;
  int __pyx_v_m;
  size_t __pyx_v_csrRowPtr;
  int __pyx_v_idxBase;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcoo2csr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_cooRowInd,&__pyx_n_s_nnz,&__pyx_n_s_m,&__pyx_n_s_csrRowPtr,&__pyx_n_s_idxBase,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cooRowInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoo2csr", 1, 6, 6, 1); __PYX_ERR(0, 537, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoo2csr", 1, 6, 6, 2); __PYX_ERR(0, 537, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoo2csr", 1, 6, 6, 3); __PYX_ERR(0, 537, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoo2csr", 1, 6, 6, 4); __PYX_ERR(0, 537, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_idxBase)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoo2csr", 1, 6, 6, 5); __PYX_ERR(0, 537, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcoo2csr") < 0)) __PYX_ERR(0, 537, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L3_error)
    __pyx_v_cooRowInd = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_cooRowInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L3_error)
    __pyx_v_idxBase = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_idxBase == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 539, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcoo2csr", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 537, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcoo2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_46xcoo2csr(__pyx_self, __pyx_v_handle, __pyx_v_cooRowInd, __pyx_v_nnz, __pyx_v_m, __pyx_v_csrRowPtr, __pyx_v_idxBase);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_46xcoo2csr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_cooRowInd, int __pyx_v_nnz, int __pyx_v_m, size_t __pyx_v_csrRowPtr, int __pyx_v_idxBase) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcoo2csr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_xcoo2csr(__pyx_v_handle, __pyx_v_cooRowInd, __pyx_v_nnz, __pyx_v_m, __pyx_v_csrRowPtr, __pyx_v_idxBase, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcoo2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":547
 * 
 * 
 * cpdef scsc2dense(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t cscSortedValA, size_t cscSortedRowIndA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_49scsc2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsc2dense(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_cscSortedValA, size_t __pyx_v_cscSortedRowIndA, size_t __pyx_v_cscSortedColPtrA, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsc2dense", 0);

  /* "cupy/cuda/cusparse.pyx":551
 *         size_t cscSortedValA, size_t cscSortedRowIndA,
 *         size_t cscSortedColPtrA, size_t A, int lda):
 *     status = cusparseScsc2dense(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <MatDescr>descrA,
 *         <const float *>cscSortedValA, <const int *>cscSortedRowIndA,
 */
  __pyx_v_status = cusparseScsc2dense(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t)__pyx_v_descrA), ((float const *)__pyx_v_cscSortedValA), ((int const *)__pyx_v_cscSortedRowIndA), ((int const *)__pyx_v_cscSortedColPtrA), ((float *)__pyx_v_A), __pyx_v_lda);

  /* "cupy/cuda/cusparse.pyx":555
 *         <const float *>cscSortedValA, <const int *>cscSortedRowIndA,
 *         <const int *>cscSortedColPtrA, <float *>A, lda)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":547
 * 
 * 
 * cpdef scsc2dense(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t cscSortedValA, size_t cscSortedRowIndA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsc2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_49scsc2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_48scsc2dense[] = "scsc2dense(size_t handle, int m, int n, size_t descrA, size_t cscSortedValA, size_t cscSortedRowIndA, size_t cscSortedColPtrA, size_t A, int lda)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_49scsc2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_cscSortedValA;
  size_t __pyx_v_cscSortedRowIndA;
  size_t __pyx_v_cscSortedColPtrA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsc2dense (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_cscSortedValA,&__pyx_n_s_cscSortedRowIndA,&__pyx_n_s_cscSortedColPtrA,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, 1); __PYX_ERR(0, 547, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, 2); __PYX_ERR(0, 547, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, 3); __PYX_ERR(0, 547, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscSortedValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, 4); __PYX_ERR(0, 547, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscSortedRowIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, 5); __PYX_ERR(0, 547, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscSortedColPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, 6); __PYX_ERR(0, 547, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, 7); __PYX_ERR(0, 547, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, 8); __PYX_ERR(0, 547, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsc2dense") < 0)) __PYX_ERR(0, 547, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L3_error)
    __pyx_v_cscSortedValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_cscSortedValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 549, __pyx_L3_error)
    __pyx_v_cscSortedRowIndA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_cscSortedRowIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 549, __pyx_L3_error)
    __pyx_v_cscSortedColPtrA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_cscSortedColPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsc2dense", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 547, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsc2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_48scsc2dense(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_cscSortedValA, __pyx_v_cscSortedRowIndA, __pyx_v_cscSortedColPtrA, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_48scsc2dense(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_cscSortedValA, size_t __pyx_v_cscSortedRowIndA, size_t __pyx_v_cscSortedColPtrA, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsc2dense", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsc2dense(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_cscSortedValA, __pyx_v_cscSortedRowIndA, __pyx_v_cscSortedColPtrA, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsc2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":558
 * 
 * 
 * cpdef dcsc2dense(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t cscSortedValA, size_t cscSortedRowIndA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_51dcsc2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsc2dense(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_cscSortedValA, size_t __pyx_v_cscSortedRowIndA, size_t __pyx_v_cscSortedColPtrA, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsc2dense", 0);

  /* "cupy/cuda/cusparse.pyx":562
 *         size_t cscSortedValA, size_t cscSortedRowIndA,
 *         size_t cscSortedColPtrA, size_t A, int lda):
 *     status = cusparseDcsc2dense(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <MatDescr>descrA,
 *         <const double *>cscSortedValA, <const int *>cscSortedRowIndA,
 */
  __pyx_v_status = cusparseDcsc2dense(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t)__pyx_v_descrA), ((double const *)__pyx_v_cscSortedValA), ((int const *)__pyx_v_cscSortedRowIndA), ((int const *)__pyx_v_cscSortedColPtrA), ((double *)__pyx_v_A), __pyx_v_lda);

  /* "cupy/cuda/cusparse.pyx":566
 *         <const double *>cscSortedValA, <const int *>cscSortedRowIndA,
 *         <const int *>cscSortedColPtrA, <double *>A, lda)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":558
 * 
 * 
 * cpdef dcsc2dense(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t cscSortedValA, size_t cscSortedRowIndA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsc2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_51dcsc2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_50dcsc2dense[] = "dcsc2dense(size_t handle, int m, int n, size_t descrA, size_t cscSortedValA, size_t cscSortedRowIndA, size_t cscSortedColPtrA, size_t A, int lda)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_51dcsc2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_cscSortedValA;
  size_t __pyx_v_cscSortedRowIndA;
  size_t __pyx_v_cscSortedColPtrA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsc2dense (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_cscSortedValA,&__pyx_n_s_cscSortedRowIndA,&__pyx_n_s_cscSortedColPtrA,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, 1); __PYX_ERR(0, 558, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, 2); __PYX_ERR(0, 558, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, 3); __PYX_ERR(0, 558, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscSortedValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, 4); __PYX_ERR(0, 558, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscSortedRowIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, 5); __PYX_ERR(0, 558, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscSortedColPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, 6); __PYX_ERR(0, 558, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, 7); __PYX_ERR(0, 558, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, 8); __PYX_ERR(0, 558, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsc2dense") < 0)) __PYX_ERR(0, 558, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 559, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 559, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 559, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 559, __pyx_L3_error)
    __pyx_v_cscSortedValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_cscSortedValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 560, __pyx_L3_error)
    __pyx_v_cscSortedRowIndA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_cscSortedRowIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 560, __pyx_L3_error)
    __pyx_v_cscSortedColPtrA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_cscSortedColPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 561, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 561, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 561, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsc2dense", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 558, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsc2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_50dcsc2dense(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_cscSortedValA, __pyx_v_cscSortedRowIndA, __pyx_v_cscSortedColPtrA, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_50dcsc2dense(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_cscSortedValA, size_t __pyx_v_cscSortedRowIndA, size_t __pyx_v_cscSortedColPtrA, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsc2dense", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsc2dense(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_cscSortedValA, __pyx_v_cscSortedRowIndA, __pyx_v_cscSortedColPtrA, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsc2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":569
 * 
 * 
 * cpdef xcsr2coo(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t csrRowPtr, int nnz, int m, size_t cooRowInd,
 *         int idxBase):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_53xcsr2coo(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcsr2coo(size_t __pyx_v_handle, size_t __pyx_v_csrRowPtr, int __pyx_v_nnz, int __pyx_v_m, size_t __pyx_v_cooRowInd, int __pyx_v_idxBase, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsr2coo", 0);

  /* "cupy/cuda/cusparse.pyx":572
 *         size_t handle, size_t csrRowPtr, int nnz, int m, size_t cooRowInd,
 *         int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcsr2coo(
 *         <Handle>handle, <const int *>csrRowPtr, nnz, m, <int *>cooRowInd,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":573
 *         int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcsr2coo(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <const int *>csrRowPtr, nnz, m, <int *>cooRowInd,
 *         <IndexBase>idxBase)
 */
  __pyx_v_status = cusparseXcsr2coo(((cusparseHandle_t)__pyx_v_handle), ((int const *)__pyx_v_csrRowPtr), __pyx_v_nnz, __pyx_v_m, ((int *)__pyx_v_cooRowInd), ((cusparseIndexBase_t)__pyx_v_idxBase));

  /* "cupy/cuda/cusparse.pyx":576
 *         <Handle>handle, <const int *>csrRowPtr, nnz, m, <int *>cooRowInd,
 *         <IndexBase>idxBase)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":569
 * 
 * 
 * cpdef xcsr2coo(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t csrRowPtr, int nnz, int m, size_t cooRowInd,
 *         int idxBase):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsr2coo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_53xcsr2coo(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_52xcsr2coo[] = "xcsr2coo(size_t handle, size_t csrRowPtr, int nnz, int m, size_t cooRowInd, int idxBase)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_53xcsr2coo(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_csrRowPtr;
  int __pyx_v_nnz;
  int __pyx_v_m;
  size_t __pyx_v_cooRowInd;
  int __pyx_v_idxBase;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcsr2coo (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_csrRowPtr,&__pyx_n_s_nnz,&__pyx_n_s_m,&__pyx_n_s_cooRowInd,&__pyx_n_s_idxBase,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsr2coo", 1, 6, 6, 1); __PYX_ERR(0, 569, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsr2coo", 1, 6, 6, 2); __PYX_ERR(0, 569, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsr2coo", 1, 6, 6, 3); __PYX_ERR(0, 569, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cooRowInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsr2coo", 1, 6, 6, 4); __PYX_ERR(0, 569, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_idxBase)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsr2coo", 1, 6, 6, 5); __PYX_ERR(0, 569, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcsr2coo") < 0)) __PYX_ERR(0, 569, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_cooRowInd = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_cooRowInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_idxBase = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_idxBase == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 571, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcsr2coo", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 569, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsr2coo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_52xcsr2coo(__pyx_self, __pyx_v_handle, __pyx_v_csrRowPtr, __pyx_v_nnz, __pyx_v_m, __pyx_v_cooRowInd, __pyx_v_idxBase);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_52xcsr2coo(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_csrRowPtr, int __pyx_v_nnz, int __pyx_v_m, size_t __pyx_v_cooRowInd, int __pyx_v_idxBase) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsr2coo", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_xcsr2coo(__pyx_v_handle, __pyx_v_csrRowPtr, __pyx_v_nnz, __pyx_v_m, __pyx_v_cooRowInd, __pyx_v_idxBase, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsr2coo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":579
 * 
 * 
 * cpdef scsr2csc(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t csrVal,
 *         size_t csrRowPtr, size_t csrColInd, size_t cscVal,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_55scsr2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsr2csc(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_cscVal, size_t __pyx_v_cscRowInd, size_t __pyx_v_cscColPtr, int __pyx_v_copyValues, int __pyx_v_idxBase, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsr2csc", 0);

  /* "cupy/cuda/cusparse.pyx":583
 *         size_t csrRowPtr, size_t csrColInd, size_t cscVal,
 *         size_t cscRowInd, size_t cscColPtr, int copyValues, int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseScsr2csc(
 *         <Handle>handle, m, n, nnz, <const float *>csrVal,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":584
 *         size_t cscRowInd, size_t cscColPtr, int copyValues, int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseScsr2csc(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, nnz, <const float *>csrVal,
 *         <const int *>csrRowPtr, <const int *>csrColInd, <float *>cscVal,
 */
  __pyx_v_status = cusparseScsr2csc(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((float const *)__pyx_v_csrVal), ((int const *)__pyx_v_csrRowPtr), ((int const *)__pyx_v_csrColInd), ((float *)__pyx_v_cscVal), ((int *)__pyx_v_cscRowInd), ((int *)__pyx_v_cscColPtr), ((cusparseAction_t)__pyx_v_copyValues), ((cusparseIndexBase_t)__pyx_v_idxBase));

  /* "cupy/cuda/cusparse.pyx":589
 *         <int *>cscRowInd, <int *>cscColPtr, <Action>copyValues,
 *         <IndexBase>idxBase)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":579
 * 
 * 
 * cpdef scsr2csc(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t csrVal,
 *         size_t csrRowPtr, size_t csrColInd, size_t cscVal,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_55scsr2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_54scsr2csc[] = "scsr2csc(size_t handle, int m, int n, int nnz, size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t cscVal, size_t cscRowInd, size_t cscColPtr, int copyValues, int idxBase)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_55scsr2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_csrVal;
  size_t __pyx_v_csrRowPtr;
  size_t __pyx_v_csrColInd;
  size_t __pyx_v_cscVal;
  size_t __pyx_v_cscRowInd;
  size_t __pyx_v_cscColPtr;
  int __pyx_v_copyValues;
  int __pyx_v_idxBase;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsr2csc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_csrVal,&__pyx_n_s_csrRowPtr,&__pyx_n_s_csrColInd,&__pyx_n_s_cscVal,&__pyx_n_s_cscRowInd,&__pyx_n_s_cscColPtr,&__pyx_n_s_copyValues,&__pyx_n_s_idxBase,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 1); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 2); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 3); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 4); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 5); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 6); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 7); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscRowInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 8); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscColPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 9); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_copyValues)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 10); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_idxBase)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, 11); __PYX_ERR(0, 579, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsr2csc") < 0)) __PYX_ERR(0, 579, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L3_error)
    __pyx_v_csrVal = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L3_error)
    __pyx_v_csrColInd = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L3_error)
    __pyx_v_cscVal = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_cscVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L3_error)
    __pyx_v_cscRowInd = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_cscRowInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
    __pyx_v_cscColPtr = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_cscColPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
    __pyx_v_copyValues = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_copyValues == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
    __pyx_v_idxBase = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_idxBase == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsr2csc", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 579, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_54scsr2csc(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_cscVal, __pyx_v_cscRowInd, __pyx_v_cscColPtr, __pyx_v_copyValues, __pyx_v_idxBase);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_54scsr2csc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_cscVal, size_t __pyx_v_cscRowInd, size_t __pyx_v_cscColPtr, int __pyx_v_copyValues, int __pyx_v_idxBase) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsr2csc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsr2csc(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_cscVal, __pyx_v_cscRowInd, __pyx_v_cscColPtr, __pyx_v_copyValues, __pyx_v_idxBase, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":592
 * 
 * 
 * cpdef dcsr2csc(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t csrVal,
 *         size_t csrRowPtr, size_t csrColInd, size_t cscVal,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_57dcsr2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsr2csc(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_cscVal, size_t __pyx_v_cscRowInd, size_t __pyx_v_cscColPtr, int __pyx_v_copyValues, int __pyx_v_idxBase, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsr2csc", 0);

  /* "cupy/cuda/cusparse.pyx":596
 *         size_t csrRowPtr, size_t csrColInd, size_t cscVal,
 *         size_t cscRowInd, size_t cscColPtr, int copyValues, int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseDcsr2csc(
 *         <Handle>handle, m, n, nnz, <const double *>csrVal,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":597
 *         size_t cscRowInd, size_t cscColPtr, int copyValues, int idxBase):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseDcsr2csc(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, nnz, <const double *>csrVal,
 *         <const int *>csrRowPtr, <const int *>csrColInd, <double *>cscVal,
 */
  __pyx_v_status = cusparseDcsr2csc(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((double const *)__pyx_v_csrVal), ((int const *)__pyx_v_csrRowPtr), ((int const *)__pyx_v_csrColInd), ((double *)__pyx_v_cscVal), ((int *)__pyx_v_cscRowInd), ((int *)__pyx_v_cscColPtr), ((cusparseAction_t)__pyx_v_copyValues), ((cusparseIndexBase_t)__pyx_v_idxBase));

  /* "cupy/cuda/cusparse.pyx":602
 *         <int *>cscRowInd, <int *>cscColPtr, <Action>copyValues,
 *         <IndexBase>idxBase)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":592
 * 
 * 
 * cpdef dcsr2csc(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t csrVal,
 *         size_t csrRowPtr, size_t csrColInd, size_t cscVal,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_57dcsr2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_56dcsr2csc[] = "dcsr2csc(size_t handle, int m, int n, int nnz, size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t cscVal, size_t cscRowInd, size_t cscColPtr, int copyValues, int idxBase)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_57dcsr2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_csrVal;
  size_t __pyx_v_csrRowPtr;
  size_t __pyx_v_csrColInd;
  size_t __pyx_v_cscVal;
  size_t __pyx_v_cscRowInd;
  size_t __pyx_v_cscColPtr;
  int __pyx_v_copyValues;
  int __pyx_v_idxBase;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsr2csc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_csrVal,&__pyx_n_s_csrRowPtr,&__pyx_n_s_csrColInd,&__pyx_n_s_cscVal,&__pyx_n_s_cscRowInd,&__pyx_n_s_cscColPtr,&__pyx_n_s_copyValues,&__pyx_n_s_idxBase,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 1); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 2); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 3); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 4); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 5); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 6); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 7); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscRowInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 8); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscColPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 9); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_copyValues)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 10); __PYX_ERR(0, 592, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_idxBase)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, 11); __PYX_ERR(0, 592, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsr2csc") < 0)) __PYX_ERR(0, 592, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L3_error)
    __pyx_v_csrVal = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L3_error)
    __pyx_v_csrColInd = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L3_error)
    __pyx_v_cscVal = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_cscVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L3_error)
    __pyx_v_cscRowInd = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_cscRowInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 595, __pyx_L3_error)
    __pyx_v_cscColPtr = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_cscColPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 595, __pyx_L3_error)
    __pyx_v_copyValues = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_copyValues == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 595, __pyx_L3_error)
    __pyx_v_idxBase = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_idxBase == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 595, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsr2csc", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 592, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_56dcsr2csc(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_cscVal, __pyx_v_cscRowInd, __pyx_v_cscColPtr, __pyx_v_copyValues, __pyx_v_idxBase);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_56dcsr2csc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_cscVal, size_t __pyx_v_cscRowInd, size_t __pyx_v_cscColPtr, int __pyx_v_copyValues, int __pyx_v_idxBase) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsr2csc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsr2csc(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_cscVal, __pyx_v_cscRowInd, __pyx_v_cscColPtr, __pyx_v_copyValues, __pyx_v_idxBase, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":605
 * 
 * 
 * cpdef scsr2dense(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t csrSortedValA, size_t csrSortedRowPtrA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_59scsr2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsr2dense(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsr2dense", 0);

  /* "cupy/cuda/cusparse.pyx":609
 *         size_t csrSortedValA, size_t csrSortedRowPtrA,
 *         size_t csrSortedColIndA, size_t A, int lda):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseScsr2dense(
 *         <Handle>handle, m, n, <MatDescr>descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":610
 *         size_t csrSortedColIndA, size_t A, int lda):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseScsr2dense(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <MatDescr>descrA,
 *         <const float *>csrSortedValA, <const int *>csrSortedRowPtrA,
 */
  __pyx_v_status = cusparseScsr2dense(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t)__pyx_v_descrA), ((float const *)__pyx_v_csrSortedValA), ((int const *)__pyx_v_csrSortedRowPtrA), ((int const *)__pyx_v_csrSortedColIndA), ((float *)__pyx_v_A), __pyx_v_lda);

  /* "cupy/cuda/cusparse.pyx":614
 *         <const float *>csrSortedValA, <const int *>csrSortedRowPtrA,
 *         <const int *>csrSortedColIndA, <float *>A, lda)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":605
 * 
 * 
 * cpdef scsr2dense(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t csrSortedValA, size_t csrSortedRowPtrA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_59scsr2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_58scsr2dense[] = "scsr2dense(size_t handle, int m, int n, size_t descrA, size_t csrSortedValA, size_t csrSortedRowPtrA, size_t csrSortedColIndA, size_t A, int lda)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_59scsr2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrSortedValA;
  size_t __pyx_v_csrSortedRowPtrA;
  size_t __pyx_v_csrSortedColIndA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsr2dense (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_csrSortedValA,&__pyx_n_s_csrSortedRowPtrA,&__pyx_n_s_csrSortedColIndA,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, 1); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, 2); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, 3); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, 4); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, 5); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, 6); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, 7); __PYX_ERR(0, 605, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, 8); __PYX_ERR(0, 605, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsr2dense") < 0)) __PYX_ERR(0, 605, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 606, __pyx_L3_error)
    __pyx_v_csrSortedValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrSortedValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 607, __pyx_L3_error)
    __pyx_v_csrSortedRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrSortedRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 607, __pyx_L3_error)
    __pyx_v_csrSortedColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrSortedColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 608, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 608, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 608, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsr2dense", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 605, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_58scsr2dense(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_58scsr2dense(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsr2dense", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsr2dense(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":617
 * 
 * 
 * cpdef dcsr2dense(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t csrSortedValA, size_t csrSortedRowPtrA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_61dcsr2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsr2dense(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsr2dense", 0);

  /* "cupy/cuda/cusparse.pyx":621
 *         size_t csrSortedValA, size_t csrSortedRowPtrA,
 *         size_t csrSortedColIndA, size_t A, int lda):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseDcsr2dense(
 *         <Handle>handle, m, n, <MatDescr>descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":622
 *         size_t csrSortedColIndA, size_t A, int lda):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseDcsr2dense(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <MatDescr>descrA,
 *         <const double *>csrSortedValA, <const int *>csrSortedRowPtrA,
 */
  __pyx_v_status = cusparseDcsr2dense(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t)__pyx_v_descrA), ((double const *)__pyx_v_csrSortedValA), ((int const *)__pyx_v_csrSortedRowPtrA), ((int const *)__pyx_v_csrSortedColIndA), ((double *)__pyx_v_A), __pyx_v_lda);

  /* "cupy/cuda/cusparse.pyx":626
 *         <const double *>csrSortedValA, <const int *>csrSortedRowPtrA,
 *         <const int *>csrSortedColIndA, <double *>A, lda)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":617
 * 
 * 
 * cpdef dcsr2dense(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t csrSortedValA, size_t csrSortedRowPtrA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_61dcsr2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_60dcsr2dense[] = "dcsr2dense(size_t handle, int m, int n, size_t descrA, size_t csrSortedValA, size_t csrSortedRowPtrA, size_t csrSortedColIndA, size_t A, int lda)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_61dcsr2dense(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrSortedValA;
  size_t __pyx_v_csrSortedRowPtrA;
  size_t __pyx_v_csrSortedColIndA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsr2dense (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_csrSortedValA,&__pyx_n_s_csrSortedRowPtrA,&__pyx_n_s_csrSortedColIndA,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, 1); __PYX_ERR(0, 617, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, 2); __PYX_ERR(0, 617, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, 3); __PYX_ERR(0, 617, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, 4); __PYX_ERR(0, 617, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, 5); __PYX_ERR(0, 617, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrSortedColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, 6); __PYX_ERR(0, 617, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, 7); __PYX_ERR(0, 617, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, 8); __PYX_ERR(0, 617, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsr2dense") < 0)) __PYX_ERR(0, 617, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_csrSortedValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrSortedValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 619, __pyx_L3_error)
    __pyx_v_csrSortedRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrSortedRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 619, __pyx_L3_error)
    __pyx_v_csrSortedColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrSortedColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 620, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 620, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 620, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsr2dense", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 617, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_60dcsr2dense(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_60dcsr2dense(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_csrSortedValA, size_t __pyx_v_csrSortedRowPtrA, size_t __pyx_v_csrSortedColIndA, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsr2dense", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsr2dense(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_csrSortedValA, __pyx_v_csrSortedRowPtrA, __pyx_v_csrSortedColIndA, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2dense", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":629
 * 
 * 
 * cpdef snnz_compress(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, size_t descr,
 *         size_t values, size_t rowPtr, size_t nnzPerRow,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_63snnz_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_snnz_compress(size_t __pyx_v_handle, int __pyx_v_m, size_t __pyx_v_descr, size_t __pyx_v_values, size_t __pyx_v_rowPtr, size_t __pyx_v_nnzPerRow, float __pyx_v_tol, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_nnz_total;
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("snnz_compress", 0);

  /* "cupy/cuda/cusparse.pyx":634
 *         float tol):
 *     cdef int nnz_total
 *     status = cusparseSnnz_compress(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, <const MatDescr>descr,
 *         <const float *>values, <const int *>rowPtr, <int *>nnzPerRow,
 */
  __pyx_v_status = cusparseSnnz_compress(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, ((cusparseMatDescr_t const )__pyx_v_descr), ((float const *)__pyx_v_values), ((int const *)__pyx_v_rowPtr), ((int *)__pyx_v_nnzPerRow), (&__pyx_v_nnz_total), __pyx_v_tol);

  /* "cupy/cuda/cusparse.pyx":638
 *         <const float *>values, <const int *>rowPtr, <int *>nnzPerRow,
 *         &nnz_total, tol)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return nnz_total
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":639
 *         &nnz_total, tol)
 *     check_status(status)
 *     return nnz_total             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nnz_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cusparse.pyx":629
 * 
 * 
 * cpdef snnz_compress(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, size_t descr,
 *         size_t values, size_t rowPtr, size_t nnzPerRow,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.snnz_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_63snnz_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_62snnz_compress[] = "snnz_compress(size_t handle, int m, size_t descr, size_t values, size_t rowPtr, size_t nnzPerRow, float tol)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_63snnz_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  size_t __pyx_v_descr;
  size_t __pyx_v_values;
  size_t __pyx_v_rowPtr;
  size_t __pyx_v_nnzPerRow;
  float __pyx_v_tol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("snnz_compress (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_descr,&__pyx_n_s_values,&__pyx_n_s_rowPtr,&__pyx_n_s_nnzPerRow,&__pyx_n_s_tol,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz_compress", 1, 7, 7, 1); __PYX_ERR(0, 629, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz_compress", 1, 7, 7, 2); __PYX_ERR(0, 629, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz_compress", 1, 7, 7, 3); __PYX_ERR(0, 629, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz_compress", 1, 7, 7, 4); __PYX_ERR(0, 629, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerRow)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz_compress", 1, 7, 7, 5); __PYX_ERR(0, 629, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz_compress", 1, 7, 7, 6); __PYX_ERR(0, 629, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "snnz_compress") < 0)) __PYX_ERR(0, 629, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_descr = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_descr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_values == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 631, __pyx_L3_error)
    __pyx_v_rowPtr = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_rowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 631, __pyx_L3_error)
    __pyx_v_nnzPerRow = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_nnzPerRow == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 631, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsFloat(values[6]); if (unlikely((__pyx_v_tol == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 632, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("snnz_compress", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 629, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.snnz_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_62snnz_compress(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_descr, __pyx_v_values, __pyx_v_rowPtr, __pyx_v_nnzPerRow, __pyx_v_tol);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_62snnz_compress(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, size_t __pyx_v_descr, size_t __pyx_v_values, size_t __pyx_v_rowPtr, size_t __pyx_v_nnzPerRow, float __pyx_v_tol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("snnz_compress", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_snnz_compress(__pyx_v_handle, __pyx_v_m, __pyx_v_descr, __pyx_v_values, __pyx_v_rowPtr, __pyx_v_nnzPerRow, __pyx_v_tol, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.snnz_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":642
 * 
 * 
 * cpdef dnnz_compress(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, size_t descr,
 *         size_t values, size_t rowPtr, size_t nnzPerRow,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_65dnnz_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dnnz_compress(size_t __pyx_v_handle, int __pyx_v_m, size_t __pyx_v_descr, size_t __pyx_v_values, size_t __pyx_v_rowPtr, size_t __pyx_v_nnzPerRow, double __pyx_v_tol, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_nnz_total;
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dnnz_compress", 0);

  /* "cupy/cuda/cusparse.pyx":647
 *         double tol):
 *     cdef int nnz_total
 *     status = cusparseDnnz_compress(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, <const MatDescr>descr,
 *         <const double *>values, <const int *>rowPtr, <int *>nnzPerRow,
 */
  __pyx_v_status = cusparseDnnz_compress(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, ((cusparseMatDescr_t const )__pyx_v_descr), ((double const *)__pyx_v_values), ((int const *)__pyx_v_rowPtr), ((int *)__pyx_v_nnzPerRow), (&__pyx_v_nnz_total), __pyx_v_tol);

  /* "cupy/cuda/cusparse.pyx":651
 *         <const double *>values, <const int *>rowPtr, <int *>nnzPerRow,
 *         &nnz_total, tol)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return nnz_total
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":652
 *         &nnz_total, tol)
 *     check_status(status)
 *     return nnz_total             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nnz_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cusparse.pyx":642
 * 
 * 
 * cpdef dnnz_compress(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, size_t descr,
 *         size_t values, size_t rowPtr, size_t nnzPerRow,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dnnz_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_65dnnz_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_64dnnz_compress[] = "dnnz_compress(size_t handle, int m, size_t descr, size_t values, size_t rowPtr, size_t nnzPerRow, double tol)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_65dnnz_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  size_t __pyx_v_descr;
  size_t __pyx_v_values;
  size_t __pyx_v_rowPtr;
  size_t __pyx_v_nnzPerRow;
  double __pyx_v_tol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dnnz_compress (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_descr,&__pyx_n_s_values,&__pyx_n_s_rowPtr,&__pyx_n_s_nnzPerRow,&__pyx_n_s_tol,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz_compress", 1, 7, 7, 1); __PYX_ERR(0, 642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz_compress", 1, 7, 7, 2); __PYX_ERR(0, 642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_values)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz_compress", 1, 7, 7, 3); __PYX_ERR(0, 642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz_compress", 1, 7, 7, 4); __PYX_ERR(0, 642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerRow)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz_compress", 1, 7, 7, 5); __PYX_ERR(0, 642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz_compress", 1, 7, 7, 6); __PYX_ERR(0, 642, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dnnz_compress") < 0)) __PYX_ERR(0, 642, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 643, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 643, __pyx_L3_error)
    __pyx_v_descr = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_descr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 643, __pyx_L3_error)
    __pyx_v_values = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_values == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 644, __pyx_L3_error)
    __pyx_v_rowPtr = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_rowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 644, __pyx_L3_error)
    __pyx_v_nnzPerRow = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_nnzPerRow == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 644, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_tol == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dnnz_compress", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 642, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dnnz_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_64dnnz_compress(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_descr, __pyx_v_values, __pyx_v_rowPtr, __pyx_v_nnzPerRow, __pyx_v_tol);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_64dnnz_compress(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, size_t __pyx_v_descr, size_t __pyx_v_values, size_t __pyx_v_rowPtr, size_t __pyx_v_nnzPerRow, double __pyx_v_tol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dnnz_compress", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dnnz_compress(__pyx_v_handle, __pyx_v_m, __pyx_v_descr, __pyx_v_values, __pyx_v_rowPtr, __pyx_v_nnzPerRow, __pyx_v_tol, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dnnz_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":655
 * 
 * 
 * cpdef scsr2csr_compress(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t inVal, size_t inColInd, size_t inRowPtr,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_67scsr2csr_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_scsr2csr_compress(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_inVal, size_t __pyx_v_inColInd, size_t __pyx_v_inRowPtr, int __pyx_v_inNnz, size_t __pyx_v_nnzPerRow, size_t __pyx_v_outVal, size_t __pyx_v_outColInd, size_t __pyx_v_outRowPtr, float __pyx_v_tol, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsr2csr_compress", 0);

  /* "cupy/cuda/cusparse.pyx":660
 *         int inNnz, size_t nnzPerRow, size_t outVal, size_t outColInd,
 *         size_t outRowPtr, float tol):
 *     status = cusparseScsr2csr_compress(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <MatDescr>descrA,
 *         <const float *>inVal, <const int *>inColInd, <const int *>inRowPtr,
 */
  __pyx_v_status = cusparseScsr2csr_compress(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t)__pyx_v_descrA), ((float const *)__pyx_v_inVal), ((int const *)__pyx_v_inColInd), ((int const *)__pyx_v_inRowPtr), __pyx_v_inNnz, ((int *)__pyx_v_nnzPerRow), ((float *)__pyx_v_outVal), ((int *)__pyx_v_outColInd), ((int *)__pyx_v_outRowPtr), __pyx_v_tol);

  /* "cupy/cuda/cusparse.pyx":665
 *         inNnz, <int *>nnzPerRow, <float *>outVal, <int *>outColInd,
 *         <int *>outRowPtr, tol)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":655
 * 
 * 
 * cpdef scsr2csr_compress(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t inVal, size_t inColInd, size_t inRowPtr,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2csr_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_67scsr2csr_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_66scsr2csr_compress[] = "scsr2csr_compress(size_t handle, int m, int n, size_t descrA, size_t inVal, size_t inColInd, size_t inRowPtr, int inNnz, size_t nnzPerRow, size_t outVal, size_t outColInd, size_t outRowPtr, float tol)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_67scsr2csr_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_inVal;
  size_t __pyx_v_inColInd;
  size_t __pyx_v_inRowPtr;
  int __pyx_v_inNnz;
  size_t __pyx_v_nnzPerRow;
  size_t __pyx_v_outVal;
  size_t __pyx_v_outColInd;
  size_t __pyx_v_outRowPtr;
  float __pyx_v_tol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsr2csr_compress (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_inVal,&__pyx_n_s_inColInd,&__pyx_n_s_inRowPtr,&__pyx_n_s_inNnz,&__pyx_n_s_nnzPerRow,&__pyx_n_s_outVal,&__pyx_n_s_outColInd,&__pyx_n_s_outRowPtr,&__pyx_n_s_tol,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 1); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 2); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 3); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 4); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 5); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 6); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inNnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 7); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerRow)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 8); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_outVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 9); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_outColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 10); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_outRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 11); __PYX_ERR(0, 655, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, 12); __PYX_ERR(0, 655, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsr2csr_compress") < 0)) __PYX_ERR(0, 655, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L3_error)
    __pyx_v_inVal = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_inVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 657, __pyx_L3_error)
    __pyx_v_inColInd = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_inColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 657, __pyx_L3_error)
    __pyx_v_inRowPtr = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_inRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 657, __pyx_L3_error)
    __pyx_v_inNnz = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_inNnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 658, __pyx_L3_error)
    __pyx_v_nnzPerRow = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_nnzPerRow == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 658, __pyx_L3_error)
    __pyx_v_outVal = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_outVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 658, __pyx_L3_error)
    __pyx_v_outColInd = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_outColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 658, __pyx_L3_error)
    __pyx_v_outRowPtr = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_outRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 659, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsFloat(values[12]); if (unlikely((__pyx_v_tol == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 659, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsr2csr_compress", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 655, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2csr_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_66scsr2csr_compress(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_inVal, __pyx_v_inColInd, __pyx_v_inRowPtr, __pyx_v_inNnz, __pyx_v_nnzPerRow, __pyx_v_outVal, __pyx_v_outColInd, __pyx_v_outRowPtr, __pyx_v_tol);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_66scsr2csr_compress(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_inVal, size_t __pyx_v_inColInd, size_t __pyx_v_inRowPtr, int __pyx_v_inNnz, size_t __pyx_v_nnzPerRow, size_t __pyx_v_outVal, size_t __pyx_v_outColInd, size_t __pyx_v_outRowPtr, float __pyx_v_tol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsr2csr_compress", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_scsr2csr_compress(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_inVal, __pyx_v_inColInd, __pyx_v_inRowPtr, __pyx_v_inNnz, __pyx_v_nnzPerRow, __pyx_v_outVal, __pyx_v_outColInd, __pyx_v_outRowPtr, __pyx_v_tol, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.scsr2csr_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":668
 * 
 * 
 * cpdef dcsr2csr_compress(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t inVal, size_t inColInd, size_t inRowPtr,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_69dcsr2csr_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dcsr2csr_compress(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_inVal, size_t __pyx_v_inColInd, size_t __pyx_v_inRowPtr, int __pyx_v_inNnz, size_t __pyx_v_nnzPerRow, size_t __pyx_v_outVal, size_t __pyx_v_outColInd, size_t __pyx_v_outRowPtr, float __pyx_v_tol, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsr2csr_compress", 0);

  /* "cupy/cuda/cusparse.pyx":673
 *         int inNnz, size_t nnzPerRow, size_t outVal, size_t outColInd,
 *         size_t outRowPtr, float tol):
 *     status = cusparseDcsr2csr_compress(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <MatDescr>descrA,
 *         <const double *>inVal, <const int *>inColInd, <const int *>inRowPtr,
 */
  __pyx_v_status = cusparseDcsr2csr_compress(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t)__pyx_v_descrA), ((double const *)__pyx_v_inVal), ((int const *)__pyx_v_inColInd), ((int const *)__pyx_v_inRowPtr), __pyx_v_inNnz, ((int *)__pyx_v_nnzPerRow), ((double *)__pyx_v_outVal), ((int *)__pyx_v_outColInd), ((int *)__pyx_v_outRowPtr), __pyx_v_tol);

  /* "cupy/cuda/cusparse.pyx":678
 *         inNnz, <int *>nnzPerRow, <double *>outVal, <int *>outColInd,
 *         <int *>outRowPtr, tol)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":668
 * 
 * 
 * cpdef dcsr2csr_compress(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t inVal, size_t inColInd, size_t inRowPtr,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2csr_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_69dcsr2csr_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_68dcsr2csr_compress[] = "dcsr2csr_compress(size_t handle, int m, int n, size_t descrA, size_t inVal, size_t inColInd, size_t inRowPtr, int inNnz, size_t nnzPerRow, size_t outVal, size_t outColInd, size_t outRowPtr, float tol)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_69dcsr2csr_compress(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_inVal;
  size_t __pyx_v_inColInd;
  size_t __pyx_v_inRowPtr;
  int __pyx_v_inNnz;
  size_t __pyx_v_nnzPerRow;
  size_t __pyx_v_outVal;
  size_t __pyx_v_outColInd;
  size_t __pyx_v_outRowPtr;
  float __pyx_v_tol;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsr2csr_compress (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_inVal,&__pyx_n_s_inColInd,&__pyx_n_s_inRowPtr,&__pyx_n_s_inNnz,&__pyx_n_s_nnzPerRow,&__pyx_n_s_outVal,&__pyx_n_s_outColInd,&__pyx_n_s_outRowPtr,&__pyx_n_s_tol,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 1); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 2); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 3); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 4); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 5); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 6); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inNnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 7); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerRow)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 8); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_outVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 9); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_outColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 10); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_outRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 11); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, 12); __PYX_ERR(0, 668, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsr2csr_compress") < 0)) __PYX_ERR(0, 668, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 669, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 669, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 669, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 669, __pyx_L3_error)
    __pyx_v_inVal = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_inVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 670, __pyx_L3_error)
    __pyx_v_inColInd = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_inColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 670, __pyx_L3_error)
    __pyx_v_inRowPtr = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_inRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 670, __pyx_L3_error)
    __pyx_v_inNnz = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_inNnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
    __pyx_v_nnzPerRow = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_nnzPerRow == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
    __pyx_v_outVal = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_outVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
    __pyx_v_outColInd = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_outColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 671, __pyx_L3_error)
    __pyx_v_outRowPtr = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_outRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsFloat(values[12]); if (unlikely((__pyx_v_tol == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsr2csr_compress", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 668, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2csr_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_68dcsr2csr_compress(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_inVal, __pyx_v_inColInd, __pyx_v_inRowPtr, __pyx_v_inNnz, __pyx_v_nnzPerRow, __pyx_v_outVal, __pyx_v_outColInd, __pyx_v_outRowPtr, __pyx_v_tol);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_68dcsr2csr_compress(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_inVal, size_t __pyx_v_inColInd, size_t __pyx_v_inRowPtr, int __pyx_v_inNnz, size_t __pyx_v_nnzPerRow, size_t __pyx_v_outVal, size_t __pyx_v_outColInd, size_t __pyx_v_outRowPtr, float __pyx_v_tol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsr2csr_compress", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dcsr2csr_compress(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_inVal, __pyx_v_inColInd, __pyx_v_inRowPtr, __pyx_v_inNnz, __pyx_v_nnzPerRow, __pyx_v_outVal, __pyx_v_outColInd, __pyx_v_outRowPtr, __pyx_v_tol, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dcsr2csr_compress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":681
 * 
 * 
 * cpdef sdense2csc(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA, size_t A,
 *         int lda, size_t nnzPerCol, size_t cscValA, size_t cscRowIndA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_71sdense2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_sdense2csc(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerCol, size_t __pyx_v_cscValA, size_t __pyx_v_cscRowIndA, size_t __pyx_v_cscColPtrA, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sdense2csc", 0);

  /* "cupy/cuda/cusparse.pyx":685
 *         int lda, size_t nnzPerCol, size_t cscValA, size_t cscRowIndA,
 *         size_t cscColPtrA):
 *     status = cusparseSdense2csc(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <const MatDescr>descrA, <const float *>A,
 *         lda, <const int *>nnzPerCol, <float *>cscValA, <int *>cscRowIndA,
 */
  __pyx_v_status = cusparseSdense2csc(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t const )__pyx_v_descrA), ((float const *)__pyx_v_A), __pyx_v_lda, ((int const *)__pyx_v_nnzPerCol), ((float *)__pyx_v_cscValA), ((int *)__pyx_v_cscRowIndA), ((int *)__pyx_v_cscColPtrA));

  /* "cupy/cuda/cusparse.pyx":689
 *         lda, <const int *>nnzPerCol, <float *>cscValA, <int *>cscRowIndA,
 *         <int *>cscColPtrA)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":681
 * 
 * 
 * cpdef sdense2csc(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA, size_t A,
 *         int lda, size_t nnzPerCol, size_t cscValA, size_t cscRowIndA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.sdense2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_71sdense2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_70sdense2csc[] = "sdense2csc(size_t handle, int m, int n, size_t descrA, size_t A, int lda, size_t nnzPerCol, size_t cscValA, size_t cscRowIndA, size_t cscColPtrA)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_71sdense2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_nnzPerCol;
  size_t __pyx_v_cscValA;
  size_t __pyx_v_cscRowIndA;
  size_t __pyx_v_cscColPtrA;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sdense2csc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_nnzPerCol,&__pyx_n_s_cscValA,&__pyx_n_s_cscRowIndA,&__pyx_n_s_cscColPtrA,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 1); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 2); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 3); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 4); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 5); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerCol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 6); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 7); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscRowIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 8); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscColPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, 9); __PYX_ERR(0, 681, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sdense2csc") < 0)) __PYX_ERR(0, 681, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L3_error)
    __pyx_v_nnzPerCol = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_nnzPerCol == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L3_error)
    __pyx_v_cscValA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_cscValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L3_error)
    __pyx_v_cscRowIndA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_cscRowIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L3_error)
    __pyx_v_cscColPtrA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_cscColPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sdense2csc", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 681, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.sdense2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_70sdense2csc(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerCol, __pyx_v_cscValA, __pyx_v_cscRowIndA, __pyx_v_cscColPtrA);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_70sdense2csc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerCol, size_t __pyx_v_cscValA, size_t __pyx_v_cscRowIndA, size_t __pyx_v_cscColPtrA) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sdense2csc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_sdense2csc(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerCol, __pyx_v_cscValA, __pyx_v_cscRowIndA, __pyx_v_cscColPtrA, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.sdense2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":692
 * 
 * 
 * cpdef ddense2csc(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA, size_t A,
 *         int lda, size_t nnzPerCol, size_t cscValA, size_t cscRowIndA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_73ddense2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_ddense2csc(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerCol, size_t __pyx_v_cscValA, size_t __pyx_v_cscRowIndA, size_t __pyx_v_cscColPtrA, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ddense2csc", 0);

  /* "cupy/cuda/cusparse.pyx":696
 *         int lda, size_t nnzPerCol, size_t cscValA, size_t cscRowIndA,
 *         size_t cscColPtrA):
 *     status = cusparseDdense2csc(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <const MatDescr>descrA, <const double *>A,
 *         lda, <const int *>nnzPerCol, <double *>cscValA, <int *>cscRowIndA,
 */
  __pyx_v_status = cusparseDdense2csc(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t const )__pyx_v_descrA), ((double const *)__pyx_v_A), __pyx_v_lda, ((int const *)__pyx_v_nnzPerCol), ((double *)__pyx_v_cscValA), ((int *)__pyx_v_cscRowIndA), ((int *)__pyx_v_cscColPtrA));

  /* "cupy/cuda/cusparse.pyx":700
 *         lda, <const int *>nnzPerCol, <double *>cscValA, <int *>cscRowIndA,
 *         <int *>cscColPtrA)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":692
 * 
 * 
 * cpdef ddense2csc(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA, size_t A,
 *         int lda, size_t nnzPerCol, size_t cscValA, size_t cscRowIndA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.ddense2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_73ddense2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_72ddense2csc[] = "ddense2csc(size_t handle, int m, int n, size_t descrA, size_t A, int lda, size_t nnzPerCol, size_t cscValA, size_t cscRowIndA, size_t cscColPtrA)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_73ddense2csc(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_nnzPerCol;
  size_t __pyx_v_cscValA;
  size_t __pyx_v_cscRowIndA;
  size_t __pyx_v_cscColPtrA;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ddense2csc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_nnzPerCol,&__pyx_n_s_cscValA,&__pyx_n_s_cscRowIndA,&__pyx_n_s_cscColPtrA,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 1); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 2); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 3); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 4); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 5); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerCol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 6); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 7); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscRowIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 8); __PYX_ERR(0, 692, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscColPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, 9); __PYX_ERR(0, 692, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ddense2csc") < 0)) __PYX_ERR(0, 692, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 693, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 693, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 693, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 693, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 693, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_nnzPerCol = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_nnzPerCol == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_cscValA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_cscValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_cscRowIndA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_cscRowIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_cscColPtrA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_cscColPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 695, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ddense2csc", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 692, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.ddense2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_72ddense2csc(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerCol, __pyx_v_cscValA, __pyx_v_cscRowIndA, __pyx_v_cscColPtrA);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_72ddense2csc(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerCol, size_t __pyx_v_cscValA, size_t __pyx_v_cscRowIndA, size_t __pyx_v_cscColPtrA) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ddense2csc", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_ddense2csc(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerCol, __pyx_v_cscValA, __pyx_v_cscRowIndA, __pyx_v_cscColPtrA, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.ddense2csc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":703
 * 
 * 
 * cpdef sdense2csr(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRow, size_t csrValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_75sdense2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_sdense2csr(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRow, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sdense2csr", 0);

  /* "cupy/cuda/cusparse.pyx":707
 *         size_t A, int lda, size_t nnzPerRow, size_t csrValA,
 *         size_t csrRowPtrA, size_t csrColIndA):
 *     status = cusparseSdense2csr(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <MatDescr>descrA,
 *         <const float *>A, lda, <const int *>nnzPerRow, <float *>csrValA,
 */
  __pyx_v_status = cusparseSdense2csr(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t)__pyx_v_descrA), ((float const *)__pyx_v_A), __pyx_v_lda, ((int const *)__pyx_v_nnzPerRow), ((float *)__pyx_v_csrValA), ((int *)__pyx_v_csrRowPtrA), ((int *)__pyx_v_csrColIndA));

  /* "cupy/cuda/cusparse.pyx":711
 *         <const float *>A, lda, <const int *>nnzPerRow, <float *>csrValA,
 *         <int *>csrRowPtrA, <int *>csrColIndA)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":703
 * 
 * 
 * cpdef sdense2csr(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRow, size_t csrValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.sdense2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_75sdense2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_74sdense2csr[] = "sdense2csr(size_t handle, int m, int n, size_t descrA, size_t A, int lda, size_t nnzPerRow, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_75sdense2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_nnzPerRow;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sdense2csr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_nnzPerRow,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 1); __PYX_ERR(0, 703, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 2); __PYX_ERR(0, 703, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 3); __PYX_ERR(0, 703, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 4); __PYX_ERR(0, 703, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 5); __PYX_ERR(0, 703, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerRow)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 6); __PYX_ERR(0, 703, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 7); __PYX_ERR(0, 703, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 8); __PYX_ERR(0, 703, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, 9); __PYX_ERR(0, 703, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sdense2csr") < 0)) __PYX_ERR(0, 703, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 704, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 704, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 704, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 704, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 705, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 705, __pyx_L3_error)
    __pyx_v_nnzPerRow = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_nnzPerRow == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 705, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 705, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 706, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 706, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sdense2csr", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 703, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.sdense2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_74sdense2csr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerRow, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_74sdense2csr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRow, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sdense2csr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_sdense2csr(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerRow, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.sdense2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":714
 * 
 * 
 * cpdef ddense2csr(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRow, size_t csrValA,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_77ddense2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_ddense2csr(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRow, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ddense2csr", 0);

  /* "cupy/cuda/cusparse.pyx":718
 *         size_t A, int lda, size_t nnzPerRow, size_t csrValA,
 *         size_t csrRowPtrA, size_t csrColIndA):
 *     status = cusparseDdense2csr(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, <MatDescr>descrA,
 *         <const double *>A, lda, <const int *>nnzPerRow, <double *>csrValA,
 */
  __pyx_v_status = cusparseDdense2csr(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t)__pyx_v_descrA), ((double const *)__pyx_v_A), __pyx_v_lda, ((int const *)__pyx_v_nnzPerRow), ((double *)__pyx_v_csrValA), ((int *)__pyx_v_csrRowPtrA), ((int *)__pyx_v_csrColIndA));

  /* "cupy/cuda/cusparse.pyx":722
 *         <const double *>A, lda, <const int *>nnzPerRow, <double *>csrValA,
 *         <int *>csrRowPtrA, <int *>csrColIndA)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":714
 * 
 * 
 * cpdef ddense2csr(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRow, size_t csrValA,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.ddense2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_77ddense2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_76ddense2csr[] = "ddense2csr(size_t handle, int m, int n, size_t descrA, size_t A, int lda, size_t nnzPerRow, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_77ddense2csr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_nnzPerRow;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ddense2csr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_nnzPerRow,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 1); __PYX_ERR(0, 714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 2); __PYX_ERR(0, 714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 3); __PYX_ERR(0, 714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 4); __PYX_ERR(0, 714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 5); __PYX_ERR(0, 714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerRow)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 6); __PYX_ERR(0, 714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 7); __PYX_ERR(0, 714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 8); __PYX_ERR(0, 714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, 9); __PYX_ERR(0, 714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ddense2csr") < 0)) __PYX_ERR(0, 714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 715, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 715, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 715, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 715, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 716, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 716, __pyx_L3_error)
    __pyx_v_nnzPerRow = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_nnzPerRow == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 716, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 716, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ddense2csr", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.ddense2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_76ddense2csr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerRow, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_76ddense2csr(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRow, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ddense2csr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_ddense2csr(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerRow, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.ddense2csr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":725
 * 
 * 
 * cpdef snnz(             # <<<<<<<<<<<<<<
 *         size_t handle, int dirA, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRowColumn, size_t nnzTotalDevHostPtr):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_79snnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_snnz(size_t __pyx_v_handle, int __pyx_v_dirA, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRowColumn, size_t __pyx_v_nnzTotalDevHostPtr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("snnz", 0);

  /* "cupy/cuda/cusparse.pyx":728
 *         size_t handle, int dirA, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRowColumn, size_t nnzTotalDevHostPtr):
 *     status = cusparseSnnz(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Direction>dirA, m, n, <const MatDescr>descrA,
 *         <const float *>A, lda, <int *>nnzPerRowColumn,
 */
  __pyx_v_status = cusparseSnnz(((cusparseHandle_t)__pyx_v_handle), ((cusparseDirection_t)__pyx_v_dirA), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t const )__pyx_v_descrA), ((float const *)__pyx_v_A), __pyx_v_lda, ((int *)__pyx_v_nnzPerRowColumn), ((int *)__pyx_v_nnzTotalDevHostPtr));

  /* "cupy/cuda/cusparse.pyx":732
 *         <const float *>A, lda, <int *>nnzPerRowColumn,
 *         <int *>nnzTotalDevHostPtr)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":725
 * 
 * 
 * cpdef snnz(             # <<<<<<<<<<<<<<
 *         size_t handle, int dirA, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRowColumn, size_t nnzTotalDevHostPtr):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.snnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_79snnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_78snnz[] = "snnz(size_t handle, int dirA, int m, int n, size_t descrA, size_t A, int lda, size_t nnzPerRowColumn, size_t nnzTotalDevHostPtr)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_79snnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_dirA;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_nnzPerRowColumn;
  size_t __pyx_v_nnzTotalDevHostPtr;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("snnz (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_dirA,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_nnzPerRowColumn,&__pyx_n_s_nnzTotalDevHostPtr,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dirA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, 1); __PYX_ERR(0, 725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, 2); __PYX_ERR(0, 725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, 3); __PYX_ERR(0, 725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, 4); __PYX_ERR(0, 725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, 5); __PYX_ERR(0, 725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, 6); __PYX_ERR(0, 725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerRowColumn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, 7); __PYX_ERR(0, 725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzTotalDevHostPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, 8); __PYX_ERR(0, 725, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "snnz") < 0)) __PYX_ERR(0, 725, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_dirA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_dirA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 727, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 727, __pyx_L3_error)
    __pyx_v_nnzPerRowColumn = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_nnzPerRowColumn == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 727, __pyx_L3_error)
    __pyx_v_nnzTotalDevHostPtr = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_nnzTotalDevHostPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 727, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("snnz", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 725, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.snnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_78snnz(__pyx_self, __pyx_v_handle, __pyx_v_dirA, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerRowColumn, __pyx_v_nnzTotalDevHostPtr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_78snnz(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_dirA, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRowColumn, size_t __pyx_v_nnzTotalDevHostPtr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("snnz", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_snnz(__pyx_v_handle, __pyx_v_dirA, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerRowColumn, __pyx_v_nnzTotalDevHostPtr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.snnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":735
 * 
 * 
 * cpdef dnnz(             # <<<<<<<<<<<<<<
 *         size_t handle, int dirA, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRowColumn, size_t nnzTotalDevHostPtr):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_81dnnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_dnnz(size_t __pyx_v_handle, int __pyx_v_dirA, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRowColumn, size_t __pyx_v_nnzTotalDevHostPtr, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dnnz", 0);

  /* "cupy/cuda/cusparse.pyx":738
 *         size_t handle, int dirA, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRowColumn, size_t nnzTotalDevHostPtr):
 *     status = cusparseDnnz(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <Direction>dirA, m, n, <const MatDescr>descrA,
 *         <const double *>A, lda, <int *>nnzPerRowColumn,
 */
  __pyx_v_status = cusparseDnnz(((cusparseHandle_t)__pyx_v_handle), ((cusparseDirection_t)__pyx_v_dirA), __pyx_v_m, __pyx_v_n, ((cusparseMatDescr_t const )__pyx_v_descrA), ((double const *)__pyx_v_A), __pyx_v_lda, ((int *)__pyx_v_nnzPerRowColumn), ((int *)__pyx_v_nnzTotalDevHostPtr));

  /* "cupy/cuda/cusparse.pyx":742
 *         <const double *>A, lda, <int *>nnzPerRowColumn,
 *         <int *>nnzTotalDevHostPtr)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":735
 * 
 * 
 * cpdef dnnz(             # <<<<<<<<<<<<<<
 *         size_t handle, int dirA, int m, int n, size_t descrA,
 *         size_t A, int lda, size_t nnzPerRowColumn, size_t nnzTotalDevHostPtr):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dnnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_81dnnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_80dnnz[] = "dnnz(size_t handle, int dirA, int m, int n, size_t descrA, size_t A, int lda, size_t nnzPerRowColumn, size_t nnzTotalDevHostPtr)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_81dnnz(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_dirA;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_descrA;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_nnzPerRowColumn;
  size_t __pyx_v_nnzTotalDevHostPtr;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dnnz (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_dirA,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_descrA,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_nnzPerRowColumn,&__pyx_n_s_nnzTotalDevHostPtr,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dirA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, 1); __PYX_ERR(0, 735, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, 2); __PYX_ERR(0, 735, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, 3); __PYX_ERR(0, 735, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, 4); __PYX_ERR(0, 735, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, 5); __PYX_ERR(0, 735, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, 6); __PYX_ERR(0, 735, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzPerRowColumn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, 7); __PYX_ERR(0, 735, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnzTotalDevHostPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, 8); __PYX_ERR(0, 735, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dnnz") < 0)) __PYX_ERR(0, 735, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L3_error)
    __pyx_v_dirA = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_dirA == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L3_error)
    __pyx_v_nnzPerRowColumn = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_nnzPerRowColumn == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L3_error)
    __pyx_v_nnzTotalDevHostPtr = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_nnzTotalDevHostPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dnnz", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 735, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.dnnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_80dnnz(__pyx_self, __pyx_v_handle, __pyx_v_dirA, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerRowColumn, __pyx_v_nnzTotalDevHostPtr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_80dnnz(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_dirA, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_descrA, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_nnzPerRowColumn, size_t __pyx_v_nnzTotalDevHostPtr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dnnz", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_dnnz(__pyx_v_handle, __pyx_v_dirA, __pyx_v_m, __pyx_v_n, __pyx_v_descrA, __pyx_v_A, __pyx_v_lda, __pyx_v_nnzPerRowColumn, __pyx_v_nnzTotalDevHostPtr, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.dnnz", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":745
 * 
 * 
 * cpdef createIdentityPermutation(             # <<<<<<<<<<<<<<
 *         size_t handle, int n, size_t p):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_83createIdentityPermutation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_createIdentityPermutation(size_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_p, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createIdentityPermutation", 0);

  /* "cupy/cuda/cusparse.pyx":747
 * cpdef createIdentityPermutation(
 *         size_t handle, int n, size_t p):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseCreateIdentityPermutation(
 *         <Handle>handle, n, <int *>p)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":748
 *         size_t handle, int n, size_t p):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseCreateIdentityPermutation(             # <<<<<<<<<<<<<<
 *         <Handle>handle, n, <int *>p)
 *     check_status(status)
 */
  __pyx_v_status = cusparseCreateIdentityPermutation(((cusparseHandle_t)__pyx_v_handle), __pyx_v_n, ((int *)__pyx_v_p));

  /* "cupy/cuda/cusparse.pyx":750
 *     status = cusparseCreateIdentityPermutation(
 *         <Handle>handle, n, <int *>p)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":745
 * 
 * 
 * cpdef createIdentityPermutation(             # <<<<<<<<<<<<<<
 *         size_t handle, int n, size_t p):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.createIdentityPermutation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_83createIdentityPermutation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_82createIdentityPermutation[] = "createIdentityPermutation(size_t handle, int n, size_t p)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_83createIdentityPermutation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_n;
  size_t __pyx_v_p;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createIdentityPermutation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_n,&__pyx_n_s_p,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("createIdentityPermutation", 1, 3, 3, 1); __PYX_ERR(0, 745, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("createIdentityPermutation", 1, 3, 3, 2); __PYX_ERR(0, 745, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "createIdentityPermutation") < 0)) __PYX_ERR(0, 745, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 746, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 746, __pyx_L3_error)
    __pyx_v_p = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_p == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 746, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("createIdentityPermutation", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 745, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.createIdentityPermutation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_82createIdentityPermutation(__pyx_self, __pyx_v_handle, __pyx_v_n, __pyx_v_p);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_82createIdentityPermutation(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createIdentityPermutation", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_createIdentityPermutation(__pyx_v_handle, __pyx_v_n, __pyx_v_p, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.createIdentityPermutation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":753
 * 
 * 
 * cpdef size_t xcoosort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t cooRows,
 *         size_t cooCols):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_85xcoosort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_xcoosort_bufferSizeExt(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cooRows, size_t __pyx_v_cooCols, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_bufferSizeInBytes;
  cusparseStatus_t __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcoosort_bufferSizeExt", 0);

  /* "cupy/cuda/cusparse.pyx":757
 *         size_t cooCols):
 *     cdef size_t bufferSizeInBytes
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcoosort_bufferSizeExt(
 *         <Handle>handle, m, n, nnz, <const int *>cooRows,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":758
 *     cdef size_t bufferSizeInBytes
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcoosort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, nnz, <const int *>cooRows,
 *         <const int *>cooCols, &bufferSizeInBytes)
 */
  __pyx_v_status = cusparseXcoosort_bufferSizeExt(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((int const *)__pyx_v_cooRows), ((int const *)__pyx_v_cooCols), (&__pyx_v_bufferSizeInBytes));

  /* "cupy/cuda/cusparse.pyx":761
 *         <Handle>handle, m, n, nnz, <const int *>cooRows,
 *         <const int *>cooCols, &bufferSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return bufferSizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":762
 *         <const int *>cooCols, &bufferSizeInBytes)
 *     check_status(status)
 *     return bufferSizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_bufferSizeInBytes;
  goto __pyx_L0;

  /* "cupy/cuda/cusparse.pyx":753
 * 
 * 
 * cpdef size_t xcoosort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t cooRows,
 *         size_t cooCols):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusparse.xcoosort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_85xcoosort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_84xcoosort_bufferSizeExt[] = "xcoosort_bufferSizeExt(size_t handle, int m, int n, int nnz, size_t cooRows, size_t cooCols) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_85xcoosort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_cooRows;
  size_t __pyx_v_cooCols;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcoosort_bufferSizeExt (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_cooRows,&__pyx_n_s_cooCols,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosort_bufferSizeExt", 1, 6, 6, 1); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosort_bufferSizeExt", 1, 6, 6, 2); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosort_bufferSizeExt", 1, 6, 6, 3); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cooRows)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosort_bufferSizeExt", 1, 6, 6, 4); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cooCols)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosort_bufferSizeExt", 1, 6, 6, 5); __PYX_ERR(0, 753, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcoosort_bufferSizeExt") < 0)) __PYX_ERR(0, 753, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_cooRows = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_cooRows == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_cooCols = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_cooCols == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcoosort_bufferSizeExt", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 753, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcoosort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_84xcoosort_bufferSizeExt(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_cooRows, __pyx_v_cooCols);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_84xcoosort_bufferSizeExt(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cooRows, size_t __pyx_v_cooCols) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcoosort_bufferSizeExt", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_f_4cupy_4cuda_8cusparse_xcoosort_bufferSizeExt(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_cooRows, __pyx_v_cooCols, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcoosort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":765
 * 
 * 
 * cpdef xcoosortByRow(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t cooRows, size_t cooCols,
 *         size_t P, size_t pBuffer):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_87xcoosortByRow(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcoosortByRow(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cooRows, size_t __pyx_v_cooCols, size_t __pyx_v_P, size_t __pyx_v_pBuffer, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcoosortByRow", 0);

  /* "cupy/cuda/cusparse.pyx":768
 *         size_t handle, int m, int n, int nnz, size_t cooRows, size_t cooCols,
 *         size_t P, size_t pBuffer):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcoosortByRow(
 *         <Handle>handle, m, n, nnz, <int *>cooRows, <int *>cooCols,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":769
 *         size_t P, size_t pBuffer):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcoosortByRow(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, nnz, <int *>cooRows, <int *>cooCols,
 *         <int *>P, <void *>pBuffer)
 */
  __pyx_v_status = cusparseXcoosortByRow(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((int *)__pyx_v_cooRows), ((int *)__pyx_v_cooCols), ((int *)__pyx_v_P), ((void *)__pyx_v_pBuffer));

  /* "cupy/cuda/cusparse.pyx":772
 *         <Handle>handle, m, n, nnz, <int *>cooRows, <int *>cooCols,
 *         <int *>P, <void *>pBuffer)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":765
 * 
 * 
 * cpdef xcoosortByRow(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t cooRows, size_t cooCols,
 *         size_t P, size_t pBuffer):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcoosortByRow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_87xcoosortByRow(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_86xcoosortByRow[] = "xcoosortByRow(size_t handle, int m, int n, int nnz, size_t cooRows, size_t cooCols, size_t P, size_t pBuffer)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_87xcoosortByRow(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_cooRows;
  size_t __pyx_v_cooCols;
  size_t __pyx_v_P;
  size_t __pyx_v_pBuffer;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcoosortByRow (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_cooRows,&__pyx_n_s_cooCols,&__pyx_n_s_P,&__pyx_n_s_pBuffer,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosortByRow", 1, 8, 8, 1); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosortByRow", 1, 8, 8, 2); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosortByRow", 1, 8, 8, 3); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cooRows)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosortByRow", 1, 8, 8, 4); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cooCols)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosortByRow", 1, 8, 8, 5); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_P)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosortByRow", 1, 8, 8, 6); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pBuffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcoosortByRow", 1, 8, 8, 7); __PYX_ERR(0, 765, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcoosortByRow") < 0)) __PYX_ERR(0, 765, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_cooRows = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_cooRows == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_cooCols = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_cooCols == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_P = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_P == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
    __pyx_v_pBuffer = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_pBuffer == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcoosortByRow", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 765, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcoosortByRow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_86xcoosortByRow(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_cooRows, __pyx_v_cooCols, __pyx_v_P, __pyx_v_pBuffer);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_86xcoosortByRow(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cooRows, size_t __pyx_v_cooCols, size_t __pyx_v_P, size_t __pyx_v_pBuffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcoosortByRow", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_xcoosortByRow(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_cooRows, __pyx_v_cooCols, __pyx_v_P, __pyx_v_pBuffer, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcoosortByRow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":775
 * 
 * 
 * cpdef size_t xcsrsort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t csrRowPtr,
 *         size_t csrColInd):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_89xcsrsort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_xcsrsort_bufferSizeExt(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_bufferSizeInBytes;
  cusparseStatus_t __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsrsort_bufferSizeExt", 0);

  /* "cupy/cuda/cusparse.pyx":779
 *         size_t csrColInd):
 *     cdef size_t bufferSizeInBytes
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcsrsort_bufferSizeExt(
 *         <Handle>handle, m, n, nnz, <const int *>csrRowPtr,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":780
 *     cdef size_t bufferSizeInBytes
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcsrsort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, nnz, <const int *>csrRowPtr,
 *         <const int *>csrColInd, &bufferSizeInBytes)
 */
  __pyx_v_status = cusparseXcsrsort_bufferSizeExt(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((int const *)__pyx_v_csrRowPtr), ((int const *)__pyx_v_csrColInd), (&__pyx_v_bufferSizeInBytes));

  /* "cupy/cuda/cusparse.pyx":783
 *         <Handle>handle, m, n, nnz, <const int *>csrRowPtr,
 *         <const int *>csrColInd, &bufferSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return bufferSizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":784
 *         <const int *>csrColInd, &bufferSizeInBytes)
 *     check_status(status)
 *     return bufferSizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_bufferSizeInBytes;
  goto __pyx_L0;

  /* "cupy/cuda/cusparse.pyx":775
 * 
 * 
 * cpdef size_t xcsrsort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t csrRowPtr,
 *         size_t csrColInd):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusparse.xcsrsort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_89xcsrsort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_88xcsrsort_bufferSizeExt[] = "xcsrsort_bufferSizeExt(size_t handle, int m, int n, int nnz, size_t csrRowPtr, size_t csrColInd) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_89xcsrsort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_csrRowPtr;
  size_t __pyx_v_csrColInd;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcsrsort_bufferSizeExt (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_csrRowPtr,&__pyx_n_s_csrColInd,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort_bufferSizeExt", 1, 6, 6, 1); __PYX_ERR(0, 775, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort_bufferSizeExt", 1, 6, 6, 2); __PYX_ERR(0, 775, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort_bufferSizeExt", 1, 6, 6, 3); __PYX_ERR(0, 775, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort_bufferSizeExt", 1, 6, 6, 4); __PYX_ERR(0, 775, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort_bufferSizeExt", 1, 6, 6, 5); __PYX_ERR(0, 775, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcsrsort_bufferSizeExt") < 0)) __PYX_ERR(0, 775, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_csrColInd = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 777, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcsrsort_bufferSizeExt", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 775, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrsort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_88xcsrsort_bufferSizeExt(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_csrRowPtr, __pyx_v_csrColInd);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_88xcsrsort_bufferSizeExt(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsrsort_bufferSizeExt", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_f_4cupy_4cuda_8cusparse_xcsrsort_bufferSizeExt(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_csrRowPtr, __pyx_v_csrColInd, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrsort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":787
 * 
 * 
 * cpdef xcsrsort(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t descrA,
 *         size_t csrRowPtr, size_t csrColInd, size_t P, size_t pBuffer):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_91xcsrsort(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcsrsort(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_P, size_t __pyx_v_pBuffer, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsrsort", 0);

  /* "cupy/cuda/cusparse.pyx":790
 *         size_t handle, int m, int n, int nnz, size_t descrA,
 *         size_t csrRowPtr, size_t csrColInd, size_t P, size_t pBuffer):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcsrsort(
 *         <Handle>handle, m, n, nnz, <const MatDescr>descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":791
 *         size_t csrRowPtr, size_t csrColInd, size_t P, size_t pBuffer):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcsrsort(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, nnz, <const MatDescr>descrA,
 *         <const int *>csrRowPtr, <int *>csrColInd, <int *>P, <void *>pBuffer)
 */
  __pyx_v_status = cusparseXcsrsort(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((int const *)__pyx_v_csrRowPtr), ((int *)__pyx_v_csrColInd), ((int *)__pyx_v_P), ((void *)__pyx_v_pBuffer));

  /* "cupy/cuda/cusparse.pyx":794
 *         <Handle>handle, m, n, nnz, <const MatDescr>descrA,
 *         <const int *>csrRowPtr, <int *>csrColInd, <int *>P, <void *>pBuffer)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":787
 * 
 * 
 * cpdef xcsrsort(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t descrA,
 *         size_t csrRowPtr, size_t csrColInd, size_t P, size_t pBuffer):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrsort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_91xcsrsort(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_90xcsrsort[] = "xcsrsort(size_t handle, int m, int n, int nnz, size_t descrA, size_t csrRowPtr, size_t csrColInd, size_t P, size_t pBuffer)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_91xcsrsort(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrRowPtr;
  size_t __pyx_v_csrColInd;
  size_t __pyx_v_P;
  size_t __pyx_v_pBuffer;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcsrsort (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrRowPtr,&__pyx_n_s_csrColInd,&__pyx_n_s_P,&__pyx_n_s_pBuffer,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, 1); __PYX_ERR(0, 787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, 2); __PYX_ERR(0, 787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, 3); __PYX_ERR(0, 787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, 4); __PYX_ERR(0, 787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, 5); __PYX_ERR(0, 787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_csrColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, 6); __PYX_ERR(0, 787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_P)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, 7); __PYX_ERR(0, 787, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pBuffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, 8); __PYX_ERR(0, 787, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcsrsort") < 0)) __PYX_ERR(0, 787, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 789, __pyx_L3_error)
    __pyx_v_csrColInd = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 789, __pyx_L3_error)
    __pyx_v_P = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_P == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 789, __pyx_L3_error)
    __pyx_v_pBuffer = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_pBuffer == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 789, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcsrsort", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 787, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrsort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_90xcsrsort(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_P, __pyx_v_pBuffer);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_90xcsrsort(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_P, size_t __pyx_v_pBuffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcsrsort", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_xcsrsort(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_P, __pyx_v_pBuffer, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcsrsort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":797
 * 
 * 
 * cpdef size_t xcscsort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t cscColPtr,
 *         size_t cscRowInd):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_93xcscsort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusparse_xcscsort_bufferSizeExt(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cscColPtr, size_t __pyx_v_cscRowInd, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_bufferSizeInBytes;
  cusparseStatus_t __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcscsort_bufferSizeExt", 0);

  /* "cupy/cuda/cusparse.pyx":801
 *         size_t cscRowInd):
 *     cdef size_t bufferSizeInBytes
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcscsort_bufferSizeExt(
 *         <Handle>handle, m, n, nnz, <const int *>cscColPtr,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":802
 *     cdef size_t bufferSizeInBytes
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcscsort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, nnz, <const int *>cscColPtr,
 *         <const int *>cscRowInd, &bufferSizeInBytes)
 */
  __pyx_v_status = cusparseXcscsort_bufferSizeExt(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((int const *)__pyx_v_cscColPtr), ((int const *)__pyx_v_cscRowInd), (&__pyx_v_bufferSizeInBytes));

  /* "cupy/cuda/cusparse.pyx":805
 *         <Handle>handle, m, n, nnz, <const int *>cscColPtr,
 *         <const int *>cscRowInd, &bufferSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return bufferSizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":806
 *         <const int *>cscRowInd, &bufferSizeInBytes)
 *     check_status(status)
 *     return bufferSizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_bufferSizeInBytes;
  goto __pyx_L0;

  /* "cupy/cuda/cusparse.pyx":797
 * 
 * 
 * cpdef size_t xcscsort_bufferSizeExt(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t cscColPtr,
 *         size_t cscRowInd):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusparse.xcscsort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_93xcscsort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_92xcscsort_bufferSizeExt[] = "xcscsort_bufferSizeExt(size_t handle, int m, int n, int nnz, size_t cscColPtr, size_t cscRowInd) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_93xcscsort_bufferSizeExt(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_cscColPtr;
  size_t __pyx_v_cscRowInd;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcscsort_bufferSizeExt (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_cscColPtr,&__pyx_n_s_cscRowInd,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort_bufferSizeExt", 1, 6, 6, 1); __PYX_ERR(0, 797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort_bufferSizeExt", 1, 6, 6, 2); __PYX_ERR(0, 797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort_bufferSizeExt", 1, 6, 6, 3); __PYX_ERR(0, 797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscColPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort_bufferSizeExt", 1, 6, 6, 4); __PYX_ERR(0, 797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscRowInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort_bufferSizeExt", 1, 6, 6, 5); __PYX_ERR(0, 797, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcscsort_bufferSizeExt") < 0)) __PYX_ERR(0, 797, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_cscColPtr = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_cscColPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_cscRowInd = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_cscRowInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 799, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcscsort_bufferSizeExt", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 797, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcscsort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_92xcscsort_bufferSizeExt(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_cscColPtr, __pyx_v_cscRowInd);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_92xcscsort_bufferSizeExt(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_cscColPtr, size_t __pyx_v_cscRowInd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcscsort_bufferSizeExt", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_f_4cupy_4cuda_8cusparse_xcscsort_bufferSizeExt(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_cscColPtr, __pyx_v_cscRowInd, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcscsort_bufferSizeExt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusparse.pyx":809
 * 
 * 
 * cpdef xcscsort(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t descrA,
 *         size_t cscColPtr, size_t cscRowInd, size_t P, size_t pBuffer):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_95xcscsort(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusparse_xcscsort(size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_cscColPtr, size_t __pyx_v_cscRowInd, size_t __pyx_v_P, size_t __pyx_v_pBuffer, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusparseStatus_t __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcscsort", 0);

  /* "cupy/cuda/cusparse.pyx":812
 *         size_t handle, int m, int n, int nnz, size_t descrA,
 *         size_t cscColPtr, size_t cscRowInd, size_t P, size_t pBuffer):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     status = cusparseXcscsort(
 *         <Handle>handle, m, n, nnz, <const MatDescr>descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":813
 *         size_t cscColPtr, size_t cscRowInd, size_t P, size_t pBuffer):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     status = cusparseXcscsort(             # <<<<<<<<<<<<<<
 *         <Handle>handle, m, n, nnz, <const MatDescr>descrA,
 *         <const int *>cscColPtr, <int *>cscRowInd, <int *>P, <void *>pBuffer)
 */
  __pyx_v_status = cusparseXcscsort(((cusparseHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((int const *)__pyx_v_cscColPtr), ((int *)__pyx_v_cscRowInd), ((int *)__pyx_v_P), ((void *)__pyx_v_pBuffer));

  /* "cupy/cuda/cusparse.pyx":816
 *         <Handle>handle, m, n, nnz, <const MatDescr>descrA,
 *         <const int *>cscColPtr, <int *>cscRowInd, <int *>P, <void *>pBuffer)
 *     check_status(status)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusparse.pyx":809
 * 
 * 
 * cpdef xcscsort(             # <<<<<<<<<<<<<<
 *         size_t handle, int m, int n, int nnz, size_t descrA,
 *         size_t cscColPtr, size_t cscRowInd, size_t P, size_t pBuffer):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcscsort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_95xcscsort(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusparse_94xcscsort[] = "xcscsort(size_t handle, int m, int n, int nnz, size_t descrA, size_t cscColPtr, size_t cscRowInd, size_t P, size_t pBuffer)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusparse_95xcscsort(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_cscColPtr;
  size_t __pyx_v_cscRowInd;
  size_t __pyx_v_P;
  size_t __pyx_v_pBuffer;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xcscsort (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_cscColPtr,&__pyx_n_s_cscRowInd,&__pyx_n_s_P,&__pyx_n_s_pBuffer,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, 1); __PYX_ERR(0, 809, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, 2); __PYX_ERR(0, 809, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, 3); __PYX_ERR(0, 809, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, 4); __PYX_ERR(0, 809, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscColPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, 5); __PYX_ERR(0, 809, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cscRowInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, 6); __PYX_ERR(0, 809, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_P)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, 7); __PYX_ERR(0, 809, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pBuffer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, 8); __PYX_ERR(0, 809, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xcscsort") < 0)) __PYX_ERR(0, 809, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L3_error)
    __pyx_v_cscColPtr = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_cscColPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 811, __pyx_L3_error)
    __pyx_v_cscRowInd = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_cscRowInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 811, __pyx_L3_error)
    __pyx_v_P = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_P == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 811, __pyx_L3_error)
    __pyx_v_pBuffer = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_pBuffer == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 811, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xcscsort", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 809, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcscsort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusparse_94xcscsort(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_cscColPtr, __pyx_v_cscRowInd, __pyx_v_P, __pyx_v_pBuffer);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusparse_94xcscsort(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_cscColPtr, size_t __pyx_v_cscRowInd, size_t __pyx_v_P, size_t __pyx_v_pBuffer) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xcscsort", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusparse_xcscsort(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_cscColPtr, __pyx_v_cscRowInd, __pyx_v_P, __pyx_v_pBuffer, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusparse.xcscsort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {"check_status", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_1check_status, METH_O, __pyx_doc_4cupy_4cuda_8cusparse_check_status},
  {"create", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_3create, METH_NOARGS, __pyx_doc_4cupy_4cuda_8cusparse_2create},
  {"createMatDescr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_5createMatDescr, METH_NOARGS, __pyx_doc_4cupy_4cuda_8cusparse_4createMatDescr},
  {"destroy", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_7destroy, METH_O, __pyx_doc_4cupy_4cuda_8cusparse_6destroy},
  {"destroyMatDescr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_9destroyMatDescr, METH_O, __pyx_doc_4cupy_4cuda_8cusparse_8destroyMatDescr},
  {"setMatIndexBase", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_11setMatIndexBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_10setMatIndexBase},
  {"setMatType", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_13setMatType, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_12setMatType},
  {"setPointerMode", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_15setPointerMode, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_14setPointerMode},
  {"setStream", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_17setStream, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_16setStream},
  {"sgthr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_19sgthr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_18sgthr},
  {"dgthr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_21dgthr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_20dgthr},
  {"scsrmv", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_23scsrmv, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_22scsrmv},
  {"dcsrmv", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_25dcsrmv, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_24dcsrmv},
  {"scsrmm", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_27scsrmm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_26scsrmm},
  {"dcsrmm", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_29dcsrmm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_28dcsrmm},
  {"scsrmm2", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_31scsrmm2, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_30scsrmm2},
  {"dcsrmm2", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_33dcsrmm2, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_32dcsrmm2},
  {"xcsrgeamNnz", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_35xcsrgeamNnz, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_34xcsrgeamNnz},
  {"scsrgeam", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_37scsrgeam, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_36scsrgeam},
  {"dcsrgeam", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_39dcsrgeam, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_38dcsrgeam},
  {"xcsrgemmNnz", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_41xcsrgemmNnz, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_40xcsrgemmNnz},
  {"scsrgemm", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_43scsrgemm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_42scsrgemm},
  {"dcsrgemm", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_45dcsrgemm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_44dcsrgemm},
  {"xcoo2csr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_47xcoo2csr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_46xcoo2csr},
  {"scsc2dense", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_49scsc2dense, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_48scsc2dense},
  {"dcsc2dense", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_51dcsc2dense, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_50dcsc2dense},
  {"xcsr2coo", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_53xcsr2coo, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_52xcsr2coo},
  {"scsr2csc", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_55scsr2csc, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_54scsr2csc},
  {"dcsr2csc", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_57dcsr2csc, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_56dcsr2csc},
  {"scsr2dense", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_59scsr2dense, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_58scsr2dense},
  {"dcsr2dense", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_61dcsr2dense, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_60dcsr2dense},
  {"snnz_compress", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_63snnz_compress, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_62snnz_compress},
  {"dnnz_compress", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_65dnnz_compress, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_64dnnz_compress},
  {"scsr2csr_compress", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_67scsr2csr_compress, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_66scsr2csr_compress},
  {"dcsr2csr_compress", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_69dcsr2csr_compress, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_68dcsr2csr_compress},
  {"sdense2csc", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_71sdense2csc, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_70sdense2csc},
  {"ddense2csc", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_73ddense2csc, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_72ddense2csc},
  {"sdense2csr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_75sdense2csr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_74sdense2csr},
  {"ddense2csr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_77ddense2csr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_76ddense2csr},
  {"snnz", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_79snnz, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_78snnz},
  {"dnnz", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_81dnnz, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_80dnnz},
  {"createIdentityPermutation", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_83createIdentityPermutation, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_82createIdentityPermutation},
  {"xcoosort_bufferSizeExt", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_85xcoosort_bufferSizeExt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_84xcoosort_bufferSizeExt},
  {"xcoosortByRow", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_87xcoosortByRow, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_86xcoosortByRow},
  {"xcsrsort_bufferSizeExt", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_89xcsrsort_bufferSizeExt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_88xcsrsort_bufferSizeExt},
  {"xcsrsort", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_91xcsrsort, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_90xcsrsort},
  {"xcscsort_bufferSizeExt", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_93xcscsort_bufferSizeExt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_92xcscsort_bufferSizeExt},
  {"xcscsort", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusparse_95xcscsort, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusparse_94xcscsort},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "cusparse",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 0, 1, 1},
  {&__pyx_n_s_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 0, 1, 1},
  {&__pyx_n_s_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_ALLOC_FAILED, __pyx_k_CUSPARSE_STATUS_ALLOC_FAILED, sizeof(__pyx_k_CUSPARSE_STATUS_ALLOC_FAILED), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_ARCH_MISMATCH, __pyx_k_CUSPARSE_STATUS_ARCH_MISMATCH, sizeof(__pyx_k_CUSPARSE_STATUS_ARCH_MISMATCH), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_EXECUTION_FAILED, __pyx_k_CUSPARSE_STATUS_EXECUTION_FAILED, sizeof(__pyx_k_CUSPARSE_STATUS_EXECUTION_FAILED), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_INTERNAL_ERROR, __pyx_k_CUSPARSE_STATUS_INTERNAL_ERROR, sizeof(__pyx_k_CUSPARSE_STATUS_INTERNAL_ERROR), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_INVALID_VALUE, __pyx_k_CUSPARSE_STATUS_INVALID_VALUE, sizeof(__pyx_k_CUSPARSE_STATUS_INVALID_VALUE), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_MAPPING_ERROR, __pyx_k_CUSPARSE_STATUS_MAPPING_ERROR, sizeof(__pyx_k_CUSPARSE_STATUS_MAPPING_ERROR), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_MATRIX_TYPE_NOT, __pyx_k_CUSPARSE_STATUS_MATRIX_TYPE_NOT, sizeof(__pyx_k_CUSPARSE_STATUS_MATRIX_TYPE_NOT), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_NOT_INITIALIZED, __pyx_k_CUSPARSE_STATUS_NOT_INITIALIZED, sizeof(__pyx_k_CUSPARSE_STATUS_NOT_INITIALIZED), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_SUCCESS, __pyx_k_CUSPARSE_STATUS_SUCCESS, sizeof(__pyx_k_CUSPARSE_STATUS_SUCCESS), 0, 0, 1, 1},
  {&__pyx_n_s_CUSPARSE_STATUS_ZERO_PIVOT, __pyx_k_CUSPARSE_STATUS_ZERO_PIVOT, sizeof(__pyx_k_CUSPARSE_STATUS_ZERO_PIVOT), 0, 0, 1, 1},
  {&__pyx_n_s_CuSparseError, __pyx_k_CuSparseError, sizeof(__pyx_k_CuSparseError), 0, 0, 1, 1},
  {&__pyx_n_s_CuSparseError___init, __pyx_k_CuSparseError___init, sizeof(__pyx_k_CuSparseError___init), 0, 0, 1, 1},
  {&__pyx_n_s_P, __pyx_k_P, sizeof(__pyx_k_P), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_alpha, __pyx_k_alpha, sizeof(__pyx_k_alpha), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_beta, __pyx_k_beta, sizeof(__pyx_k_beta), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_cooCols, __pyx_k_cooCols, sizeof(__pyx_k_cooCols), 0, 0, 1, 1},
  {&__pyx_n_s_cooRowInd, __pyx_k_cooRowInd, sizeof(__pyx_k_cooRowInd), 0, 0, 1, 1},
  {&__pyx_n_s_cooRows, __pyx_k_cooRows, sizeof(__pyx_k_cooRows), 0, 0, 1, 1},
  {&__pyx_n_s_copyValues, __pyx_k_copyValues, sizeof(__pyx_k_copyValues), 0, 0, 1, 1},
  {&__pyx_n_s_cscColPtr, __pyx_k_cscColPtr, sizeof(__pyx_k_cscColPtr), 0, 0, 1, 1},
  {&__pyx_n_s_cscColPtrA, __pyx_k_cscColPtrA, sizeof(__pyx_k_cscColPtrA), 0, 0, 1, 1},
  {&__pyx_n_s_cscRowInd, __pyx_k_cscRowInd, sizeof(__pyx_k_cscRowInd), 0, 0, 1, 1},
  {&__pyx_n_s_cscRowIndA, __pyx_k_cscRowIndA, sizeof(__pyx_k_cscRowIndA), 0, 0, 1, 1},
  {&__pyx_n_s_cscSortedColPtrA, __pyx_k_cscSortedColPtrA, sizeof(__pyx_k_cscSortedColPtrA), 0, 0, 1, 1},
  {&__pyx_n_s_cscSortedRowIndA, __pyx_k_cscSortedRowIndA, sizeof(__pyx_k_cscSortedRowIndA), 0, 0, 1, 1},
  {&__pyx_n_s_cscSortedValA, __pyx_k_cscSortedValA, sizeof(__pyx_k_cscSortedValA), 0, 0, 1, 1},
  {&__pyx_n_s_cscVal, __pyx_k_cscVal, sizeof(__pyx_k_cscVal), 0, 0, 1, 1},
  {&__pyx_n_s_cscValA, __pyx_k_cscValA, sizeof(__pyx_k_cscValA), 0, 0, 1, 1},
  {&__pyx_n_s_csrColInd, __pyx_k_csrColInd, sizeof(__pyx_k_csrColInd), 0, 0, 1, 1},
  {&__pyx_n_s_csrColIndA, __pyx_k_csrColIndA, sizeof(__pyx_k_csrColIndA), 0, 0, 1, 1},
  {&__pyx_n_s_csrColIndB, __pyx_k_csrColIndB, sizeof(__pyx_k_csrColIndB), 0, 0, 1, 1},
  {&__pyx_n_s_csrColIndC, __pyx_k_csrColIndC, sizeof(__pyx_k_csrColIndC), 0, 0, 1, 1},
  {&__pyx_n_s_csrRowPtr, __pyx_k_csrRowPtr, sizeof(__pyx_k_csrRowPtr), 0, 0, 1, 1},
  {&__pyx_n_s_csrRowPtrA, __pyx_k_csrRowPtrA, sizeof(__pyx_k_csrRowPtrA), 0, 0, 1, 1},
  {&__pyx_n_s_csrRowPtrB, __pyx_k_csrRowPtrB, sizeof(__pyx_k_csrRowPtrB), 0, 0, 1, 1},
  {&__pyx_n_s_csrRowPtrC, __pyx_k_csrRowPtrC, sizeof(__pyx_k_csrRowPtrC), 0, 0, 1, 1},
  {&__pyx_n_s_csrSortedColIndA, __pyx_k_csrSortedColIndA, sizeof(__pyx_k_csrSortedColIndA), 0, 0, 1, 1},
  {&__pyx_n_s_csrSortedRowPtrA, __pyx_k_csrSortedRowPtrA, sizeof(__pyx_k_csrSortedRowPtrA), 0, 0, 1, 1},
  {&__pyx_n_s_csrSortedValA, __pyx_k_csrSortedValA, sizeof(__pyx_k_csrSortedValA), 0, 0, 1, 1},
  {&__pyx_n_s_csrVal, __pyx_k_csrVal, sizeof(__pyx_k_csrVal), 0, 0, 1, 1},
  {&__pyx_n_s_csrValA, __pyx_k_csrValA, sizeof(__pyx_k_csrValA), 0, 0, 1, 1},
  {&__pyx_n_s_csrValB, __pyx_k_csrValB, sizeof(__pyx_k_csrValB), 0, 0, 1, 1},
  {&__pyx_n_s_csrValC, __pyx_k_csrValC, sizeof(__pyx_k_csrValC), 0, 0, 1, 1},
  {&__pyx_n_s_cupy_cuda_cusparse, __pyx_k_cupy_cuda_cusparse, sizeof(__pyx_k_cupy_cuda_cusparse), 0, 0, 1, 1},
  {&__pyx_kp_s_cupy_cuda_cusparse_pyx, __pyx_k_cupy_cuda_cusparse_pyx, sizeof(__pyx_k_cupy_cuda_cusparse_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_descr, __pyx_k_descr, sizeof(__pyx_k_descr), 0, 0, 1, 1},
  {&__pyx_n_s_descrA, __pyx_k_descrA, sizeof(__pyx_k_descrA), 0, 0, 1, 1},
  {&__pyx_n_s_descrB, __pyx_k_descrB, sizeof(__pyx_k_descrB), 0, 0, 1, 1},
  {&__pyx_n_s_descrC, __pyx_k_descrC, sizeof(__pyx_k_descrC), 0, 0, 1, 1},
  {&__pyx_n_s_dirA, __pyx_k_dirA, sizeof(__pyx_k_dirA), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_handle, __pyx_k_handle, sizeof(__pyx_k_handle), 0, 0, 1, 1},
  {&__pyx_n_s_idxBase, __pyx_k_idxBase, sizeof(__pyx_k_idxBase), 0, 0, 1, 1},
  {&__pyx_n_s_inColInd, __pyx_k_inColInd, sizeof(__pyx_k_inColInd), 0, 0, 1, 1},
  {&__pyx_n_s_inNnz, __pyx_k_inNnz, sizeof(__pyx_k_inNnz), 0, 0, 1, 1},
  {&__pyx_n_s_inRowPtr, __pyx_k_inRowPtr, sizeof(__pyx_k_inRowPtr), 0, 0, 1, 1},
  {&__pyx_n_s_inVal, __pyx_k_inVal, sizeof(__pyx_k_inVal), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_lda, __pyx_k_lda, sizeof(__pyx_k_lda), 0, 0, 1, 1},
  {&__pyx_n_s_ldb, __pyx_k_ldb, sizeof(__pyx_k_ldb), 0, 0, 1, 1},
  {&__pyx_n_s_ldc, __pyx_k_ldc, sizeof(__pyx_k_ldc), 0, 0, 1, 1},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_nnz, __pyx_k_nnz, sizeof(__pyx_k_nnz), 0, 0, 1, 1},
  {&__pyx_n_s_nnzA, __pyx_k_nnzA, sizeof(__pyx_k_nnzA), 0, 0, 1, 1},
  {&__pyx_n_s_nnzB, __pyx_k_nnzB, sizeof(__pyx_k_nnzB), 0, 0, 1, 1},
  {&__pyx_n_s_nnzPerCol, __pyx_k_nnzPerCol, sizeof(__pyx_k_nnzPerCol), 0, 0, 1, 1},
  {&__pyx_n_s_nnzPerRow, __pyx_k_nnzPerRow, sizeof(__pyx_k_nnzPerRow), 0, 0, 1, 1},
  {&__pyx_n_s_nnzPerRowColumn, __pyx_k_nnzPerRowColumn, sizeof(__pyx_k_nnzPerRowColumn), 0, 0, 1, 1},
  {&__pyx_n_s_nnzTotalDevHostPtr, __pyx_k_nnzTotalDevHostPtr, sizeof(__pyx_k_nnzTotalDevHostPtr), 0, 0, 1, 1},
  {&__pyx_n_s_outColInd, __pyx_k_outColInd, sizeof(__pyx_k_outColInd), 0, 0, 1, 1},
  {&__pyx_n_s_outRowPtr, __pyx_k_outRowPtr, sizeof(__pyx_k_outRowPtr), 0, 0, 1, 1},
  {&__pyx_n_s_outVal, __pyx_k_outVal, sizeof(__pyx_k_outVal), 0, 0, 1, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_pBuffer, __pyx_k_pBuffer, sizeof(__pyx_k_pBuffer), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_rowPtr, __pyx_k_rowPtr, sizeof(__pyx_k_rowPtr), 0, 0, 1, 1},
  {&__pyx_kp_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_status, __pyx_k_status, sizeof(__pyx_k_status), 0, 0, 1, 1},
  {&__pyx_n_s_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_tol, __pyx_k_tol, sizeof(__pyx_k_tol), 0, 0, 1, 1},
  {&__pyx_n_s_transA, __pyx_k_transA, sizeof(__pyx_k_transA), 0, 0, 1, 1},
  {&__pyx_n_s_transB, __pyx_k_transB, sizeof(__pyx_k_transB), 0, 0, 1, 1},
  {&__pyx_n_s_typ, __pyx_k_typ, sizeof(__pyx_k_typ), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_xInd, __pyx_k_xInd, sizeof(__pyx_k_xInd), 0, 0, 1, 1},
  {&__pyx_n_s_xVal, __pyx_k_xVal, sizeof(__pyx_k_xVal), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 249, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 253, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "cupy/cuda/cusparse.pyx":251
 * class CuSparseError(RuntimeError):
 * 
 *     def __init__(self, int status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         super(CuSparseError, self).__init__('%s' % (STATUS[status]))
 */
  __pyx_tuple_ = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_status); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_codeobj__2 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple_, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cupy_cuda_cusparse_pyx, __pyx_n_s_init, 251, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initcusparse(void); /*proto*/
PyMODINIT_FUNC initcusparse(void)
#else
PyMODINIT_FUNC PyInit_cusparse(void); /*proto*/
PyMODINIT_FUNC PyInit_cusparse(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_cusparse(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cusparse", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_cupy__cuda__cusparse) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cupy.cuda.cusparse")) {
      if (unlikely(PyDict_SetItemString(modules, "cupy.cuda.cusparse", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  __pyx_v_4cupy_4cuda_8cusparse_STATUS = ((PyObject*)Py_None); Py_INCREF(Py_None);
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("create", (void (*)(void))__pyx_f_4cupy_4cuda_8cusparse_create, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroy", (void (*)(void))__pyx_f_4cupy_4cuda_8cusparse_destroy, "void (size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("cupy.cuda.stream"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_current_stream_ptr", (void (**)(void))&__pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr, "size_t (void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "cupy/cuda/cusparse.pyx":236
 * 
 * cdef dict STATUS = {
 *     0: 'CUSPARSE_STATUS_SUCCESS',             # <<<<<<<<<<<<<<
 *     1: 'CUSPARSE_STATUS_NOT_INITIALIZED',
 *     2: 'CUSPARSE_STATUS_ALLOC_FAILED',
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_0, __pyx_n_s_CUSPARSE_STATUS_SUCCESS) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_1, __pyx_n_s_CUSPARSE_STATUS_NOT_INITIALIZED) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_2, __pyx_n_s_CUSPARSE_STATUS_ALLOC_FAILED) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_3, __pyx_n_s_CUSPARSE_STATUS_INVALID_VALUE) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_4, __pyx_n_s_CUSPARSE_STATUS_ARCH_MISMATCH) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_5, __pyx_n_s_CUSPARSE_STATUS_MAPPING_ERROR) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_6, __pyx_n_s_CUSPARSE_STATUS_EXECUTION_FAILED) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_7, __pyx_n_s_CUSPARSE_STATUS_INTERNAL_ERROR) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_8, __pyx_n_s_CUSPARSE_STATUS_MATRIX_TYPE_NOT) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_9, __pyx_n_s_CUSPARSE_STATUS_ZERO_PIVOT) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_4cupy_4cuda_8cusparse_STATUS);
  __Pyx_DECREF_SET(__pyx_v_4cupy_4cuda_8cusparse_STATUS, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "cupy/cuda/cusparse.pyx":249
 * 
 * 
 * class CuSparseError(RuntimeError):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int status):
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_builtin_RuntimeError);
  __Pyx_GIVEREF(__pyx_builtin_RuntimeError);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_builtin_RuntimeError);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_2, __pyx_n_s_CuSparseError, __pyx_n_s_CuSparseError, (PyObject *) NULL, __pyx_n_s_cupy_cuda_cusparse, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "cupy/cuda/cusparse.pyx":251
 * class CuSparseError(RuntimeError):
 * 
 *     def __init__(self, int status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         super(CuSparseError, self).__init__('%s' % (STATUS[status]))
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4cupy_4cuda_8cusparse_13CuSparseError_1__init__, 0, __pyx_n_s_CuSparseError___init, NULL, __pyx_n_s_cupy_cuda_cusparse, __pyx_d, ((PyObject *)__pyx_codeobj__2)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_init, __pyx_t_5) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cupy/cuda/cusparse.pyx":249
 * 
 * 
 * class CuSparseError(RuntimeError):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int status):
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_CuSparseError, __pyx_t_2, __pyx_t_4, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CuSparseError, __pyx_t_5) < 0) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cupy/cuda/cusparse.pyx":1
 * cimport cython             # <<<<<<<<<<<<<<
 * 
 * from cupy.cuda cimport driver
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_ACTION_NUMERIC);
    if (unlikely(!wrapped)) __PYX_ERR(1, 25, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_ACTION_NUMERIC", wrapped) < 0) __PYX_ERR(1, 25, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_ACTION_SYMBOLIC);
    if (unlikely(!wrapped)) __PYX_ERR(1, 24, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_ACTION_SYMBOLIC", wrapped) < 0) __PYX_ERR(1, 24, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_DIRECTION_COLUMN);
    if (unlikely(!wrapped)) __PYX_ERR(1, 40, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_DIRECTION_COLUMN", wrapped) < 0) __PYX_ERR(1, 40, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_DIRECTION_ROW);
    if (unlikely(!wrapped)) __PYX_ERR(1, 39, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_DIRECTION_ROW", wrapped) < 0) __PYX_ERR(1, 39, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_INDEX_BASE_ONE);
    if (unlikely(!wrapped)) __PYX_ERR(1, 28, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_INDEX_BASE_ONE", wrapped) < 0) __PYX_ERR(1, 28, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_INDEX_BASE_ZERO);
    if (unlikely(!wrapped)) __PYX_ERR(1, 27, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_INDEX_BASE_ZERO", wrapped) < 0) __PYX_ERR(1, 27, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_MATRIX_TYPE_GENERAL);
    if (unlikely(!wrapped)) __PYX_ERR(1, 30, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_MATRIX_TYPE_GENERAL", wrapped) < 0) __PYX_ERR(1, 30, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_MATRIX_TYPE_HERMITIAN);
    if (unlikely(!wrapped)) __PYX_ERR(1, 32, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_MATRIX_TYPE_HERMITIAN", wrapped) < 0) __PYX_ERR(1, 32, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_MATRIX_TYPE_SYMMETRIC);
    if (unlikely(!wrapped)) __PYX_ERR(1, 31, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_MATRIX_TYPE_SYMMETRIC", wrapped) < 0) __PYX_ERR(1, 31, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_MATRIX_TYPE_TRIANGULAR);
    if (unlikely(!wrapped)) __PYX_ERR(1, 33, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_MATRIX_TYPE_TRIANGULAR", wrapped) < 0) __PYX_ERR(1, 33, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE);
    if (unlikely(!wrapped)) __PYX_ERR(1, 37, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE", wrapped) < 0) __PYX_ERR(1, 37, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_OPERATION_NON_TRANSPOSE);
    if (unlikely(!wrapped)) __PYX_ERR(1, 35, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_OPERATION_NON_TRANSPOSE", wrapped) < 0) __PYX_ERR(1, 35, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_OPERATION_TRANSPOSE);
    if (unlikely(!wrapped)) __PYX_ERR(1, 36, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_OPERATION_TRANSPOSE", wrapped) < 0) __PYX_ERR(1, 36, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_POINTER_MODE_DEVICE);
    if (unlikely(!wrapped)) __PYX_ERR(1, 22, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_POINTER_MODE_DEVICE", wrapped) < 0) __PYX_ERR(1, 22, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusparse_CUSPARSE_POINTER_MODE_HOST);
    if (unlikely(!wrapped)) __PYX_ERR(1, 21, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSPARSE_POINTER_MODE_HOST", wrapped) < 0) __PYX_ERR(1, 21, __pyx_L1_error);
  }

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init cupy.cuda.cusparse", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cupy.cuda.cusparse");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
    static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* CalculateMetaclass */
    static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* FetchCommonType */
    static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
    static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* Py3ClassCreate */
        static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
        static int __Pyx_CLineForTraceback(int c_line) {
#ifdef CYTHON_CLINE_IN_TRACEBACK
    return ((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0;
#else
    PyObject *use_cline;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = PyDict_GetItem(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *ptype, *pvalue, *ptraceback;
      PyObject *use_cline_obj;
      PyErr_Fetch(&ptype, &pvalue, &ptraceback);
      use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        use_cline = NULL;
      }
      PyErr_Restore(ptype, pvalue, ptraceback);
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    return c_line;
#endif
}

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
        #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
        static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE cusparseIndexBase_t __Pyx_PyInt_As_cusparseIndexBase_t(PyObject *x) {
    const cusparseIndexBase_t neg_one = (cusparseIndexBase_t) -1, const_zero = (cusparseIndexBase_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(cusparseIndexBase_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (cusparseIndexBase_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (cusparseIndexBase_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(cusparseIndexBase_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) >= 2 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) (((((cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(cusparseIndexBase_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) >= 3 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) (((((((cusparseIndexBase_t)digits[2]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(cusparseIndexBase_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) >= 4 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) (((((((((cusparseIndexBase_t)digits[3]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[2]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (cusparseIndexBase_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(cusparseIndexBase_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(cusparseIndexBase_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(cusparseIndexBase_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(cusparseIndexBase_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (cusparseIndexBase_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(cusparseIndexBase_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) - 1 > 2 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) (((cusparseIndexBase_t)-1)*(((((cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(cusparseIndexBase_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) - 1 > 2 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) ((((((cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(cusparseIndexBase_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) - 1 > 3 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) (((cusparseIndexBase_t)-1)*(((((((cusparseIndexBase_t)digits[2]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(cusparseIndexBase_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) - 1 > 3 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) ((((((((cusparseIndexBase_t)digits[2]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(cusparseIndexBase_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) - 1 > 4 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) (((cusparseIndexBase_t)-1)*(((((((((cusparseIndexBase_t)digits[3]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[2]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(cusparseIndexBase_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseIndexBase_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseIndexBase_t) - 1 > 4 * PyLong_SHIFT) {
                            return (cusparseIndexBase_t) ((((((((((cusparseIndexBase_t)digits[3]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[2]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[1]) << PyLong_SHIFT) | (cusparseIndexBase_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(cusparseIndexBase_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(cusparseIndexBase_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(cusparseIndexBase_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(cusparseIndexBase_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            cusparseIndexBase_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (cusparseIndexBase_t) -1;
        }
    } else {
        cusparseIndexBase_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (cusparseIndexBase_t) -1;
        val = __Pyx_PyInt_As_cusparseIndexBase_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to cusparseIndexBase_t");
    return (cusparseIndexBase_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to cusparseIndexBase_t");
    return (cusparseIndexBase_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE cusparseMatrixType_t __Pyx_PyInt_As_cusparseMatrixType_t(PyObject *x) {
    const cusparseMatrixType_t neg_one = (cusparseMatrixType_t) -1, const_zero = (cusparseMatrixType_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(cusparseMatrixType_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (cusparseMatrixType_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (cusparseMatrixType_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(cusparseMatrixType_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) >= 2 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) (((((cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(cusparseMatrixType_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) >= 3 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) (((((((cusparseMatrixType_t)digits[2]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(cusparseMatrixType_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) >= 4 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) (((((((((cusparseMatrixType_t)digits[3]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[2]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (cusparseMatrixType_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(cusparseMatrixType_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(cusparseMatrixType_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(cusparseMatrixType_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(cusparseMatrixType_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (cusparseMatrixType_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(cusparseMatrixType_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) - 1 > 2 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) (((cusparseMatrixType_t)-1)*(((((cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(cusparseMatrixType_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) - 1 > 2 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) ((((((cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(cusparseMatrixType_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) - 1 > 3 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) (((cusparseMatrixType_t)-1)*(((((((cusparseMatrixType_t)digits[2]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(cusparseMatrixType_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) - 1 > 3 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) ((((((((cusparseMatrixType_t)digits[2]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(cusparseMatrixType_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) - 1 > 4 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) (((cusparseMatrixType_t)-1)*(((((((((cusparseMatrixType_t)digits[3]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[2]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(cusparseMatrixType_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(cusparseMatrixType_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(cusparseMatrixType_t) - 1 > 4 * PyLong_SHIFT) {
                            return (cusparseMatrixType_t) ((((((((((cusparseMatrixType_t)digits[3]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[2]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[1]) << PyLong_SHIFT) | (cusparseMatrixType_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(cusparseMatrixType_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(cusparseMatrixType_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(cusparseMatrixType_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(cusparseMatrixType_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            cusparseMatrixType_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (cusparseMatrixType_t) -1;
        }
    } else {
        cusparseMatrixType_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (cusparseMatrixType_t) -1;
        val = __Pyx_PyInt_As_cusparseMatrixType_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to cusparseMatrixType_t");
    return (cusparseMatrixType_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to cusparseMatrixType_t");
    return (cusparseMatrixType_t) -1;
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
        static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
        static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
        static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
        #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* FunctionImport */
        #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
        static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            PyErr_Clear();
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
