/* Generated by Cython 0.26.1 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "_GLIBCXX_USE_CXX11_ABI",
                "0"
            ],
            [
                "_FORCE_INLINES",
                "1"
            ],
            [
                "CUPY_NO_CUDA",
                "1"
            ]
        ],
        "depends": [
            "cupy/cuda/cupy_cudnn.h"
        ],
        "include_dirs": [
            "cupy/cuda",
            "/usr/local/cuda/include"
        ],
        "language": "c++",
        "library_dirs": [
            "/usr/local/cuda/lib64"
        ],
        "name": "cupy.cuda.cudnn",
        "runtime_library_dirs": [
            "/usr/local/cuda/lib64"
        ],
        "sources": [
            "cupy/cuda/cudnn.pyx"
        ]
    },
    "module_name": "cupy.cuda.cudnn"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_26_1"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX < 0x030700A0 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject **args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #ifdef __cplusplus
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough) || (defined(__GNUC__) && defined(__attribute__))
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cupy__cuda__cudnn
#define __PYX_HAVE_API__cupy__cuda__cudnn
#include <vector>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include "cupy_cudnn.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "cupy/cuda/cudnn.pyx",
  "stringsource",
  "cupy/cuda/cudnn.pxd",
};

/*--- Type declarations ---*/

/* "driver.pxd":20
 *     ctypedef int CUjitInputType 'CUjitInputType'
 * 
 * cpdef enum:             # <<<<<<<<<<<<<<
 *     CU_JIT_INPUT_CUBIN = 0
 *     CU_JIT_INPUT_PTX = 1
 */
enum  {
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_CUBIN = 0,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_PTX = 1,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_FATBINARY = 2,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_OBJECT = 3,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_LIBRARY = 4
};

/* "cupy/cuda/cudnn.pxd":6
 * ###############################################################################
 * 
 * cpdef enum:             # <<<<<<<<<<<<<<
 *     CUDNN_DATA_FLOAT = 0
 *     CUDNN_DATA_DOUBLE = 1
 */
enum  {
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_DATA_FLOAT = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_DATA_DOUBLE = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_DATA_HALF = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_DEFAULT_MATH = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_TENSOR_OP_MATH = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_NOT_PROPAGATE_NAN = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_PROPAGATE_NAN = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_TENSOR_NCHW = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_TENSOR_NHWC = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_IMAGE = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_SAME_HW = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_FEATURE_MAP = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_SAME_CHW = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_SAME_C = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_FULL_TENSOR = 3,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CROSS_CORRELATION = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_NO_WORKSPACE = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_PREFER_FASTEST = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_SPECIFY_WORKSPACE_LIMIT = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_GEMM = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_GEMM = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_DIRECT = 3,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_FFT = 4,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_FFT_TILING = 5,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD = 6,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD_NONFUSED = 7,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_NO_WORKSPACE = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_PREFER_FASTEST = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_SPECIFY_WORKSPACE_LIMIT = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_0 = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_1 = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_FFT = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_3 = 3,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_WINOGRAD = 4,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_WINOGRAD_NONFUSED = 5,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_NO_WORKSPACE = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_PREFER_FASTEST = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_SPECIFY_WORKSPACE_LIMIT = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_0 = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_1 = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_FFT = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_FFT_TILING = 3,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_WINOGRAD = 4,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_WINOGRAD_NONFUSED = 5,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_FAST = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_ACCURATE = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_LOG = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_MODE_INSTANCE = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_MODE_CHANNEL = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_POOLING_MAX = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_POOLING_AVERAGE_COUNT_INCLUDE_PADDING = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_POOLING_AVERAGE_COUNT_EXCLUDE_PADDING = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ACTIVATION_SIGMOID = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ACTIVATION_RELU = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ACTIVATION_TANH = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_ACTIVATION_CLIPPED_RELU = 3,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_LRN_CROSS_CHANNEL_DIM1 = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_DIVNORM_PRECOMPUTED_MEANS = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_BATCHNORM_PER_ACTIVATION = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_BATCHNORM_SPATIAL = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_RELU = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_TANH = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_LSTM = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_GRU = 3,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_UNIDIRECTIONAL = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_BIDIRECTIONAL = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_ALGO_STANDARD = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_ALGO_PERSIST_STATIC = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_ALGO_PERSIST_DYNAMIC = 2,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_LINEAR_INPUT = 0,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_SKIP_INPUT = 1,
  __pyx_e_4cupy_4cuda_5cudnn_CUDNN_SAMPLER_BILINEAR = 0
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#else
    #define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
static int __Pyx_CLineForTraceback(int c_line);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_cudnnDataType_t(cudnnDataType_t value);

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_cudnnTensorFormat_t(cudnnTensorFormat_t value);

static PyObject* __pyx_convert__to_py_cudnnConvolutionFwdAlgoPerf_t(cudnnConvolutionFwdAlgoPerf_t s);
static PyObject* __pyx_convert__to_py_cudnnConvolutionBwdFilterAlgoPerf_t(cudnnConvolutionBwdFilterAlgoPerf_t s);
static PyObject* __pyx_convert__to_py_cudnnConvolutionBwdDataAlgoPerf_t(cudnnConvolutionBwdDataAlgoPerf_t s);
/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_As_unsigned_PY_LONG_LONG(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cython' */

/* Module declarations from 'libcpp' */

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'cupy.cuda' */

/* Module declarations from 'cupy.cuda.driver' */

/* Module declarations from 'cupy.cuda.stream' */
static size_t (*__pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr)(void); /*proto*/

/* Module declarations from 'cupy.cuda.cudnn' */
static PyObject *__pyx_v_4cupy_4cuda_5cudnn_STATUS = 0;
static size_t __pyx_f_4cupy_4cuda_5cudnn_getVersion(int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_create(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroy(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setStream(size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getStream(size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createTensorDescriptor(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setTensor4dDescriptor(size_t, int, int, int, int, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setTensor4dDescriptorEx(size_t, int, int, int, int, int, int, int, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setTensorNdDescriptor(size_t, int, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyTensorDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_addTensor_v3(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createFilterDescriptor(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setFilter4dDescriptor_v4(size_t, int, int, int, int, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setFilterNdDescriptor_v4(size_t, int, int, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getFilterNdDescriptor(size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyFilterDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createConvolutionDescriptor(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolutionMathType(size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getConvolutionMathType(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolutionGroupCount(size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_5cudnn_getConvolutionGroupCount(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolution2dDescriptor_v4(size_t, int, int, int, int, int, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolution2dDescriptor_v5(size_t, int, int, int, int, int, int, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolutionNdDescriptor_v3(size_t, int, size_t, size_t, size_t, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyConvolutionDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionForwardAlgorithm(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionForwardAlgorithmEx(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_5cudnn_getConvolutionForwardAlgorithm(size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getConvolutionForwardWorkspaceSize(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_convolutionForward(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardBias(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardFilterAlgorithm(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardFilterAlgorithmEx(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardFilterAlgorithm(size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardFilterWorkspaceSize(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardFilter_v3(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardDataAlgorithm(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardDataAlgorithmEx(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardDataAlgorithm(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardDataWorkspaceSize(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardData_v3(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createPoolingDescriptor(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setPooling2dDescriptor_v4(size_t, int, int, int, int, int, int, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setPoolingNdDescriptor_v4(size_t, int, int, int, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyPoolingDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_poolingForward(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_poolingBackward(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_deriveBNTensorDescriptor(size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationForwardTraining(size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, double, size_t, size_t, double, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationForwardInference(size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, double, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationBackward(size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, double, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createActivationDescriptor(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setActivationDescriptor(size_t, int, int, double, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyActivationDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_softmaxForward(size_t, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_softmaxBackward(size_t, int, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_activationForward_v4(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_activationBackward_v4(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createDropoutDescriptor(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyDropoutDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_dropoutGetStatesSize(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setDropoutDescriptor(size_t, size_t, float, size_t, size_t, unsigned PY_LONG_LONG, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getDropoutReserveSpaceSize(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_dropoutForward(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_dropoutBackward(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createRNNDescriptor(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyRNNDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createPersistentRNNPlan(size_t, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setPersistentRNNPlan(size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyPersistentRNNPlan(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setRNNDescriptor_v5(size_t, int, int, size_t, int, int, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setRNNDescriptor_v6(size_t, size_t, int, int, size_t, int, int, int, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNWorkspaceSize(size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNTrainingReserveSize(size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNParamsSize(size_t, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNLinLayerMatrixParams(size_t, size_t, int, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNLinLayerBiasParams(size_t, size_t, int, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_RNNForwardInference(size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_RNNForwardTraining(size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_RNNBackwardData(size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_RNNBackwardWeights(size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createSpatialTransformerDescriptor(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroySpatialTransformerDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setSpatialTransformerDescriptor(size_t, size_t, int, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_spatialTfGridGeneratorForward(size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_spatialTfGridGeneratorBackward(size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_spatialTfSamplerForward(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_spatialTfSamplerBackward(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4cupy_4cuda_5cudnn_check_status(int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getTensor4dDescriptor(size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_convert_vector_to_py_int(const std::vector<int>  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t(const std::vector<cudnnConvolutionFwdAlgoPerf_t>  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t(const std::vector<cudnnConvolutionBwdFilterAlgoPerf_t>  &); /*proto*/
static PyObject *__pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t(const std::vector<cudnnConvolutionBwdDataAlgoPerf_t>  &); /*proto*/
#define __Pyx_MODULE_NAME "cupy.cuda.cudnn"
int __pyx_module_is_main_cupy__cuda__cudnn = 0;

/* Implementation of 'cupy.cuda.cudnn' */
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_range;
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_u[] = "u";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_cx[] = "cx";
static const char __pyx_k_cy[] = "cy";
static const char __pyx_k_dw[] = "dw";
static const char __pyx_k_dx[] = "dx";
static const char __pyx_k_dy[] = "dy";
static const char __pyx_k_hx[] = "hx";
static const char __pyx_k_hy[] = "hy";
static const char __pyx_k_dcx[] = "dcx";
static const char __pyx_k_dcy[] = "dcy";
static const char __pyx_k_dhx[] = "dhx";
static const char __pyx_k_dhy[] = "dhy";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_s_s[] = "%s: %s";
static const char __pyx_k_algo[] = "algo";
static const char __pyx_k_beta[] = "beta";
static const char __pyx_k_dimA[] = "dimA";
static const char __pyx_k_grid[] = "grid";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_padA[] = "padA";
static const char __pyx_k_plan[] = "plan";
static const char __pyx_k_seed[] = "seed";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_alpha[] = "alpha";
static const char __pyx_k_bDesc[] = "bDesc";
static const char __pyx_k_dgrid[] = "dgrid";
static const char __pyx_k_layer[] = "layer";
static const char __pyx_k_pad_h[] = "pad_h";
static const char __pyx_k_pad_w[] = "pad_w";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_theta[] = "theta";
static const char __pyx_k_wDesc[] = "wDesc";
static const char __pyx_k_xDesc[] = "xDesc";
static const char __pyx_k_yDesc[] = "yDesc";
static const char __pyx_k_bnBias[] = "bnBias";
static const char __pyx_k_cxDesc[] = "cxDesc";
static const char __pyx_k_cyDesc[] = "cyDesc";
static const char __pyx_k_dtheta[] = "dtheta";
static const char __pyx_k_dwDesc[] = "dwDesc";
static const char __pyx_k_dxData[] = "dxData";
static const char __pyx_k_dxDesc[] = "dxDesc";
static const char __pyx_k_dyData[] = "dyData";
static const char __pyx_k_dyDesc[] = "dyDesc";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_handle[] = "handle";
static const char __pyx_k_hxDesc[] = "hxDesc";
static const char __pyx_k_hyDesc[] = "hyDesc";
static const char __pyx_k_memory[] = "memory";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_nbDims[] = "nbDims";
static const char __pyx_k_stDesc[] = "stDesc";
static const char __pyx_k_states[] = "states";
static const char __pyx_k_status[] = "status";
static const char __pyx_k_stream[] = "stream";
static const char __pyx_k_bnScale[] = "bnScale";
static const char __pyx_k_cStride[] = "cStride";
static const char __pyx_k_dcxDesc[] = "dcxDesc";
static const char __pyx_k_dcyDesc[] = "dcyDesc";
static const char __pyx_k_dhxDesc[] = "dhxDesc";
static const char __pyx_k_dhyDesc[] = "dhyDesc";
static const char __pyx_k_dropout[] = "dropout";
static const char __pyx_k_dstData[] = "dstData";
static const char __pyx_k_dstDesc[] = "dstDesc";
static const char __pyx_k_epsilon[] = "epsilon";
static const char __pyx_k_hStride[] = "hStride";
static const char __pyx_k_nStride[] = "nStride";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_rnnDesc[] = "rnnDesc";
static const char __pyx_k_srcData[] = "srcData";
static const char __pyx_k_srcDesc[] = "srcDesc";
static const char __pyx_k_strideA[] = "strideA";
static const char __pyx_k_wStride[] = "wStride";
static const char __pyx_k_convDesc[] = "convDesc";
static const char __pyx_k_dataType[] = "dataType";
static const char __pyx_k_destData[] = "destData";
static const char __pyx_k_destDesc[] = "destDesc";
static const char __pyx_k_diffData[] = "diffData";
static const char __pyx_k_diffDesc[] = "diffDesc";
static const char __pyx_k_gradData[] = "gradData";
static const char __pyx_k_gradDesc[] = "gradDesc";
static const char __pyx_k_mathType[] = "mathType";
static const char __pyx_k_paddingA[] = "paddingA";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_algorithm[] = "algorithm";
static const char __pyx_k_betaDgrid[] = "betaDgrid";
static const char __pyx_k_dilationA[] = "dilationA";
static const char __pyx_k_direction[] = "direction";
static const char __pyx_k_inputMode[] = "inputMode";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_minibatch[] = "minibatch";
static const char __pyx_k_numLayers[] = "numLayers";
static const char __pyx_k_savedMean[] = "savedMean";
static const char __pyx_k_seqLength[] = "seqLength";
static const char __pyx_k_workSpace[] = "workSpace";
static const char __pyx_k_workspace[] = "workspace";
static const char __pyx_k_CuDNNError[] = "CuDNNError";
static const char __pyx_k_alphaDgrid[] = "alphaDgrid";
static const char __pyx_k_dilation_h[] = "dilation_h";
static const char __pyx_k_dilation_w[] = "dilation_w";
static const char __pyx_k_filterData[] = "filterData";
static const char __pyx_k_filterDesc[] = "filterDesc";
static const char __pyx_k_filterDimA[] = "filterDimA";
static const char __pyx_k_groupCount[] = "groupCount";
static const char __pyx_k_hiddenSize[] = "hiddenSize";
static const char __pyx_k_linLayerID[] = "linLayerID";
static const char __pyx_k_preference[] = "preference";
static const char __pyx_k_reluNanOpt[] = "reluNanOpt";
static const char __pyx_k_tensorDesc[] = "tensorDesc";
static const char __pyx_k_windowDimA[] = "windowDimA";
static const char __pyx_k_arrayLength[] = "arrayLength";
static const char __pyx_k_computeType[] = "computeType";
static const char __pyx_k_dropoutDesc[] = "dropoutDesc";
static const char __pyx_k_linLayerMat[] = "linLayerMat";
static const char __pyx_k_poolingDesc[] = "poolingDesc";
static const char __pyx_k_reluCeiling[] = "reluCeiling";
static const char __pyx_k_samplerType[] = "samplerType";
static const char __pyx_k_srcDiffData[] = "srcDiffData";
static const char __pyx_k_srcDiffDesc[] = "srcDiffDesc";
static const char __pyx_k_windowWidth[] = "windowWidth";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_betaDataDiff[] = "betaDataDiff";
static const char __pyx_k_destDiffData[] = "destDiffData";
static const char __pyx_k_destDiffDesc[] = "destDiffDesc";
static const char __pyx_k_linLayerBias[] = "linLayerBias";
static const char __pyx_k_reserveSpace[] = "reserveSpace";
static const char __pyx_k_windowHeight[] = "windowHeight";
static const char __pyx_k_alphaDataDiff[] = "alphaDataDiff";
static const char __pyx_k_betaParamDiff[] = "betaParamDiff";
static const char __pyx_k_dBnBiasResult[] = "dBnBiasResult";
static const char __pyx_k_derivedBnDesc[] = "derivedBnDesc";
static const char __pyx_k_estimatedMean[] = "estimatedMean";
static const char __pyx_k_filterStrideA[] = "filterStrideA";
static const char __pyx_k_activationDesc[] = "activationDesc";
static const char __pyx_k_alphaParamDiff[] = "alphaParamDiff";
static const char __pyx_k_dBnScaleResult[] = "dBnScaleResult";
static const char __pyx_k_resultSaveMean[] = "resultSaveMean";
static const char __pyx_k_verticalStride[] = "verticalStride";
static const char __pyx_k_cupy_cuda_cudnn[] = "cupy.cuda.cudnn";
static const char __pyx_k_linLayerMatDesc[] = "linLayerMatDesc";
static const char __pyx_k_nbDimsRequested[] = "nbDimsRequested";
static const char __pyx_k_verticalPadding[] = "verticalPadding";
static const char __pyx_k_dBnScaleBiasDesc[] = "dBnScaleBiasDesc";
static const char __pyx_k_horizontalStride[] = "horizontalStride";
static const char __pyx_k_linLayerBiasDesc[] = "linLayerBiasDesc";
static const char __pyx_k_maxpoolingNanOpt[] = "maxpoolingNanOpt";
static const char __pyx_k_savedInvVariance[] = "savedInvVariance";
static const char __pyx_k_stateSizeInBytes[] = "stateSizeInBytes";
static const char __pyx_k_CuDNNError___init[] = "CuDNNError.__init__";
static const char __pyx_k_estimatedVariance[] = "estimatedVariance";
static const char __pyx_k_horizontalPadding[] = "horizontalPadding";
static const char __pyx_k_resultRunningMean[] = "resultRunningMean";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_memoryLimitInbytes[] = "memoryLimitInbytes";
static const char __pyx_k_requestedAlgoCount[] = "requestedAlgoCount";
static const char __pyx_k_cupy_cuda_cudnn_pyx[] = "cupy/cuda/cudnn.pyx";
static const char __pyx_k_CUDNN_STATUS_SUCCESS[] = "CUDNN_STATUS_SUCCESS";
static const char __pyx_k_workSpaceSizeInBytes[] = "workSpaceSizeInBytes";
static const char __pyx_k_Thin_wrapper_of_cuDNN[] = "Thin wrapper of cuDNN.";
static const char __pyx_k_resultRunningVariance[] = "resultRunningVariance";
static const char __pyx_k_resultSaveInvVariance[] = "resultSaveInvVariance";
static const char __pyx_k_CUDNN_STATUS_BAD_PARAM[] = "CUDNN_STATUS_BAD_PARAM";
static const char __pyx_k_bnScaleBiasMeanVarDesc[] = "bnScaleBiasMeanVarDesc";
static const char __pyx_k_reserveSpaceSizeInBytes[] = "reserveSpaceSizeInBytes";
static const char __pyx_k_exponentialAverageFactor[] = "exponentialAverageFactor";
static const char __pyx_k_CUDNN_STATUS_ALLOC_FAILED[] = "CUDNN_STATUS_ALLOC_FAILED";
static const char __pyx_k_CUDNN_STATUS_ARCH_MISMATCH[] = "CUDNN_STATUS_ARCH_MISMATCH";
static const char __pyx_k_CUDNN_STATUS_INVALID_VALUE[] = "CUDNN_STATUS_INVALID_VALUE";
static const char __pyx_k_CUDNN_STATUS_LICENSE_ERROR[] = "CUDNN_STATUS_LICENSE_ERROR";
static const char __pyx_k_CUDNN_STATUS_MAPPING_ERROR[] = "CUDNN_STATUS_MAPPING_ERROR";
static const char __pyx_k_CUDNN_STATUS_NOT_SUPPORTED[] = "CUDNN_STATUS_NOT_SUPPORTED";
static const char __pyx_k_CUDNN_STATUS_INTERNAL_ERROR[] = "CUDNN_STATUS_INTERNAL_ERROR";
static const char __pyx_k_CUDNN_STATUS_NOT_INITIALIZED[] = "CUDNN_STATUS_NOT_INITIALIZED";
static const char __pyx_k_CUDNN_STATUS_EXECUTION_FAILED[] = "CUDNN_STATUS_EXECUTION_FAILED";
static PyObject *__pyx_n_s_CUDNN_STATUS_ALLOC_FAILED;
static PyObject *__pyx_n_s_CUDNN_STATUS_ARCH_MISMATCH;
static PyObject *__pyx_n_s_CUDNN_STATUS_BAD_PARAM;
static PyObject *__pyx_n_s_CUDNN_STATUS_EXECUTION_FAILED;
static PyObject *__pyx_n_s_CUDNN_STATUS_INTERNAL_ERROR;
static PyObject *__pyx_n_s_CUDNN_STATUS_INVALID_VALUE;
static PyObject *__pyx_n_s_CUDNN_STATUS_LICENSE_ERROR;
static PyObject *__pyx_n_s_CUDNN_STATUS_MAPPING_ERROR;
static PyObject *__pyx_n_s_CUDNN_STATUS_NOT_INITIALIZED;
static PyObject *__pyx_n_s_CUDNN_STATUS_NOT_SUPPORTED;
static PyObject *__pyx_n_s_CUDNN_STATUS_SUCCESS;
static PyObject *__pyx_n_s_CuDNNError;
static PyObject *__pyx_n_s_CuDNNError___init;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_activationDesc;
static PyObject *__pyx_n_s_algo;
static PyObject *__pyx_n_s_algorithm;
static PyObject *__pyx_n_s_alpha;
static PyObject *__pyx_n_s_alphaDataDiff;
static PyObject *__pyx_n_s_alphaDgrid;
static PyObject *__pyx_n_s_alphaParamDiff;
static PyObject *__pyx_n_s_arrayLength;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_bDesc;
static PyObject *__pyx_n_s_beta;
static PyObject *__pyx_n_s_betaDataDiff;
static PyObject *__pyx_n_s_betaDgrid;
static PyObject *__pyx_n_s_betaParamDiff;
static PyObject *__pyx_n_s_bnBias;
static PyObject *__pyx_n_s_bnScale;
static PyObject *__pyx_n_s_bnScaleBiasMeanVarDesc;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_cStride;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_computeType;
static PyObject *__pyx_n_s_convDesc;
static PyObject *__pyx_n_s_cupy_cuda_cudnn;
static PyObject *__pyx_kp_s_cupy_cuda_cudnn_pyx;
static PyObject *__pyx_n_s_cx;
static PyObject *__pyx_n_s_cxDesc;
static PyObject *__pyx_n_s_cy;
static PyObject *__pyx_n_s_cyDesc;
static PyObject *__pyx_n_s_dBnBiasResult;
static PyObject *__pyx_n_s_dBnScaleBiasDesc;
static PyObject *__pyx_n_s_dBnScaleResult;
static PyObject *__pyx_n_s_dataType;
static PyObject *__pyx_n_s_dcx;
static PyObject *__pyx_n_s_dcxDesc;
static PyObject *__pyx_n_s_dcy;
static PyObject *__pyx_n_s_dcyDesc;
static PyObject *__pyx_n_s_derivedBnDesc;
static PyObject *__pyx_n_s_destData;
static PyObject *__pyx_n_s_destDesc;
static PyObject *__pyx_n_s_destDiffData;
static PyObject *__pyx_n_s_destDiffDesc;
static PyObject *__pyx_n_s_dgrid;
static PyObject *__pyx_n_s_dhx;
static PyObject *__pyx_n_s_dhxDesc;
static PyObject *__pyx_n_s_dhy;
static PyObject *__pyx_n_s_dhyDesc;
static PyObject *__pyx_n_s_diffData;
static PyObject *__pyx_n_s_diffDesc;
static PyObject *__pyx_n_s_dilationA;
static PyObject *__pyx_n_s_dilation_h;
static PyObject *__pyx_n_s_dilation_w;
static PyObject *__pyx_n_s_dimA;
static PyObject *__pyx_n_s_direction;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dropout;
static PyObject *__pyx_n_s_dropoutDesc;
static PyObject *__pyx_n_s_dstData;
static PyObject *__pyx_n_s_dstDesc;
static PyObject *__pyx_n_s_dtheta;
static PyObject *__pyx_n_s_dw;
static PyObject *__pyx_n_s_dwDesc;
static PyObject *__pyx_n_s_dx;
static PyObject *__pyx_n_s_dxData;
static PyObject *__pyx_n_s_dxDesc;
static PyObject *__pyx_n_s_dy;
static PyObject *__pyx_n_s_dyData;
static PyObject *__pyx_n_s_dyDesc;
static PyObject *__pyx_n_s_epsilon;
static PyObject *__pyx_n_s_estimatedMean;
static PyObject *__pyx_n_s_estimatedVariance;
static PyObject *__pyx_n_s_exponentialAverageFactor;
static PyObject *__pyx_n_s_filterData;
static PyObject *__pyx_n_s_filterDesc;
static PyObject *__pyx_n_s_filterDimA;
static PyObject *__pyx_n_s_filterStrideA;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_gradData;
static PyObject *__pyx_n_s_gradDesc;
static PyObject *__pyx_n_s_grid;
static PyObject *__pyx_n_s_groupCount;
static PyObject *__pyx_n_s_h;
static PyObject *__pyx_n_s_hStride;
static PyObject *__pyx_n_s_handle;
static PyObject *__pyx_n_s_hiddenSize;
static PyObject *__pyx_n_s_horizontalPadding;
static PyObject *__pyx_n_s_horizontalStride;
static PyObject *__pyx_n_s_hx;
static PyObject *__pyx_n_s_hxDesc;
static PyObject *__pyx_n_s_hy;
static PyObject *__pyx_n_s_hyDesc;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_inputMode;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_layer;
static PyObject *__pyx_n_s_linLayerBias;
static PyObject *__pyx_n_s_linLayerBiasDesc;
static PyObject *__pyx_n_s_linLayerID;
static PyObject *__pyx_n_s_linLayerMat;
static PyObject *__pyx_n_s_linLayerMatDesc;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mathType;
static PyObject *__pyx_n_s_maxpoolingNanOpt;
static PyObject *__pyx_n_s_memory;
static PyObject *__pyx_n_s_memoryLimitInbytes;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_minibatch;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_msg;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_nStride;
static PyObject *__pyx_n_s_nbDims;
static PyObject *__pyx_n_s_nbDimsRequested;
static PyObject *__pyx_n_s_numLayers;
static PyObject *__pyx_n_s_padA;
static PyObject *__pyx_n_s_pad_h;
static PyObject *__pyx_n_s_pad_w;
static PyObject *__pyx_n_s_paddingA;
static PyObject *__pyx_n_s_plan;
static PyObject *__pyx_n_s_poolingDesc;
static PyObject *__pyx_n_s_preference;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reluCeiling;
static PyObject *__pyx_n_s_reluNanOpt;
static PyObject *__pyx_n_s_requestedAlgoCount;
static PyObject *__pyx_n_s_reserveSpace;
static PyObject *__pyx_n_s_reserveSpaceSizeInBytes;
static PyObject *__pyx_n_s_resultRunningMean;
static PyObject *__pyx_n_s_resultRunningVariance;
static PyObject *__pyx_n_s_resultSaveInvVariance;
static PyObject *__pyx_n_s_resultSaveMean;
static PyObject *__pyx_n_s_rnnDesc;
static PyObject *__pyx_kp_s_s_s;
static PyObject *__pyx_n_s_samplerType;
static PyObject *__pyx_n_s_savedInvVariance;
static PyObject *__pyx_n_s_savedMean;
static PyObject *__pyx_n_s_seed;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_seqLength;
static PyObject *__pyx_n_s_srcData;
static PyObject *__pyx_n_s_srcDesc;
static PyObject *__pyx_n_s_srcDiffData;
static PyObject *__pyx_n_s_srcDiffDesc;
static PyObject *__pyx_n_s_stDesc;
static PyObject *__pyx_n_s_stateSizeInBytes;
static PyObject *__pyx_n_s_states;
static PyObject *__pyx_n_s_status;
static PyObject *__pyx_n_s_stream;
static PyObject *__pyx_n_s_strideA;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_tensorDesc;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theta;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_u;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_verticalPadding;
static PyObject *__pyx_n_s_verticalStride;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_s_wDesc;
static PyObject *__pyx_n_s_wStride;
static PyObject *__pyx_n_s_windowDimA;
static PyObject *__pyx_n_s_windowHeight;
static PyObject *__pyx_n_s_windowWidth;
static PyObject *__pyx_n_s_workSpace;
static PyObject *__pyx_n_s_workSpaceSizeInBytes;
static PyObject *__pyx_n_s_workspace;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_xDesc;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_yDesc;
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_10CuDNNError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, int __pyx_v_status); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_check_status(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_status); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_2getVersion(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_4create(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_6destroy(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_8setStream(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_10getStream(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_12createTensorDescriptor(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_14setTensor4dDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc, int __pyx_v_format, int __pyx_v_dataType, int __pyx_v_n, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_16setTensor4dDescriptorEx(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc, int __pyx_v_dataType, int __pyx_v_n, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w, int __pyx_v_nStride, int __pyx_v_cStride, int __pyx_v_hStride, int __pyx_v_wStride); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_18getTensor4dDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_20setTensorNdDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc, int __pyx_v_dataType, int __pyx_v_nbDims, size_t __pyx_v_dimA, size_t __pyx_v_strideA); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_22destroyTensorDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_24addTensor_v3(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_bDesc, size_t __pyx_v_b, size_t __pyx_v_beta, size_t __pyx_v_yDesc, size_t __pyx_v_y); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_26createFilterDescriptor(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_28setFilter4dDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_filterDesc, int __pyx_v_dataType, int __pyx_v_format, int __pyx_v_k, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_30setFilterNdDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_filterDesc, int __pyx_v_dataType, int __pyx_v_format, int __pyx_v_nbDims, size_t __pyx_v_filterDimA); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_32getFilterNdDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_wDesc, int __pyx_v_nbDimsRequested); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_34destroyFilterDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_filterDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_36createConvolutionDescriptor(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_38setConvolutionMathType(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, size_t __pyx_v_mathType); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_40getConvolutionMathType(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_42setConvolutionGroupCount(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, int __pyx_v_groupCount); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_44getConvolutionGroupCount(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_46setConvolution2dDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, int __pyx_v_pad_h, int __pyx_v_pad_w, int __pyx_v_u, int __pyx_v_v, int __pyx_v_dilation_h, int __pyx_v_dilation_w, int __pyx_v_mode); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_48setConvolution2dDescriptor_v5(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, int __pyx_v_pad_h, int __pyx_v_pad_w, int __pyx_v_u, int __pyx_v_v, int __pyx_v_dilation_h, int __pyx_v_dilation_w, int __pyx_v_mode, size_t __pyx_v_computeType); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_50setConvolutionNdDescriptor_v3(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, int __pyx_v_arrayLength, size_t __pyx_v_padA, size_t __pyx_v_filterStrideA, size_t __pyx_v_dilationA, int __pyx_v_mode, int __pyx_v_dataType); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_52destroyConvolutionDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_54findConvolutionForwardAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_convDesc, size_t __pyx_v_yDesc, int __pyx_v_requestedAlgoCount); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_56findConvolutionForwardAlgorithmEx(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_convDesc, size_t __pyx_v_yDesc, size_t __pyx_v_y, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_58getConvolutionForwardAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_filterDesc, size_t __pyx_v_convDesc, size_t __pyx_v_destDesc, int __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_60getConvolutionForwardWorkspaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_filterDesc, size_t __pyx_v_convDesc, size_t __pyx_v_destDesc, int __pyx_v_algo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_62convolutionForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_filterDesc, size_t __pyx_v_filterData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_destDesc, size_t __pyx_v_destData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_64convolutionBackwardBias(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_destDesc, size_t __pyx_v_destData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_66findConvolutionBackwardFilterAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_convDesc, size_t __pyx_v_dwDesc, int __pyx_v_requestedAlgoCount); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_68findConvolutionBackwardFilterAlgorithmEx(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_convDesc, size_t __pyx_v_dwDesc, size_t __pyx_v_dw, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_70getConvolutionBackwardFilterAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_filterDesc, int __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_72getConvolutionBackwardFilterWorkspaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_filterDesc, int __pyx_v_algo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_74convolutionBackwardFilter_v3(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_diffDesc, size_t __pyx_v_diffData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_gradDesc, size_t __pyx_v_gradData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_76findConvolutionBackwardDataAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_wDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_convDesc, size_t __pyx_v_dxDesc, int __pyx_v_requestedAlgoCount); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_78findConvolutionBackwardDataAlgorithmEx(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_convDesc, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_80getConvolutionBackwardDataAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_filterDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_gradDesc, size_t __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_82getConvolutionBackwardDataWorkspaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_filterDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_gradDesc, int __pyx_v_algo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_84convolutionBackwardData_v3(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_filterDesc, size_t __pyx_v_filterData, size_t __pyx_v_diffDesc, size_t __pyx_v_diffData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_gradDesc, size_t __pyx_v_gradData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_86createPoolingDescriptor(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_88setPooling2dDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_poolingDesc, int __pyx_v_mode, int __pyx_v_maxpoolingNanOpt, int __pyx_v_windowHeight, int __pyx_v_windowWidth, int __pyx_v_verticalPadding, int __pyx_v_horizontalPadding, int __pyx_v_verticalStride, int __pyx_v_horizontalStride); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_90setPoolingNdDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_poolingDesc, int __pyx_v_mode, int __pyx_v_maxpoolingNanOpt, int __pyx_v_nbDims, size_t __pyx_v_windowDimA, size_t __pyx_v_paddingA, size_t __pyx_v_strideA); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_92destroyPoolingDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_poolingDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_94poolingForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_poolingDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_96poolingBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_poolingDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_destDesc, size_t __pyx_v_destData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_98deriveBNTensorDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_derivedBnDesc, size_t __pyx_v_xDesc, int __pyx_v_mode); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_100batchNormalizationForwardTraining(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_beta, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_bnScaleBiasMeanVarDesc, size_t __pyx_v_bnScale, size_t __pyx_v_bnBias, double __pyx_v_exponentialAverageFactor, size_t __pyx_v_resultRunningMean, size_t __pyx_v_resultRunningVariance, double __pyx_v_epsilon, size_t __pyx_v_resultSaveMean, size_t __pyx_v_resultSaveInvVariance); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_102batchNormalizationForwardInference(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_beta, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_bnScaleBiasMeanVarDesc, size_t __pyx_v_bnScale, size_t __pyx_v_bnBias, size_t __pyx_v_estimatedMean, size_t __pyx_v_estimatedVariance, double __pyx_v_epsilon); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_104batchNormalizationBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alphaDataDiff, size_t __pyx_v_betaDataDiff, size_t __pyx_v_alphaParamDiff, size_t __pyx_v_betaParamDiff, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_dBnScaleBiasDesc, size_t __pyx_v_bnScale, size_t __pyx_v_dBnScaleResult, size_t __pyx_v_dBnBiasResult, double __pyx_v_epsilon, size_t __pyx_v_savedMean, size_t __pyx_v_savedInvVariance); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_106createActivationDescriptor(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_108setActivationDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_activationDesc, int __pyx_v_mode, int __pyx_v_reluNanOpt, double __pyx_v_reluCeiling); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_110destroyActivationDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_activationDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_112softmaxForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_algorithm, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_114softmaxBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_algorithm, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_116activationForward_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_activationDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_118activationBackward_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_activationDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_destDesc, size_t __pyx_v_destData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_120createDropoutDescriptor(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_122destroyDropoutDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_dropoutDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_124dropoutGetStatesSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_126setDropoutDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_dropoutDesc, size_t __pyx_v_handle, float __pyx_v_dropout, size_t __pyx_v_states, size_t __pyx_v_stateSizeInBytes, unsigned PY_LONG_LONG __pyx_v_seed); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_128getDropoutReserveSpaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_xDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_130dropoutForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_dropoutDesc, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_132dropoutBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_dropoutDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_dyData, size_t __pyx_v_dxDesc, size_t __pyx_v_dxData, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_134createRNNDescriptor(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_136destroyRNNDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_rnnDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_138createPersistentRNNPlan(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_rnnDesc, int __pyx_v_minibatch, int __pyx_v_dataType); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_140setPersistentRNNPlan(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_rnnDesc, size_t __pyx_v_plan); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_142destroyPersistentRNNPlan(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_plan); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_144setRNNDescriptor_v5(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_rnnDesc, int __pyx_v_hiddenSize, int __pyx_v_numLayers, size_t __pyx_v_dropoutDesc, int __pyx_v_inputMode, int __pyx_v_direction, int __pyx_v_mode, int __pyx_v_dataType); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_146setRNNDescriptor_v6(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_hiddenSize, int __pyx_v_numLayers, size_t __pyx_v_dropoutDesc, int __pyx_v_inputMode, int __pyx_v_direction, int __pyx_v_mode, int __pyx_v_algo, int __pyx_v_dataType); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_148getRNNWorkspaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_150getRNNTrainingReserveSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_152getRNNParamsSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, size_t __pyx_v_xDesc, int __pyx_v_dataType); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_154getRNNLinLayerMatrixParams(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_layer, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_w, int __pyx_v_linLayerID, size_t __pyx_v_linLayerMatDesc, size_t __pyx_v_linLayerMat); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_156getRNNLinLayerBiasParams(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_layer, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_w, int __pyx_v_linLayerID, size_t __pyx_v_linLayerBiasDesc, size_t __pyx_v_linLayerBias); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_158RNNForwardInference(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_hyDesc, size_t __pyx_v_hy, size_t __pyx_v_cyDesc, size_t __pyx_v_cy, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_160RNNForwardTraining(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_hyDesc, size_t __pyx_v_hy, size_t __pyx_v_cyDesc, size_t __pyx_v_cy, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_162RNNBackwardData(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_dhyDesc, size_t __pyx_v_dhy, size_t __pyx_v_dcyDesc, size_t __pyx_v_dcy, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_dhxDesc, size_t __pyx_v_dhx, size_t __pyx_v_dcxDesc, size_t __pyx_v_dcx, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_164RNNBackwardWeights(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_dwDesc, size_t __pyx_v_dw, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_166createSpatialTransformerDescriptor(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_168destroySpatialTransformerDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_stDesc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_170setSpatialTransformerDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_stDesc, size_t __pyx_v_samplerType, int __pyx_v_dataType, int __pyx_v_nbDims, size_t __pyx_v_dimA); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_172spatialTfGridGeneratorForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_theta, size_t __pyx_v_grid); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_174spatialTfGridGeneratorBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_dgrid, size_t __pyx_v_dtheta); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_176spatialTfSamplerForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_alpha, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_grid, size_t __pyx_v_beta, size_t __pyx_v_yDesc, size_t __pyx_v_y); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_178spatialTfSamplerBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_alpha, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_alphaDgrid, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_grid, size_t __pyx_v_betaDgrid, size_t __pyx_v_dgrid); /* proto */
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_10;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_codeobj__2;

/* "cupy/cuda/cudnn.pyx":448
 * class CuDNNError(RuntimeError):
 * 
 *     def __init__(self, int status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         msg = cudnnGetErrorString(<Status>status)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_10CuDNNError_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_10CuDNNError___init__[] = "CuDNNError.__init__(self, int status)";
static PyMethodDef __pyx_mdef_4cupy_4cuda_5cudnn_10CuDNNError_1__init__ = {"__init__", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_10CuDNNError_1__init__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_10CuDNNError___init__};
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_10CuDNNError_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  int __pyx_v_status;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_status,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_status)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 448, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 448, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_status = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_status == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 448, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 448, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.CuDNNError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_10CuDNNError___init__(__pyx_self, __pyx_v_self, __pyx_v_status);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_10CuDNNError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, int __pyx_v_status) {
  char const *__pyx_v_msg;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "cupy/cuda/cudnn.pyx":449
 * 
 *     def __init__(self, int status):
 *         self.status = status             # <<<<<<<<<<<<<<
 *         msg = cudnnGetErrorString(<Status>status)
 *         super(CuDNNError, self).__init__('%s: %s' % (STATUS[status], msg))
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_status, __pyx_t_1) < 0) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":450
 *     def __init__(self, int status):
 *         self.status = status
 *         msg = cudnnGetErrorString(<Status>status)             # <<<<<<<<<<<<<<
 *         super(CuDNNError, self).__init__('%s: %s' % (STATUS[status], msg))
 * 
 */
  __pyx_v_msg = cudnnGetErrorString(((cudnnStatus_t)__pyx_v_status));

  /* "cupy/cuda/cudnn.pyx":451
 *         self.status = status
 *         msg = cudnnGetErrorString(<Status>status)
 *         super(CuDNNError, self).__init__('%s: %s' % (STATUS[status], msg))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_CuDNNError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_4cupy_4cuda_5cudnn_STATUS == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 451, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_4cupy_4cuda_5cudnn_STATUS, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_msg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":448
 * class CuDNNError(RuntimeError):
 * 
 *     def __init__(self, int status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         msg = cudnnGetErrorString(<Status>status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cupy.cuda.cudnn.CuDNNError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":455
 * 
 * @cython.profile(False)
 * cpdef inline check_status(int status):             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         raise CuDNNError(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4cupy_4cuda_5cudnn_check_status(int __pyx_v_status, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("check_status", 0);

  /* "cupy/cuda/cudnn.pyx":456
 * @cython.profile(False)
 * cpdef inline check_status(int status):
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         raise CuDNNError(status)
 * 
 */
  __pyx_t_1 = ((__pyx_v_status != 0) != 0);
  if (__pyx_t_1) {

    /* "cupy/cuda/cudnn.pyx":457
 * cpdef inline check_status(int status):
 *     if status != 0:
 *         raise CuDNNError(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_CuDNNError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 457, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 457, __pyx_L1_error)

    /* "cupy/cuda/cudnn.pyx":456
 * @cython.profile(False)
 * cpdef inline check_status(int status):
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         raise CuDNNError(status)
 * 
 */
  }

  /* "cupy/cuda/cudnn.pyx":455
 * 
 * @cython.profile(False)
 * cpdef inline check_status(int status):             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         raise CuDNNError(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cupy.cuda.cudnn.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_check_status[] = "check_status(int status)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status) {
  int __pyx_v_status;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_status (wrapper)", 0);
  assert(__pyx_arg_status); {
    __pyx_v_status = __Pyx_PyInt_As_int(__pyx_arg_status); if (unlikely((__pyx_v_status == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 455, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_check_status(__pyx_self, ((int)__pyx_v_status));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_check_status(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_status) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("check_status", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":464
 * ###############################################################################
 * 
 * cpdef size_t getVersion() except *:             # <<<<<<<<<<<<<<
 *     return cudnnGetVersion()
 * 
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_3getVersion(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getVersion(CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getVersion", 0);

  /* "cupy/cuda/cudnn.pyx":465
 * 
 * cpdef size_t getVersion() except *:
 *     return cudnnGetVersion()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = cudnnGetVersion();
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":464
 * ###############################################################################
 * 
 * cpdef size_t getVersion() except *:             # <<<<<<<<<<<<<<
 *     return cudnnGetVersion()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_3getVersion(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_2getVersion[] = "getVersion() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_3getVersion(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getVersion (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_2getVersion(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_2getVersion(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getVersion", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getVersion(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getVersion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":472
 * ###############################################################################
 * 
 * cpdef size_t create() except *:             # <<<<<<<<<<<<<<
 *     cdef Handle handle
 *     with nogil:
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_5create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_create(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnHandle_t __pyx_v_handle;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("create", 0);

  /* "cupy/cuda/cudnn.pyx":474
 * cpdef size_t create() except *:
 *     cdef Handle handle
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnCreate(&handle)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":475
 *     cdef Handle handle
 *     with nogil:
 *         status = cudnnCreate(&handle)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>handle
 */
        __pyx_v_status = cudnnCreate((&__pyx_v_handle));
      }

      /* "cupy/cuda/cudnn.pyx":474
 * cpdef size_t create() except *:
 *     cdef Handle handle
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnCreate(&handle)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":476
 *     with nogil:
 *         status = cudnnCreate(&handle)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>handle
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":477
 *         status = cudnnCreate(&handle)
 *     check_status(status)
 *     return <size_t>handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_handle);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":472
 * ###############################################################################
 * 
 * cpdef size_t create() except *:             # <<<<<<<<<<<<<<
 *     cdef Handle handle
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_5create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_4create[] = "create() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_5create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_4create(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_4create(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("create", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_create(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 472, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":480
 * 
 * 
 * cpdef destroy(size_t handle):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnDestroy(<Handle>handle)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_7destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroy(size_t __pyx_v_handle, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroy", 0);

  /* "cupy/cuda/cudnn.pyx":481
 * 
 * cpdef destroy(size_t handle):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnDestroy(<Handle>handle)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":482
 * cpdef destroy(size_t handle):
 *     with nogil:
 *         status = cudnnDestroy(<Handle>handle)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
        __pyx_v_status = cudnnDestroy(((cudnnHandle_t)__pyx_v_handle));
      }

      /* "cupy/cuda/cudnn.pyx":481
 * 
 * cpdef destroy(size_t handle):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnDestroy(<Handle>handle)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":483
 *     with nogil:
 *         status = cudnnDestroy(<Handle>handle)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":480
 * 
 * 
 * cpdef destroy(size_t handle):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnDestroy(<Handle>handle)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_7destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_6destroy[] = "destroy(size_t handle)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_7destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle) {
  size_t __pyx_v_handle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroy (wrapper)", 0);
  assert(__pyx_arg_handle); {
    __pyx_v_handle = __Pyx_PyInt_As_size_t(__pyx_arg_handle); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 480, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_6destroy(__pyx_self, ((size_t)__pyx_v_handle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_6destroy(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroy(__pyx_v_handle, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":486
 * 
 * 
 * cpdef setStream(size_t handle, size_t stream):             # <<<<<<<<<<<<<<
 *     status = cudnnSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_9setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setStream(size_t __pyx_v_handle, size_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setStream", 0);

  /* "cupy/cuda/cudnn.pyx":487
 * 
 * cpdef setStream(size_t handle, size_t stream):
 *     status = cudnnSetStream(<Handle>handle, <driver.Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cudnnSetStream(((cudnnHandle_t)__pyx_v_handle), ((struct CUstream_st*)__pyx_v_stream));

  /* "cupy/cuda/cudnn.pyx":488
 * cpdef setStream(size_t handle, size_t stream):
 *     status = cudnnSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":486
 * 
 * 
 * cpdef setStream(size_t handle, size_t stream):             # <<<<<<<<<<<<<<
 *     status = cudnnSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_9setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_8setStream[] = "setStream(size_t handle, size_t stream)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_9setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_stream;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setStream (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_stream,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setStream", 1, 2, 2, 1); __PYX_ERR(0, 486, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setStream") < 0)) __PYX_ERR(0, 486, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 486, __pyx_L3_error)
    __pyx_v_stream = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_stream == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 486, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setStream", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 486, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_8setStream(__pyx_self, __pyx_v_handle, __pyx_v_stream);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_8setStream(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setStream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_v_stream, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":491
 * 
 * 
 * cpdef size_t getStream(size_t handle) except *:             # <<<<<<<<<<<<<<
 *     cdef driver.Stream stream
 *     status = cudnnGetStream(<Handle>handle, &stream)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_11getStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getStream(size_t __pyx_v_handle, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct CUstream_st* __pyx_v_stream;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getStream", 0);

  /* "cupy/cuda/cudnn.pyx":493
 * cpdef size_t getStream(size_t handle) except *:
 *     cdef driver.Stream stream
 *     status = cudnnGetStream(<Handle>handle, &stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>stream
 */
  __pyx_v_status = cudnnGetStream(((cudnnHandle_t)__pyx_v_handle), (&__pyx_v_stream));

  /* "cupy/cuda/cudnn.pyx":494
 *     cdef driver.Stream stream
 *     status = cudnnGetStream(<Handle>handle, &stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>stream
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":495
 *     status = cudnnGetStream(<Handle>handle, &stream)
 *     check_status(status)
 *     return <size_t>stream             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_stream);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":491
 * 
 * 
 * cpdef size_t getStream(size_t handle) except *:             # <<<<<<<<<<<<<<
 *     cdef driver.Stream stream
 *     status = cudnnGetStream(<Handle>handle, &stream)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_11getStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_10getStream[] = "getStream(size_t handle) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_11getStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle) {
  size_t __pyx_v_handle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getStream (wrapper)", 0);
  assert(__pyx_arg_handle); {
    __pyx_v_handle = __Pyx_PyInt_As_size_t(__pyx_arg_handle); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 491, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_10getStream(__pyx_self, ((size_t)__pyx_v_handle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_10getStream(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getStream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getStream(__pyx_v_handle, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 491, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":502
 * ###############################################################################
 * 
 * cpdef size_t createTensorDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef TensorDescriptor descriptor
 *     status = cudnnCreateTensorDescriptor(&descriptor)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_13createTensorDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createTensorDescriptor(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnTensorDescriptor_t __pyx_v_descriptor;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createTensorDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":504
 * cpdef size_t createTensorDescriptor() except *:
 *     cdef TensorDescriptor descriptor
 *     status = cudnnCreateTensorDescriptor(&descriptor)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>descriptor
 */
  __pyx_v_status = cudnnCreateTensorDescriptor((&__pyx_v_descriptor));

  /* "cupy/cuda/cudnn.pyx":505
 *     cdef TensorDescriptor descriptor
 *     status = cudnnCreateTensorDescriptor(&descriptor)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>descriptor
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":506
 *     status = cudnnCreateTensorDescriptor(&descriptor)
 *     check_status(status)
 *     return <size_t>descriptor             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_descriptor);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":502
 * ###############################################################################
 * 
 * cpdef size_t createTensorDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef TensorDescriptor descriptor
 *     status = cudnnCreateTensorDescriptor(&descriptor)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createTensorDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_13createTensorDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_12createTensorDescriptor[] = "createTensorDescriptor() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_13createTensorDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createTensorDescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_12createTensorDescriptor(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_12createTensorDescriptor(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createTensorDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createTensorDescriptor(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 502, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createTensorDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":509
 * 
 * 
 * cpdef setTensor4dDescriptor(size_t tensorDesc, int format, int dataType,             # <<<<<<<<<<<<<<
 *                             int n, int c, int h, int w):
 *     status = cudnnSetTensor4dDescriptor(
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_15setTensor4dDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setTensor4dDescriptor(size_t __pyx_v_tensorDesc, int __pyx_v_format, int __pyx_v_dataType, int __pyx_v_n, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setTensor4dDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":511
 * cpdef setTensor4dDescriptor(size_t tensorDesc, int format, int dataType,
 *                             int n, int c, int h, int w):
 *     status = cudnnSetTensor4dDescriptor(             # <<<<<<<<<<<<<<
 *         <TensorDescriptor>tensorDesc, <TensorFormat>format,
 *         <DataType>dataType, n, c, h, w)
 */
  __pyx_v_status = cudnnSetTensor4dDescriptor(((cudnnTensorDescriptor_t)__pyx_v_tensorDesc), ((cudnnTensorFormat_t)__pyx_v_format), ((cudnnDataType_t)__pyx_v_dataType), __pyx_v_n, __pyx_v_c, __pyx_v_h, __pyx_v_w);

  /* "cupy/cuda/cudnn.pyx":514
 *         <TensorDescriptor>tensorDesc, <TensorFormat>format,
 *         <DataType>dataType, n, c, h, w)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":509
 * 
 * 
 * cpdef setTensor4dDescriptor(size_t tensorDesc, int format, int dataType,             # <<<<<<<<<<<<<<
 *                             int n, int c, int h, int w):
 *     status = cudnnSetTensor4dDescriptor(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensor4dDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_15setTensor4dDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_14setTensor4dDescriptor[] = "setTensor4dDescriptor(size_t tensorDesc, int format, int dataType, int n, int c, int h, int w)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_15setTensor4dDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_tensorDesc;
  int __pyx_v_format;
  int __pyx_v_dataType;
  int __pyx_v_n;
  int __pyx_v_c;
  int __pyx_v_h;
  int __pyx_v_w;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setTensor4dDescriptor (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tensorDesc,&__pyx_n_s_format,&__pyx_n_s_dataType,&__pyx_n_s_n,&__pyx_n_s_c,&__pyx_n_s_h,&__pyx_n_s_w,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tensorDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptor", 1, 7, 7, 1); __PYX_ERR(0, 509, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptor", 1, 7, 7, 2); __PYX_ERR(0, 509, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptor", 1, 7, 7, 3); __PYX_ERR(0, 509, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_c)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptor", 1, 7, 7, 4); __PYX_ERR(0, 509, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptor", 1, 7, 7, 5); __PYX_ERR(0, 509, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptor", 1, 7, 7, 6); __PYX_ERR(0, 509, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setTensor4dDescriptor") < 0)) __PYX_ERR(0, 509, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_tensorDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_tensorDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L3_error)
    __pyx_v_format = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_format == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 510, __pyx_L3_error)
    __pyx_v_c = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_c == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 510, __pyx_L3_error)
    __pyx_v_h = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_h == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 510, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_w == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 510, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptor", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 509, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensor4dDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_14setTensor4dDescriptor(__pyx_self, __pyx_v_tensorDesc, __pyx_v_format, __pyx_v_dataType, __pyx_v_n, __pyx_v_c, __pyx_v_h, __pyx_v_w);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_14setTensor4dDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc, int __pyx_v_format, int __pyx_v_dataType, int __pyx_v_n, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setTensor4dDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setTensor4dDescriptor(__pyx_v_tensorDesc, __pyx_v_format, __pyx_v_dataType, __pyx_v_n, __pyx_v_c, __pyx_v_h, __pyx_v_w, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensor4dDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":517
 * 
 * 
 * cpdef setTensor4dDescriptorEx(size_t tensorDesc, int dataType,             # <<<<<<<<<<<<<<
 *                               int n, int c, int h, int w, int nStride,
 *                               int cStride, int hStride, int wStride):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_17setTensor4dDescriptorEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setTensor4dDescriptorEx(size_t __pyx_v_tensorDesc, int __pyx_v_dataType, int __pyx_v_n, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w, int __pyx_v_nStride, int __pyx_v_cStride, int __pyx_v_hStride, int __pyx_v_wStride, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setTensor4dDescriptorEx", 0);

  /* "cupy/cuda/cudnn.pyx":520
 *                               int n, int c, int h, int w, int nStride,
 *                               int cStride, int hStride, int wStride):
 *     status = cudnnSetTensor4dDescriptorEx(             # <<<<<<<<<<<<<<
 *         <TensorDescriptor>tensorDesc, <DataType>dataType, n, c, h, w,
 *         nStride, cStride, hStride, wStride)
 */
  __pyx_v_status = cudnnSetTensor4dDescriptorEx(((cudnnTensorDescriptor_t)__pyx_v_tensorDesc), ((cudnnDataType_t)__pyx_v_dataType), __pyx_v_n, __pyx_v_c, __pyx_v_h, __pyx_v_w, __pyx_v_nStride, __pyx_v_cStride, __pyx_v_hStride, __pyx_v_wStride);

  /* "cupy/cuda/cudnn.pyx":523
 *         <TensorDescriptor>tensorDesc, <DataType>dataType, n, c, h, w,
 *         nStride, cStride, hStride, wStride)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":517
 * 
 * 
 * cpdef setTensor4dDescriptorEx(size_t tensorDesc, int dataType,             # <<<<<<<<<<<<<<
 *                               int n, int c, int h, int w, int nStride,
 *                               int cStride, int hStride, int wStride):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensor4dDescriptorEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_17setTensor4dDescriptorEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_16setTensor4dDescriptorEx[] = "setTensor4dDescriptorEx(size_t tensorDesc, int dataType, int n, int c, int h, int w, int nStride, int cStride, int hStride, int wStride)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_17setTensor4dDescriptorEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_tensorDesc;
  int __pyx_v_dataType;
  int __pyx_v_n;
  int __pyx_v_c;
  int __pyx_v_h;
  int __pyx_v_w;
  int __pyx_v_nStride;
  int __pyx_v_cStride;
  int __pyx_v_hStride;
  int __pyx_v_wStride;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setTensor4dDescriptorEx (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tensorDesc,&__pyx_n_s_dataType,&__pyx_n_s_n,&__pyx_n_s_c,&__pyx_n_s_h,&__pyx_n_s_w,&__pyx_n_s_nStride,&__pyx_n_s_cStride,&__pyx_n_s_hStride,&__pyx_n_s_wStride,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tensorDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 1); __PYX_ERR(0, 517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 2); __PYX_ERR(0, 517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_c)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 3); __PYX_ERR(0, 517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 4); __PYX_ERR(0, 517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 5); __PYX_ERR(0, 517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nStride)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 6); __PYX_ERR(0, 517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cStride)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 7); __PYX_ERR(0, 517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hStride)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 8); __PYX_ERR(0, 517, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wStride)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, 9); __PYX_ERR(0, 517, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setTensor4dDescriptorEx") < 0)) __PYX_ERR(0, 517, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_tensorDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_tensorDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_c = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_c == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_h = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_h == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_w == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_nStride = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_nStride == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L3_error)
    __pyx_v_cStride = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_cStride == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 519, __pyx_L3_error)
    __pyx_v_hStride = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_hStride == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 519, __pyx_L3_error)
    __pyx_v_wStride = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_wStride == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 519, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setTensor4dDescriptorEx", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 517, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensor4dDescriptorEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_16setTensor4dDescriptorEx(__pyx_self, __pyx_v_tensorDesc, __pyx_v_dataType, __pyx_v_n, __pyx_v_c, __pyx_v_h, __pyx_v_w, __pyx_v_nStride, __pyx_v_cStride, __pyx_v_hStride, __pyx_v_wStride);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_16setTensor4dDescriptorEx(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc, int __pyx_v_dataType, int __pyx_v_n, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w, int __pyx_v_nStride, int __pyx_v_cStride, int __pyx_v_hStride, int __pyx_v_wStride) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setTensor4dDescriptorEx", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setTensor4dDescriptorEx(__pyx_v_tensorDesc, __pyx_v_dataType, __pyx_v_n, __pyx_v_c, __pyx_v_h, __pyx_v_w, __pyx_v_nStride, __pyx_v_cStride, __pyx_v_hStride, __pyx_v_wStride, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensor4dDescriptorEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":526
 * 
 * 
 * cpdef getTensor4dDescriptor(size_t tensorDesc):             # <<<<<<<<<<<<<<
 *     cdef DataType dataType
 *     cdef int n, c, h, w, nStride, cStride, hStride, wStride
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_19getTensor4dDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_tensorDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getTensor4dDescriptor(size_t __pyx_v_tensorDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnDataType_t __pyx_v_dataType;
  int __pyx_v_n;
  int __pyx_v_c;
  int __pyx_v_h;
  int __pyx_v_w;
  int __pyx_v_nStride;
  int __pyx_v_cStride;
  int __pyx_v_hStride;
  int __pyx_v_wStride;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("getTensor4dDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":529
 *     cdef DataType dataType
 *     cdef int n, c, h, w, nStride, cStride, hStride, wStride
 *     status = cudnnGetTensor4dDescriptor(             # <<<<<<<<<<<<<<
 *         <TensorDescriptor>tensorDesc, &dataType,
 *         &n, &c, &h, &w, &nStride, &cStride, &hStride, &wStride)
 */
  __pyx_v_status = cudnnGetTensor4dDescriptor(((cudnnTensorDescriptor_t)__pyx_v_tensorDesc), (&__pyx_v_dataType), (&__pyx_v_n), (&__pyx_v_c), (&__pyx_v_h), (&__pyx_v_w), (&__pyx_v_nStride), (&__pyx_v_cStride), (&__pyx_v_hStride), (&__pyx_v_wStride));

  /* "cupy/cuda/cudnn.pyx":532
 *         <TensorDescriptor>tensorDesc, &dataType,
 *         &n, &c, &h, &w, &nStride, &cStride, &hStride, &wStride)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return dataType, n, c, h, w, nStride, cStride, hStride, wStride
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":533
 *         &n, &c, &h, &w, &nStride, &cStride, &hStride, &wStride)
 *     check_status(status)
 *     return dataType, n, c, h, w, nStride, cStride, hStride, wStride             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_cudnnDataType_t(__pyx_v_dataType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_c); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_h); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_w); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_nStride); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_cStride); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_hStride); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_wStride); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyTuple_New(9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 7, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 8, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":526
 * 
 * 
 * cpdef getTensor4dDescriptor(size_t tensorDesc):             # <<<<<<<<<<<<<<
 *     cdef DataType dataType
 *     cdef int n, c, h, w, nStride, cStride, hStride, wStride
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getTensor4dDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_19getTensor4dDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_tensorDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_18getTensor4dDescriptor[] = "getTensor4dDescriptor(size_t tensorDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_19getTensor4dDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_tensorDesc) {
  size_t __pyx_v_tensorDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getTensor4dDescriptor (wrapper)", 0);
  assert(__pyx_arg_tensorDesc); {
    __pyx_v_tensorDesc = __Pyx_PyInt_As_size_t(__pyx_arg_tensorDesc); if (unlikely((__pyx_v_tensorDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 526, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getTensor4dDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_18getTensor4dDescriptor(__pyx_self, ((size_t)__pyx_v_tensorDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_18getTensor4dDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getTensor4dDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getTensor4dDescriptor(__pyx_v_tensorDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getTensor4dDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":536
 * 
 * 
 * cpdef setTensorNdDescriptor(size_t tensorDesc, int dataType, int nbDims,             # <<<<<<<<<<<<<<
 *                             size_t dimA, size_t strideA):
 *     status = cudnnSetTensorNdDescriptor(
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_21setTensorNdDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setTensorNdDescriptor(size_t __pyx_v_tensorDesc, int __pyx_v_dataType, int __pyx_v_nbDims, size_t __pyx_v_dimA, size_t __pyx_v_strideA, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setTensorNdDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":538
 * cpdef setTensorNdDescriptor(size_t tensorDesc, int dataType, int nbDims,
 *                             size_t dimA, size_t strideA):
 *     status = cudnnSetTensorNdDescriptor(             # <<<<<<<<<<<<<<
 *         <TensorDescriptor>tensorDesc, <DataType>dataType, nbDims,
 *         <int*>dimA, <int*>strideA)
 */
  __pyx_v_status = cudnnSetTensorNdDescriptor(((cudnnTensorDescriptor_t)__pyx_v_tensorDesc), ((cudnnDataType_t)__pyx_v_dataType), __pyx_v_nbDims, ((int *)__pyx_v_dimA), ((int *)__pyx_v_strideA));

  /* "cupy/cuda/cudnn.pyx":541
 *         <TensorDescriptor>tensorDesc, <DataType>dataType, nbDims,
 *         <int*>dimA, <int*>strideA)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":536
 * 
 * 
 * cpdef setTensorNdDescriptor(size_t tensorDesc, int dataType, int nbDims,             # <<<<<<<<<<<<<<
 *                             size_t dimA, size_t strideA):
 *     status = cudnnSetTensorNdDescriptor(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensorNdDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_21setTensorNdDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_20setTensorNdDescriptor[] = "setTensorNdDescriptor(size_t tensorDesc, int dataType, int nbDims, size_t dimA, size_t strideA)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_21setTensorNdDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_tensorDesc;
  int __pyx_v_dataType;
  int __pyx_v_nbDims;
  size_t __pyx_v_dimA;
  size_t __pyx_v_strideA;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setTensorNdDescriptor (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tensorDesc,&__pyx_n_s_dataType,&__pyx_n_s_nbDims,&__pyx_n_s_dimA,&__pyx_n_s_strideA,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tensorDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensorNdDescriptor", 1, 5, 5, 1); __PYX_ERR(0, 536, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nbDims)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensorNdDescriptor", 1, 5, 5, 2); __PYX_ERR(0, 536, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dimA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensorNdDescriptor", 1, 5, 5, 3); __PYX_ERR(0, 536, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setTensorNdDescriptor", 1, 5, 5, 4); __PYX_ERR(0, 536, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setTensorNdDescriptor") < 0)) __PYX_ERR(0, 536, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_tensorDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_tensorDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L3_error)
    __pyx_v_nbDims = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nbDims == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L3_error)
    __pyx_v_dimA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_dimA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 537, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_strideA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 537, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setTensorNdDescriptor", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 536, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensorNdDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_20setTensorNdDescriptor(__pyx_self, __pyx_v_tensorDesc, __pyx_v_dataType, __pyx_v_nbDims, __pyx_v_dimA, __pyx_v_strideA);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_20setTensorNdDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc, int __pyx_v_dataType, int __pyx_v_nbDims, size_t __pyx_v_dimA, size_t __pyx_v_strideA) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setTensorNdDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setTensorNdDescriptor(__pyx_v_tensorDesc, __pyx_v_dataType, __pyx_v_nbDims, __pyx_v_dimA, __pyx_v_strideA, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setTensorNdDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":544
 * 
 * 
 * cpdef destroyTensorDescriptor(size_t tensorDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyTensorDescriptor(<TensorDescriptor>tensorDesc)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_23destroyTensorDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_tensorDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyTensorDescriptor(size_t __pyx_v_tensorDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyTensorDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":545
 * 
 * cpdef destroyTensorDescriptor(size_t tensorDesc):
 *     status = cudnnDestroyTensorDescriptor(<TensorDescriptor>tensorDesc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cudnnDestroyTensorDescriptor(((cudnnTensorDescriptor_t)__pyx_v_tensorDesc));

  /* "cupy/cuda/cudnn.pyx":546
 * cpdef destroyTensorDescriptor(size_t tensorDesc):
 *     status = cudnnDestroyTensorDescriptor(<TensorDescriptor>tensorDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":544
 * 
 * 
 * cpdef destroyTensorDescriptor(size_t tensorDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyTensorDescriptor(<TensorDescriptor>tensorDesc)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyTensorDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_23destroyTensorDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_tensorDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_22destroyTensorDescriptor[] = "destroyTensorDescriptor(size_t tensorDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_23destroyTensorDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_tensorDesc) {
  size_t __pyx_v_tensorDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyTensorDescriptor (wrapper)", 0);
  assert(__pyx_arg_tensorDesc); {
    __pyx_v_tensorDesc = __Pyx_PyInt_As_size_t(__pyx_arg_tensorDesc); if (unlikely((__pyx_v_tensorDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 544, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyTensorDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_22destroyTensorDescriptor(__pyx_self, ((size_t)__pyx_v_tensorDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_22destroyTensorDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_tensorDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyTensorDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroyTensorDescriptor(__pyx_v_tensorDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyTensorDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":549
 * 
 * 
 * cpdef addTensor_v3(size_t handle, size_t alpha, size_t bDesc,             # <<<<<<<<<<<<<<
 *                    size_t b, size_t beta, size_t yDesc, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_25addTensor_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_addTensor_v3(size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_bDesc, size_t __pyx_v_b, size_t __pyx_v_beta, size_t __pyx_v_yDesc, size_t __pyx_v_y, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("addTensor_v3", 0);

  /* "cupy/cuda/cudnn.pyx":551
 * cpdef addTensor_v3(size_t handle, size_t alpha, size_t bDesc,
 *                    size_t b, size_t beta, size_t yDesc, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnAddTensor_v3(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":552
 *                    size_t b, size_t beta, size_t yDesc, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnAddTensor_v3(
 *             <Handle>handle, <void*>alpha, <TensorDescriptor>bDesc,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":553
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnAddTensor_v3(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <void*>alpha, <TensorDescriptor>bDesc,
 *             <void*>b, <void*>beta, <TensorDescriptor>yDesc, <void*>y)
 */
        __pyx_v_status = cudnnAddTensor_v3(((cudnnHandle_t)__pyx_v_handle), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_bDesc), ((void *)__pyx_v_b), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_yDesc), ((void *)__pyx_v_y));
      }

      /* "cupy/cuda/cudnn.pyx":552
 *                    size_t b, size_t beta, size_t yDesc, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnAddTensor_v3(
 *             <Handle>handle, <void*>alpha, <TensorDescriptor>bDesc,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":556
 *             <Handle>handle, <void*>alpha, <TensorDescriptor>bDesc,
 *             <void*>b, <void*>beta, <TensorDescriptor>yDesc, <void*>y)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":549
 * 
 * 
 * cpdef addTensor_v3(size_t handle, size_t alpha, size_t bDesc,             # <<<<<<<<<<<<<<
 *                    size_t b, size_t beta, size_t yDesc, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.addTensor_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_25addTensor_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_24addTensor_v3[] = "addTensor_v3(size_t handle, size_t alpha, size_t bDesc, size_t b, size_t beta, size_t yDesc, size_t y)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_25addTensor_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_alpha;
  size_t __pyx_v_bDesc;
  size_t __pyx_v_b;
  size_t __pyx_v_beta;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addTensor_v3 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_alpha,&__pyx_n_s_bDesc,&__pyx_n_s_b,&__pyx_n_s_beta,&__pyx_n_s_yDesc,&__pyx_n_s_y,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addTensor_v3", 1, 7, 7, 1); __PYX_ERR(0, 549, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addTensor_v3", 1, 7, 7, 2); __PYX_ERR(0, 549, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addTensor_v3", 1, 7, 7, 3); __PYX_ERR(0, 549, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addTensor_v3", 1, 7, 7, 4); __PYX_ERR(0, 549, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addTensor_v3", 1, 7, 7, 5); __PYX_ERR(0, 549, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addTensor_v3", 1, 7, 7, 6); __PYX_ERR(0, 549, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "addTensor_v3") < 0)) __PYX_ERR(0, 549, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 549, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 549, __pyx_L3_error)
    __pyx_v_bDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_bDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 549, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 550, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addTensor_v3", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 549, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.addTensor_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_24addTensor_v3(__pyx_self, __pyx_v_handle, __pyx_v_alpha, __pyx_v_bDesc, __pyx_v_b, __pyx_v_beta, __pyx_v_yDesc, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_24addTensor_v3(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_bDesc, size_t __pyx_v_b, size_t __pyx_v_beta, size_t __pyx_v_yDesc, size_t __pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("addTensor_v3", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_addTensor_v3(__pyx_v_handle, __pyx_v_alpha, __pyx_v_bDesc, __pyx_v_b, __pyx_v_beta, __pyx_v_yDesc, __pyx_v_y, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.addTensor_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":563
 * ###############################################################################
 * 
 * cpdef size_t createFilterDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef FilterDescriptor desc
 *     status = cudnnCreateFilterDescriptor(&desc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_27createFilterDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createFilterDescriptor(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnFilterDescriptor_t __pyx_v_desc;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createFilterDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":565
 * cpdef size_t createFilterDescriptor() except *:
 *     cdef FilterDescriptor desc
 *     status = cudnnCreateFilterDescriptor(&desc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>desc
 */
  __pyx_v_status = cudnnCreateFilterDescriptor((&__pyx_v_desc));

  /* "cupy/cuda/cudnn.pyx":566
 *     cdef FilterDescriptor desc
 *     status = cudnnCreateFilterDescriptor(&desc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>desc
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":567
 *     status = cudnnCreateFilterDescriptor(&desc)
 *     check_status(status)
 *     return <size_t>desc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_desc);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":563
 * ###############################################################################
 * 
 * cpdef size_t createFilterDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef FilterDescriptor desc
 *     status = cudnnCreateFilterDescriptor(&desc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createFilterDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_27createFilterDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_26createFilterDescriptor[] = "createFilterDescriptor() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_27createFilterDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createFilterDescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_26createFilterDescriptor(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_26createFilterDescriptor(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createFilterDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createFilterDescriptor(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 563, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createFilterDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":570
 * 
 * 
 * cpdef setFilter4dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t filterDesc, int dataType,
 *         int format, int k, int c, int h, int w):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_29setFilter4dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setFilter4dDescriptor_v4(size_t __pyx_v_filterDesc, int __pyx_v_dataType, int __pyx_v_format, int __pyx_v_k, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setFilter4dDescriptor_v4", 0);

  /* "cupy/cuda/cudnn.pyx":573
 *         size_t filterDesc, int dataType,
 *         int format, int k, int c, int h, int w):
 *     status = cudnnSetFilter4dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         <FilterDescriptor>filterDesc, <DataType> dataType,
 *         <TensorFormat> format, k, c, h, w)
 */
  __pyx_v_status = cudnnSetFilter4dDescriptor_v4(((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((cudnnDataType_t)__pyx_v_dataType), ((cudnnTensorFormat_t)__pyx_v_format), __pyx_v_k, __pyx_v_c, __pyx_v_h, __pyx_v_w);

  /* "cupy/cuda/cudnn.pyx":576
 *         <FilterDescriptor>filterDesc, <DataType> dataType,
 *         <TensorFormat> format, k, c, h, w)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":570
 * 
 * 
 * cpdef setFilter4dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t filterDesc, int dataType,
 *         int format, int k, int c, int h, int w):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setFilter4dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_29setFilter4dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_28setFilter4dDescriptor_v4[] = "setFilter4dDescriptor_v4(size_t filterDesc, int dataType, int format, int k, int c, int h, int w)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_29setFilter4dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_filterDesc;
  int __pyx_v_dataType;
  int __pyx_v_format;
  int __pyx_v_k;
  int __pyx_v_c;
  int __pyx_v_h;
  int __pyx_v_w;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setFilter4dDescriptor_v4 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filterDesc,&__pyx_n_s_dataType,&__pyx_n_s_format,&__pyx_n_s_k,&__pyx_n_s_c,&__pyx_n_s_h,&__pyx_n_s_w,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilter4dDescriptor_v4", 1, 7, 7, 1); __PYX_ERR(0, 570, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilter4dDescriptor_v4", 1, 7, 7, 2); __PYX_ERR(0, 570, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilter4dDescriptor_v4", 1, 7, 7, 3); __PYX_ERR(0, 570, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_c)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilter4dDescriptor_v4", 1, 7, 7, 4); __PYX_ERR(0, 570, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilter4dDescriptor_v4", 1, 7, 7, 5); __PYX_ERR(0, 570, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilter4dDescriptor_v4", 1, 7, 7, 6); __PYX_ERR(0, 570, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setFilter4dDescriptor_v4") < 0)) __PYX_ERR(0, 570, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 571, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 571, __pyx_L3_error)
    __pyx_v_format = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_format == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L3_error)
    __pyx_v_c = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_c == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L3_error)
    __pyx_v_h = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_h == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_w == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setFilter4dDescriptor_v4", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 570, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setFilter4dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_28setFilter4dDescriptor_v4(__pyx_self, __pyx_v_filterDesc, __pyx_v_dataType, __pyx_v_format, __pyx_v_k, __pyx_v_c, __pyx_v_h, __pyx_v_w);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_28setFilter4dDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_filterDesc, int __pyx_v_dataType, int __pyx_v_format, int __pyx_v_k, int __pyx_v_c, int __pyx_v_h, int __pyx_v_w) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setFilter4dDescriptor_v4", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setFilter4dDescriptor_v4(__pyx_v_filterDesc, __pyx_v_dataType, __pyx_v_format, __pyx_v_k, __pyx_v_c, __pyx_v_h, __pyx_v_w, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setFilter4dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":579
 * 
 * 
 * cpdef setFilterNdDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t filterDesc, int dataType,
 *         int format, int nbDims, size_t filterDimA):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_31setFilterNdDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setFilterNdDescriptor_v4(size_t __pyx_v_filterDesc, int __pyx_v_dataType, int __pyx_v_format, int __pyx_v_nbDims, size_t __pyx_v_filterDimA, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setFilterNdDescriptor_v4", 0);

  /* "cupy/cuda/cudnn.pyx":582
 *         size_t filterDesc, int dataType,
 *         int format, int nbDims, size_t filterDimA):
 *     status = cudnnSetFilterNdDescriptor_v4(             # <<<<<<<<<<<<<<
 *         <FilterDescriptor>filterDesc, <DataType>dataType,
 *         <TensorFormat>format, nbDims, <int*>filterDimA)
 */
  __pyx_v_status = cudnnSetFilterNdDescriptor_v4(((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((cudnnDataType_t)__pyx_v_dataType), ((cudnnTensorFormat_t)__pyx_v_format), __pyx_v_nbDims, ((int *)__pyx_v_filterDimA));

  /* "cupy/cuda/cudnn.pyx":585
 *         <FilterDescriptor>filterDesc, <DataType>dataType,
 *         <TensorFormat>format, nbDims, <int*>filterDimA)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":579
 * 
 * 
 * cpdef setFilterNdDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t filterDesc, int dataType,
 *         int format, int nbDims, size_t filterDimA):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setFilterNdDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_31setFilterNdDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_30setFilterNdDescriptor_v4[] = "setFilterNdDescriptor_v4(size_t filterDesc, int dataType, int format, int nbDims, size_t filterDimA)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_31setFilterNdDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_filterDesc;
  int __pyx_v_dataType;
  int __pyx_v_format;
  int __pyx_v_nbDims;
  size_t __pyx_v_filterDimA;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setFilterNdDescriptor_v4 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filterDesc,&__pyx_n_s_dataType,&__pyx_n_s_format,&__pyx_n_s_nbDims,&__pyx_n_s_filterDimA,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilterNdDescriptor_v4", 1, 5, 5, 1); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilterNdDescriptor_v4", 1, 5, 5, 2); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nbDims)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilterNdDescriptor_v4", 1, 5, 5, 3); __PYX_ERR(0, 579, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDimA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFilterNdDescriptor_v4", 1, 5, 5, 4); __PYX_ERR(0, 579, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setFilterNdDescriptor_v4") < 0)) __PYX_ERR(0, 579, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 580, __pyx_L3_error)
    __pyx_v_format = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_format == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L3_error)
    __pyx_v_nbDims = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nbDims == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L3_error)
    __pyx_v_filterDimA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_filterDimA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setFilterNdDescriptor_v4", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 579, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setFilterNdDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_30setFilterNdDescriptor_v4(__pyx_self, __pyx_v_filterDesc, __pyx_v_dataType, __pyx_v_format, __pyx_v_nbDims, __pyx_v_filterDimA);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_30setFilterNdDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_filterDesc, int __pyx_v_dataType, int __pyx_v_format, int __pyx_v_nbDims, size_t __pyx_v_filterDimA) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setFilterNdDescriptor_v4", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setFilterNdDescriptor_v4(__pyx_v_filterDesc, __pyx_v_dataType, __pyx_v_format, __pyx_v_nbDims, __pyx_v_filterDimA, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setFilterNdDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":588
 * 
 * 
 * cpdef getFilterNdDescriptor(size_t wDesc, int nbDimsRequested):             # <<<<<<<<<<<<<<
 *     cdef DataType dataType
 *     cdef TensorFormat format
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_33getFilterNdDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getFilterNdDescriptor(size_t __pyx_v_wDesc, int __pyx_v_nbDimsRequested, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnDataType_t __pyx_v_dataType;
  cudnnTensorFormat_t __pyx_v_format;
  int __pyx_v_nbDims;
  std::vector<int>  __pyx_v_filterDimA;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("getFilterNdDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":593
 *     cdef int nbDims
 *     cdef vector.vector[int] filterDimA
 *     filterDimA.resize(nbDimsRequested)             # <<<<<<<<<<<<<<
 * 
 *     status = cudnnGetFilterNdDescriptor_v4(
 */
  try {
    __pyx_v_filterDimA.resize(__pyx_v_nbDimsRequested);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 593, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":595
 *     filterDimA.resize(nbDimsRequested)
 * 
 *     status = cudnnGetFilterNdDescriptor_v4(             # <<<<<<<<<<<<<<
 *         <FilterDescriptor>wDesc, nbDimsRequested, &dataType,
 *         &format, &nbDims, &filterDimA[0])
 */
  __pyx_v_status = cudnnGetFilterNdDescriptor_v4(((cudnnFilterDescriptor_t)__pyx_v_wDesc), __pyx_v_nbDimsRequested, (&__pyx_v_dataType), (&__pyx_v_format), (&__pyx_v_nbDims), (&(__pyx_v_filterDimA[0])));

  /* "cupy/cuda/cudnn.pyx":598
 *         <FilterDescriptor>wDesc, nbDimsRequested, &dataType,
 *         &format, &nbDims, &filterDimA[0])
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return dataType, format, nbDims, tuple(filterDimA)
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":599
 *         &format, &nbDims, &filterDimA[0])
 *     check_status(status)
 *     return dataType, format, nbDims, tuple(filterDimA)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_cudnnDataType_t(__pyx_v_dataType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_cudnnTensorFormat_t(__pyx_v_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nbDims); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_convert_vector_to_py_int(__pyx_v_filterDimA); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PySequence_Tuple(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":588
 * 
 * 
 * cpdef getFilterNdDescriptor(size_t wDesc, int nbDimsRequested):             # <<<<<<<<<<<<<<
 *     cdef DataType dataType
 *     cdef TensorFormat format
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getFilterNdDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_33getFilterNdDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_32getFilterNdDescriptor[] = "getFilterNdDescriptor(size_t wDesc, int nbDimsRequested)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_33getFilterNdDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_wDesc;
  int __pyx_v_nbDimsRequested;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getFilterNdDescriptor (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_wDesc,&__pyx_n_s_nbDimsRequested,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nbDimsRequested)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getFilterNdDescriptor", 1, 2, 2, 1); __PYX_ERR(0, 588, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getFilterNdDescriptor") < 0)) __PYX_ERR(0, 588, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L3_error)
    __pyx_v_nbDimsRequested = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_nbDimsRequested == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getFilterNdDescriptor", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 588, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getFilterNdDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_32getFilterNdDescriptor(__pyx_self, __pyx_v_wDesc, __pyx_v_nbDimsRequested);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_32getFilterNdDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_wDesc, int __pyx_v_nbDimsRequested) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getFilterNdDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getFilterNdDescriptor(__pyx_v_wDesc, __pyx_v_nbDimsRequested, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getFilterNdDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":602
 * 
 * 
 * cpdef destroyFilterDescriptor(size_t filterDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyFilterDescriptor(<FilterDescriptor>filterDesc)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_35destroyFilterDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_filterDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyFilterDescriptor(size_t __pyx_v_filterDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyFilterDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":603
 * 
 * cpdef destroyFilterDescriptor(size_t filterDesc):
 *     status = cudnnDestroyFilterDescriptor(<FilterDescriptor>filterDesc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cudnnDestroyFilterDescriptor(((cudnnFilterDescriptor_t)__pyx_v_filterDesc));

  /* "cupy/cuda/cudnn.pyx":604
 * cpdef destroyFilterDescriptor(size_t filterDesc):
 *     status = cudnnDestroyFilterDescriptor(<FilterDescriptor>filterDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":602
 * 
 * 
 * cpdef destroyFilterDescriptor(size_t filterDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyFilterDescriptor(<FilterDescriptor>filterDesc)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyFilterDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_35destroyFilterDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_filterDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_34destroyFilterDescriptor[] = "destroyFilterDescriptor(size_t filterDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_35destroyFilterDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_filterDesc) {
  size_t __pyx_v_filterDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyFilterDescriptor (wrapper)", 0);
  assert(__pyx_arg_filterDesc); {
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(__pyx_arg_filterDesc); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 602, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyFilterDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_34destroyFilterDescriptor(__pyx_self, ((size_t)__pyx_v_filterDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_34destroyFilterDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_filterDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyFilterDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroyFilterDescriptor(__pyx_v_filterDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyFilterDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":611
 * ###############################################################################
 * 
 * cpdef size_t createConvolutionDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef ConvolutionDescriptor desc
 *     status = cudnnCreateConvolutionDescriptor(&desc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_37createConvolutionDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createConvolutionDescriptor(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnConvolutionDescriptor_t __pyx_v_desc;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createConvolutionDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":613
 * cpdef size_t createConvolutionDescriptor() except *:
 *     cdef ConvolutionDescriptor desc
 *     status = cudnnCreateConvolutionDescriptor(&desc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>desc
 */
  __pyx_v_status = cudnnCreateConvolutionDescriptor((&__pyx_v_desc));

  /* "cupy/cuda/cudnn.pyx":614
 *     cdef ConvolutionDescriptor desc
 *     status = cudnnCreateConvolutionDescriptor(&desc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>desc
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":615
 *     status = cudnnCreateConvolutionDescriptor(&desc)
 *     check_status(status)
 *     return <size_t>desc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_desc);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":611
 * ###############################################################################
 * 
 * cpdef size_t createConvolutionDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef ConvolutionDescriptor desc
 *     status = cudnnCreateConvolutionDescriptor(&desc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createConvolutionDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_37createConvolutionDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_36createConvolutionDescriptor[] = "createConvolutionDescriptor() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_37createConvolutionDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createConvolutionDescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_36createConvolutionDescriptor(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_36createConvolutionDescriptor(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createConvolutionDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createConvolutionDescriptor(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 611, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createConvolutionDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":618
 * 
 * 
 * cpdef setConvolutionMathType(size_t convDesc, size_t mathType):             # <<<<<<<<<<<<<<
 *     status = cudnnSetConvolutionMathType(
 *         <ConvolutionDescriptor>convDesc, <MathType>mathType)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_39setConvolutionMathType(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolutionMathType(size_t __pyx_v_convDesc, size_t __pyx_v_mathType, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolutionMathType", 0);

  /* "cupy/cuda/cudnn.pyx":619
 * 
 * cpdef setConvolutionMathType(size_t convDesc, size_t mathType):
 *     status = cudnnSetConvolutionMathType(             # <<<<<<<<<<<<<<
 *         <ConvolutionDescriptor>convDesc, <MathType>mathType)
 *     check_status(status)
 */
  __pyx_v_status = cudnnSetConvolutionMathType(((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnMathType_t)__pyx_v_mathType));

  /* "cupy/cuda/cudnn.pyx":621
 *     status = cudnnSetConvolutionMathType(
 *         <ConvolutionDescriptor>convDesc, <MathType>mathType)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":618
 * 
 * 
 * cpdef setConvolutionMathType(size_t convDesc, size_t mathType):             # <<<<<<<<<<<<<<
 *     status = cudnnSetConvolutionMathType(
 *         <ConvolutionDescriptor>convDesc, <MathType>mathType)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionMathType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_39setConvolutionMathType(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_38setConvolutionMathType[] = "setConvolutionMathType(size_t convDesc, size_t mathType)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_39setConvolutionMathType(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_convDesc;
  size_t __pyx_v_mathType;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setConvolutionMathType (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_convDesc,&__pyx_n_s_mathType,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mathType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolutionMathType", 1, 2, 2, 1); __PYX_ERR(0, 618, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setConvolutionMathType") < 0)) __PYX_ERR(0, 618, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
    __pyx_v_mathType = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_mathType == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 618, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setConvolutionMathType", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 618, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionMathType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_38setConvolutionMathType(__pyx_self, __pyx_v_convDesc, __pyx_v_mathType);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_38setConvolutionMathType(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, size_t __pyx_v_mathType) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolutionMathType", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setConvolutionMathType(__pyx_v_convDesc, __pyx_v_mathType, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionMathType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":624
 * 
 * 
 * cpdef size_t getConvolutionMathType(size_t convDesc) except *:             # <<<<<<<<<<<<<<
 *     cdef MathType mathType
 *     status = cudnnGetConvolutionMathType(
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_41getConvolutionMathType(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getConvolutionMathType(size_t __pyx_v_convDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnMathType_t __pyx_v_mathType;
  CYTHON_UNUSED int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionMathType", 0);

  /* "cupy/cuda/cudnn.pyx":626
 * cpdef size_t getConvolutionMathType(size_t convDesc) except *:
 *     cdef MathType mathType
 *     status = cudnnGetConvolutionMathType(             # <<<<<<<<<<<<<<
 *         <ConvolutionDescriptor>convDesc, &mathType)
 *     return <size_t>mathType
 */
  __pyx_v_status = cudnnGetConvolutionMathType(((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), (&__pyx_v_mathType));

  /* "cupy/cuda/cudnn.pyx":628
 *     status = cudnnGetConvolutionMathType(
 *         <ConvolutionDescriptor>convDesc, &mathType)
 *     return <size_t>mathType             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_mathType);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":624
 * 
 * 
 * cpdef size_t getConvolutionMathType(size_t convDesc) except *:             # <<<<<<<<<<<<<<
 *     cdef MathType mathType
 *     status = cudnnGetConvolutionMathType(
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_41getConvolutionMathType(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_40getConvolutionMathType[] = "getConvolutionMathType(size_t convDesc) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_41getConvolutionMathType(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc) {
  size_t __pyx_v_convDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionMathType (wrapper)", 0);
  assert(__pyx_arg_convDesc); {
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(__pyx_arg_convDesc); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionMathType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_40getConvolutionMathType(__pyx_self, ((size_t)__pyx_v_convDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_40getConvolutionMathType(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionMathType", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getConvolutionMathType(__pyx_v_convDesc, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionMathType", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":631
 * 
 * 
 * cpdef setConvolutionGroupCount(size_t convDesc, int groupCount):             # <<<<<<<<<<<<<<
 *     status = cudnnSetConvolutionGroupCount(
 *         <ConvolutionDescriptor>convDesc, groupCount)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_43setConvolutionGroupCount(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolutionGroupCount(size_t __pyx_v_convDesc, int __pyx_v_groupCount, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolutionGroupCount", 0);

  /* "cupy/cuda/cudnn.pyx":632
 * 
 * cpdef setConvolutionGroupCount(size_t convDesc, int groupCount):
 *     status = cudnnSetConvolutionGroupCount(             # <<<<<<<<<<<<<<
 *         <ConvolutionDescriptor>convDesc, groupCount)
 *     check_status(status)
 */
  __pyx_v_status = cudnnSetConvolutionGroupCount(((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), __pyx_v_groupCount);

  /* "cupy/cuda/cudnn.pyx":634
 *     status = cudnnSetConvolutionGroupCount(
 *         <ConvolutionDescriptor>convDesc, groupCount)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":631
 * 
 * 
 * cpdef setConvolutionGroupCount(size_t convDesc, int groupCount):             # <<<<<<<<<<<<<<
 *     status = cudnnSetConvolutionGroupCount(
 *         <ConvolutionDescriptor>convDesc, groupCount)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionGroupCount", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_43setConvolutionGroupCount(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_42setConvolutionGroupCount[] = "setConvolutionGroupCount(size_t convDesc, int groupCount)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_43setConvolutionGroupCount(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_convDesc;
  int __pyx_v_groupCount;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setConvolutionGroupCount (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_convDesc,&__pyx_n_s_groupCount,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_groupCount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolutionGroupCount", 1, 2, 2, 1); __PYX_ERR(0, 631, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setConvolutionGroupCount") < 0)) __PYX_ERR(0, 631, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 631, __pyx_L3_error)
    __pyx_v_groupCount = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_groupCount == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 631, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setConvolutionGroupCount", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 631, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionGroupCount", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_42setConvolutionGroupCount(__pyx_self, __pyx_v_convDesc, __pyx_v_groupCount);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_42setConvolutionGroupCount(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, int __pyx_v_groupCount) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolutionGroupCount", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setConvolutionGroupCount(__pyx_v_convDesc, __pyx_v_groupCount, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionGroupCount", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":637
 * 
 * 
 * cpdef int getConvolutionGroupCount(size_t convDesc) except *:             # <<<<<<<<<<<<<<
 *     cdef int groupCount
 *     status = cudnnGetConvolutionGroupCount(
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_45getConvolutionGroupCount(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc); /*proto*/
static int __pyx_f_4cupy_4cuda_5cudnn_getConvolutionGroupCount(size_t __pyx_v_convDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_groupCount;
  CYTHON_UNUSED int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionGroupCount", 0);

  /* "cupy/cuda/cudnn.pyx":639
 * cpdef int getConvolutionGroupCount(size_t convDesc) except *:
 *     cdef int groupCount
 *     status = cudnnGetConvolutionGroupCount(             # <<<<<<<<<<<<<<
 *         <ConvolutionDescriptor>convDesc, &groupCount)
 *     return groupCount
 */
  __pyx_v_status = cudnnGetConvolutionGroupCount(((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), (&__pyx_v_groupCount));

  /* "cupy/cuda/cudnn.pyx":641
 *     status = cudnnGetConvolutionGroupCount(
 *         <ConvolutionDescriptor>convDesc, &groupCount)
 *     return groupCount             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_groupCount;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":637
 * 
 * 
 * cpdef int getConvolutionGroupCount(size_t convDesc) except *:             # <<<<<<<<<<<<<<
 *     cdef int groupCount
 *     status = cudnnGetConvolutionGroupCount(
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_45getConvolutionGroupCount(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_44getConvolutionGroupCount[] = "getConvolutionGroupCount(size_t convDesc) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_45getConvolutionGroupCount(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc) {
  size_t __pyx_v_convDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionGroupCount (wrapper)", 0);
  assert(__pyx_arg_convDesc); {
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(__pyx_arg_convDesc); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 637, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionGroupCount", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_44getConvolutionGroupCount(__pyx_self, ((size_t)__pyx_v_convDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_44getConvolutionGroupCount(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionGroupCount", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getConvolutionGroupCount(__pyx_v_convDesc, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 637, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionGroupCount", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":644
 * 
 * 
 * cpdef setConvolution2dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t convDesc, int pad_h, int pad_w, int u, int v, int dilation_h,
 *         int dilation_w, int mode):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_47setConvolution2dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolution2dDescriptor_v4(size_t __pyx_v_convDesc, int __pyx_v_pad_h, int __pyx_v_pad_w, int __pyx_v_u, int __pyx_v_v, int __pyx_v_dilation_h, int __pyx_v_dilation_w, int __pyx_v_mode, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolution2dDescriptor_v4", 0);

  /* "cupy/cuda/cudnn.pyx":647
 *         size_t convDesc, int pad_h, int pad_w, int u, int v, int dilation_h,
 *         int dilation_w, int mode):
 *     status = cudnnSetConvolution2dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         <ConvolutionDescriptor>convDesc, pad_h, pad_w, u, v, dilation_h,
 *         dilation_w, <ConvolutionMode>mode)
 */
  __pyx_v_status = cudnnSetConvolution2dDescriptor_v4(((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), __pyx_v_pad_h, __pyx_v_pad_w, __pyx_v_u, __pyx_v_v, __pyx_v_dilation_h, __pyx_v_dilation_w, ((cudnnConvolutionMode_t)__pyx_v_mode));

  /* "cupy/cuda/cudnn.pyx":650
 *         <ConvolutionDescriptor>convDesc, pad_h, pad_w, u, v, dilation_h,
 *         dilation_w, <ConvolutionMode>mode)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":644
 * 
 * 
 * cpdef setConvolution2dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t convDesc, int pad_h, int pad_w, int u, int v, int dilation_h,
 *         int dilation_w, int mode):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolution2dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_47setConvolution2dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_46setConvolution2dDescriptor_v4[] = "setConvolution2dDescriptor_v4(size_t convDesc, int pad_h, int pad_w, int u, int v, int dilation_h, int dilation_w, int mode)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_47setConvolution2dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_convDesc;
  int __pyx_v_pad_h;
  int __pyx_v_pad_w;
  int __pyx_v_u;
  int __pyx_v_v;
  int __pyx_v_dilation_h;
  int __pyx_v_dilation_w;
  int __pyx_v_mode;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setConvolution2dDescriptor_v4 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_convDesc,&__pyx_n_s_pad_h,&__pyx_n_s_pad_w,&__pyx_n_s_u,&__pyx_n_s_v,&__pyx_n_s_dilation_h,&__pyx_n_s_dilation_w,&__pyx_n_s_mode,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pad_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v4", 1, 8, 8, 1); __PYX_ERR(0, 644, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pad_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v4", 1, 8, 8, 2); __PYX_ERR(0, 644, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_u)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v4", 1, 8, 8, 3); __PYX_ERR(0, 644, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v4", 1, 8, 8, 4); __PYX_ERR(0, 644, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dilation_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v4", 1, 8, 8, 5); __PYX_ERR(0, 644, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dilation_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v4", 1, 8, 8, 6); __PYX_ERR(0, 644, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v4", 1, 8, 8, 7); __PYX_ERR(0, 644, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setConvolution2dDescriptor_v4") < 0)) __PYX_ERR(0, 644, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L3_error)
    __pyx_v_pad_h = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_pad_h == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L3_error)
    __pyx_v_pad_w = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_pad_w == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L3_error)
    __pyx_v_u = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_u == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L3_error)
    __pyx_v_v = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_v == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L3_error)
    __pyx_v_dilation_h = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_dilation_h == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L3_error)
    __pyx_v_dilation_w = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_dilation_w == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 646, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 646, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v4", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 644, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolution2dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_46setConvolution2dDescriptor_v4(__pyx_self, __pyx_v_convDesc, __pyx_v_pad_h, __pyx_v_pad_w, __pyx_v_u, __pyx_v_v, __pyx_v_dilation_h, __pyx_v_dilation_w, __pyx_v_mode);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_46setConvolution2dDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, int __pyx_v_pad_h, int __pyx_v_pad_w, int __pyx_v_u, int __pyx_v_v, int __pyx_v_dilation_h, int __pyx_v_dilation_w, int __pyx_v_mode) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolution2dDescriptor_v4", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setConvolution2dDescriptor_v4(__pyx_v_convDesc, __pyx_v_pad_h, __pyx_v_pad_w, __pyx_v_u, __pyx_v_v, __pyx_v_dilation_h, __pyx_v_dilation_w, __pyx_v_mode, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolution2dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":653
 * 
 * 
 * cpdef setConvolution2dDescriptor_v5(             # <<<<<<<<<<<<<<
 *         size_t convDesc, int pad_h, int pad_w, int u, int v, int dilation_h,
 *         int dilation_w, int mode, size_t computeType):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_49setConvolution2dDescriptor_v5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolution2dDescriptor_v5(size_t __pyx_v_convDesc, int __pyx_v_pad_h, int __pyx_v_pad_w, int __pyx_v_u, int __pyx_v_v, int __pyx_v_dilation_h, int __pyx_v_dilation_w, int __pyx_v_mode, size_t __pyx_v_computeType, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolution2dDescriptor_v5", 0);

  /* "cupy/cuda/cudnn.pyx":656
 *         size_t convDesc, int pad_h, int pad_w, int u, int v, int dilation_h,
 *         int dilation_w, int mode, size_t computeType):
 *     status = cudnnSetConvolution2dDescriptor_v5(             # <<<<<<<<<<<<<<
 *         <ConvolutionDescriptor>convDesc, pad_h, pad_w, u, v, dilation_h,
 *         dilation_w, <ConvolutionMode>mode, <DataType>computeType)
 */
  __pyx_v_status = cudnnSetConvolution2dDescriptor_v5(((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), __pyx_v_pad_h, __pyx_v_pad_w, __pyx_v_u, __pyx_v_v, __pyx_v_dilation_h, __pyx_v_dilation_w, ((cudnnConvolutionMode_t)__pyx_v_mode), ((cudnnDataType_t)__pyx_v_computeType));

  /* "cupy/cuda/cudnn.pyx":659
 *         <ConvolutionDescriptor>convDesc, pad_h, pad_w, u, v, dilation_h,
 *         dilation_w, <ConvolutionMode>mode, <DataType>computeType)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":653
 * 
 * 
 * cpdef setConvolution2dDescriptor_v5(             # <<<<<<<<<<<<<<
 *         size_t convDesc, int pad_h, int pad_w, int u, int v, int dilation_h,
 *         int dilation_w, int mode, size_t computeType):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolution2dDescriptor_v5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_49setConvolution2dDescriptor_v5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_48setConvolution2dDescriptor_v5[] = "setConvolution2dDescriptor_v5(size_t convDesc, int pad_h, int pad_w, int u, int v, int dilation_h, int dilation_w, int mode, size_t computeType)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_49setConvolution2dDescriptor_v5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_convDesc;
  int __pyx_v_pad_h;
  int __pyx_v_pad_w;
  int __pyx_v_u;
  int __pyx_v_v;
  int __pyx_v_dilation_h;
  int __pyx_v_dilation_w;
  int __pyx_v_mode;
  size_t __pyx_v_computeType;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setConvolution2dDescriptor_v5 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_convDesc,&__pyx_n_s_pad_h,&__pyx_n_s_pad_w,&__pyx_n_s_u,&__pyx_n_s_v,&__pyx_n_s_dilation_h,&__pyx_n_s_dilation_w,&__pyx_n_s_mode,&__pyx_n_s_computeType,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pad_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, 1); __PYX_ERR(0, 653, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_pad_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, 2); __PYX_ERR(0, 653, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_u)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, 3); __PYX_ERR(0, 653, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, 4); __PYX_ERR(0, 653, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dilation_h)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, 5); __PYX_ERR(0, 653, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dilation_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, 6); __PYX_ERR(0, 653, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, 7); __PYX_ERR(0, 653, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_computeType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, 8); __PYX_ERR(0, 653, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setConvolution2dDescriptor_v5") < 0)) __PYX_ERR(0, 653, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L3_error)
    __pyx_v_pad_h = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_pad_h == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L3_error)
    __pyx_v_pad_w = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_pad_w == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L3_error)
    __pyx_v_u = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_u == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L3_error)
    __pyx_v_v = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_v == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L3_error)
    __pyx_v_dilation_h = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_dilation_h == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L3_error)
    __pyx_v_dilation_w = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_dilation_w == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 655, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 655, __pyx_L3_error)
    __pyx_v_computeType = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_computeType == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 655, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setConvolution2dDescriptor_v5", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 653, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolution2dDescriptor_v5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_48setConvolution2dDescriptor_v5(__pyx_self, __pyx_v_convDesc, __pyx_v_pad_h, __pyx_v_pad_w, __pyx_v_u, __pyx_v_v, __pyx_v_dilation_h, __pyx_v_dilation_w, __pyx_v_mode, __pyx_v_computeType);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_48setConvolution2dDescriptor_v5(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, int __pyx_v_pad_h, int __pyx_v_pad_w, int __pyx_v_u, int __pyx_v_v, int __pyx_v_dilation_h, int __pyx_v_dilation_w, int __pyx_v_mode, size_t __pyx_v_computeType) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolution2dDescriptor_v5", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setConvolution2dDescriptor_v5(__pyx_v_convDesc, __pyx_v_pad_h, __pyx_v_pad_w, __pyx_v_u, __pyx_v_v, __pyx_v_dilation_h, __pyx_v_dilation_w, __pyx_v_mode, __pyx_v_computeType, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolution2dDescriptor_v5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":662
 * 
 * 
 * cpdef setConvolutionNdDescriptor_v3(             # <<<<<<<<<<<<<<
 *         size_t convDesc, int arrayLength, size_t padA, size_t filterStrideA,
 *         size_t dilationA, int mode, int dataType):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_51setConvolutionNdDescriptor_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setConvolutionNdDescriptor_v3(size_t __pyx_v_convDesc, int __pyx_v_arrayLength, size_t __pyx_v_padA, size_t __pyx_v_filterStrideA, size_t __pyx_v_dilationA, int __pyx_v_mode, int __pyx_v_dataType, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolutionNdDescriptor_v3", 0);

  /* "cupy/cuda/cudnn.pyx":665
 *         size_t convDesc, int arrayLength, size_t padA, size_t filterStrideA,
 *         size_t dilationA, int mode, int dataType):
 *     status = cudnnSetConvolutionNdDescriptor_v3(             # <<<<<<<<<<<<<<
 *         <ConvolutionDescriptor>convDesc, arrayLength, <int*>padA,
 *         <int*>filterStrideA, <int*>dilationA, <ConvolutionMode>mode,
 */
  __pyx_v_status = cudnnSetConvolutionNdDescriptor_v3(((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), __pyx_v_arrayLength, ((int *)__pyx_v_padA), ((int *)__pyx_v_filterStrideA), ((int *)__pyx_v_dilationA), ((cudnnConvolutionMode_t)__pyx_v_mode), ((cudnnDataType_t)__pyx_v_dataType));

  /* "cupy/cuda/cudnn.pyx":669
 *         <int*>filterStrideA, <int*>dilationA, <ConvolutionMode>mode,
 *         <DataType>dataType)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":662
 * 
 * 
 * cpdef setConvolutionNdDescriptor_v3(             # <<<<<<<<<<<<<<
 *         size_t convDesc, int arrayLength, size_t padA, size_t filterStrideA,
 *         size_t dilationA, int mode, int dataType):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionNdDescriptor_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_51setConvolutionNdDescriptor_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_50setConvolutionNdDescriptor_v3[] = "setConvolutionNdDescriptor_v3(size_t convDesc, int arrayLength, size_t padA, size_t filterStrideA, size_t dilationA, int mode, int dataType)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_51setConvolutionNdDescriptor_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_convDesc;
  int __pyx_v_arrayLength;
  size_t __pyx_v_padA;
  size_t __pyx_v_filterStrideA;
  size_t __pyx_v_dilationA;
  int __pyx_v_mode;
  int __pyx_v_dataType;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setConvolutionNdDescriptor_v3 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_convDesc,&__pyx_n_s_arrayLength,&__pyx_n_s_padA,&__pyx_n_s_filterStrideA,&__pyx_n_s_dilationA,&__pyx_n_s_mode,&__pyx_n_s_dataType,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_arrayLength)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolutionNdDescriptor_v3", 1, 7, 7, 1); __PYX_ERR(0, 662, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_padA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolutionNdDescriptor_v3", 1, 7, 7, 2); __PYX_ERR(0, 662, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterStrideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolutionNdDescriptor_v3", 1, 7, 7, 3); __PYX_ERR(0, 662, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dilationA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolutionNdDescriptor_v3", 1, 7, 7, 4); __PYX_ERR(0, 662, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolutionNdDescriptor_v3", 1, 7, 7, 5); __PYX_ERR(0, 662, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setConvolutionNdDescriptor_v3", 1, 7, 7, 6); __PYX_ERR(0, 662, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setConvolutionNdDescriptor_v3") < 0)) __PYX_ERR(0, 662, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L3_error)
    __pyx_v_arrayLength = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_arrayLength == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L3_error)
    __pyx_v_padA = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_padA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L3_error)
    __pyx_v_filterStrideA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_filterStrideA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L3_error)
    __pyx_v_dilationA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dilationA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 664, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 664, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 664, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setConvolutionNdDescriptor_v3", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 662, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionNdDescriptor_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_50setConvolutionNdDescriptor_v3(__pyx_self, __pyx_v_convDesc, __pyx_v_arrayLength, __pyx_v_padA, __pyx_v_filterStrideA, __pyx_v_dilationA, __pyx_v_mode, __pyx_v_dataType);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_50setConvolutionNdDescriptor_v3(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc, int __pyx_v_arrayLength, size_t __pyx_v_padA, size_t __pyx_v_filterStrideA, size_t __pyx_v_dilationA, int __pyx_v_mode, int __pyx_v_dataType) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setConvolutionNdDescriptor_v3", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setConvolutionNdDescriptor_v3(__pyx_v_convDesc, __pyx_v_arrayLength, __pyx_v_padA, __pyx_v_filterStrideA, __pyx_v_dilationA, __pyx_v_mode, __pyx_v_dataType, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setConvolutionNdDescriptor_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":672
 * 
 * 
 * cpdef destroyConvolutionDescriptor(size_t convDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyConvolutionDescriptor(
 *         <ConvolutionDescriptor>convDesc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_53destroyConvolutionDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyConvolutionDescriptor(size_t __pyx_v_convDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyConvolutionDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":673
 * 
 * cpdef destroyConvolutionDescriptor(size_t convDesc):
 *     status = cudnnDestroyConvolutionDescriptor(             # <<<<<<<<<<<<<<
 *         <ConvolutionDescriptor>convDesc)
 *     check_status(status)
 */
  __pyx_v_status = cudnnDestroyConvolutionDescriptor(((cudnnConvolutionDescriptor_t)__pyx_v_convDesc));

  /* "cupy/cuda/cudnn.pyx":675
 *     status = cudnnDestroyConvolutionDescriptor(
 *         <ConvolutionDescriptor>convDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":672
 * 
 * 
 * cpdef destroyConvolutionDescriptor(size_t convDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyConvolutionDescriptor(
 *         <ConvolutionDescriptor>convDesc)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyConvolutionDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_53destroyConvolutionDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_52destroyConvolutionDescriptor[] = "destroyConvolutionDescriptor(size_t convDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_53destroyConvolutionDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_convDesc) {
  size_t __pyx_v_convDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyConvolutionDescriptor (wrapper)", 0);
  assert(__pyx_arg_convDesc); {
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(__pyx_arg_convDesc); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 672, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyConvolutionDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_52destroyConvolutionDescriptor(__pyx_self, ((size_t)__pyx_v_convDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_52destroyConvolutionDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_convDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyConvolutionDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroyConvolutionDescriptor(__pyx_v_convDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyConvolutionDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":678
 * 
 * 
 * cpdef findConvolutionForwardAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t xDesc, size_t wDesc, size_t convDesc,
 *         size_t yDesc, int requestedAlgoCount):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_55findConvolutionForwardAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionForwardAlgorithm(size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_convDesc, size_t __pyx_v_yDesc, int __pyx_v_requestedAlgoCount, CYTHON_UNUSED int __pyx_skip_dispatch) {
  std::vector<cudnnConvolutionFwdAlgoPerf_t>  __pyx_v_perfResults;
  int __pyx_v_returnedAlgoCount;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionForwardAlgorithm", 0);

  /* "cupy/cuda/cudnn.pyx":683
 *     cdef vector.vector[ConvolutionFwdAlgoPerf] perfResults
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)             # <<<<<<<<<<<<<<
 *     status = cudnnFindConvolutionForwardAlgorithm(
 *         <Handle> handle, <TensorDescriptor>xDesc, <FilterDescriptor>wDesc,
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_requestedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 683, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":684
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)
 *     status = cudnnFindConvolutionForwardAlgorithm(             # <<<<<<<<<<<<<<
 *         <Handle> handle, <TensorDescriptor>xDesc, <FilterDescriptor>wDesc,
 *         <ConvolutionDescriptor>convDesc, <TensorDescriptor>yDesc,
 */
  __pyx_v_status = cudnnFindConvolutionForwardAlgorithm(((cudnnHandle_t)__pyx_v_handle), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnTensorDescriptor_t)__pyx_v_yDesc), __pyx_v_requestedAlgoCount, (&__pyx_v_returnedAlgoCount), (&(__pyx_v_perfResults[0])));

  /* "cupy/cuda/cudnn.pyx":688
 *         <ConvolutionDescriptor>convDesc, <TensorDescriptor>yDesc,
 *         requestedAlgoCount, &returnedAlgoCount, &perfResults[0])
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":689
 *         requestedAlgoCount, &returnedAlgoCount, &perfResults[0])
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)             # <<<<<<<<<<<<<<
 *     return perfResults
 * 
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_returnedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 689, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":690
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t(__pyx_v_perfResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":678
 * 
 * 
 * cpdef findConvolutionForwardAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t xDesc, size_t wDesc, size_t convDesc,
 *         size_t yDesc, int requestedAlgoCount):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionForwardAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_55findConvolutionForwardAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_54findConvolutionForwardAlgorithm[] = "findConvolutionForwardAlgorithm(size_t handle, size_t xDesc, size_t wDesc, size_t convDesc, size_t yDesc, int requestedAlgoCount)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_55findConvolutionForwardAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_yDesc;
  int __pyx_v_requestedAlgoCount;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findConvolutionForwardAlgorithm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_xDesc,&__pyx_n_s_wDesc,&__pyx_n_s_convDesc,&__pyx_n_s_yDesc,&__pyx_n_s_requestedAlgoCount,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithm", 1, 6, 6, 1); __PYX_ERR(0, 678, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithm", 1, 6, 6, 2); __PYX_ERR(0, 678, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithm", 1, 6, 6, 3); __PYX_ERR(0, 678, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithm", 1, 6, 6, 4); __PYX_ERR(0, 678, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_requestedAlgoCount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithm", 1, 6, 6, 5); __PYX_ERR(0, 678, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findConvolutionForwardAlgorithm") < 0)) __PYX_ERR(0, 678, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L3_error)
    __pyx_v_requestedAlgoCount = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_requestedAlgoCount == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithm", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 678, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionForwardAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_54findConvolutionForwardAlgorithm(__pyx_self, __pyx_v_handle, __pyx_v_xDesc, __pyx_v_wDesc, __pyx_v_convDesc, __pyx_v_yDesc, __pyx_v_requestedAlgoCount);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_54findConvolutionForwardAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_convDesc, size_t __pyx_v_yDesc, int __pyx_v_requestedAlgoCount) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionForwardAlgorithm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_findConvolutionForwardAlgorithm(__pyx_v_handle, __pyx_v_xDesc, __pyx_v_wDesc, __pyx_v_convDesc, __pyx_v_yDesc, __pyx_v_requestedAlgoCount, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionForwardAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":693
 * 
 * 
 * cpdef findConvolutionForwardAlgorithmEx(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t xDesc, size_t x, size_t wDesc, size_t w,
 *         size_t convDesc, size_t yDesc, size_t y, int requestedAlgoCount,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_57findConvolutionForwardAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionForwardAlgorithmEx(size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_convDesc, size_t __pyx_v_yDesc, size_t __pyx_v_y, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  std::vector<cudnnConvolutionFwdAlgoPerf_t>  __pyx_v_perfResults;
  int __pyx_v_returnedAlgoCount;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionForwardAlgorithmEx", 0);

  /* "cupy/cuda/cudnn.pyx":699
 *     cdef vector.vector[ConvolutionFwdAlgoPerf] perfResults
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)             # <<<<<<<<<<<<<<
 *     status = cudnnFindConvolutionForwardAlgorithmEx(
 *         <Handle> handle, <TensorDescriptor>xDesc, <void*>x,
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_requestedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 699, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":700
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)
 *     status = cudnnFindConvolutionForwardAlgorithmEx(             # <<<<<<<<<<<<<<
 *         <Handle> handle, <TensorDescriptor>xDesc, <void*>x,
 *         <FilterDescriptor>wDesc, <void*>w, <ConvolutionDescriptor>convDesc,
 */
  __pyx_v_status = cudnnFindConvolutionForwardAlgorithmEx(((cudnnHandle_t)__pyx_v_handle), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((void *)__pyx_v_x), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((void *)__pyx_v_w), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnTensorDescriptor_t)__pyx_v_yDesc), ((void *)__pyx_v_y), __pyx_v_requestedAlgoCount, (&__pyx_v_returnedAlgoCount), (&(__pyx_v_perfResults[0])), ((void *)__pyx_v_workSpace), __pyx_v_workSpaceSizeInBytes);

  /* "cupy/cuda/cudnn.pyx":706
 *         &returnedAlgoCount, &perfResults[0], <void*>workSpace,
 *         workSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":707
 *         workSpaceSizeInBytes)
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)             # <<<<<<<<<<<<<<
 *     return perfResults
 * 
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_returnedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 707, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":708
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t(__pyx_v_perfResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":693
 * 
 * 
 * cpdef findConvolutionForwardAlgorithmEx(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t xDesc, size_t x, size_t wDesc, size_t w,
 *         size_t convDesc, size_t yDesc, size_t y, int requestedAlgoCount,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionForwardAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_57findConvolutionForwardAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_56findConvolutionForwardAlgorithmEx[] = "findConvolutionForwardAlgorithmEx(size_t handle, size_t xDesc, size_t x, size_t wDesc, size_t w, size_t convDesc, size_t yDesc, size_t y, int requestedAlgoCount, size_t workSpace, size_t workSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_57findConvolutionForwardAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_w;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  int __pyx_v_requestedAlgoCount;
  size_t __pyx_v_workSpace;
  size_t __pyx_v_workSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findConvolutionForwardAlgorithmEx (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_wDesc,&__pyx_n_s_w,&__pyx_n_s_convDesc,&__pyx_n_s_yDesc,&__pyx_n_s_y,&__pyx_n_s_requestedAlgoCount,&__pyx_n_s_workSpace,&__pyx_n_s_workSpaceSizeInBytes,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 1); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 2); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 3); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 4); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 5); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 6); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 7); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_requestedAlgoCount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 8); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 9); __PYX_ERR(0, 693, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, 10); __PYX_ERR(0, 693, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findConvolutionForwardAlgorithmEx") < 0)) __PYX_ERR(0, 693, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_w == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 694, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 695, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 695, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 695, __pyx_L3_error)
    __pyx_v_requestedAlgoCount = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_requestedAlgoCount == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 695, __pyx_L3_error)
    __pyx_v_workSpace = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_workSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 696, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findConvolutionForwardAlgorithmEx", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 693, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionForwardAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_56findConvolutionForwardAlgorithmEx(__pyx_self, __pyx_v_handle, __pyx_v_xDesc, __pyx_v_x, __pyx_v_wDesc, __pyx_v_w, __pyx_v_convDesc, __pyx_v_yDesc, __pyx_v_y, __pyx_v_requestedAlgoCount, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_56findConvolutionForwardAlgorithmEx(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_convDesc, size_t __pyx_v_yDesc, size_t __pyx_v_y, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionForwardAlgorithmEx", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_findConvolutionForwardAlgorithmEx(__pyx_v_handle, __pyx_v_xDesc, __pyx_v_x, __pyx_v_wDesc, __pyx_v_w, __pyx_v_convDesc, __pyx_v_yDesc, __pyx_v_y, __pyx_v_requestedAlgoCount, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionForwardAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":711
 * 
 * 
 * cpdef int getConvolutionForwardAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t srcDesc, size_t filterDesc, size_t convDesc,
 *         size_t destDesc, int preference, size_t memoryLimitInbytes) except *:
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_59getConvolutionForwardAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_5cudnn_getConvolutionForwardAlgorithm(size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_filterDesc, size_t __pyx_v_convDesc, size_t __pyx_v_destDesc, int __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnConvolutionFwdAlgo_t __pyx_v_algo;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionForwardAlgorithm", 0);

  /* "cupy/cuda/cudnn.pyx":715
 *         size_t destDesc, int preference, size_t memoryLimitInbytes) except *:
 *     cdef ConvolutionFwdAlgo algo
 *     status = cudnnGetConvolutionForwardAlgorithm(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <TensorDescriptor>srcDesc,
 *         <FilterDescriptor>filterDesc, <ConvolutionDescriptor>convDesc,
 */
  __pyx_v_status = cudnnGetConvolutionForwardAlgorithm(((cudnnHandle_t)__pyx_v_handle), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnTensorDescriptor_t)__pyx_v_destDesc), ((cudnnConvolutionFwdPreference_t)__pyx_v_preference), __pyx_v_memoryLimitInbytes, (&__pyx_v_algo));

  /* "cupy/cuda/cudnn.pyx":720
 *         <TensorDescriptor>destDesc, <ConvolutionFwdPreference>preference,
 *         memoryLimitInbytes, &algo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return algo
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":721
 *         memoryLimitInbytes, &algo)
 *     check_status(status)
 *     return algo             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_algo;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":711
 * 
 * 
 * cpdef int getConvolutionForwardAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t srcDesc, size_t filterDesc, size_t convDesc,
 *         size_t destDesc, int preference, size_t memoryLimitInbytes) except *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionForwardAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_59getConvolutionForwardAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_58getConvolutionForwardAlgorithm[] = "getConvolutionForwardAlgorithm(size_t handle, size_t srcDesc, size_t filterDesc, size_t convDesc, size_t destDesc, int preference, size_t memoryLimitInbytes) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_59getConvolutionForwardAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_filterDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_destDesc;
  int __pyx_v_preference;
  size_t __pyx_v_memoryLimitInbytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionForwardAlgorithm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_srcDesc,&__pyx_n_s_filterDesc,&__pyx_n_s_convDesc,&__pyx_n_s_destDesc,&__pyx_n_s_preference,&__pyx_n_s_memoryLimitInbytes,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardAlgorithm", 1, 7, 7, 1); __PYX_ERR(0, 711, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardAlgorithm", 1, 7, 7, 2); __PYX_ERR(0, 711, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardAlgorithm", 1, 7, 7, 3); __PYX_ERR(0, 711, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardAlgorithm", 1, 7, 7, 4); __PYX_ERR(0, 711, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_preference)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardAlgorithm", 1, 7, 7, 5); __PYX_ERR(0, 711, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_memoryLimitInbytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardAlgorithm", 1, 7, 7, 6); __PYX_ERR(0, 711, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getConvolutionForwardAlgorithm") < 0)) __PYX_ERR(0, 711, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L3_error)
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L3_error)
    __pyx_v_destDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_destDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 713, __pyx_L3_error)
    __pyx_v_preference = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_preference == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 713, __pyx_L3_error)
    __pyx_v_memoryLimitInbytes = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_memoryLimitInbytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 713, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getConvolutionForwardAlgorithm", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 711, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionForwardAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_58getConvolutionForwardAlgorithm(__pyx_self, __pyx_v_handle, __pyx_v_srcDesc, __pyx_v_filterDesc, __pyx_v_convDesc, __pyx_v_destDesc, __pyx_v_preference, __pyx_v_memoryLimitInbytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_58getConvolutionForwardAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_filterDesc, size_t __pyx_v_convDesc, size_t __pyx_v_destDesc, int __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionForwardAlgorithm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getConvolutionForwardAlgorithm(__pyx_v_handle, __pyx_v_srcDesc, __pyx_v_filterDesc, __pyx_v_convDesc, __pyx_v_destDesc, __pyx_v_preference, __pyx_v_memoryLimitInbytes, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 711, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionForwardAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":724
 * 
 * 
 * cpdef size_t getConvolutionForwardWorkspaceSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t srcDesc, size_t filterDesc, size_t convDesc,
 *         size_t destDesc, int algo) except *:
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_61getConvolutionForwardWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getConvolutionForwardWorkspaceSize(size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_filterDesc, size_t __pyx_v_convDesc, size_t __pyx_v_destDesc, int __pyx_v_algo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_sizeInBytes;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionForwardWorkspaceSize", 0);

  /* "cupy/cuda/cudnn.pyx":728
 *         size_t destDesc, int algo) except *:
 *     cdef size_t sizeInBytes
 *     status = cudnnGetConvolutionForwardWorkspaceSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <TensorDescriptor>srcDesc,
 *         <FilterDescriptor>filterDesc, <ConvolutionDescriptor> convDesc,
 */
  __pyx_v_status = cudnnGetConvolutionForwardWorkspaceSize(((cudnnHandle_t)__pyx_v_handle), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnTensorDescriptor_t)__pyx_v_destDesc), ((cudnnConvolutionFwdAlgo_t)__pyx_v_algo), (&__pyx_v_sizeInBytes));

  /* "cupy/cuda/cudnn.pyx":732
 *         <FilterDescriptor>filterDesc, <ConvolutionDescriptor> convDesc,
 *         <TensorDescriptor>destDesc, <ConvolutionFwdAlgo>algo, &sizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return sizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":733
 *         <TensorDescriptor>destDesc, <ConvolutionFwdAlgo>algo, &sizeInBytes)
 *     check_status(status)
 *     return sizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_sizeInBytes;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":724
 * 
 * 
 * cpdef size_t getConvolutionForwardWorkspaceSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t srcDesc, size_t filterDesc, size_t convDesc,
 *         size_t destDesc, int algo) except *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionForwardWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_61getConvolutionForwardWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_60getConvolutionForwardWorkspaceSize[] = "getConvolutionForwardWorkspaceSize(size_t handle, size_t srcDesc, size_t filterDesc, size_t convDesc, size_t destDesc, int algo) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_61getConvolutionForwardWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_filterDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_destDesc;
  int __pyx_v_algo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionForwardWorkspaceSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_srcDesc,&__pyx_n_s_filterDesc,&__pyx_n_s_convDesc,&__pyx_n_s_destDesc,&__pyx_n_s_algo,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardWorkspaceSize", 1, 6, 6, 1); __PYX_ERR(0, 724, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardWorkspaceSize", 1, 6, 6, 2); __PYX_ERR(0, 724, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardWorkspaceSize", 1, 6, 6, 3); __PYX_ERR(0, 724, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardWorkspaceSize", 1, 6, 6, 4); __PYX_ERR(0, 724, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionForwardWorkspaceSize", 1, 6, 6, 5); __PYX_ERR(0, 724, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getConvolutionForwardWorkspaceSize") < 0)) __PYX_ERR(0, 724, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 725, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 725, __pyx_L3_error)
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 725, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 725, __pyx_L3_error)
    __pyx_v_destDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_destDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
    __pyx_v_algo = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_algo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getConvolutionForwardWorkspaceSize", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 724, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionForwardWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_60getConvolutionForwardWorkspaceSize(__pyx_self, __pyx_v_handle, __pyx_v_srcDesc, __pyx_v_filterDesc, __pyx_v_convDesc, __pyx_v_destDesc, __pyx_v_algo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_60getConvolutionForwardWorkspaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_filterDesc, size_t __pyx_v_convDesc, size_t __pyx_v_destDesc, int __pyx_v_algo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionForwardWorkspaceSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getConvolutionForwardWorkspaceSize(__pyx_v_handle, __pyx_v_srcDesc, __pyx_v_filterDesc, __pyx_v_convDesc, __pyx_v_destDesc, __pyx_v_algo, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 724, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionForwardWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":736
 * 
 * 
 * cpdef convolutionForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t alpha, size_t srcDesc, size_t srcData,
 *         size_t filterDesc, size_t filterData, size_t convDesc, int algo,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_63convolutionForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_convolutionForward(size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_filterDesc, size_t __pyx_v_filterData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_destDesc, size_t __pyx_v_destData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("convolutionForward", 0);

  /* "cupy/cuda/cudnn.pyx":741
 *         size_t workSpace, size_t workSpaceSizeInBytes, size_t beta,
 *         size_t destDesc, size_t destData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnConvolutionForward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":742
 *         size_t destDesc, size_t destData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnConvolutionForward(
 *             <Handle>handle, <void*>alpha,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":743
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnConvolutionForward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <void*>alpha,
 *             <TensorDescriptor>srcDesc, <void*>srcData,
 */
        __pyx_v_status = cudnnConvolutionForward(((cudnnHandle_t)__pyx_v_handle), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((void *)__pyx_v_filterData), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnConvolutionFwdAlgo_t)__pyx_v_algo), ((void *)__pyx_v_workSpace), __pyx_v_workSpaceSizeInBytes, ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_destDesc), ((void *)__pyx_v_destData));
      }

      /* "cupy/cuda/cudnn.pyx":742
 *         size_t destDesc, size_t destData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnConvolutionForward(
 *             <Handle>handle, <void*>alpha,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":750
 *             <void*>workSpace, workSpaceSizeInBytes, <void*>beta,
 *             <TensorDescriptor>destDesc, <void*>destData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":736
 * 
 * 
 * cpdef convolutionForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t alpha, size_t srcDesc, size_t srcData,
 *         size_t filterDesc, size_t filterData, size_t convDesc, int algo,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_63convolutionForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_62convolutionForward[] = "convolutionForward(size_t handle, size_t alpha, size_t srcDesc, size_t srcData, size_t filterDesc, size_t filterData, size_t convDesc, int algo, size_t workSpace, size_t workSpaceSizeInBytes, size_t beta, size_t destDesc, size_t destData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_63convolutionForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_filterDesc;
  size_t __pyx_v_filterData;
  size_t __pyx_v_convDesc;
  int __pyx_v_algo;
  size_t __pyx_v_workSpace;
  size_t __pyx_v_workSpaceSizeInBytes;
  size_t __pyx_v_beta;
  size_t __pyx_v_destDesc;
  size_t __pyx_v_destData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convolutionForward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_filterDesc,&__pyx_n_s_filterData,&__pyx_n_s_convDesc,&__pyx_n_s_algo,&__pyx_n_s_workSpace,&__pyx_n_s_workSpaceSizeInBytes,&__pyx_n_s_beta,&__pyx_n_s_destDesc,&__pyx_n_s_destData,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 1); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 2); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 3); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 4); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 5); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 6); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 7); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 8); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 9); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 10); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 11); __PYX_ERR(0, 736, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, 12); __PYX_ERR(0, 736, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convolutionForward") < 0)) __PYX_ERR(0, 736, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 737, __pyx_L3_error)
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_filterData = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_filterData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_algo = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_algo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_workSpace = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_workSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L3_error)
    __pyx_v_destDesc = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_destDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 740, __pyx_L3_error)
    __pyx_v_destData = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_destData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 740, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convolutionForward", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 736, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_62convolutionForward(__pyx_self, __pyx_v_handle, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_filterDesc, __pyx_v_filterData, __pyx_v_convDesc, __pyx_v_algo, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, __pyx_v_beta, __pyx_v_destDesc, __pyx_v_destData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_62convolutionForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_filterDesc, size_t __pyx_v_filterData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_destDesc, size_t __pyx_v_destData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("convolutionForward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_convolutionForward(__pyx_v_handle, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_filterDesc, __pyx_v_filterData, __pyx_v_convDesc, __pyx_v_algo, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, __pyx_v_beta, __pyx_v_destDesc, __pyx_v_destData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":753
 * 
 * 
 * cpdef convolutionBackwardBias(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t alpha, size_t srcDesc, size_t srcData,
 *         size_t beta, size_t destDesc, size_t destData):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_65convolutionBackwardBias(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardBias(size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_destDesc, size_t __pyx_v_destData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("convolutionBackwardBias", 0);

  /* "cupy/cuda/cudnn.pyx":756
 *         size_t handle, size_t alpha, size_t srcDesc, size_t srcData,
 *         size_t beta, size_t destDesc, size_t destData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnConvolutionBackwardBias(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":757
 *         size_t beta, size_t destDesc, size_t destData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnConvolutionBackwardBias(
 *             <Handle>handle, <void*>alpha,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":758
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnConvolutionBackwardBias(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <void*>alpha,
 *             <TensorDescriptor>srcDesc, <void*>srcData, <void*>beta,
 */
        __pyx_v_status = cudnnConvolutionBackwardBias(((cudnnHandle_t)__pyx_v_handle), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_destDesc), ((void *)__pyx_v_destData));
      }

      /* "cupy/cuda/cudnn.pyx":757
 *         size_t beta, size_t destDesc, size_t destData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnConvolutionBackwardBias(
 *             <Handle>handle, <void*>alpha,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":762
 *             <TensorDescriptor>srcDesc, <void*>srcData, <void*>beta,
 *             <TensorDescriptor>destDesc, <void*>destData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":753
 * 
 * 
 * cpdef convolutionBackwardBias(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t alpha, size_t srcDesc, size_t srcData,
 *         size_t beta, size_t destDesc, size_t destData):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardBias", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_65convolutionBackwardBias(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_64convolutionBackwardBias[] = "convolutionBackwardBias(size_t handle, size_t alpha, size_t srcDesc, size_t srcData, size_t beta, size_t destDesc, size_t destData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_65convolutionBackwardBias(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_beta;
  size_t __pyx_v_destDesc;
  size_t __pyx_v_destData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convolutionBackwardBias (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_beta,&__pyx_n_s_destDesc,&__pyx_n_s_destData,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardBias", 1, 7, 7, 1); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardBias", 1, 7, 7, 2); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardBias", 1, 7, 7, 3); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardBias", 1, 7, 7, 4); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardBias", 1, 7, 7, 5); __PYX_ERR(0, 753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardBias", 1, 7, 7, 6); __PYX_ERR(0, 753, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convolutionBackwardBias") < 0)) __PYX_ERR(0, 753, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L3_error)
    __pyx_v_destDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_destDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L3_error)
    __pyx_v_destData = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_destData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convolutionBackwardBias", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 753, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardBias", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_64convolutionBackwardBias(__pyx_self, __pyx_v_handle, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_beta, __pyx_v_destDesc, __pyx_v_destData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_64convolutionBackwardBias(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_destDesc, size_t __pyx_v_destData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("convolutionBackwardBias", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardBias(__pyx_v_handle, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_beta, __pyx_v_destDesc, __pyx_v_destData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardBias", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":765
 * 
 * 
 * cpdef findConvolutionBackwardFilterAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t xDesc, size_t dyDesc, size_t convDesc,
 *         size_t dwDesc, int requestedAlgoCount):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_67findConvolutionBackwardFilterAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardFilterAlgorithm(size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_convDesc, size_t __pyx_v_dwDesc, int __pyx_v_requestedAlgoCount, CYTHON_UNUSED int __pyx_skip_dispatch) {
  std::vector<cudnnConvolutionBwdFilterAlgoPerf_t>  __pyx_v_perfResults;
  int __pyx_v_returnedAlgoCount;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionBackwardFilterAlgorithm", 0);

  /* "cupy/cuda/cudnn.pyx":770
 *     cdef vector.vector[ConvolutionBwdFilterAlgoPerf] perfResults
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)             # <<<<<<<<<<<<<<
 *     status = cudnnFindConvolutionBackwardFilterAlgorithm(
 *         <Handle> handle, <TensorDescriptor>xDesc, <TensorDescriptor>dyDesc,
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_requestedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 770, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":771
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)
 *     status = cudnnFindConvolutionBackwardFilterAlgorithm(             # <<<<<<<<<<<<<<
 *         <Handle> handle, <TensorDescriptor>xDesc, <TensorDescriptor>dyDesc,
 *         <ConvolutionDescriptor>convDesc, <FilterDescriptor>dwDesc,
 */
  __pyx_v_status = cudnnFindConvolutionBackwardFilterAlgorithm(((cudnnHandle_t)__pyx_v_handle), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((cudnnTensorDescriptor_t)__pyx_v_dyDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnFilterDescriptor_t)__pyx_v_dwDesc), __pyx_v_requestedAlgoCount, (&__pyx_v_returnedAlgoCount), (&(__pyx_v_perfResults[0])));

  /* "cupy/cuda/cudnn.pyx":775
 *         <ConvolutionDescriptor>convDesc, <FilterDescriptor>dwDesc,
 *         requestedAlgoCount, &returnedAlgoCount, &perfResults[0])
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":776
 *         requestedAlgoCount, &returnedAlgoCount, &perfResults[0])
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)             # <<<<<<<<<<<<<<
 *     return perfResults
 * 
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_returnedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 776, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":777
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t(__pyx_v_perfResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":765
 * 
 * 
 * cpdef findConvolutionBackwardFilterAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t xDesc, size_t dyDesc, size_t convDesc,
 *         size_t dwDesc, int requestedAlgoCount):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardFilterAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_67findConvolutionBackwardFilterAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_66findConvolutionBackwardFilterAlgorithm[] = "findConvolutionBackwardFilterAlgorithm(size_t handle, size_t xDesc, size_t dyDesc, size_t convDesc, size_t dwDesc, int requestedAlgoCount)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_67findConvolutionBackwardFilterAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_dyDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_dwDesc;
  int __pyx_v_requestedAlgoCount;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findConvolutionBackwardFilterAlgorithm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_xDesc,&__pyx_n_s_dyDesc,&__pyx_n_s_convDesc,&__pyx_n_s_dwDesc,&__pyx_n_s_requestedAlgoCount,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithm", 1, 6, 6, 1); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithm", 1, 6, 6, 2); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithm", 1, 6, 6, 3); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dwDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithm", 1, 6, 6, 4); __PYX_ERR(0, 765, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_requestedAlgoCount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithm", 1, 6, 6, 5); __PYX_ERR(0, 765, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findConvolutionBackwardFilterAlgorithm") < 0)) __PYX_ERR(0, 765, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_dyDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_dyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_dwDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dwDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
    __pyx_v_requestedAlgoCount = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_requestedAlgoCount == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithm", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 765, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardFilterAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_66findConvolutionBackwardFilterAlgorithm(__pyx_self, __pyx_v_handle, __pyx_v_xDesc, __pyx_v_dyDesc, __pyx_v_convDesc, __pyx_v_dwDesc, __pyx_v_requestedAlgoCount);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_66findConvolutionBackwardFilterAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_convDesc, size_t __pyx_v_dwDesc, int __pyx_v_requestedAlgoCount) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionBackwardFilterAlgorithm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardFilterAlgorithm(__pyx_v_handle, __pyx_v_xDesc, __pyx_v_dyDesc, __pyx_v_convDesc, __pyx_v_dwDesc, __pyx_v_requestedAlgoCount, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardFilterAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":780
 * 
 * 
 * cpdef findConvolutionBackwardFilterAlgorithmEx(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t xDesc, size_t x, size_t dyDesc, size_t dy,
 *         size_t convDesc, size_t dwDesc, size_t dw, int requestedAlgoCount,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_69findConvolutionBackwardFilterAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardFilterAlgorithmEx(size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_convDesc, size_t __pyx_v_dwDesc, size_t __pyx_v_dw, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  std::vector<cudnnConvolutionBwdFilterAlgoPerf_t>  __pyx_v_perfResults;
  int __pyx_v_returnedAlgoCount;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionBackwardFilterAlgorithmEx", 0);

  /* "cupy/cuda/cudnn.pyx":786
 *     cdef vector.vector[ConvolutionBwdFilterAlgoPerf] perfResults
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)             # <<<<<<<<<<<<<<
 *     status = cudnnFindConvolutionBackwardFilterAlgorithmEx(
 *         <Handle> handle, <TensorDescriptor>xDesc, <void*>x,
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_requestedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 786, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":787
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)
 *     status = cudnnFindConvolutionBackwardFilterAlgorithmEx(             # <<<<<<<<<<<<<<
 *         <Handle> handle, <TensorDescriptor>xDesc, <void*>x,
 *         <TensorDescriptor>dyDesc, <void*>dy, <ConvolutionDescriptor>convDesc,
 */
  __pyx_v_status = cudnnFindConvolutionBackwardFilterAlgorithmEx(((cudnnHandle_t)__pyx_v_handle), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((void *)__pyx_v_x), ((cudnnTensorDescriptor_t)__pyx_v_dyDesc), ((void *)__pyx_v_dy), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnFilterDescriptor_t)__pyx_v_dwDesc), ((void *)__pyx_v_dw), __pyx_v_requestedAlgoCount, (&__pyx_v_returnedAlgoCount), (&(__pyx_v_perfResults[0])), ((void *)__pyx_v_workSpace), __pyx_v_workSpaceSizeInBytes);

  /* "cupy/cuda/cudnn.pyx":793
 *         requestedAlgoCount, &returnedAlgoCount, &perfResults[0],
 *         <void*>workSpace, workSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":794
 *         <void*>workSpace, workSpaceSizeInBytes)
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)             # <<<<<<<<<<<<<<
 *     return perfResults
 * 
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_returnedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 794, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":795
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t(__pyx_v_perfResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":780
 * 
 * 
 * cpdef findConvolutionBackwardFilterAlgorithmEx(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t xDesc, size_t x, size_t dyDesc, size_t dy,
 *         size_t convDesc, size_t dwDesc, size_t dw, int requestedAlgoCount,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardFilterAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_69findConvolutionBackwardFilterAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_68findConvolutionBackwardFilterAlgorithmEx[] = "findConvolutionBackwardFilterAlgorithmEx(size_t handle, size_t xDesc, size_t x, size_t dyDesc, size_t dy, size_t convDesc, size_t dwDesc, size_t dw, int requestedAlgoCount, size_t workSpace, size_t workSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_69findConvolutionBackwardFilterAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_dyDesc;
  size_t __pyx_v_dy;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_dwDesc;
  size_t __pyx_v_dw;
  int __pyx_v_requestedAlgoCount;
  size_t __pyx_v_workSpace;
  size_t __pyx_v_workSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findConvolutionBackwardFilterAlgorithmEx (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_dyDesc,&__pyx_n_s_dy,&__pyx_n_s_convDesc,&__pyx_n_s_dwDesc,&__pyx_n_s_dw,&__pyx_n_s_requestedAlgoCount,&__pyx_n_s_workSpace,&__pyx_n_s_workSpaceSizeInBytes,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 1); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 2); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 3); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 4); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 5); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dwDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 6); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dw)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 7); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_requestedAlgoCount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 8); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 9); __PYX_ERR(0, 780, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, 10); __PYX_ERR(0, 780, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findConvolutionBackwardFilterAlgorithmEx") < 0)) __PYX_ERR(0, 780, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 781, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 781, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 781, __pyx_L3_error)
    __pyx_v_dyDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_dyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 781, __pyx_L3_error)
    __pyx_v_dy = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 781, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 782, __pyx_L3_error)
    __pyx_v_dwDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_dwDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 782, __pyx_L3_error)
    __pyx_v_dw = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_dw == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 782, __pyx_L3_error)
    __pyx_v_requestedAlgoCount = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_requestedAlgoCount == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 782, __pyx_L3_error)
    __pyx_v_workSpace = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_workSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardFilterAlgorithmEx", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 780, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardFilterAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_68findConvolutionBackwardFilterAlgorithmEx(__pyx_self, __pyx_v_handle, __pyx_v_xDesc, __pyx_v_x, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_convDesc, __pyx_v_dwDesc, __pyx_v_dw, __pyx_v_requestedAlgoCount, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_68findConvolutionBackwardFilterAlgorithmEx(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_convDesc, size_t __pyx_v_dwDesc, size_t __pyx_v_dw, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionBackwardFilterAlgorithmEx", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardFilterAlgorithmEx(__pyx_v_handle, __pyx_v_xDesc, __pyx_v_x, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_convDesc, __pyx_v_dwDesc, __pyx_v_dw, __pyx_v_requestedAlgoCount, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardFilterAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":798
 * 
 * 
 * cpdef int getConvolutionBackwardFilterAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t srcDesc, size_t diffDesc, size_t convDesc,
 *         size_t filterDesc, int preference, size_t memoryLimitInbytes) except *:
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_71getConvolutionBackwardFilterAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardFilterAlgorithm(size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_filterDesc, int __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnConvolutionBwdFilterAlgo_t __pyx_v_algo;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionBackwardFilterAlgorithm", 0);

  /* "cupy/cuda/cudnn.pyx":802
 *         size_t filterDesc, int preference, size_t memoryLimitInbytes) except *:
 *     cdef ConvolutionBwdFilterAlgo algo
 *     status = cudnnGetConvolutionBackwardFilterAlgorithm(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <TensorDescriptor>srcDesc,
 *         <TensorDescriptor>diffDesc, <ConvolutionDescriptor>convDesc,
 */
  __pyx_v_status = cudnnGetConvolutionBackwardFilterAlgorithm(((cudnnHandle_t)__pyx_v_handle), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((cudnnTensorDescriptor_t)__pyx_v_diffDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((cudnnConvolutionBwdFilterPreference_t)__pyx_v_preference), __pyx_v_memoryLimitInbytes, (&__pyx_v_algo));

  /* "cupy/cuda/cudnn.pyx":808
 *         <ConvolutionBwdFilterPreference>preference,
 *         memoryLimitInbytes, &algo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return algo
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":809
 *         memoryLimitInbytes, &algo)
 *     check_status(status)
 *     return algo             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_algo;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":798
 * 
 * 
 * cpdef int getConvolutionBackwardFilterAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t srcDesc, size_t diffDesc, size_t convDesc,
 *         size_t filterDesc, int preference, size_t memoryLimitInbytes) except *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardFilterAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_71getConvolutionBackwardFilterAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_70getConvolutionBackwardFilterAlgorithm[] = "getConvolutionBackwardFilterAlgorithm(size_t handle, size_t srcDesc, size_t diffDesc, size_t convDesc, size_t filterDesc, int preference, size_t memoryLimitInbytes) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_71getConvolutionBackwardFilterAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_diffDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_filterDesc;
  int __pyx_v_preference;
  size_t __pyx_v_memoryLimitInbytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionBackwardFilterAlgorithm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_srcDesc,&__pyx_n_s_diffDesc,&__pyx_n_s_convDesc,&__pyx_n_s_filterDesc,&__pyx_n_s_preference,&__pyx_n_s_memoryLimitInbytes,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterAlgorithm", 1, 7, 7, 1); __PYX_ERR(0, 798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_diffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterAlgorithm", 1, 7, 7, 2); __PYX_ERR(0, 798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterAlgorithm", 1, 7, 7, 3); __PYX_ERR(0, 798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterAlgorithm", 1, 7, 7, 4); __PYX_ERR(0, 798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_preference)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterAlgorithm", 1, 7, 7, 5); __PYX_ERR(0, 798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_memoryLimitInbytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterAlgorithm", 1, 7, 7, 6); __PYX_ERR(0, 798, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getConvolutionBackwardFilterAlgorithm") < 0)) __PYX_ERR(0, 798, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 799, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 799, __pyx_L3_error)
    __pyx_v_diffDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_diffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 799, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 799, __pyx_L3_error)
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 800, __pyx_L3_error)
    __pyx_v_preference = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_preference == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 800, __pyx_L3_error)
    __pyx_v_memoryLimitInbytes = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_memoryLimitInbytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 800, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterAlgorithm", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 798, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardFilterAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_70getConvolutionBackwardFilterAlgorithm(__pyx_self, __pyx_v_handle, __pyx_v_srcDesc, __pyx_v_diffDesc, __pyx_v_convDesc, __pyx_v_filterDesc, __pyx_v_preference, __pyx_v_memoryLimitInbytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_70getConvolutionBackwardFilterAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_filterDesc, int __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionBackwardFilterAlgorithm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardFilterAlgorithm(__pyx_v_handle, __pyx_v_srcDesc, __pyx_v_diffDesc, __pyx_v_convDesc, __pyx_v_filterDesc, __pyx_v_preference, __pyx_v_memoryLimitInbytes, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardFilterAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":812
 * 
 * 
 * cpdef size_t getConvolutionBackwardFilterWorkspaceSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t srcDesc, size_t diffDesc, size_t convDesc,
 *         size_t filterDesc, int algo) except *:
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_73getConvolutionBackwardFilterWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardFilterWorkspaceSize(size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_filterDesc, int __pyx_v_algo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_sizeInBytes;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionBackwardFilterWorkspaceSize", 0);

  /* "cupy/cuda/cudnn.pyx":816
 *         size_t filterDesc, int algo) except *:
 *     cdef size_t sizeInBytes
 *     status = cudnnGetConvolutionBackwardFilterWorkspaceSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <TensorDescriptor>srcDesc,
 *         <TensorDescriptor>diffDesc, <ConvolutionDescriptor> convDesc,
 */
  __pyx_v_status = cudnnGetConvolutionBackwardFilterWorkspaceSize(((cudnnHandle_t)__pyx_v_handle), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((cudnnTensorDescriptor_t)__pyx_v_diffDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((cudnnConvolutionBwdFilterAlgo_t)__pyx_v_algo), (&__pyx_v_sizeInBytes));

  /* "cupy/cuda/cudnn.pyx":821
 *         <FilterDescriptor>filterDesc, <ConvolutionBwdFilterAlgo>algo,
 *         &sizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return sizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":822
 *         &sizeInBytes)
 *     check_status(status)
 *     return sizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_sizeInBytes;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":812
 * 
 * 
 * cpdef size_t getConvolutionBackwardFilterWorkspaceSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t srcDesc, size_t diffDesc, size_t convDesc,
 *         size_t filterDesc, int algo) except *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardFilterWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_73getConvolutionBackwardFilterWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_72getConvolutionBackwardFilterWorkspaceSize[] = "getConvolutionBackwardFilterWorkspaceSize(size_t handle, size_t srcDesc, size_t diffDesc, size_t convDesc, size_t filterDesc, int algo) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_73getConvolutionBackwardFilterWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_diffDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_filterDesc;
  int __pyx_v_algo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionBackwardFilterWorkspaceSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_srcDesc,&__pyx_n_s_diffDesc,&__pyx_n_s_convDesc,&__pyx_n_s_filterDesc,&__pyx_n_s_algo,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterWorkspaceSize", 1, 6, 6, 1); __PYX_ERR(0, 812, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_diffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterWorkspaceSize", 1, 6, 6, 2); __PYX_ERR(0, 812, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterWorkspaceSize", 1, 6, 6, 3); __PYX_ERR(0, 812, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterWorkspaceSize", 1, 6, 6, 4); __PYX_ERR(0, 812, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterWorkspaceSize", 1, 6, 6, 5); __PYX_ERR(0, 812, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getConvolutionBackwardFilterWorkspaceSize") < 0)) __PYX_ERR(0, 812, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 813, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 813, __pyx_L3_error)
    __pyx_v_diffDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_diffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 813, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 813, __pyx_L3_error)
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 814, __pyx_L3_error)
    __pyx_v_algo = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_algo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 814, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardFilterWorkspaceSize", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 812, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardFilterWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_72getConvolutionBackwardFilterWorkspaceSize(__pyx_self, __pyx_v_handle, __pyx_v_srcDesc, __pyx_v_diffDesc, __pyx_v_convDesc, __pyx_v_filterDesc, __pyx_v_algo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_72getConvolutionBackwardFilterWorkspaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_srcDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_filterDesc, int __pyx_v_algo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionBackwardFilterWorkspaceSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardFilterWorkspaceSize(__pyx_v_handle, __pyx_v_srcDesc, __pyx_v_diffDesc, __pyx_v_convDesc, __pyx_v_filterDesc, __pyx_v_algo, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 812, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardFilterWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":825
 * 
 * 
 * cpdef convolutionBackwardFilter_v3(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t alpha, size_t srcDesc, size_t srcData,
 *         size_t diffDesc, size_t diffData, size_t convDesc, int algo,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_75convolutionBackwardFilter_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardFilter_v3(size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_diffDesc, size_t __pyx_v_diffData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_gradDesc, size_t __pyx_v_gradData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("convolutionBackwardFilter_v3", 0);

  /* "cupy/cuda/cudnn.pyx":830
 *         size_t workSpace, size_t workSpaceSizeInBytes, size_t beta,
 *         size_t gradDesc, size_t gradData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnConvolutionBackwardFilter_v3(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":831
 *         size_t gradDesc, size_t gradData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnConvolutionBackwardFilter_v3(
 *             <Handle>handle, <void*>alpha,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":832
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnConvolutionBackwardFilter_v3(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <void*>alpha,
 *             <TensorDescriptor>srcDesc, <void*>srcData,
 */
        __pyx_v_status = cudnnConvolutionBackwardFilter_v3(((cudnnHandle_t)__pyx_v_handle), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((cudnnTensorDescriptor_t)__pyx_v_diffDesc), ((void *)__pyx_v_diffData), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnConvolutionBwdFilterAlgo_t)__pyx_v_algo), ((void *)__pyx_v_workSpace), __pyx_v_workSpaceSizeInBytes, ((void *)__pyx_v_beta), ((cudnnFilterDescriptor_t)__pyx_v_gradDesc), ((void *)__pyx_v_gradData));
      }

      /* "cupy/cuda/cudnn.pyx":831
 *         size_t gradDesc, size_t gradData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnConvolutionBackwardFilter_v3(
 *             <Handle>handle, <void*>alpha,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":839
 *             <void*>workSpace, workSpaceSizeInBytes, <void*>beta,
 *             <FilterDescriptor>gradDesc, <void*>gradData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":825
 * 
 * 
 * cpdef convolutionBackwardFilter_v3(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t alpha, size_t srcDesc, size_t srcData,
 *         size_t diffDesc, size_t diffData, size_t convDesc, int algo,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardFilter_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_75convolutionBackwardFilter_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_74convolutionBackwardFilter_v3[] = "convolutionBackwardFilter_v3(size_t handle, size_t alpha, size_t srcDesc, size_t srcData, size_t diffDesc, size_t diffData, size_t convDesc, int algo, size_t workSpace, size_t workSpaceSizeInBytes, size_t beta, size_t gradDesc, size_t gradData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_75convolutionBackwardFilter_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_diffDesc;
  size_t __pyx_v_diffData;
  size_t __pyx_v_convDesc;
  int __pyx_v_algo;
  size_t __pyx_v_workSpace;
  size_t __pyx_v_workSpaceSizeInBytes;
  size_t __pyx_v_beta;
  size_t __pyx_v_gradDesc;
  size_t __pyx_v_gradData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convolutionBackwardFilter_v3 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_diffDesc,&__pyx_n_s_diffData,&__pyx_n_s_convDesc,&__pyx_n_s_algo,&__pyx_n_s_workSpace,&__pyx_n_s_workSpaceSizeInBytes,&__pyx_n_s_beta,&__pyx_n_s_gradDesc,&__pyx_n_s_gradData,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 1); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 2); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 3); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_diffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 4); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_diffData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 5); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 6); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 7); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 8); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 9); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 10); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_gradDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 11); __PYX_ERR(0, 825, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_gradData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, 12); __PYX_ERR(0, 825, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convolutionBackwardFilter_v3") < 0)) __PYX_ERR(0, 825, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 826, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 826, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 826, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 826, __pyx_L3_error)
    __pyx_v_diffDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_diffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 827, __pyx_L3_error)
    __pyx_v_diffData = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_diffData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 827, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 827, __pyx_L3_error)
    __pyx_v_algo = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_algo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 827, __pyx_L3_error)
    __pyx_v_workSpace = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_workSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_gradDesc = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_gradDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 829, __pyx_L3_error)
    __pyx_v_gradData = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_gradData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 829, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convolutionBackwardFilter_v3", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 825, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardFilter_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_74convolutionBackwardFilter_v3(__pyx_self, __pyx_v_handle, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_diffDesc, __pyx_v_diffData, __pyx_v_convDesc, __pyx_v_algo, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, __pyx_v_beta, __pyx_v_gradDesc, __pyx_v_gradData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_74convolutionBackwardFilter_v3(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_diffDesc, size_t __pyx_v_diffData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_gradDesc, size_t __pyx_v_gradData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("convolutionBackwardFilter_v3", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardFilter_v3(__pyx_v_handle, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_diffDesc, __pyx_v_diffData, __pyx_v_convDesc, __pyx_v_algo, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, __pyx_v_beta, __pyx_v_gradDesc, __pyx_v_gradData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardFilter_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":842
 * 
 * 
 * cpdef findConvolutionBackwardDataAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t wDesc, size_t dyDesc, size_t convDesc,
 *         size_t dxDesc, int requestedAlgoCount):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_77findConvolutionBackwardDataAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardDataAlgorithm(size_t __pyx_v_handle, size_t __pyx_v_wDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_convDesc, size_t __pyx_v_dxDesc, int __pyx_v_requestedAlgoCount, CYTHON_UNUSED int __pyx_skip_dispatch) {
  std::vector<cudnnConvolutionBwdDataAlgoPerf_t>  __pyx_v_perfResults;
  int __pyx_v_returnedAlgoCount;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionBackwardDataAlgorithm", 0);

  /* "cupy/cuda/cudnn.pyx":847
 *     cdef vector.vector[ConvolutionBwdDataAlgoPerf] perfResults
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)             # <<<<<<<<<<<<<<
 *     status = cudnnFindConvolutionBackwardDataAlgorithm(
 *         <Handle> handle, <FilterDescriptor>wDesc, <TensorDescriptor>dyDesc,
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_requestedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 847, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":848
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)
 *     status = cudnnFindConvolutionBackwardDataAlgorithm(             # <<<<<<<<<<<<<<
 *         <Handle> handle, <FilterDescriptor>wDesc, <TensorDescriptor>dyDesc,
 *         <ConvolutionDescriptor>convDesc, <TensorDescriptor>dxDesc,
 */
  __pyx_v_status = cudnnFindConvolutionBackwardDataAlgorithm(((cudnnHandle_t)__pyx_v_handle), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((cudnnTensorDescriptor_t)__pyx_v_dyDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnTensorDescriptor_t)__pyx_v_dxDesc), __pyx_v_requestedAlgoCount, (&__pyx_v_returnedAlgoCount), (&(__pyx_v_perfResults[0])));

  /* "cupy/cuda/cudnn.pyx":852
 *         <ConvolutionDescriptor>convDesc, <TensorDescriptor>dxDesc,
 *         requestedAlgoCount, &returnedAlgoCount, &perfResults[0])
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":853
 *         requestedAlgoCount, &returnedAlgoCount, &perfResults[0])
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)             # <<<<<<<<<<<<<<
 *     return perfResults
 * 
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_returnedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 853, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":854
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t(__pyx_v_perfResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":842
 * 
 * 
 * cpdef findConvolutionBackwardDataAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t wDesc, size_t dyDesc, size_t convDesc,
 *         size_t dxDesc, int requestedAlgoCount):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardDataAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_77findConvolutionBackwardDataAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_76findConvolutionBackwardDataAlgorithm[] = "findConvolutionBackwardDataAlgorithm(size_t handle, size_t wDesc, size_t dyDesc, size_t convDesc, size_t dxDesc, int requestedAlgoCount)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_77findConvolutionBackwardDataAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_dyDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_dxDesc;
  int __pyx_v_requestedAlgoCount;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findConvolutionBackwardDataAlgorithm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_wDesc,&__pyx_n_s_dyDesc,&__pyx_n_s_convDesc,&__pyx_n_s_dxDesc,&__pyx_n_s_requestedAlgoCount,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithm", 1, 6, 6, 1); __PYX_ERR(0, 842, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithm", 1, 6, 6, 2); __PYX_ERR(0, 842, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithm", 1, 6, 6, 3); __PYX_ERR(0, 842, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithm", 1, 6, 6, 4); __PYX_ERR(0, 842, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_requestedAlgoCount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithm", 1, 6, 6, 5); __PYX_ERR(0, 842, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findConvolutionBackwardDataAlgorithm") < 0)) __PYX_ERR(0, 842, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 843, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 843, __pyx_L3_error)
    __pyx_v_dyDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_dyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 843, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 843, __pyx_L3_error)
    __pyx_v_dxDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 844, __pyx_L3_error)
    __pyx_v_requestedAlgoCount = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_requestedAlgoCount == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 844, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithm", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 842, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardDataAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_76findConvolutionBackwardDataAlgorithm(__pyx_self, __pyx_v_handle, __pyx_v_wDesc, __pyx_v_dyDesc, __pyx_v_convDesc, __pyx_v_dxDesc, __pyx_v_requestedAlgoCount);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_76findConvolutionBackwardDataAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_wDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_convDesc, size_t __pyx_v_dxDesc, int __pyx_v_requestedAlgoCount) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionBackwardDataAlgorithm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardDataAlgorithm(__pyx_v_handle, __pyx_v_wDesc, __pyx_v_dyDesc, __pyx_v_convDesc, __pyx_v_dxDesc, __pyx_v_requestedAlgoCount, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardDataAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":857
 * 
 * 
 * cpdef findConvolutionBackwardDataAlgorithmEx(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t wDesc, size_t w, size_t dyDesc, size_t dy,
 *         size_t convDesc, size_t dxDesc, size_t dx,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_79findConvolutionBackwardDataAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardDataAlgorithmEx(size_t __pyx_v_handle, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_convDesc, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  std::vector<cudnnConvolutionBwdDataAlgoPerf_t>  __pyx_v_perfResults;
  int __pyx_v_returnedAlgoCount;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionBackwardDataAlgorithmEx", 0);

  /* "cupy/cuda/cudnn.pyx":863
 *     cdef vector.vector[ConvolutionBwdDataAlgoPerf] perfResults
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)             # <<<<<<<<<<<<<<
 *     status = cudnnFindConvolutionBackwardDataAlgorithmEx(
 *         <Handle> handle, <FilterDescriptor>wDesc, <void*>w,
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_requestedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 863, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":864
 *     cdef int returnedAlgoCount
 *     perfResults.resize(requestedAlgoCount)
 *     status = cudnnFindConvolutionBackwardDataAlgorithmEx(             # <<<<<<<<<<<<<<
 *         <Handle> handle, <FilterDescriptor>wDesc, <void*>w,
 *         <TensorDescriptor>dyDesc, <void*>dy, <ConvolutionDescriptor>convDesc,
 */
  __pyx_v_status = cudnnFindConvolutionBackwardDataAlgorithmEx(((cudnnHandle_t)__pyx_v_handle), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((void *)__pyx_v_w), ((cudnnTensorDescriptor_t)__pyx_v_dyDesc), ((void *)__pyx_v_dy), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnTensorDescriptor_t)__pyx_v_dxDesc), ((void *)__pyx_v_dx), __pyx_v_requestedAlgoCount, (&__pyx_v_returnedAlgoCount), (&(__pyx_v_perfResults[0])), ((void *)__pyx_v_workSpace), __pyx_v_workSpaceSizeInBytes);

  /* "cupy/cuda/cudnn.pyx":870
 *         requestedAlgoCount, &returnedAlgoCount, &perfResults[0],
 *         <void*>workSpace, workSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":871
 *         <void*>workSpace, workSpaceSizeInBytes)
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)             # <<<<<<<<<<<<<<
 *     return perfResults
 * 
 */
  try {
    __pyx_v_perfResults.resize(__pyx_v_returnedAlgoCount);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 871, __pyx_L1_error)
  }

  /* "cupy/cuda/cudnn.pyx":872
 *     check_status(status)
 *     perfResults.resize(returnedAlgoCount)
 *     return perfResults             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t(__pyx_v_perfResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":857
 * 
 * 
 * cpdef findConvolutionBackwardDataAlgorithmEx(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t wDesc, size_t w, size_t dyDesc, size_t dy,
 *         size_t convDesc, size_t dxDesc, size_t dx,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardDataAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_79findConvolutionBackwardDataAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_78findConvolutionBackwardDataAlgorithmEx[] = "findConvolutionBackwardDataAlgorithmEx(size_t handle, size_t wDesc, size_t w, size_t dyDesc, size_t dy, size_t convDesc, size_t dxDesc, size_t dx, int requestedAlgoCount, size_t workSpace, size_t workSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_79findConvolutionBackwardDataAlgorithmEx(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_w;
  size_t __pyx_v_dyDesc;
  size_t __pyx_v_dy;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_dxDesc;
  size_t __pyx_v_dx;
  int __pyx_v_requestedAlgoCount;
  size_t __pyx_v_workSpace;
  size_t __pyx_v_workSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("findConvolutionBackwardDataAlgorithmEx (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_wDesc,&__pyx_n_s_w,&__pyx_n_s_dyDesc,&__pyx_n_s_dy,&__pyx_n_s_convDesc,&__pyx_n_s_dxDesc,&__pyx_n_s_dx,&__pyx_n_s_requestedAlgoCount,&__pyx_n_s_workSpace,&__pyx_n_s_workSpaceSizeInBytes,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 1); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 2); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 3); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 4); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 5); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 6); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 7); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_requestedAlgoCount)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 8); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 9); __PYX_ERR(0, 857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, 10); __PYX_ERR(0, 857, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "findConvolutionBackwardDataAlgorithmEx") < 0)) __PYX_ERR(0, 857, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_w == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_dyDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_dyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_dy = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 859, __pyx_L3_error)
    __pyx_v_dxDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_dxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 859, __pyx_L3_error)
    __pyx_v_dx = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_dx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 859, __pyx_L3_error)
    __pyx_v_requestedAlgoCount = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_requestedAlgoCount == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 860, __pyx_L3_error)
    __pyx_v_workSpace = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_workSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 860, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 860, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("findConvolutionBackwardDataAlgorithmEx", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 857, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardDataAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_78findConvolutionBackwardDataAlgorithmEx(__pyx_self, __pyx_v_handle, __pyx_v_wDesc, __pyx_v_w, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_convDesc, __pyx_v_dxDesc, __pyx_v_dx, __pyx_v_requestedAlgoCount, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_78findConvolutionBackwardDataAlgorithmEx(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_convDesc, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, int __pyx_v_requestedAlgoCount, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("findConvolutionBackwardDataAlgorithmEx", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardDataAlgorithmEx(__pyx_v_handle, __pyx_v_wDesc, __pyx_v_w, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_convDesc, __pyx_v_dxDesc, __pyx_v_dx, __pyx_v_requestedAlgoCount, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.findConvolutionBackwardDataAlgorithmEx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":875
 * 
 * 
 * cpdef int getConvolutionBackwardDataAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t filterDesc, size_t diffDesc, size_t convDesc,
 *         size_t gradDesc, size_t preference,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_81getConvolutionBackwardDataAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardDataAlgorithm(size_t __pyx_v_handle, size_t __pyx_v_filterDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_gradDesc, size_t __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnConvolutionBwdDataAlgo_t __pyx_v_algo;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionBackwardDataAlgorithm", 0);

  /* "cupy/cuda/cudnn.pyx":880
 *         size_t memoryLimitInbytes) except *:
 *     cdef ConvolutionBwdDataAlgo algo
 *     status = cudnnGetConvolutionBackwardDataAlgorithm(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <FilterDescriptor>filterDesc,
 *         <TensorDescriptor>diffDesc, <ConvolutionDescriptor>convDesc,
 */
  __pyx_v_status = cudnnGetConvolutionBackwardDataAlgorithm(((cudnnHandle_t)__pyx_v_handle), ((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((cudnnTensorDescriptor_t)__pyx_v_diffDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnTensorDescriptor_t)__pyx_v_gradDesc), ((cudnnConvolutionBwdDataPreference_t)__pyx_v_preference), __pyx_v_memoryLimitInbytes, (&__pyx_v_algo));

  /* "cupy/cuda/cudnn.pyx":885
 *         <TensorDescriptor>gradDesc, <ConvolutionBwdDataPreference>preference,
 *         memoryLimitInbytes, &algo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return algo
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":886
 *         memoryLimitInbytes, &algo)
 *     check_status(status)
 *     return algo             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_algo;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":875
 * 
 * 
 * cpdef int getConvolutionBackwardDataAlgorithm(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t filterDesc, size_t diffDesc, size_t convDesc,
 *         size_t gradDesc, size_t preference,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardDataAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_81getConvolutionBackwardDataAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_80getConvolutionBackwardDataAlgorithm[] = "getConvolutionBackwardDataAlgorithm(size_t handle, size_t filterDesc, size_t diffDesc, size_t convDesc, size_t gradDesc, size_t preference, size_t memoryLimitInbytes) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_81getConvolutionBackwardDataAlgorithm(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_filterDesc;
  size_t __pyx_v_diffDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_gradDesc;
  size_t __pyx_v_preference;
  size_t __pyx_v_memoryLimitInbytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionBackwardDataAlgorithm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_filterDesc,&__pyx_n_s_diffDesc,&__pyx_n_s_convDesc,&__pyx_n_s_gradDesc,&__pyx_n_s_preference,&__pyx_n_s_memoryLimitInbytes,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataAlgorithm", 1, 7, 7, 1); __PYX_ERR(0, 875, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_diffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataAlgorithm", 1, 7, 7, 2); __PYX_ERR(0, 875, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataAlgorithm", 1, 7, 7, 3); __PYX_ERR(0, 875, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_gradDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataAlgorithm", 1, 7, 7, 4); __PYX_ERR(0, 875, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_preference)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataAlgorithm", 1, 7, 7, 5); __PYX_ERR(0, 875, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_memoryLimitInbytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataAlgorithm", 1, 7, 7, 6); __PYX_ERR(0, 875, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getConvolutionBackwardDataAlgorithm") < 0)) __PYX_ERR(0, 875, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_diffDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_diffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_gradDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_gradDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 877, __pyx_L3_error)
    __pyx_v_preference = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_preference == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 877, __pyx_L3_error)
    __pyx_v_memoryLimitInbytes = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_memoryLimitInbytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 878, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataAlgorithm", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 875, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardDataAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_80getConvolutionBackwardDataAlgorithm(__pyx_self, __pyx_v_handle, __pyx_v_filterDesc, __pyx_v_diffDesc, __pyx_v_convDesc, __pyx_v_gradDesc, __pyx_v_preference, __pyx_v_memoryLimitInbytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_80getConvolutionBackwardDataAlgorithm(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_filterDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_gradDesc, size_t __pyx_v_preference, size_t __pyx_v_memoryLimitInbytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionBackwardDataAlgorithm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardDataAlgorithm(__pyx_v_handle, __pyx_v_filterDesc, __pyx_v_diffDesc, __pyx_v_convDesc, __pyx_v_gradDesc, __pyx_v_preference, __pyx_v_memoryLimitInbytes, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 875, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardDataAlgorithm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":889
 * 
 * 
 * cpdef size_t getConvolutionBackwardDataWorkspaceSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t filterDesc, size_t diffDesc, size_t convDesc,
 *         size_t gradDesc, int algo) except *:
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_83getConvolutionBackwardDataWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardDataWorkspaceSize(size_t __pyx_v_handle, size_t __pyx_v_filterDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_gradDesc, int __pyx_v_algo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_sizeInBytes;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionBackwardDataWorkspaceSize", 0);

  /* "cupy/cuda/cudnn.pyx":893
 *         size_t gradDesc, int algo) except *:
 *     cdef size_t sizeInBytes
 *     status = cudnnGetConvolutionBackwardDataWorkspaceSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <FilterDescriptor>filterDesc,
 *         <TensorDescriptor>diffDesc,
 */
  __pyx_v_status = cudnnGetConvolutionBackwardDataWorkspaceSize(((cudnnHandle_t)__pyx_v_handle), ((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((cudnnTensorDescriptor_t)__pyx_v_diffDesc), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnTensorDescriptor_t)__pyx_v_gradDesc), ((cudnnConvolutionBwdDataAlgo_t)__pyx_v_algo), (&__pyx_v_sizeInBytes));

  /* "cupy/cuda/cudnn.pyx":898
 *         <ConvolutionDescriptor>convDesc, <TensorDescriptor>gradDesc,
 *         <ConvolutionBwdDataAlgo>algo, &sizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return sizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":899
 *         <ConvolutionBwdDataAlgo>algo, &sizeInBytes)
 *     check_status(status)
 *     return sizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_sizeInBytes;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":889
 * 
 * 
 * cpdef size_t getConvolutionBackwardDataWorkspaceSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t filterDesc, size_t diffDesc, size_t convDesc,
 *         size_t gradDesc, int algo) except *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardDataWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_83getConvolutionBackwardDataWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_82getConvolutionBackwardDataWorkspaceSize[] = "getConvolutionBackwardDataWorkspaceSize(size_t handle, size_t filterDesc, size_t diffDesc, size_t convDesc, size_t gradDesc, int algo) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_83getConvolutionBackwardDataWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_filterDesc;
  size_t __pyx_v_diffDesc;
  size_t __pyx_v_convDesc;
  size_t __pyx_v_gradDesc;
  int __pyx_v_algo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConvolutionBackwardDataWorkspaceSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_filterDesc,&__pyx_n_s_diffDesc,&__pyx_n_s_convDesc,&__pyx_n_s_gradDesc,&__pyx_n_s_algo,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataWorkspaceSize", 1, 6, 6, 1); __PYX_ERR(0, 889, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_diffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataWorkspaceSize", 1, 6, 6, 2); __PYX_ERR(0, 889, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataWorkspaceSize", 1, 6, 6, 3); __PYX_ERR(0, 889, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_gradDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataWorkspaceSize", 1, 6, 6, 4); __PYX_ERR(0, 889, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataWorkspaceSize", 1, 6, 6, 5); __PYX_ERR(0, 889, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getConvolutionBackwardDataWorkspaceSize") < 0)) __PYX_ERR(0, 889, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 890, __pyx_L3_error)
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 890, __pyx_L3_error)
    __pyx_v_diffDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_diffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 890, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 890, __pyx_L3_error)
    __pyx_v_gradDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_gradDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 891, __pyx_L3_error)
    __pyx_v_algo = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_algo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 891, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getConvolutionBackwardDataWorkspaceSize", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 889, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardDataWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_82getConvolutionBackwardDataWorkspaceSize(__pyx_self, __pyx_v_handle, __pyx_v_filterDesc, __pyx_v_diffDesc, __pyx_v_convDesc, __pyx_v_gradDesc, __pyx_v_algo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_82getConvolutionBackwardDataWorkspaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_filterDesc, size_t __pyx_v_diffDesc, size_t __pyx_v_convDesc, size_t __pyx_v_gradDesc, int __pyx_v_algo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getConvolutionBackwardDataWorkspaceSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardDataWorkspaceSize(__pyx_v_handle, __pyx_v_filterDesc, __pyx_v_diffDesc, __pyx_v_convDesc, __pyx_v_gradDesc, __pyx_v_algo, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 889, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getConvolutionBackwardDataWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":902
 * 
 * 
 * cpdef convolutionBackwardData_v3(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t alpha, size_t filterDesc, size_t filterData,
 *         size_t diffDesc, size_t diffData, size_t convDesc, int algo,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_85convolutionBackwardData_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardData_v3(size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_filterDesc, size_t __pyx_v_filterData, size_t __pyx_v_diffDesc, size_t __pyx_v_diffData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_gradDesc, size_t __pyx_v_gradData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("convolutionBackwardData_v3", 0);

  /* "cupy/cuda/cudnn.pyx":907
 *         size_t workSpace, size_t workSpaceSizeInBytes, size_t beta,
 *         size_t gradDesc, size_t gradData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnConvolutionBackwardData_v3(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":908
 *         size_t gradDesc, size_t gradData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnConvolutionBackwardData_v3(
 *             <Handle>handle, <void*>alpha,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":909
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnConvolutionBackwardData_v3(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <void*>alpha,
 *             <FilterDescriptor>filterDesc, <void*>filterData,
 */
        __pyx_v_status = cudnnConvolutionBackwardData_v3(((cudnnHandle_t)__pyx_v_handle), ((void *)__pyx_v_alpha), ((cudnnFilterDescriptor_t)__pyx_v_filterDesc), ((void *)__pyx_v_filterData), ((cudnnTensorDescriptor_t)__pyx_v_diffDesc), ((void *)__pyx_v_diffData), ((cudnnConvolutionDescriptor_t)__pyx_v_convDesc), ((cudnnConvolutionBwdDataAlgo_t)__pyx_v_algo), ((void *)__pyx_v_workSpace), __pyx_v_workSpaceSizeInBytes, ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_gradDesc), ((void *)__pyx_v_gradData));
      }

      /* "cupy/cuda/cudnn.pyx":908
 *         size_t gradDesc, size_t gradData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnConvolutionBackwardData_v3(
 *             <Handle>handle, <void*>alpha,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":916
 *             <void*>workSpace, workSpaceSizeInBytes, <void*>beta,
 *             <TensorDescriptor>gradDesc, <void*>gradData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * ###############################################################################
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":902
 * 
 * 
 * cpdef convolutionBackwardData_v3(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t alpha, size_t filterDesc, size_t filterData,
 *         size_t diffDesc, size_t diffData, size_t convDesc, int algo,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardData_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_85convolutionBackwardData_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_84convolutionBackwardData_v3[] = "convolutionBackwardData_v3(size_t handle, size_t alpha, size_t filterDesc, size_t filterData, size_t diffDesc, size_t diffData, size_t convDesc, int algo, size_t workSpace, size_t workSpaceSizeInBytes, size_t beta, size_t gradDesc, size_t gradData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_85convolutionBackwardData_v3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_alpha;
  size_t __pyx_v_filterDesc;
  size_t __pyx_v_filterData;
  size_t __pyx_v_diffDesc;
  size_t __pyx_v_diffData;
  size_t __pyx_v_convDesc;
  int __pyx_v_algo;
  size_t __pyx_v_workSpace;
  size_t __pyx_v_workSpaceSizeInBytes;
  size_t __pyx_v_beta;
  size_t __pyx_v_gradDesc;
  size_t __pyx_v_gradData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convolutionBackwardData_v3 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_alpha,&__pyx_n_s_filterDesc,&__pyx_n_s_filterData,&__pyx_n_s_diffDesc,&__pyx_n_s_diffData,&__pyx_n_s_convDesc,&__pyx_n_s_algo,&__pyx_n_s_workSpace,&__pyx_n_s_workSpaceSizeInBytes,&__pyx_n_s_beta,&__pyx_n_s_gradDesc,&__pyx_n_s_gradData,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 1); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 2); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filterData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 3); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_diffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 4); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_diffData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 5); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_convDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 6); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 7); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 8); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 9); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 10); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_gradDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 11); __PYX_ERR(0, 902, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_gradData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, 12); __PYX_ERR(0, 902, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convolutionBackwardData_v3") < 0)) __PYX_ERR(0, 902, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 903, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 903, __pyx_L3_error)
    __pyx_v_filterDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_filterDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 903, __pyx_L3_error)
    __pyx_v_filterData = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_filterData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 903, __pyx_L3_error)
    __pyx_v_diffDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_diffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L3_error)
    __pyx_v_diffData = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_diffData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L3_error)
    __pyx_v_convDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_convDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L3_error)
    __pyx_v_algo = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_algo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 904, __pyx_L3_error)
    __pyx_v_workSpace = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_workSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 905, __pyx_L3_error)
    __pyx_v_gradDesc = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_gradDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 906, __pyx_L3_error)
    __pyx_v_gradData = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_gradData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 906, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convolutionBackwardData_v3", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 902, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardData_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_84convolutionBackwardData_v3(__pyx_self, __pyx_v_handle, __pyx_v_alpha, __pyx_v_filterDesc, __pyx_v_filterData, __pyx_v_diffDesc, __pyx_v_diffData, __pyx_v_convDesc, __pyx_v_algo, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, __pyx_v_beta, __pyx_v_gradDesc, __pyx_v_gradData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_84convolutionBackwardData_v3(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_alpha, size_t __pyx_v_filterDesc, size_t __pyx_v_filterData, size_t __pyx_v_diffDesc, size_t __pyx_v_diffData, size_t __pyx_v_convDesc, int __pyx_v_algo, size_t __pyx_v_workSpace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_beta, size_t __pyx_v_gradDesc, size_t __pyx_v_gradData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("convolutionBackwardData_v3", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardData_v3(__pyx_v_handle, __pyx_v_alpha, __pyx_v_filterDesc, __pyx_v_filterData, __pyx_v_diffDesc, __pyx_v_diffData, __pyx_v_convDesc, __pyx_v_algo, __pyx_v_workSpace, __pyx_v_workSpaceSizeInBytes, __pyx_v_beta, __pyx_v_gradDesc, __pyx_v_gradData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.convolutionBackwardData_v3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":922
 * ###############################################################################
 * 
 * cpdef size_t createPoolingDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef PoolingDescriptor desc
 *     status = cudnnCreatePoolingDescriptor(&desc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_87createPoolingDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createPoolingDescriptor(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnPoolingDescriptor_t __pyx_v_desc;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createPoolingDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":924
 * cpdef size_t createPoolingDescriptor() except *:
 *     cdef PoolingDescriptor desc
 *     status = cudnnCreatePoolingDescriptor(&desc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>desc
 */
  __pyx_v_status = cudnnCreatePoolingDescriptor((&__pyx_v_desc));

  /* "cupy/cuda/cudnn.pyx":925
 *     cdef PoolingDescriptor desc
 *     status = cudnnCreatePoolingDescriptor(&desc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>desc
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":926
 *     status = cudnnCreatePoolingDescriptor(&desc)
 *     check_status(status)
 *     return <size_t>desc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_desc);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":922
 * ###############################################################################
 * 
 * cpdef size_t createPoolingDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef PoolingDescriptor desc
 *     status = cudnnCreatePoolingDescriptor(&desc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createPoolingDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_87createPoolingDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_86createPoolingDescriptor[] = "createPoolingDescriptor() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_87createPoolingDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createPoolingDescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_86createPoolingDescriptor(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_86createPoolingDescriptor(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createPoolingDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createPoolingDescriptor(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 922, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createPoolingDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":929
 * 
 * 
 * cpdef setPooling2dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t poolingDesc, int mode, int maxpoolingNanOpt, int windowHeight,
 *         int windowWidth, int verticalPadding, int horizontalPadding,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_89setPooling2dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setPooling2dDescriptor_v4(size_t __pyx_v_poolingDesc, int __pyx_v_mode, int __pyx_v_maxpoolingNanOpt, int __pyx_v_windowHeight, int __pyx_v_windowWidth, int __pyx_v_verticalPadding, int __pyx_v_horizontalPadding, int __pyx_v_verticalStride, int __pyx_v_horizontalStride, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setPooling2dDescriptor_v4", 0);

  /* "cupy/cuda/cudnn.pyx":933
 *         int windowWidth, int verticalPadding, int horizontalPadding,
 *         int verticalStride, int horizontalStride):
 *     status = cudnnSetPooling2dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         <PoolingDescriptor>poolingDesc, <PoolingMode>mode,
 *         <NanPropagation>maxpoolingNanOpt, windowHeight, windowWidth,
 */
  __pyx_v_status = cudnnSetPooling2dDescriptor_v4(((cudnnPoolingDescriptor_t)__pyx_v_poolingDesc), ((cudnnPoolingMode_t)__pyx_v_mode), ((cudnnNanPropagation_t)__pyx_v_maxpoolingNanOpt), __pyx_v_windowHeight, __pyx_v_windowWidth, __pyx_v_verticalPadding, __pyx_v_horizontalPadding, __pyx_v_verticalStride, __pyx_v_horizontalStride);

  /* "cupy/cuda/cudnn.pyx":937
 *         <NanPropagation>maxpoolingNanOpt, windowHeight, windowWidth,
 *         verticalPadding, horizontalPadding, verticalStride, horizontalStride)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":929
 * 
 * 
 * cpdef setPooling2dDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t poolingDesc, int mode, int maxpoolingNanOpt, int windowHeight,
 *         int windowWidth, int verticalPadding, int horizontalPadding,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPooling2dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_89setPooling2dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_88setPooling2dDescriptor_v4[] = "setPooling2dDescriptor_v4(size_t poolingDesc, int mode, int maxpoolingNanOpt, int windowHeight, int windowWidth, int verticalPadding, int horizontalPadding, int verticalStride, int horizontalStride)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_89setPooling2dDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_poolingDesc;
  int __pyx_v_mode;
  int __pyx_v_maxpoolingNanOpt;
  int __pyx_v_windowHeight;
  int __pyx_v_windowWidth;
  int __pyx_v_verticalPadding;
  int __pyx_v_horizontalPadding;
  int __pyx_v_verticalStride;
  int __pyx_v_horizontalStride;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setPooling2dDescriptor_v4 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_poolingDesc,&__pyx_n_s_mode,&__pyx_n_s_maxpoolingNanOpt,&__pyx_n_s_windowHeight,&__pyx_n_s_windowWidth,&__pyx_n_s_verticalPadding,&__pyx_n_s_horizontalPadding,&__pyx_n_s_verticalStride,&__pyx_n_s_horizontalStride,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_poolingDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, 1); __PYX_ERR(0, 929, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_maxpoolingNanOpt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, 2); __PYX_ERR(0, 929, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_windowHeight)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, 3); __PYX_ERR(0, 929, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_windowWidth)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, 4); __PYX_ERR(0, 929, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_verticalPadding)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, 5); __PYX_ERR(0, 929, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_horizontalPadding)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, 6); __PYX_ERR(0, 929, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_verticalStride)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, 7); __PYX_ERR(0, 929, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_horizontalStride)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, 8); __PYX_ERR(0, 929, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setPooling2dDescriptor_v4") < 0)) __PYX_ERR(0, 929, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_poolingDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_poolingDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
    __pyx_v_maxpoolingNanOpt = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_maxpoolingNanOpt == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
    __pyx_v_windowHeight = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_windowHeight == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 930, __pyx_L3_error)
    __pyx_v_windowWidth = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_windowWidth == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 931, __pyx_L3_error)
    __pyx_v_verticalPadding = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_verticalPadding == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 931, __pyx_L3_error)
    __pyx_v_horizontalPadding = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_horizontalPadding == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 931, __pyx_L3_error)
    __pyx_v_verticalStride = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_verticalStride == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 932, __pyx_L3_error)
    __pyx_v_horizontalStride = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_horizontalStride == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 932, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setPooling2dDescriptor_v4", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 929, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPooling2dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_88setPooling2dDescriptor_v4(__pyx_self, __pyx_v_poolingDesc, __pyx_v_mode, __pyx_v_maxpoolingNanOpt, __pyx_v_windowHeight, __pyx_v_windowWidth, __pyx_v_verticalPadding, __pyx_v_horizontalPadding, __pyx_v_verticalStride, __pyx_v_horizontalStride);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_88setPooling2dDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_poolingDesc, int __pyx_v_mode, int __pyx_v_maxpoolingNanOpt, int __pyx_v_windowHeight, int __pyx_v_windowWidth, int __pyx_v_verticalPadding, int __pyx_v_horizontalPadding, int __pyx_v_verticalStride, int __pyx_v_horizontalStride) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setPooling2dDescriptor_v4", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setPooling2dDescriptor_v4(__pyx_v_poolingDesc, __pyx_v_mode, __pyx_v_maxpoolingNanOpt, __pyx_v_windowHeight, __pyx_v_windowWidth, __pyx_v_verticalPadding, __pyx_v_horizontalPadding, __pyx_v_verticalStride, __pyx_v_horizontalStride, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPooling2dDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":940
 * 
 * 
 * cpdef setPoolingNdDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t poolingDesc, int mode, int maxpoolingNanOpt, int nbDims,
 *         size_t windowDimA, size_t paddingA, size_t strideA):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_91setPoolingNdDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setPoolingNdDescriptor_v4(size_t __pyx_v_poolingDesc, int __pyx_v_mode, int __pyx_v_maxpoolingNanOpt, int __pyx_v_nbDims, size_t __pyx_v_windowDimA, size_t __pyx_v_paddingA, size_t __pyx_v_strideA, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setPoolingNdDescriptor_v4", 0);

  /* "cupy/cuda/cudnn.pyx":943
 *         size_t poolingDesc, int mode, int maxpoolingNanOpt, int nbDims,
 *         size_t windowDimA, size_t paddingA, size_t strideA):
 *     status = cudnnSetPoolingNdDescriptor_v4(             # <<<<<<<<<<<<<<
 *         <PoolingDescriptor>poolingDesc, <PoolingMode>mode,
 *         <NanPropagation>maxpoolingNanOpt, nbDims,
 */
  __pyx_v_status = cudnnSetPoolingNdDescriptor_v4(((cudnnPoolingDescriptor_t)__pyx_v_poolingDesc), ((cudnnPoolingMode_t)__pyx_v_mode), ((cudnnNanPropagation_t)__pyx_v_maxpoolingNanOpt), __pyx_v_nbDims, ((int *)__pyx_v_windowDimA), ((int *)__pyx_v_paddingA), ((int *)__pyx_v_strideA));

  /* "cupy/cuda/cudnn.pyx":947
 *         <NanPropagation>maxpoolingNanOpt, nbDims,
 *         <int*>windowDimA, <int*>paddingA, <int*>strideA)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":940
 * 
 * 
 * cpdef setPoolingNdDescriptor_v4(             # <<<<<<<<<<<<<<
 *         size_t poolingDesc, int mode, int maxpoolingNanOpt, int nbDims,
 *         size_t windowDimA, size_t paddingA, size_t strideA):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPoolingNdDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_91setPoolingNdDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_90setPoolingNdDescriptor_v4[] = "setPoolingNdDescriptor_v4(size_t poolingDesc, int mode, int maxpoolingNanOpt, int nbDims, size_t windowDimA, size_t paddingA, size_t strideA)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_91setPoolingNdDescriptor_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_poolingDesc;
  int __pyx_v_mode;
  int __pyx_v_maxpoolingNanOpt;
  int __pyx_v_nbDims;
  size_t __pyx_v_windowDimA;
  size_t __pyx_v_paddingA;
  size_t __pyx_v_strideA;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setPoolingNdDescriptor_v4 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_poolingDesc,&__pyx_n_s_mode,&__pyx_n_s_maxpoolingNanOpt,&__pyx_n_s_nbDims,&__pyx_n_s_windowDimA,&__pyx_n_s_paddingA,&__pyx_n_s_strideA,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_poolingDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPoolingNdDescriptor_v4", 1, 7, 7, 1); __PYX_ERR(0, 940, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_maxpoolingNanOpt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPoolingNdDescriptor_v4", 1, 7, 7, 2); __PYX_ERR(0, 940, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nbDims)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPoolingNdDescriptor_v4", 1, 7, 7, 3); __PYX_ERR(0, 940, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_windowDimA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPoolingNdDescriptor_v4", 1, 7, 7, 4); __PYX_ERR(0, 940, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_paddingA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPoolingNdDescriptor_v4", 1, 7, 7, 5); __PYX_ERR(0, 940, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPoolingNdDescriptor_v4", 1, 7, 7, 6); __PYX_ERR(0, 940, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setPoolingNdDescriptor_v4") < 0)) __PYX_ERR(0, 940, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_poolingDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_poolingDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 941, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 941, __pyx_L3_error)
    __pyx_v_maxpoolingNanOpt = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_maxpoolingNanOpt == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 941, __pyx_L3_error)
    __pyx_v_nbDims = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nbDims == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 941, __pyx_L3_error)
    __pyx_v_windowDimA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_windowDimA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_paddingA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_paddingA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_strideA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setPoolingNdDescriptor_v4", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 940, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPoolingNdDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_90setPoolingNdDescriptor_v4(__pyx_self, __pyx_v_poolingDesc, __pyx_v_mode, __pyx_v_maxpoolingNanOpt, __pyx_v_nbDims, __pyx_v_windowDimA, __pyx_v_paddingA, __pyx_v_strideA);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_90setPoolingNdDescriptor_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_poolingDesc, int __pyx_v_mode, int __pyx_v_maxpoolingNanOpt, int __pyx_v_nbDims, size_t __pyx_v_windowDimA, size_t __pyx_v_paddingA, size_t __pyx_v_strideA) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setPoolingNdDescriptor_v4", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setPoolingNdDescriptor_v4(__pyx_v_poolingDesc, __pyx_v_mode, __pyx_v_maxpoolingNanOpt, __pyx_v_nbDims, __pyx_v_windowDimA, __pyx_v_paddingA, __pyx_v_strideA, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPoolingNdDescriptor_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":950
 * 
 * 
 * cpdef destroyPoolingDescriptor(size_t poolingDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyPoolingDescriptor(<PoolingDescriptor>poolingDesc)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_93destroyPoolingDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_poolingDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyPoolingDescriptor(size_t __pyx_v_poolingDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyPoolingDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":951
 * 
 * cpdef destroyPoolingDescriptor(size_t poolingDesc):
 *     status = cudnnDestroyPoolingDescriptor(<PoolingDescriptor>poolingDesc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cudnnDestroyPoolingDescriptor(((cudnnPoolingDescriptor_t)__pyx_v_poolingDesc));

  /* "cupy/cuda/cudnn.pyx":952
 * cpdef destroyPoolingDescriptor(size_t poolingDesc):
 *     status = cudnnDestroyPoolingDescriptor(<PoolingDescriptor>poolingDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":950
 * 
 * 
 * cpdef destroyPoolingDescriptor(size_t poolingDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyPoolingDescriptor(<PoolingDescriptor>poolingDesc)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyPoolingDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_93destroyPoolingDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_poolingDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_92destroyPoolingDescriptor[] = "destroyPoolingDescriptor(size_t poolingDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_93destroyPoolingDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_poolingDesc) {
  size_t __pyx_v_poolingDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyPoolingDescriptor (wrapper)", 0);
  assert(__pyx_arg_poolingDesc); {
    __pyx_v_poolingDesc = __Pyx_PyInt_As_size_t(__pyx_arg_poolingDesc); if (unlikely((__pyx_v_poolingDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 950, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyPoolingDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_92destroyPoolingDescriptor(__pyx_self, ((size_t)__pyx_v_poolingDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_92destroyPoolingDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_poolingDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyPoolingDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroyPoolingDescriptor(__pyx_v_poolingDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyPoolingDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":955
 * 
 * 
 * cpdef poolingForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t poolingDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_95poolingForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_poolingForward(size_t __pyx_v_handle, size_t __pyx_v_poolingDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("poolingForward", 0);

  /* "cupy/cuda/cudnn.pyx":958
 *         size_t handle, size_t poolingDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnPoolingForward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":959
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnPoolingForward(
 *             <Handle>handle, <PoolingDescriptor>poolingDesc, <void*>alpha,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":960
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnPoolingForward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <PoolingDescriptor>poolingDesc, <void*>alpha,
 *             <TensorDescriptor>srcDesc, <void*>srcData, <void*>beta,
 */
        __pyx_v_status = cudnnPoolingForward(((cudnnHandle_t)__pyx_v_handle), ((cudnnPoolingDescriptor_t)__pyx_v_poolingDesc), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_dstDesc), ((void *)__pyx_v_dstData));
      }

      /* "cupy/cuda/cudnn.pyx":959
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnPoolingForward(
 *             <Handle>handle, <PoolingDescriptor>poolingDesc, <void*>alpha,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":964
 *             <TensorDescriptor>srcDesc, <void*>srcData, <void*>beta,
 *             <TensorDescriptor>dstDesc, <void*>dstData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":955
 * 
 * 
 * cpdef poolingForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t poolingDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.poolingForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_95poolingForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_94poolingForward[] = "poolingForward(size_t handle, size_t poolingDesc, size_t alpha, size_t srcDesc, size_t srcData, size_t beta, size_t dstDesc, size_t dstData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_95poolingForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_poolingDesc;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_beta;
  size_t __pyx_v_dstDesc;
  size_t __pyx_v_dstData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("poolingForward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_poolingDesc,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_beta,&__pyx_n_s_dstDesc,&__pyx_n_s_dstData,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_poolingDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingForward", 1, 8, 8, 1); __PYX_ERR(0, 955, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingForward", 1, 8, 8, 2); __PYX_ERR(0, 955, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingForward", 1, 8, 8, 3); __PYX_ERR(0, 955, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingForward", 1, 8, 8, 4); __PYX_ERR(0, 955, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingForward", 1, 8, 8, 5); __PYX_ERR(0, 955, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dstDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingForward", 1, 8, 8, 6); __PYX_ERR(0, 955, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dstData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingForward", 1, 8, 8, 7); __PYX_ERR(0, 955, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "poolingForward") < 0)) __PYX_ERR(0, 955, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 956, __pyx_L3_error)
    __pyx_v_poolingDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_poolingDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 956, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 956, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 956, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 957, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 957, __pyx_L3_error)
    __pyx_v_dstDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_dstDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 957, __pyx_L3_error)
    __pyx_v_dstData = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_dstData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 957, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("poolingForward", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 955, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.poolingForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_94poolingForward(__pyx_self, __pyx_v_handle, __pyx_v_poolingDesc, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_beta, __pyx_v_dstDesc, __pyx_v_dstData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_94poolingForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_poolingDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("poolingForward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_poolingForward(__pyx_v_handle, __pyx_v_poolingDesc, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_beta, __pyx_v_dstDesc, __pyx_v_dstData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.poolingForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":967
 * 
 * 
 * cpdef poolingBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t poolingDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t srcDiffDesc, size_t srcDiffData,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_97poolingBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_poolingBackward(size_t __pyx_v_handle, size_t __pyx_v_poolingDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_destDesc, size_t __pyx_v_destData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("poolingBackward", 0);

  /* "cupy/cuda/cudnn.pyx":972
 *         size_t destDesc, size_t destData, size_t beta, size_t destDiffDesc,
 *         size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnPoolingBackward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":973
 *         size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnPoolingBackward(
 *             <Handle>handle, <PoolingDescriptor>poolingDesc, <void*>alpha,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":974
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnPoolingBackward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <PoolingDescriptor>poolingDesc, <void*>alpha,
 *             <TensorDescriptor>srcDesc, <void*>srcData,
 */
        __pyx_v_status = cudnnPoolingBackward(((cudnnHandle_t)__pyx_v_handle), ((cudnnPoolingDescriptor_t)__pyx_v_poolingDesc), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((cudnnTensorDescriptor_t)__pyx_v_srcDiffDesc), ((void *)__pyx_v_srcDiffData), ((cudnnTensorDescriptor_t)__pyx_v_destDesc), ((void *)__pyx_v_destData), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_destDiffDesc), ((void *)__pyx_v_destDiffData));
      }

      /* "cupy/cuda/cudnn.pyx":973
 *         size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnPoolingBackward(
 *             <Handle>handle, <PoolingDescriptor>poolingDesc, <void*>alpha,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":980
 *             <TensorDescriptor>destDesc, <void*>destData, <void*>beta,
 *             <TensorDescriptor>destDiffDesc, <void*>destDiffData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * ###############################################################################
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":967
 * 
 * 
 * cpdef poolingBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t poolingDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t srcDiffDesc, size_t srcDiffData,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.poolingBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_97poolingBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_96poolingBackward[] = "poolingBackward(size_t handle, size_t poolingDesc, size_t alpha, size_t srcDesc, size_t srcData, size_t srcDiffDesc, size_t srcDiffData, size_t destDesc, size_t destData, size_t beta, size_t destDiffDesc, size_t destDiffData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_97poolingBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_poolingDesc;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_srcDiffDesc;
  size_t __pyx_v_srcDiffData;
  size_t __pyx_v_destDesc;
  size_t __pyx_v_destData;
  size_t __pyx_v_beta;
  size_t __pyx_v_destDiffDesc;
  size_t __pyx_v_destDiffData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("poolingBackward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_poolingDesc,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_srcDiffDesc,&__pyx_n_s_srcDiffData,&__pyx_n_s_destDesc,&__pyx_n_s_destData,&__pyx_n_s_beta,&__pyx_n_s_destDiffDesc,&__pyx_n_s_destDiffData,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_poolingDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 1); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 2); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 3); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 4); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDiffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 5); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDiffData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 6); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 7); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 8); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 9); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDiffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 10); __PYX_ERR(0, 967, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDiffData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, 11); __PYX_ERR(0, 967, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "poolingBackward") < 0)) __PYX_ERR(0, 967, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 968, __pyx_L3_error)
    __pyx_v_poolingDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_poolingDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 968, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 968, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 968, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L3_error)
    __pyx_v_srcDiffDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_srcDiffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L3_error)
    __pyx_v_srcDiffData = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_srcDiffData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L3_error)
    __pyx_v_destDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_destDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 970, __pyx_L3_error)
    __pyx_v_destData = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_destData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 970, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 970, __pyx_L3_error)
    __pyx_v_destDiffDesc = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_destDiffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 970, __pyx_L3_error)
    __pyx_v_destDiffData = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_destDiffData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 971, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("poolingBackward", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 967, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.poolingBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_96poolingBackward(__pyx_self, __pyx_v_handle, __pyx_v_poolingDesc, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_srcDiffDesc, __pyx_v_srcDiffData, __pyx_v_destDesc, __pyx_v_destData, __pyx_v_beta, __pyx_v_destDiffDesc, __pyx_v_destDiffData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_96poolingBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_poolingDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_destDesc, size_t __pyx_v_destData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("poolingBackward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_poolingBackward(__pyx_v_handle, __pyx_v_poolingDesc, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_srcDiffDesc, __pyx_v_srcDiffData, __pyx_v_destDesc, __pyx_v_destData, __pyx_v_beta, __pyx_v_destDiffDesc, __pyx_v_destDiffData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.poolingBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":986
 * ###############################################################################
 * 
 * cpdef deriveBNTensorDescriptor(             # <<<<<<<<<<<<<<
 *         size_t derivedBnDesc, size_t xDesc, int mode):
 *     status = cudnnDeriveBNTensorDescriptor(
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_99deriveBNTensorDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_deriveBNTensorDescriptor(size_t __pyx_v_derivedBnDesc, size_t __pyx_v_xDesc, int __pyx_v_mode, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("deriveBNTensorDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":988
 * cpdef deriveBNTensorDescriptor(
 *         size_t derivedBnDesc, size_t xDesc, int mode):
 *     status = cudnnDeriveBNTensorDescriptor(             # <<<<<<<<<<<<<<
 *         <TensorDescriptor>derivedBnDesc, <TensorDescriptor>xDesc,
 *         <BatchNormMode> mode)
 */
  __pyx_v_status = cudnnDeriveBNTensorDescriptor(((cudnnTensorDescriptor_t)__pyx_v_derivedBnDesc), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((cudnnBatchNormMode_t)__pyx_v_mode));

  /* "cupy/cuda/cudnn.pyx":991
 *         <TensorDescriptor>derivedBnDesc, <TensorDescriptor>xDesc,
 *         <BatchNormMode> mode)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":986
 * ###############################################################################
 * 
 * cpdef deriveBNTensorDescriptor(             # <<<<<<<<<<<<<<
 *         size_t derivedBnDesc, size_t xDesc, int mode):
 *     status = cudnnDeriveBNTensorDescriptor(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.deriveBNTensorDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_99deriveBNTensorDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_98deriveBNTensorDescriptor[] = "deriveBNTensorDescriptor(size_t derivedBnDesc, size_t xDesc, int mode)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_99deriveBNTensorDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_derivedBnDesc;
  size_t __pyx_v_xDesc;
  int __pyx_v_mode;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("deriveBNTensorDescriptor (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_derivedBnDesc,&__pyx_n_s_xDesc,&__pyx_n_s_mode,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_derivedBnDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("deriveBNTensorDescriptor", 1, 3, 3, 1); __PYX_ERR(0, 986, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("deriveBNTensorDescriptor", 1, 3, 3, 2); __PYX_ERR(0, 986, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "deriveBNTensorDescriptor") < 0)) __PYX_ERR(0, 986, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_derivedBnDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_derivedBnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 987, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 987, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 987, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("deriveBNTensorDescriptor", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 986, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.deriveBNTensorDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_98deriveBNTensorDescriptor(__pyx_self, __pyx_v_derivedBnDesc, __pyx_v_xDesc, __pyx_v_mode);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_98deriveBNTensorDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_derivedBnDesc, size_t __pyx_v_xDesc, int __pyx_v_mode) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("deriveBNTensorDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_deriveBNTensorDescriptor(__pyx_v_derivedBnDesc, __pyx_v_xDesc, __pyx_v_mode, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.deriveBNTensorDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":994
 * 
 * 
 * cpdef batchNormalizationForwardTraining(             # <<<<<<<<<<<<<<
 *         size_t handle, int mode,
 *         size_t alpha, size_t beta, size_t xDesc,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_101batchNormalizationForwardTraining(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationForwardTraining(size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_beta, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_bnScaleBiasMeanVarDesc, size_t __pyx_v_bnScale, size_t __pyx_v_bnBias, double __pyx_v_exponentialAverageFactor, size_t __pyx_v_resultRunningMean, size_t __pyx_v_resultRunningVariance, double __pyx_v_epsilon, size_t __pyx_v_resultSaveMean, size_t __pyx_v_resultSaveInvVariance, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("batchNormalizationForwardTraining", 0);

  /* "cupy/cuda/cudnn.pyx":1002
 *         size_t resultRunningMean, size_t resultRunningVariance,
 *         double epsilon, size_t resultSaveMean, size_t resultSaveInvVariance):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnBatchNormalizationForwardTraining(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1003
 *         double epsilon, size_t resultSaveMean, size_t resultSaveInvVariance):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnBatchNormalizationForwardTraining(
 *             <Handle>handle, <BatchNormMode> mode,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1004
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnBatchNormalizationForwardTraining(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <BatchNormMode> mode,
 *             <void*>alpha, <void*>beta, <TensorDescriptor>xDesc,
 */
        __pyx_v_status = cudnnBatchNormalizationForwardTraining(((cudnnHandle_t)__pyx_v_handle), ((cudnnBatchNormMode_t)__pyx_v_mode), ((void *)__pyx_v_alpha), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((void *)__pyx_v_x), ((cudnnTensorDescriptor_t)__pyx_v_yDesc), ((void *)__pyx_v_y), ((cudnnTensorDescriptor_t)__pyx_v_bnScaleBiasMeanVarDesc), ((void *)__pyx_v_bnScale), ((void *)__pyx_v_bnBias), __pyx_v_exponentialAverageFactor, ((void *)__pyx_v_resultRunningMean), ((void *)__pyx_v_resultRunningVariance), __pyx_v_epsilon, ((void *)__pyx_v_resultSaveMean), ((void *)__pyx_v_resultSaveInvVariance));
      }

      /* "cupy/cuda/cudnn.pyx":1003
 *         double epsilon, size_t resultSaveMean, size_t resultSaveInvVariance):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnBatchNormalizationForwardTraining(
 *             <Handle>handle, <BatchNormMode> mode,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1012
 *             <void*>resultRunningMean, <void*>resultRunningVariance,
 *             epsilon, <void*>resultSaveMean, <void*>resultSaveInvVariance)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":994
 * 
 * 
 * cpdef batchNormalizationForwardTraining(             # <<<<<<<<<<<<<<
 *         size_t handle, int mode,
 *         size_t alpha, size_t beta, size_t xDesc,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationForwardTraining", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_101batchNormalizationForwardTraining(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_100batchNormalizationForwardTraining[] = "batchNormalizationForwardTraining(size_t handle, int mode, size_t alpha, size_t beta, size_t xDesc, size_t x, size_t yDesc, size_t y, size_t bnScaleBiasMeanVarDesc, size_t bnScale, size_t bnBias, double exponentialAverageFactor, size_t resultRunningMean, size_t resultRunningVariance, double epsilon, size_t resultSaveMean, size_t resultSaveInvVariance)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_101batchNormalizationForwardTraining(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_mode;
  size_t __pyx_v_alpha;
  size_t __pyx_v_beta;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  size_t __pyx_v_bnScaleBiasMeanVarDesc;
  size_t __pyx_v_bnScale;
  size_t __pyx_v_bnBias;
  double __pyx_v_exponentialAverageFactor;
  size_t __pyx_v_resultRunningMean;
  size_t __pyx_v_resultRunningVariance;
  double __pyx_v_epsilon;
  size_t __pyx_v_resultSaveMean;
  size_t __pyx_v_resultSaveInvVariance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("batchNormalizationForwardTraining (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_mode,&__pyx_n_s_alpha,&__pyx_n_s_beta,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_yDesc,&__pyx_n_s_y,&__pyx_n_s_bnScaleBiasMeanVarDesc,&__pyx_n_s_bnScale,&__pyx_n_s_bnBias,&__pyx_n_s_exponentialAverageFactor,&__pyx_n_s_resultRunningMean,&__pyx_n_s_resultRunningVariance,&__pyx_n_s_epsilon,&__pyx_n_s_resultSaveMean,&__pyx_n_s_resultSaveInvVariance,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 1); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 2); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 3); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 4); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 5); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 6); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 7); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bnScaleBiasMeanVarDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 8); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bnScale)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 9); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bnBias)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 10); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_exponentialAverageFactor)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 11); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_resultRunningMean)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 12); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_resultRunningVariance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 13); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_epsilon)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 14); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_resultSaveMean)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 15); __PYX_ERR(0, 994, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_resultSaveInvVariance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, 16); __PYX_ERR(0, 994, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "batchNormalizationForwardTraining") < 0)) __PYX_ERR(0, 994, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 995, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 995, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 996, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 997, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 997, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 997, __pyx_L3_error)
    __pyx_v_bnScaleBiasMeanVarDesc = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_bnScaleBiasMeanVarDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 998, __pyx_L3_error)
    __pyx_v_bnScale = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_bnScale == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 998, __pyx_L3_error)
    __pyx_v_bnBias = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_bnBias == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 999, __pyx_L3_error)
    __pyx_v_exponentialAverageFactor = __pyx_PyFloat_AsDouble(values[11]); if (unlikely((__pyx_v_exponentialAverageFactor == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 999, __pyx_L3_error)
    __pyx_v_resultRunningMean = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_resultRunningMean == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1000, __pyx_L3_error)
    __pyx_v_resultRunningVariance = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_resultRunningVariance == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1000, __pyx_L3_error)
    __pyx_v_epsilon = __pyx_PyFloat_AsDouble(values[14]); if (unlikely((__pyx_v_epsilon == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1001, __pyx_L3_error)
    __pyx_v_resultSaveMean = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_resultSaveMean == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1001, __pyx_L3_error)
    __pyx_v_resultSaveInvVariance = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_resultSaveInvVariance == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1001, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardTraining", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 994, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationForwardTraining", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_100batchNormalizationForwardTraining(__pyx_self, __pyx_v_handle, __pyx_v_mode, __pyx_v_alpha, __pyx_v_beta, __pyx_v_xDesc, __pyx_v_x, __pyx_v_yDesc, __pyx_v_y, __pyx_v_bnScaleBiasMeanVarDesc, __pyx_v_bnScale, __pyx_v_bnBias, __pyx_v_exponentialAverageFactor, __pyx_v_resultRunningMean, __pyx_v_resultRunningVariance, __pyx_v_epsilon, __pyx_v_resultSaveMean, __pyx_v_resultSaveInvVariance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_100batchNormalizationForwardTraining(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_beta, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_bnScaleBiasMeanVarDesc, size_t __pyx_v_bnScale, size_t __pyx_v_bnBias, double __pyx_v_exponentialAverageFactor, size_t __pyx_v_resultRunningMean, size_t __pyx_v_resultRunningVariance, double __pyx_v_epsilon, size_t __pyx_v_resultSaveMean, size_t __pyx_v_resultSaveInvVariance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("batchNormalizationForwardTraining", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_batchNormalizationForwardTraining(__pyx_v_handle, __pyx_v_mode, __pyx_v_alpha, __pyx_v_beta, __pyx_v_xDesc, __pyx_v_x, __pyx_v_yDesc, __pyx_v_y, __pyx_v_bnScaleBiasMeanVarDesc, __pyx_v_bnScale, __pyx_v_bnBias, __pyx_v_exponentialAverageFactor, __pyx_v_resultRunningMean, __pyx_v_resultRunningVariance, __pyx_v_epsilon, __pyx_v_resultSaveMean, __pyx_v_resultSaveInvVariance, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationForwardTraining", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1015
 * 
 * 
 * cpdef batchNormalizationForwardInference(             # <<<<<<<<<<<<<<
 *         size_t handle, int mode,
 *         size_t alpha, size_t beta, size_t xDesc,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_103batchNormalizationForwardInference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationForwardInference(size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_beta, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_bnScaleBiasMeanVarDesc, size_t __pyx_v_bnScale, size_t __pyx_v_bnBias, size_t __pyx_v_estimatedMean, size_t __pyx_v_estimatedVariance, double __pyx_v_epsilon, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("batchNormalizationForwardInference", 0);

  /* "cupy/cuda/cudnn.pyx":1022
 *         size_t bnBias, size_t estimatedMean, size_t estimatedVariance,
 *         double epsilon):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnBatchNormalizationForwardInference(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1023
 *         double epsilon):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnBatchNormalizationForwardInference(
 *             <Handle>handle, <BatchNormMode> mode,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1024
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnBatchNormalizationForwardInference(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <BatchNormMode> mode,
 *             <void*>alpha, <void*>beta, <TensorDescriptor>xDesc,
 */
        __pyx_v_status = cudnnBatchNormalizationForwardInference(((cudnnHandle_t)__pyx_v_handle), ((cudnnBatchNormMode_t)__pyx_v_mode), ((void *)__pyx_v_alpha), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((void *)__pyx_v_x), ((cudnnTensorDescriptor_t)__pyx_v_yDesc), ((void *)__pyx_v_y), ((cudnnTensorDescriptor_t)__pyx_v_bnScaleBiasMeanVarDesc), ((void *)__pyx_v_bnScale), ((void *)__pyx_v_bnBias), ((void *)__pyx_v_estimatedMean), ((void *)__pyx_v_estimatedVariance), __pyx_v_epsilon);
      }

      /* "cupy/cuda/cudnn.pyx":1023
 *         double epsilon):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnBatchNormalizationForwardInference(
 *             <Handle>handle, <BatchNormMode> mode,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1031
 *             <void*>bnBias, <void*>estimatedMean, <void*>estimatedVariance,
 *             epsilon)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1015
 * 
 * 
 * cpdef batchNormalizationForwardInference(             # <<<<<<<<<<<<<<
 *         size_t handle, int mode,
 *         size_t alpha, size_t beta, size_t xDesc,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationForwardInference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_103batchNormalizationForwardInference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_102batchNormalizationForwardInference[] = "batchNormalizationForwardInference(size_t handle, int mode, size_t alpha, size_t beta, size_t xDesc, size_t x, size_t yDesc, size_t y, size_t bnScaleBiasMeanVarDesc, size_t bnScale, size_t bnBias, size_t estimatedMean, size_t estimatedVariance, double epsilon)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_103batchNormalizationForwardInference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_mode;
  size_t __pyx_v_alpha;
  size_t __pyx_v_beta;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  size_t __pyx_v_bnScaleBiasMeanVarDesc;
  size_t __pyx_v_bnScale;
  size_t __pyx_v_bnBias;
  size_t __pyx_v_estimatedMean;
  size_t __pyx_v_estimatedVariance;
  double __pyx_v_epsilon;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("batchNormalizationForwardInference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_mode,&__pyx_n_s_alpha,&__pyx_n_s_beta,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_yDesc,&__pyx_n_s_y,&__pyx_n_s_bnScaleBiasMeanVarDesc,&__pyx_n_s_bnScale,&__pyx_n_s_bnBias,&__pyx_n_s_estimatedMean,&__pyx_n_s_estimatedVariance,&__pyx_n_s_epsilon,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 1); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 2); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 3); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 4); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 5); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 6); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 7); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bnScaleBiasMeanVarDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 8); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bnScale)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 9); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bnBias)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 10); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_estimatedMean)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 11); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_estimatedVariance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 12); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_epsilon)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, 13); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "batchNormalizationForwardInference") < 0)) __PYX_ERR(0, 1015, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1018, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1018, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1018, __pyx_L3_error)
    __pyx_v_bnScaleBiasMeanVarDesc = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_bnScaleBiasMeanVarDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1019, __pyx_L3_error)
    __pyx_v_bnScale = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_bnScale == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1019, __pyx_L3_error)
    __pyx_v_bnBias = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_bnBias == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L3_error)
    __pyx_v_estimatedMean = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_estimatedMean == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L3_error)
    __pyx_v_estimatedVariance = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_estimatedVariance == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L3_error)
    __pyx_v_epsilon = __pyx_PyFloat_AsDouble(values[13]); if (unlikely((__pyx_v_epsilon == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1021, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("batchNormalizationForwardInference", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1015, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationForwardInference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_102batchNormalizationForwardInference(__pyx_self, __pyx_v_handle, __pyx_v_mode, __pyx_v_alpha, __pyx_v_beta, __pyx_v_xDesc, __pyx_v_x, __pyx_v_yDesc, __pyx_v_y, __pyx_v_bnScaleBiasMeanVarDesc, __pyx_v_bnScale, __pyx_v_bnBias, __pyx_v_estimatedMean, __pyx_v_estimatedVariance, __pyx_v_epsilon);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_102batchNormalizationForwardInference(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_beta, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_bnScaleBiasMeanVarDesc, size_t __pyx_v_bnScale, size_t __pyx_v_bnBias, size_t __pyx_v_estimatedMean, size_t __pyx_v_estimatedVariance, double __pyx_v_epsilon) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("batchNormalizationForwardInference", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_batchNormalizationForwardInference(__pyx_v_handle, __pyx_v_mode, __pyx_v_alpha, __pyx_v_beta, __pyx_v_xDesc, __pyx_v_x, __pyx_v_yDesc, __pyx_v_y, __pyx_v_bnScaleBiasMeanVarDesc, __pyx_v_bnScale, __pyx_v_bnBias, __pyx_v_estimatedMean, __pyx_v_estimatedVariance, __pyx_v_epsilon, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationForwardInference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1034
 * 
 * 
 * cpdef batchNormalizationBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, int mode,
 *         size_t alphaDataDiff, size_t betaDataDiff,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_105batchNormalizationBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationBackward(size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alphaDataDiff, size_t __pyx_v_betaDataDiff, size_t __pyx_v_alphaParamDiff, size_t __pyx_v_betaParamDiff, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_dBnScaleBiasDesc, size_t __pyx_v_bnScale, size_t __pyx_v_dBnScaleResult, size_t __pyx_v_dBnBiasResult, double __pyx_v_epsilon, size_t __pyx_v_savedMean, size_t __pyx_v_savedInvVariance, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("batchNormalizationBackward", 0);

  /* "cupy/cuda/cudnn.pyx":1043
 *         size_t dBnScaleResult, size_t dBnBiasResult,
 *         double epsilon, size_t savedMean, size_t savedInvVariance):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnBatchNormalizationBackward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1044
 *         double epsilon, size_t savedMean, size_t savedInvVariance):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnBatchNormalizationBackward(
 *             <Handle>handle, <BatchNormMode>mode,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1045
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnBatchNormalizationBackward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <BatchNormMode>mode,
 *             <void*>alphaDataDiff, <void*>betaDataDiff,
 */
        __pyx_v_status = cudnnBatchNormalizationBackward(((cudnnHandle_t)__pyx_v_handle), ((cudnnBatchNormMode_t)__pyx_v_mode), ((void *)__pyx_v_alphaDataDiff), ((void *)__pyx_v_betaDataDiff), ((void *)__pyx_v_alphaParamDiff), ((void *)__pyx_v_betaParamDiff), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((void *)__pyx_v_x), ((cudnnTensorDescriptor_t)__pyx_v_dyDesc), ((void *)__pyx_v_dy), ((cudnnTensorDescriptor_t)__pyx_v_dxDesc), ((void *)__pyx_v_dx), ((cudnnTensorDescriptor_t)__pyx_v_dBnScaleBiasDesc), ((void *)__pyx_v_bnScale), ((void *)__pyx_v_dBnScaleResult), ((void *)__pyx_v_dBnBiasResult), __pyx_v_epsilon, ((void *)__pyx_v_savedMean), ((void *)__pyx_v_savedInvVariance));
      }

      /* "cupy/cuda/cudnn.pyx":1044
 *         double epsilon, size_t savedMean, size_t savedInvVariance):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnBatchNormalizationBackward(
 *             <Handle>handle, <BatchNormMode>mode,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1055
 *             <void*>dBnScaleResult, <void*>dBnBiasResult,
 *             epsilon, <void*>savedMean, <void*>savedInvVariance)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * ###############################################################################
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1034
 * 
 * 
 * cpdef batchNormalizationBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, int mode,
 *         size_t alphaDataDiff, size_t betaDataDiff,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_105batchNormalizationBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_104batchNormalizationBackward[] = "batchNormalizationBackward(size_t handle, int mode, size_t alphaDataDiff, size_t betaDataDiff, size_t alphaParamDiff, size_t betaParamDiff, size_t xDesc, size_t x, size_t dyDesc, size_t dy, size_t dxDesc, size_t dx, size_t dBnScaleBiasDesc, size_t bnScale, size_t dBnScaleResult, size_t dBnBiasResult, double epsilon, size_t savedMean, size_t savedInvVariance)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_105batchNormalizationBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_mode;
  size_t __pyx_v_alphaDataDiff;
  size_t __pyx_v_betaDataDiff;
  size_t __pyx_v_alphaParamDiff;
  size_t __pyx_v_betaParamDiff;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_dyDesc;
  size_t __pyx_v_dy;
  size_t __pyx_v_dxDesc;
  size_t __pyx_v_dx;
  size_t __pyx_v_dBnScaleBiasDesc;
  size_t __pyx_v_bnScale;
  size_t __pyx_v_dBnScaleResult;
  size_t __pyx_v_dBnBiasResult;
  double __pyx_v_epsilon;
  size_t __pyx_v_savedMean;
  size_t __pyx_v_savedInvVariance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("batchNormalizationBackward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_mode,&__pyx_n_s_alphaDataDiff,&__pyx_n_s_betaDataDiff,&__pyx_n_s_alphaParamDiff,&__pyx_n_s_betaParamDiff,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_dyDesc,&__pyx_n_s_dy,&__pyx_n_s_dxDesc,&__pyx_n_s_dx,&__pyx_n_s_dBnScaleBiasDesc,&__pyx_n_s_bnScale,&__pyx_n_s_dBnScaleResult,&__pyx_n_s_dBnBiasResult,&__pyx_n_s_epsilon,&__pyx_n_s_savedMean,&__pyx_n_s_savedInvVariance,0};
    PyObject* values[19] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 1); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alphaDataDiff)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 2); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_betaDataDiff)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 3); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alphaParamDiff)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 4); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_betaParamDiff)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 5); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 6); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 7); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 8); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 9); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 10); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 11); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dBnScaleBiasDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 12); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bnScale)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 13); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dBnScaleResult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 14); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dBnBiasResult)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 15); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_epsilon)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 16); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_savedMean)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 17); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_savedInvVariance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, 18); __PYX_ERR(0, 1034, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "batchNormalizationBackward") < 0)) __PYX_ERR(0, 1034, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 19) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1035, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1035, __pyx_L3_error)
    __pyx_v_alphaDataDiff = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alphaDataDiff == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1036, __pyx_L3_error)
    __pyx_v_betaDataDiff = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_betaDataDiff == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1036, __pyx_L3_error)
    __pyx_v_alphaParamDiff = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_alphaParamDiff == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L3_error)
    __pyx_v_betaParamDiff = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_betaParamDiff == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_dyDesc = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_dyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_dy = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_dy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1039, __pyx_L3_error)
    __pyx_v_dxDesc = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_dxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1039, __pyx_L3_error)
    __pyx_v_dx = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_dx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1039, __pyx_L3_error)
    __pyx_v_dBnScaleBiasDesc = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_dBnScaleBiasDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1040, __pyx_L3_error)
    __pyx_v_bnScale = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_bnScale == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1040, __pyx_L3_error)
    __pyx_v_dBnScaleResult = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_dBnScaleResult == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1041, __pyx_L3_error)
    __pyx_v_dBnBiasResult = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_dBnBiasResult == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1041, __pyx_L3_error)
    __pyx_v_epsilon = __pyx_PyFloat_AsDouble(values[16]); if (unlikely((__pyx_v_epsilon == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1042, __pyx_L3_error)
    __pyx_v_savedMean = __Pyx_PyInt_As_size_t(values[17]); if (unlikely((__pyx_v_savedMean == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1042, __pyx_L3_error)
    __pyx_v_savedInvVariance = __Pyx_PyInt_As_size_t(values[18]); if (unlikely((__pyx_v_savedInvVariance == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1042, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("batchNormalizationBackward", 1, 19, 19, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1034, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_104batchNormalizationBackward(__pyx_self, __pyx_v_handle, __pyx_v_mode, __pyx_v_alphaDataDiff, __pyx_v_betaDataDiff, __pyx_v_alphaParamDiff, __pyx_v_betaParamDiff, __pyx_v_xDesc, __pyx_v_x, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_dxDesc, __pyx_v_dx, __pyx_v_dBnScaleBiasDesc, __pyx_v_bnScale, __pyx_v_dBnScaleResult, __pyx_v_dBnBiasResult, __pyx_v_epsilon, __pyx_v_savedMean, __pyx_v_savedInvVariance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_104batchNormalizationBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_mode, size_t __pyx_v_alphaDataDiff, size_t __pyx_v_betaDataDiff, size_t __pyx_v_alphaParamDiff, size_t __pyx_v_betaParamDiff, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_dBnScaleBiasDesc, size_t __pyx_v_bnScale, size_t __pyx_v_dBnScaleResult, size_t __pyx_v_dBnBiasResult, double __pyx_v_epsilon, size_t __pyx_v_savedMean, size_t __pyx_v_savedInvVariance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("batchNormalizationBackward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_batchNormalizationBackward(__pyx_v_handle, __pyx_v_mode, __pyx_v_alphaDataDiff, __pyx_v_betaDataDiff, __pyx_v_alphaParamDiff, __pyx_v_betaParamDiff, __pyx_v_xDesc, __pyx_v_x, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_dxDesc, __pyx_v_dx, __pyx_v_dBnScaleBiasDesc, __pyx_v_bnScale, __pyx_v_dBnScaleResult, __pyx_v_dBnBiasResult, __pyx_v_epsilon, __pyx_v_savedMean, __pyx_v_savedInvVariance, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.batchNormalizationBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1061
 * ###############################################################################
 * 
 * cpdef size_t createActivationDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef ActivationDescriptor activationDesc
 *     status = cudnnCreateActivationDescriptor(&activationDesc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_107createActivationDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createActivationDescriptor(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnActivationDescriptor_t __pyx_v_activationDesc;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createActivationDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1063
 * cpdef size_t createActivationDescriptor() except *:
 *     cdef ActivationDescriptor activationDesc
 *     status = cudnnCreateActivationDescriptor(&activationDesc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>activationDesc
 */
  __pyx_v_status = cudnnCreateActivationDescriptor((&__pyx_v_activationDesc));

  /* "cupy/cuda/cudnn.pyx":1064
 *     cdef ActivationDescriptor activationDesc
 *     status = cudnnCreateActivationDescriptor(&activationDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>activationDesc
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1065
 *     status = cudnnCreateActivationDescriptor(&activationDesc)
 *     check_status(status)
 *     return <size_t>activationDesc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_activationDesc);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1061
 * ###############################################################################
 * 
 * cpdef size_t createActivationDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef ActivationDescriptor activationDesc
 *     status = cudnnCreateActivationDescriptor(&activationDesc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createActivationDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_107createActivationDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_106createActivationDescriptor[] = "createActivationDescriptor() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_107createActivationDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createActivationDescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_106createActivationDescriptor(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_106createActivationDescriptor(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createActivationDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createActivationDescriptor(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1061, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createActivationDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1068
 * 
 * 
 * cpdef setActivationDescriptor(             # <<<<<<<<<<<<<<
 *         size_t activationDesc, int mode, int reluNanOpt, double reluCeiling):
 *     status = cudnnSetActivationDescriptor(
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_109setActivationDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setActivationDescriptor(size_t __pyx_v_activationDesc, int __pyx_v_mode, int __pyx_v_reluNanOpt, double __pyx_v_reluCeiling, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setActivationDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1070
 * cpdef setActivationDescriptor(
 *         size_t activationDesc, int mode, int reluNanOpt, double reluCeiling):
 *     status = cudnnSetActivationDescriptor(             # <<<<<<<<<<<<<<
 *         <ActivationDescriptor>activationDesc, <ActivationMode>mode,
 *         <NanPropagation>reluNanOpt, reluCeiling)
 */
  __pyx_v_status = cudnnSetActivationDescriptor(((cudnnActivationDescriptor_t)__pyx_v_activationDesc), ((cudnnActivationMode_t)__pyx_v_mode), ((cudnnNanPropagation_t)__pyx_v_reluNanOpt), __pyx_v_reluCeiling);

  /* "cupy/cuda/cudnn.pyx":1073
 *         <ActivationDescriptor>activationDesc, <ActivationMode>mode,
 *         <NanPropagation>reluNanOpt, reluCeiling)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1068
 * 
 * 
 * cpdef setActivationDescriptor(             # <<<<<<<<<<<<<<
 *         size_t activationDesc, int mode, int reluNanOpt, double reluCeiling):
 *     status = cudnnSetActivationDescriptor(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setActivationDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_109setActivationDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_108setActivationDescriptor[] = "setActivationDescriptor(size_t activationDesc, int mode, int reluNanOpt, double reluCeiling)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_109setActivationDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_activationDesc;
  int __pyx_v_mode;
  int __pyx_v_reluNanOpt;
  double __pyx_v_reluCeiling;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setActivationDescriptor (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_activationDesc,&__pyx_n_s_mode,&__pyx_n_s_reluNanOpt,&__pyx_n_s_reluCeiling,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_activationDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setActivationDescriptor", 1, 4, 4, 1); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reluNanOpt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setActivationDescriptor", 1, 4, 4, 2); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reluCeiling)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setActivationDescriptor", 1, 4, 4, 3); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setActivationDescriptor") < 0)) __PYX_ERR(0, 1068, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_activationDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_activationDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L3_error)
    __pyx_v_reluNanOpt = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_reluNanOpt == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L3_error)
    __pyx_v_reluCeiling = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_reluCeiling == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setActivationDescriptor", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1068, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setActivationDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_108setActivationDescriptor(__pyx_self, __pyx_v_activationDesc, __pyx_v_mode, __pyx_v_reluNanOpt, __pyx_v_reluCeiling);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_108setActivationDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_activationDesc, int __pyx_v_mode, int __pyx_v_reluNanOpt, double __pyx_v_reluCeiling) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setActivationDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setActivationDescriptor(__pyx_v_activationDesc, __pyx_v_mode, __pyx_v_reluNanOpt, __pyx_v_reluCeiling, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setActivationDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1076
 * 
 * 
 * cpdef destroyActivationDescriptor(size_t activationDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyActivationDescriptor(
 *         <ActivationDescriptor>activationDesc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_111destroyActivationDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_activationDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyActivationDescriptor(size_t __pyx_v_activationDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyActivationDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1077
 * 
 * cpdef destroyActivationDescriptor(size_t activationDesc):
 *     status = cudnnDestroyActivationDescriptor(             # <<<<<<<<<<<<<<
 *         <ActivationDescriptor>activationDesc)
 *     check_status(status)
 */
  __pyx_v_status = cudnnDestroyActivationDescriptor(((cudnnActivationDescriptor_t)__pyx_v_activationDesc));

  /* "cupy/cuda/cudnn.pyx":1079
 *     status = cudnnDestroyActivationDescriptor(
 *         <ActivationDescriptor>activationDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1076
 * 
 * 
 * cpdef destroyActivationDescriptor(size_t activationDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyActivationDescriptor(
 *         <ActivationDescriptor>activationDesc)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyActivationDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_111destroyActivationDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_activationDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_110destroyActivationDescriptor[] = "destroyActivationDescriptor(size_t activationDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_111destroyActivationDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_activationDesc) {
  size_t __pyx_v_activationDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyActivationDescriptor (wrapper)", 0);
  assert(__pyx_arg_activationDesc); {
    __pyx_v_activationDesc = __Pyx_PyInt_As_size_t(__pyx_arg_activationDesc); if (unlikely((__pyx_v_activationDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1076, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyActivationDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_110destroyActivationDescriptor(__pyx_self, ((size_t)__pyx_v_activationDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_110destroyActivationDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_activationDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyActivationDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroyActivationDescriptor(__pyx_v_activationDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyActivationDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1082
 * 
 * 
 * cpdef softmaxForward(             # <<<<<<<<<<<<<<
 *         size_t handle, int algorithm, int mode, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_113softmaxForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_softmaxForward(size_t __pyx_v_handle, int __pyx_v_algorithm, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("softmaxForward", 0);

  /* "cupy/cuda/cudnn.pyx":1085
 *         size_t handle, int algorithm, int mode, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnSoftmaxForward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1086
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSoftmaxForward(
 *             <Handle>handle, <SoftmaxAlgorithm>algorithm, <SoftmaxMode>mode,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1087
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnSoftmaxForward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SoftmaxAlgorithm>algorithm, <SoftmaxMode>mode,
 *             <void*>alpha, <TensorDescriptor>srcDesc, <void*>srcData,
 */
        __pyx_v_status = cudnnSoftmaxForward(((cudnnHandle_t)__pyx_v_handle), ((cudnnSoftmaxAlgorithm_t)__pyx_v_algorithm), ((cudnnSoftmaxMode_t)__pyx_v_mode), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_dstDesc), ((void *)__pyx_v_dstData));
      }

      /* "cupy/cuda/cudnn.pyx":1086
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSoftmaxForward(
 *             <Handle>handle, <SoftmaxAlgorithm>algorithm, <SoftmaxMode>mode,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1091
 *             <void*>alpha, <TensorDescriptor>srcDesc, <void*>srcData,
 *             <void*>beta, <TensorDescriptor>dstDesc, <void*>dstData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1082
 * 
 * 
 * cpdef softmaxForward(             # <<<<<<<<<<<<<<
 *         size_t handle, int algorithm, int mode, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.softmaxForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_113softmaxForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_112softmaxForward[] = "softmaxForward(size_t handle, int algorithm, int mode, size_t alpha, size_t srcDesc, size_t srcData, size_t beta, size_t dstDesc, size_t dstData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_113softmaxForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_algorithm;
  int __pyx_v_mode;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_beta;
  size_t __pyx_v_dstDesc;
  size_t __pyx_v_dstData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("softmaxForward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_algorithm,&__pyx_n_s_mode,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_beta,&__pyx_n_s_dstDesc,&__pyx_n_s_dstData,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algorithm)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, 1); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, 2); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, 3); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, 4); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, 5); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, 6); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dstDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, 7); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dstData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, 8); __PYX_ERR(0, 1082, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "softmaxForward") < 0)) __PYX_ERR(0, 1082, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1083, __pyx_L3_error)
    __pyx_v_algorithm = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_algorithm == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1083, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1083, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1083, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1083, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
    __pyx_v_dstDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_dstDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
    __pyx_v_dstData = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_dstData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("softmaxForward", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1082, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.softmaxForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_112softmaxForward(__pyx_self, __pyx_v_handle, __pyx_v_algorithm, __pyx_v_mode, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_beta, __pyx_v_dstDesc, __pyx_v_dstData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_112softmaxForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_algorithm, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("softmaxForward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_softmaxForward(__pyx_v_handle, __pyx_v_algorithm, __pyx_v_mode, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_beta, __pyx_v_dstDesc, __pyx_v_dstData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.softmaxForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1094
 * 
 * 
 * cpdef softmaxBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, int algorithm, int mode, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t srcDiffDesc, size_t srcDiffData, size_t beta,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_115softmaxBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_softmaxBackward(size_t __pyx_v_handle, int __pyx_v_algorithm, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("softmaxBackward", 0);

  /* "cupy/cuda/cudnn.pyx":1098
 *         size_t srcData, size_t srcDiffDesc, size_t srcDiffData, size_t beta,
 *         size_t destDiffDesc, size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnSoftmaxBackward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1099
 *         size_t destDiffDesc, size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSoftmaxBackward(
 *             <Handle>handle, <SoftmaxAlgorithm>algorithm, <SoftmaxMode>mode,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1100
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnSoftmaxBackward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SoftmaxAlgorithm>algorithm, <SoftmaxMode>mode,
 *             <void*>alpha, <TensorDescriptor>srcDesc, <void*>srcData,
 */
        __pyx_v_status = cudnnSoftmaxBackward(((cudnnHandle_t)__pyx_v_handle), ((cudnnSoftmaxAlgorithm_t)__pyx_v_algorithm), ((cudnnSoftmaxMode_t)__pyx_v_mode), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((cudnnTensorDescriptor_t)__pyx_v_srcDiffDesc), ((void *)__pyx_v_srcDiffData), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_destDiffDesc), ((void *)__pyx_v_destDiffData));
      }

      /* "cupy/cuda/cudnn.pyx":1099
 *         size_t destDiffDesc, size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSoftmaxBackward(
 *             <Handle>handle, <SoftmaxAlgorithm>algorithm, <SoftmaxMode>mode,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1105
 *             <TensorDescriptor>srcDiffDesc, <void*>srcDiffData, <void*>beta,
 *             <TensorDescriptor>destDiffDesc, <void*>destDiffData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1094
 * 
 * 
 * cpdef softmaxBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, int algorithm, int mode, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t srcDiffDesc, size_t srcDiffData, size_t beta,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.softmaxBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_115softmaxBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_114softmaxBackward[] = "softmaxBackward(size_t handle, int algorithm, int mode, size_t alpha, size_t srcDesc, size_t srcData, size_t srcDiffDesc, size_t srcDiffData, size_t beta, size_t destDiffDesc, size_t destDiffData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_115softmaxBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  int __pyx_v_algorithm;
  int __pyx_v_mode;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_srcDiffDesc;
  size_t __pyx_v_srcDiffData;
  size_t __pyx_v_beta;
  size_t __pyx_v_destDiffDesc;
  size_t __pyx_v_destDiffData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("softmaxBackward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_algorithm,&__pyx_n_s_mode,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_srcDiffDesc,&__pyx_n_s_srcDiffData,&__pyx_n_s_beta,&__pyx_n_s_destDiffDesc,&__pyx_n_s_destDiffData,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algorithm)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 1); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 2); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 3); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 4); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 5); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDiffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 6); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDiffData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 7); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 8); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDiffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 9); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDiffData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, 10); __PYX_ERR(0, 1094, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "softmaxBackward") < 0)) __PYX_ERR(0, 1094, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1095, __pyx_L3_error)
    __pyx_v_algorithm = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_algorithm == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1095, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1095, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1095, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1095, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1096, __pyx_L3_error)
    __pyx_v_srcDiffDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_srcDiffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1096, __pyx_L3_error)
    __pyx_v_srcDiffData = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_srcDiffData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1096, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1096, __pyx_L3_error)
    __pyx_v_destDiffDesc = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_destDiffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1097, __pyx_L3_error)
    __pyx_v_destDiffData = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_destDiffData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1097, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("softmaxBackward", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1094, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.softmaxBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_114softmaxBackward(__pyx_self, __pyx_v_handle, __pyx_v_algorithm, __pyx_v_mode, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_srcDiffDesc, __pyx_v_srcDiffData, __pyx_v_beta, __pyx_v_destDiffDesc, __pyx_v_destDiffData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_114softmaxBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, int __pyx_v_algorithm, int __pyx_v_mode, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("softmaxBackward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_softmaxBackward(__pyx_v_handle, __pyx_v_algorithm, __pyx_v_mode, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_srcDiffDesc, __pyx_v_srcDiffData, __pyx_v_beta, __pyx_v_destDiffDesc, __pyx_v_destDiffData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.softmaxBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1108
 * 
 * 
 * cpdef activationForward_v4(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t activationDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_117activationForward_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_activationForward_v4(size_t __pyx_v_handle, size_t __pyx_v_activationDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("activationForward_v4", 0);

  /* "cupy/cuda/cudnn.pyx":1111
 *         size_t handle, size_t activationDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnActivationForward_v4(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1112
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnActivationForward_v4(
 *             <Handle>handle, <ActivationDescriptor>activationDesc, <void*>alpha,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1113
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnActivationForward_v4(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <ActivationDescriptor>activationDesc, <void*>alpha,
 *             <TensorDescriptor>srcDesc, <void*>srcData, <void*>beta,
 */
        __pyx_v_status = cudnnActivationForward_v4(((cudnnHandle_t)__pyx_v_handle), ((cudnnActivationDescriptor_t)__pyx_v_activationDesc), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_dstDesc), ((void *)__pyx_v_dstData));
      }

      /* "cupy/cuda/cudnn.pyx":1112
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnActivationForward_v4(
 *             <Handle>handle, <ActivationDescriptor>activationDesc, <void*>alpha,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1117
 *             <TensorDescriptor>srcDesc, <void*>srcData, <void*>beta,
 *             <TensorDescriptor>dstDesc, <void*>dstData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1108
 * 
 * 
 * cpdef activationForward_v4(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t activationDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t beta, size_t dstDesc, size_t dstData):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.activationForward_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_117activationForward_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_116activationForward_v4[] = "activationForward_v4(size_t handle, size_t activationDesc, size_t alpha, size_t srcDesc, size_t srcData, size_t beta, size_t dstDesc, size_t dstData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_117activationForward_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_activationDesc;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_beta;
  size_t __pyx_v_dstDesc;
  size_t __pyx_v_dstData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("activationForward_v4 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_activationDesc,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_beta,&__pyx_n_s_dstDesc,&__pyx_n_s_dstData,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_activationDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationForward_v4", 1, 8, 8, 1); __PYX_ERR(0, 1108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationForward_v4", 1, 8, 8, 2); __PYX_ERR(0, 1108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationForward_v4", 1, 8, 8, 3); __PYX_ERR(0, 1108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationForward_v4", 1, 8, 8, 4); __PYX_ERR(0, 1108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationForward_v4", 1, 8, 8, 5); __PYX_ERR(0, 1108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dstDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationForward_v4", 1, 8, 8, 6); __PYX_ERR(0, 1108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dstData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationForward_v4", 1, 8, 8, 7); __PYX_ERR(0, 1108, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "activationForward_v4") < 0)) __PYX_ERR(0, 1108, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1109, __pyx_L3_error)
    __pyx_v_activationDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_activationDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1109, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1109, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1109, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
    __pyx_v_dstDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_dstDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
    __pyx_v_dstData = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_dstData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1110, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("activationForward_v4", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1108, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.activationForward_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_116activationForward_v4(__pyx_self, __pyx_v_handle, __pyx_v_activationDesc, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_beta, __pyx_v_dstDesc, __pyx_v_dstData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_116activationForward_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_activationDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_beta, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("activationForward_v4", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_activationForward_v4(__pyx_v_handle, __pyx_v_activationDesc, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_beta, __pyx_v_dstDesc, __pyx_v_dstData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.activationForward_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1120
 * 
 * 
 * cpdef activationBackward_v4(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t activationDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t srcDiffDesc, size_t srcDiffData,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_119activationBackward_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_activationBackward_v4(size_t __pyx_v_handle, size_t __pyx_v_activationDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_destDesc, size_t __pyx_v_destData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("activationBackward_v4", 0);

  /* "cupy/cuda/cudnn.pyx":1125
 *         size_t destDesc, size_t destData, size_t beta, size_t destDiffDesc,
 *         size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnActivationBackward_v4(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1126
 *         size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnActivationBackward_v4(
 *             <Handle>handle, <ActivationDescriptor>activationDesc, <void*>alpha,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1127
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnActivationBackward_v4(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <ActivationDescriptor>activationDesc, <void*>alpha,
 *             <TensorDescriptor>srcDesc, <void*>srcData,
 */
        __pyx_v_status = cudnnActivationBackward_v4(((cudnnHandle_t)__pyx_v_handle), ((cudnnActivationDescriptor_t)__pyx_v_activationDesc), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((cudnnTensorDescriptor_t)__pyx_v_srcDiffDesc), ((void *)__pyx_v_srcDiffData), ((cudnnTensorDescriptor_t)__pyx_v_destDesc), ((void *)__pyx_v_destData), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_destDiffDesc), ((void *)__pyx_v_destDiffData));
      }

      /* "cupy/cuda/cudnn.pyx":1126
 *         size_t destDiffData):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnActivationBackward_v4(
 *             <Handle>handle, <ActivationDescriptor>activationDesc, <void*>alpha,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1133
 *             <TensorDescriptor>destDesc, <void*>destData, <void*>beta,
 *             <TensorDescriptor>destDiffDesc, <void*>destDiffData)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1120
 * 
 * 
 * cpdef activationBackward_v4(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t activationDesc, size_t alpha, size_t srcDesc,
 *         size_t srcData, size_t srcDiffDesc, size_t srcDiffData,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.activationBackward_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_119activationBackward_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_118activationBackward_v4[] = "activationBackward_v4(size_t handle, size_t activationDesc, size_t alpha, size_t srcDesc, size_t srcData, size_t srcDiffDesc, size_t srcDiffData, size_t destDesc, size_t destData, size_t beta, size_t destDiffDesc, size_t destDiffData)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_119activationBackward_v4(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_activationDesc;
  size_t __pyx_v_alpha;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_srcDiffDesc;
  size_t __pyx_v_srcDiffData;
  size_t __pyx_v_destDesc;
  size_t __pyx_v_destData;
  size_t __pyx_v_beta;
  size_t __pyx_v_destDiffDesc;
  size_t __pyx_v_destDiffData;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("activationBackward_v4 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_activationDesc,&__pyx_n_s_alpha,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_srcDiffDesc,&__pyx_n_s_srcDiffData,&__pyx_n_s_destDesc,&__pyx_n_s_destData,&__pyx_n_s_beta,&__pyx_n_s_destDiffDesc,&__pyx_n_s_destDiffData,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_activationDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 1); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 2); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 3); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 4); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDiffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 5); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDiffData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 6); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 7); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 8); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 9); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDiffDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 10); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_destDiffData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, 11); __PYX_ERR(0, 1120, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "activationBackward_v4") < 0)) __PYX_ERR(0, 1120, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1121, __pyx_L3_error)
    __pyx_v_activationDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_activationDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1121, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1121, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1121, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
    __pyx_v_srcDiffDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_srcDiffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
    __pyx_v_srcDiffData = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_srcDiffData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
    __pyx_v_destDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_destDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1123, __pyx_L3_error)
    __pyx_v_destData = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_destData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1123, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1123, __pyx_L3_error)
    __pyx_v_destDiffDesc = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_destDiffDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1123, __pyx_L3_error)
    __pyx_v_destDiffData = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_destDiffData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1124, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("activationBackward_v4", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1120, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.activationBackward_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_118activationBackward_v4(__pyx_self, __pyx_v_handle, __pyx_v_activationDesc, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_srcDiffDesc, __pyx_v_srcDiffData, __pyx_v_destDesc, __pyx_v_destData, __pyx_v_beta, __pyx_v_destDiffDesc, __pyx_v_destDiffData);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_118activationBackward_v4(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_activationDesc, size_t __pyx_v_alpha, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_srcDiffDesc, size_t __pyx_v_srcDiffData, size_t __pyx_v_destDesc, size_t __pyx_v_destData, size_t __pyx_v_beta, size_t __pyx_v_destDiffDesc, size_t __pyx_v_destDiffData) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("activationBackward_v4", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_activationBackward_v4(__pyx_v_handle, __pyx_v_activationDesc, __pyx_v_alpha, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_srcDiffDesc, __pyx_v_srcDiffData, __pyx_v_destDesc, __pyx_v_destData, __pyx_v_beta, __pyx_v_destDiffDesc, __pyx_v_destDiffData, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.activationBackward_v4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1140
 * ###############################################################################
 * 
 * cpdef size_t createDropoutDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef DropoutDescriptor desc
 *     status = cudnnCreateDropoutDescriptor(&desc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_121createDropoutDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createDropoutDescriptor(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnDropoutDescriptor_t __pyx_v_desc;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createDropoutDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1142
 * cpdef size_t createDropoutDescriptor() except *:
 *     cdef DropoutDescriptor desc
 *     status = cudnnCreateDropoutDescriptor(&desc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>desc
 */
  __pyx_v_status = cudnnCreateDropoutDescriptor((&__pyx_v_desc));

  /* "cupy/cuda/cudnn.pyx":1143
 *     cdef DropoutDescriptor desc
 *     status = cudnnCreateDropoutDescriptor(&desc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>desc
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1144
 *     status = cudnnCreateDropoutDescriptor(&desc)
 *     check_status(status)
 *     return <size_t>desc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_desc);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1140
 * ###############################################################################
 * 
 * cpdef size_t createDropoutDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef DropoutDescriptor desc
 *     status = cudnnCreateDropoutDescriptor(&desc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createDropoutDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_121createDropoutDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_120createDropoutDescriptor[] = "createDropoutDescriptor() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_121createDropoutDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createDropoutDescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_120createDropoutDescriptor(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_120createDropoutDescriptor(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createDropoutDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createDropoutDescriptor(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1140, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createDropoutDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1147
 * 
 * 
 * cpdef destroyDropoutDescriptor(size_t dropoutDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyDropoutDescriptor(<DropoutDescriptor>dropoutDesc)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_123destroyDropoutDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_dropoutDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyDropoutDescriptor(size_t __pyx_v_dropoutDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyDropoutDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1148
 * 
 * cpdef destroyDropoutDescriptor(size_t dropoutDesc):
 *     status = cudnnDestroyDropoutDescriptor(<DropoutDescriptor>dropoutDesc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cudnnDestroyDropoutDescriptor(((cudnnDropoutDescriptor_t)__pyx_v_dropoutDesc));

  /* "cupy/cuda/cudnn.pyx":1149
 * cpdef destroyDropoutDescriptor(size_t dropoutDesc):
 *     status = cudnnDestroyDropoutDescriptor(<DropoutDescriptor>dropoutDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1147
 * 
 * 
 * cpdef destroyDropoutDescriptor(size_t dropoutDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyDropoutDescriptor(<DropoutDescriptor>dropoutDesc)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyDropoutDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_123destroyDropoutDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_dropoutDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_122destroyDropoutDescriptor[] = "destroyDropoutDescriptor(size_t dropoutDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_123destroyDropoutDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_dropoutDesc) {
  size_t __pyx_v_dropoutDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyDropoutDescriptor (wrapper)", 0);
  assert(__pyx_arg_dropoutDesc); {
    __pyx_v_dropoutDesc = __Pyx_PyInt_As_size_t(__pyx_arg_dropoutDesc); if (unlikely((__pyx_v_dropoutDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1147, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyDropoutDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_122destroyDropoutDescriptor(__pyx_self, ((size_t)__pyx_v_dropoutDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_122destroyDropoutDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_dropoutDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyDropoutDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroyDropoutDescriptor(__pyx_v_dropoutDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyDropoutDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1152
 * 
 * 
 * cpdef size_t dropoutGetStatesSize(size_t handle) except *:             # <<<<<<<<<<<<<<
 *     cdef size_t sizeInBytes
 *     status = cudnnDropoutGetStatesSize(
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_125dropoutGetStatesSize(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_dropoutGetStatesSize(size_t __pyx_v_handle, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_sizeInBytes;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dropoutGetStatesSize", 0);

  /* "cupy/cuda/cudnn.pyx":1154
 * cpdef size_t dropoutGetStatesSize(size_t handle) except *:
 *     cdef size_t sizeInBytes
 *     status = cudnnDropoutGetStatesSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, &sizeInBytes)
 *     check_status(status)
 */
  __pyx_v_status = cudnnDropoutGetStatesSize(((cudnnHandle_t)__pyx_v_handle), (&__pyx_v_sizeInBytes));

  /* "cupy/cuda/cudnn.pyx":1156
 *     status = cudnnDropoutGetStatesSize(
 *         <Handle>handle, &sizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return sizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1157
 *         <Handle>handle, &sizeInBytes)
 *     check_status(status)
 *     return sizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_sizeInBytes;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1152
 * 
 * 
 * cpdef size_t dropoutGetStatesSize(size_t handle) except *:             # <<<<<<<<<<<<<<
 *     cdef size_t sizeInBytes
 *     status = cudnnDropoutGetStatesSize(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutGetStatesSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_125dropoutGetStatesSize(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_124dropoutGetStatesSize[] = "dropoutGetStatesSize(size_t handle) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_125dropoutGetStatesSize(PyObject *__pyx_self, PyObject *__pyx_arg_handle) {
  size_t __pyx_v_handle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dropoutGetStatesSize (wrapper)", 0);
  assert(__pyx_arg_handle); {
    __pyx_v_handle = __Pyx_PyInt_As_size_t(__pyx_arg_handle); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1152, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutGetStatesSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_124dropoutGetStatesSize(__pyx_self, ((size_t)__pyx_v_handle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_124dropoutGetStatesSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dropoutGetStatesSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_dropoutGetStatesSize(__pyx_v_handle, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1152, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutGetStatesSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1160
 * 
 * 
 * cpdef setDropoutDescriptor(             # <<<<<<<<<<<<<<
 *         size_t dropoutDesc, size_t handle, float dropout,
 *         size_t states, size_t stateSizeInBytes, unsigned long long seed):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_127setDropoutDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setDropoutDescriptor(size_t __pyx_v_dropoutDesc, size_t __pyx_v_handle, float __pyx_v_dropout, size_t __pyx_v_states, size_t __pyx_v_stateSizeInBytes, unsigned PY_LONG_LONG __pyx_v_seed, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setDropoutDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1163
 *         size_t dropoutDesc, size_t handle, float dropout,
 *         size_t states, size_t stateSizeInBytes, unsigned long long seed):
 *     status = cudnnSetDropoutDescriptor(             # <<<<<<<<<<<<<<
 *         <DropoutDescriptor>dropoutDesc, <Handle>handle, dropout,
 *         <void*>states, stateSizeInBytes, seed)
 */
  __pyx_v_status = cudnnSetDropoutDescriptor(((cudnnDropoutDescriptor_t)__pyx_v_dropoutDesc), ((cudnnHandle_t)__pyx_v_handle), __pyx_v_dropout, ((void *)__pyx_v_states), __pyx_v_stateSizeInBytes, __pyx_v_seed);

  /* "cupy/cuda/cudnn.pyx":1166
 *         <DropoutDescriptor>dropoutDesc, <Handle>handle, dropout,
 *         <void*>states, stateSizeInBytes, seed)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1160
 * 
 * 
 * cpdef setDropoutDescriptor(             # <<<<<<<<<<<<<<
 *         size_t dropoutDesc, size_t handle, float dropout,
 *         size_t states, size_t stateSizeInBytes, unsigned long long seed):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setDropoutDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_127setDropoutDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_126setDropoutDescriptor[] = "setDropoutDescriptor(size_t dropoutDesc, size_t handle, float dropout, size_t states, size_t stateSizeInBytes, unsigned long long seed)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_127setDropoutDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_dropoutDesc;
  size_t __pyx_v_handle;
  float __pyx_v_dropout;
  size_t __pyx_v_states;
  size_t __pyx_v_stateSizeInBytes;
  unsigned PY_LONG_LONG __pyx_v_seed;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setDropoutDescriptor (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dropoutDesc,&__pyx_n_s_handle,&__pyx_n_s_dropout,&__pyx_n_s_states,&__pyx_n_s_stateSizeInBytes,&__pyx_n_s_seed,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dropoutDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setDropoutDescriptor", 1, 6, 6, 1); __PYX_ERR(0, 1160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dropout)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setDropoutDescriptor", 1, 6, 6, 2); __PYX_ERR(0, 1160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_states)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setDropoutDescriptor", 1, 6, 6, 3); __PYX_ERR(0, 1160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stateSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setDropoutDescriptor", 1, 6, 6, 4); __PYX_ERR(0, 1160, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setDropoutDescriptor", 1, 6, 6, 5); __PYX_ERR(0, 1160, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setDropoutDescriptor") < 0)) __PYX_ERR(0, 1160, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_dropoutDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_dropoutDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1161, __pyx_L3_error)
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1161, __pyx_L3_error)
    __pyx_v_dropout = __pyx_PyFloat_AsFloat(values[2]); if (unlikely((__pyx_v_dropout == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1161, __pyx_L3_error)
    __pyx_v_states = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_states == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1162, __pyx_L3_error)
    __pyx_v_stateSizeInBytes = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_stateSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1162, __pyx_L3_error)
    __pyx_v_seed = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(values[5]); if (unlikely((__pyx_v_seed == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1162, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setDropoutDescriptor", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1160, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setDropoutDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_126setDropoutDescriptor(__pyx_self, __pyx_v_dropoutDesc, __pyx_v_handle, __pyx_v_dropout, __pyx_v_states, __pyx_v_stateSizeInBytes, __pyx_v_seed);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_126setDropoutDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_dropoutDesc, size_t __pyx_v_handle, float __pyx_v_dropout, size_t __pyx_v_states, size_t __pyx_v_stateSizeInBytes, unsigned PY_LONG_LONG __pyx_v_seed) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setDropoutDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setDropoutDescriptor(__pyx_v_dropoutDesc, __pyx_v_handle, __pyx_v_dropout, __pyx_v_states, __pyx_v_stateSizeInBytes, __pyx_v_seed, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setDropoutDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1169
 * 
 * 
 * cpdef size_t getDropoutReserveSpaceSize(size_t xDesc):             # <<<<<<<<<<<<<<
 *     cdef size_t sizeInBytes
 *     status = cudnnDropoutGetReserveSpaceSize(
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_129getDropoutReserveSpaceSize(PyObject *__pyx_self, PyObject *__pyx_arg_xDesc); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_getDropoutReserveSpaceSize(size_t __pyx_v_xDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_sizeInBytes;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getDropoutReserveSpaceSize", 0);

  /* "cupy/cuda/cudnn.pyx":1171
 * cpdef size_t getDropoutReserveSpaceSize(size_t xDesc):
 *     cdef size_t sizeInBytes
 *     status = cudnnDropoutGetReserveSpaceSize(             # <<<<<<<<<<<<<<
 *         <TensorDescriptor>xDesc, &sizeInBytes)
 *     check_status(status)
 */
  __pyx_v_status = cudnnDropoutGetReserveSpaceSize(((cudnnTensorDescriptor_t)__pyx_v_xDesc), (&__pyx_v_sizeInBytes));

  /* "cupy/cuda/cudnn.pyx":1173
 *     status = cudnnDropoutGetReserveSpaceSize(
 *         <TensorDescriptor>xDesc, &sizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return sizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1174
 *         <TensorDescriptor>xDesc, &sizeInBytes)
 *     check_status(status)
 *     return sizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_sizeInBytes;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1169
 * 
 * 
 * cpdef size_t getDropoutReserveSpaceSize(size_t xDesc):             # <<<<<<<<<<<<<<
 *     cdef size_t sizeInBytes
 *     status = cudnnDropoutGetReserveSpaceSize(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cudnn.getDropoutReserveSpaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_129getDropoutReserveSpaceSize(PyObject *__pyx_self, PyObject *__pyx_arg_xDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_128getDropoutReserveSpaceSize[] = "getDropoutReserveSpaceSize(size_t xDesc) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_129getDropoutReserveSpaceSize(PyObject *__pyx_self, PyObject *__pyx_arg_xDesc) {
  size_t __pyx_v_xDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getDropoutReserveSpaceSize (wrapper)", 0);
  assert(__pyx_arg_xDesc); {
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(__pyx_arg_xDesc); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1169, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getDropoutReserveSpaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_128getDropoutReserveSpaceSize(__pyx_self, ((size_t)__pyx_v_xDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_128getDropoutReserveSpaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_xDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getDropoutReserveSpaceSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_f_4cupy_4cuda_5cudnn_getDropoutReserveSpaceSize(__pyx_v_xDesc, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getDropoutReserveSpaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1177
 * 
 * 
 * cpdef dropoutForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t dropoutDesc,
 *         size_t srcDesc, size_t srcData,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_131dropoutForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_dropoutForward(size_t __pyx_v_handle, size_t __pyx_v_dropoutDesc, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dropoutForward", 0);

  /* "cupy/cuda/cudnn.pyx":1182
 *         size_t dstDesc, size_t dstData,
 *         size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnDropoutForward(
 *             <Handle>handle, <DropoutDescriptor>dropoutDesc,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1183
 *         size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     with nogil:
 *         status = cudnnDropoutForward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <DropoutDescriptor>dropoutDesc,
 *             <TensorDescriptor>srcDesc, <void*>srcData,
 */
        __pyx_v_status = cudnnDropoutForward(((cudnnHandle_t)__pyx_v_handle), ((cudnnDropoutDescriptor_t)__pyx_v_dropoutDesc), ((cudnnTensorDescriptor_t)__pyx_v_srcDesc), ((void *)__pyx_v_srcData), ((cudnnTensorDescriptor_t)__pyx_v_dstDesc), ((void *)__pyx_v_dstData), ((void *)__pyx_v_reserveSpace), __pyx_v_reserveSpaceSizeInBytes);
      }

      /* "cupy/cuda/cudnn.pyx":1182
 *         size_t dstDesc, size_t dstData,
 *         size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnDropoutForward(
 *             <Handle>handle, <DropoutDescriptor>dropoutDesc,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1188
 *             <TensorDescriptor>dstDesc, <void*>dstData,
 *             <void*>reserveSpace, reserveSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1177
 * 
 * 
 * cpdef dropoutForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t dropoutDesc,
 *         size_t srcDesc, size_t srcData,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_131dropoutForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_130dropoutForward[] = "dropoutForward(size_t handle, size_t dropoutDesc, size_t srcDesc, size_t srcData, size_t dstDesc, size_t dstData, size_t reserveSpace, size_t reserveSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_131dropoutForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_dropoutDesc;
  size_t __pyx_v_srcDesc;
  size_t __pyx_v_srcData;
  size_t __pyx_v_dstDesc;
  size_t __pyx_v_dstData;
  size_t __pyx_v_reserveSpace;
  size_t __pyx_v_reserveSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dropoutForward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_dropoutDesc,&__pyx_n_s_srcDesc,&__pyx_n_s_srcData,&__pyx_n_s_dstDesc,&__pyx_n_s_dstData,&__pyx_n_s_reserveSpace,&__pyx_n_s_reserveSpaceSizeInBytes,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dropoutDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutForward", 1, 8, 8, 1); __PYX_ERR(0, 1177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutForward", 1, 8, 8, 2); __PYX_ERR(0, 1177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_srcData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutForward", 1, 8, 8, 3); __PYX_ERR(0, 1177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dstDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutForward", 1, 8, 8, 4); __PYX_ERR(0, 1177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dstData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutForward", 1, 8, 8, 5); __PYX_ERR(0, 1177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutForward", 1, 8, 8, 6); __PYX_ERR(0, 1177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutForward", 1, 8, 8, 7); __PYX_ERR(0, 1177, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dropoutForward") < 0)) __PYX_ERR(0, 1177, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1178, __pyx_L3_error)
    __pyx_v_dropoutDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_dropoutDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1178, __pyx_L3_error)
    __pyx_v_srcDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_srcDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1179, __pyx_L3_error)
    __pyx_v_srcData = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_srcData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1179, __pyx_L3_error)
    __pyx_v_dstDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dstDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L3_error)
    __pyx_v_dstData = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_dstData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L3_error)
    __pyx_v_reserveSpace = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_reserveSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1181, __pyx_L3_error)
    __pyx_v_reserveSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_reserveSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1181, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dropoutForward", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1177, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_130dropoutForward(__pyx_self, __pyx_v_handle, __pyx_v_dropoutDesc, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_dstDesc, __pyx_v_dstData, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_130dropoutForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_dropoutDesc, size_t __pyx_v_srcDesc, size_t __pyx_v_srcData, size_t __pyx_v_dstDesc, size_t __pyx_v_dstData, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dropoutForward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_dropoutForward(__pyx_v_handle, __pyx_v_dropoutDesc, __pyx_v_srcDesc, __pyx_v_srcData, __pyx_v_dstDesc, __pyx_v_dstData, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1191
 * 
 * 
 * cpdef dropoutBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t dropoutDesc,
 *         size_t dyDesc, size_t dyData,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_133dropoutBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_dropoutBackward(size_t __pyx_v_handle, size_t __pyx_v_dropoutDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_dyData, size_t __pyx_v_dxDesc, size_t __pyx_v_dxData, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dropoutBackward", 0);

  /* "cupy/cuda/cudnn.pyx":1196
 *         size_t dxDesc, size_t dxData,
 *         size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnDropoutBackward(
 *             <Handle>handle, <DropoutDescriptor>dropoutDesc,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1197
 *         size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     with nogil:
 *         status = cudnnDropoutBackward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <DropoutDescriptor>dropoutDesc,
 *             <TensorDescriptor>dyDesc, <void*>dyData,
 */
        __pyx_v_status = cudnnDropoutBackward(((cudnnHandle_t)__pyx_v_handle), ((cudnnDropoutDescriptor_t)__pyx_v_dropoutDesc), ((cudnnTensorDescriptor_t)__pyx_v_dyDesc), ((void *)__pyx_v_dyData), ((cudnnTensorDescriptor_t)__pyx_v_dxDesc), ((void *)__pyx_v_dxData), ((void *)__pyx_v_reserveSpace), __pyx_v_reserveSpaceSizeInBytes);
      }

      /* "cupy/cuda/cudnn.pyx":1196
 *         size_t dxDesc, size_t dxData,
 *         size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnDropoutBackward(
 *             <Handle>handle, <DropoutDescriptor>dropoutDesc,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1202
 *             <TensorDescriptor>dxDesc, <void*>dxData,
 *             <void*>reserveSpace, reserveSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1191
 * 
 * 
 * cpdef dropoutBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t dropoutDesc,
 *         size_t dyDesc, size_t dyData,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_133dropoutBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_132dropoutBackward[] = "dropoutBackward(size_t handle, size_t dropoutDesc, size_t dyDesc, size_t dyData, size_t dxDesc, size_t dxData, size_t reserveSpace, size_t reserveSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_133dropoutBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_dropoutDesc;
  size_t __pyx_v_dyDesc;
  size_t __pyx_v_dyData;
  size_t __pyx_v_dxDesc;
  size_t __pyx_v_dxData;
  size_t __pyx_v_reserveSpace;
  size_t __pyx_v_reserveSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dropoutBackward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_dropoutDesc,&__pyx_n_s_dyDesc,&__pyx_n_s_dyData,&__pyx_n_s_dxDesc,&__pyx_n_s_dxData,&__pyx_n_s_reserveSpace,&__pyx_n_s_reserveSpaceSizeInBytes,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dropoutDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutBackward", 1, 8, 8, 1); __PYX_ERR(0, 1191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutBackward", 1, 8, 8, 2); __PYX_ERR(0, 1191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutBackward", 1, 8, 8, 3); __PYX_ERR(0, 1191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutBackward", 1, 8, 8, 4); __PYX_ERR(0, 1191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dxData)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutBackward", 1, 8, 8, 5); __PYX_ERR(0, 1191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutBackward", 1, 8, 8, 6); __PYX_ERR(0, 1191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dropoutBackward", 1, 8, 8, 7); __PYX_ERR(0, 1191, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dropoutBackward") < 0)) __PYX_ERR(0, 1191, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1192, __pyx_L3_error)
    __pyx_v_dropoutDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_dropoutDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1192, __pyx_L3_error)
    __pyx_v_dyDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_dyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L3_error)
    __pyx_v_dyData = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_dyData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L3_error)
    __pyx_v_dxDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_dxData = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_dxData == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_reserveSpace = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_reserveSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1195, __pyx_L3_error)
    __pyx_v_reserveSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_reserveSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1195, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dropoutBackward", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1191, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_132dropoutBackward(__pyx_self, __pyx_v_handle, __pyx_v_dropoutDesc, __pyx_v_dyDesc, __pyx_v_dyData, __pyx_v_dxDesc, __pyx_v_dxData, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_132dropoutBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_dropoutDesc, size_t __pyx_v_dyDesc, size_t __pyx_v_dyData, size_t __pyx_v_dxDesc, size_t __pyx_v_dxData, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dropoutBackward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_dropoutBackward(__pyx_v_handle, __pyx_v_dropoutDesc, __pyx_v_dyDesc, __pyx_v_dyData, __pyx_v_dxDesc, __pyx_v_dxData, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.dropoutBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1209
 * ###############################################################################
 * 
 * cpdef size_t createRNNDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef RNNDescriptor desc
 *     status = cudnnCreateRNNDescriptor(&desc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_135createRNNDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createRNNDescriptor(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnRNNDescriptor_t __pyx_v_desc;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createRNNDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1211
 * cpdef size_t createRNNDescriptor() except *:
 *     cdef RNNDescriptor desc
 *     status = cudnnCreateRNNDescriptor(&desc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>desc
 */
  __pyx_v_status = cudnnCreateRNNDescriptor((&__pyx_v_desc));

  /* "cupy/cuda/cudnn.pyx":1212
 *     cdef RNNDescriptor desc
 *     status = cudnnCreateRNNDescriptor(&desc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>desc
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1213
 *     status = cudnnCreateRNNDescriptor(&desc)
 *     check_status(status)
 *     return <size_t>desc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_desc);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1209
 * ###############################################################################
 * 
 * cpdef size_t createRNNDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef RNNDescriptor desc
 *     status = cudnnCreateRNNDescriptor(&desc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createRNNDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_135createRNNDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_134createRNNDescriptor[] = "createRNNDescriptor() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_135createRNNDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createRNNDescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_134createRNNDescriptor(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_134createRNNDescriptor(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createRNNDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createRNNDescriptor(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1209, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createRNNDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1216
 * 
 * 
 * cpdef destroyRNNDescriptor(size_t rnnDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyRNNDescriptor(<RNNDescriptor>rnnDesc)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_137destroyRNNDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_rnnDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyRNNDescriptor(size_t __pyx_v_rnnDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyRNNDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1217
 * 
 * cpdef destroyRNNDescriptor(size_t rnnDesc):
 *     status = cudnnDestroyRNNDescriptor(<RNNDescriptor>rnnDesc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cudnnDestroyRNNDescriptor(((cudnnRNNDescriptor_t)__pyx_v_rnnDesc));

  /* "cupy/cuda/cudnn.pyx":1218
 * cpdef destroyRNNDescriptor(size_t rnnDesc):
 *     status = cudnnDestroyRNNDescriptor(<RNNDescriptor>rnnDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1216
 * 
 * 
 * cpdef destroyRNNDescriptor(size_t rnnDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyRNNDescriptor(<RNNDescriptor>rnnDesc)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyRNNDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_137destroyRNNDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_rnnDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_136destroyRNNDescriptor[] = "destroyRNNDescriptor(size_t rnnDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_137destroyRNNDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_rnnDesc) {
  size_t __pyx_v_rnnDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyRNNDescriptor (wrapper)", 0);
  assert(__pyx_arg_rnnDesc); {
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(__pyx_arg_rnnDesc); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1216, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyRNNDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_136destroyRNNDescriptor(__pyx_self, ((size_t)__pyx_v_rnnDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_136destroyRNNDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_rnnDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyRNNDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroyRNNDescriptor(__pyx_v_rnnDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyRNNDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1221
 * 
 * 
 * cpdef size_t createPersistentRNNPlan(size_t rnnDesc, int minibatch,             # <<<<<<<<<<<<<<
 *                                      int dataType) except *:
 *     cdef PersistentRNNPlan plan
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_139createPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createPersistentRNNPlan(size_t __pyx_v_rnnDesc, int __pyx_v_minibatch, int __pyx_v_dataType, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnPersistentRNNPlan_t __pyx_v_plan;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createPersistentRNNPlan", 0);

  /* "cupy/cuda/cudnn.pyx":1224
 *                                      int dataType) except *:
 *     cdef PersistentRNNPlan plan
 *     status = cudnnCreatePersistentRNNPlan(             # <<<<<<<<<<<<<<
 *         <RNNDescriptor>rnnDesc,
 *         <int>minibatch, <DataType>dataType, &plan)
 */
  __pyx_v_status = cudnnCreatePersistentRNNPlan(((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), ((int)__pyx_v_minibatch), ((cudnnDataType_t)__pyx_v_dataType), (&__pyx_v_plan));

  /* "cupy/cuda/cudnn.pyx":1227
 *         <RNNDescriptor>rnnDesc,
 *         <int>minibatch, <DataType>dataType, &plan)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>plan
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1228
 *         <int>minibatch, <DataType>dataType, &plan)
 *     check_status(status)
 *     return <size_t>plan             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_plan);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1221
 * 
 * 
 * cpdef size_t createPersistentRNNPlan(size_t rnnDesc, int minibatch,             # <<<<<<<<<<<<<<
 *                                      int dataType) except *:
 *     cdef PersistentRNNPlan plan
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_139createPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_138createPersistentRNNPlan[] = "createPersistentRNNPlan(size_t rnnDesc, int minibatch, int dataType) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_139createPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_rnnDesc;
  int __pyx_v_minibatch;
  int __pyx_v_dataType;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createPersistentRNNPlan (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rnnDesc,&__pyx_n_s_minibatch,&__pyx_n_s_dataType,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_minibatch)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("createPersistentRNNPlan", 1, 3, 3, 1); __PYX_ERR(0, 1221, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("createPersistentRNNPlan", 1, 3, 3, 2); __PYX_ERR(0, 1221, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "createPersistentRNNPlan") < 0)) __PYX_ERR(0, 1221, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1221, __pyx_L3_error)
    __pyx_v_minibatch = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_minibatch == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1221, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1222, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("createPersistentRNNPlan", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1221, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.createPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_138createPersistentRNNPlan(__pyx_self, __pyx_v_rnnDesc, __pyx_v_minibatch, __pyx_v_dataType);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_138createPersistentRNNPlan(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_rnnDesc, int __pyx_v_minibatch, int __pyx_v_dataType) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createPersistentRNNPlan", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createPersistentRNNPlan(__pyx_v_rnnDesc, __pyx_v_minibatch, __pyx_v_dataType, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1221, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1231
 * 
 * 
 * cpdef setPersistentRNNPlan(size_t rnnDesc, size_t plan):             # <<<<<<<<<<<<<<
 *     status = cudnnSetPersistentRNNPlan(
 *         <RNNDescriptor>rnnDesc, <PersistentRNNPlan>plan)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_141setPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setPersistentRNNPlan(size_t __pyx_v_rnnDesc, size_t __pyx_v_plan, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setPersistentRNNPlan", 0);

  /* "cupy/cuda/cudnn.pyx":1232
 * 
 * cpdef setPersistentRNNPlan(size_t rnnDesc, size_t plan):
 *     status = cudnnSetPersistentRNNPlan(             # <<<<<<<<<<<<<<
 *         <RNNDescriptor>rnnDesc, <PersistentRNNPlan>plan)
 *     check_status(status)
 */
  __pyx_v_status = cudnnSetPersistentRNNPlan(((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), ((cudnnPersistentRNNPlan_t)__pyx_v_plan));

  /* "cupy/cuda/cudnn.pyx":1234
 *     status = cudnnSetPersistentRNNPlan(
 *         <RNNDescriptor>rnnDesc, <PersistentRNNPlan>plan)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1231
 * 
 * 
 * cpdef setPersistentRNNPlan(size_t rnnDesc, size_t plan):             # <<<<<<<<<<<<<<
 *     status = cudnnSetPersistentRNNPlan(
 *         <RNNDescriptor>rnnDesc, <PersistentRNNPlan>plan)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_141setPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_140setPersistentRNNPlan[] = "setPersistentRNNPlan(size_t rnnDesc, size_t plan)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_141setPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_rnnDesc;
  size_t __pyx_v_plan;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setPersistentRNNPlan (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rnnDesc,&__pyx_n_s_plan,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_plan)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPersistentRNNPlan", 1, 2, 2, 1); __PYX_ERR(0, 1231, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setPersistentRNNPlan") < 0)) __PYX_ERR(0, 1231, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1231, __pyx_L3_error)
    __pyx_v_plan = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_plan == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1231, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setPersistentRNNPlan", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1231, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_140setPersistentRNNPlan(__pyx_self, __pyx_v_rnnDesc, __pyx_v_plan);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_140setPersistentRNNPlan(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_rnnDesc, size_t __pyx_v_plan) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setPersistentRNNPlan", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setPersistentRNNPlan(__pyx_v_rnnDesc, __pyx_v_plan, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1237
 * 
 * 
 * cpdef destroyPersistentRNNPlan(size_t plan):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyPersistentRNNPlan(<PersistentRNNPlan>plan)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_143destroyPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_arg_plan); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroyPersistentRNNPlan(size_t __pyx_v_plan, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyPersistentRNNPlan", 0);

  /* "cupy/cuda/cudnn.pyx":1238
 * 
 * cpdef destroyPersistentRNNPlan(size_t plan):
 *     status = cudnnDestroyPersistentRNNPlan(<PersistentRNNPlan>plan)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cudnnDestroyPersistentRNNPlan(((cudnnPersistentRNNPlan_t)__pyx_v_plan));

  /* "cupy/cuda/cudnn.pyx":1239
 * cpdef destroyPersistentRNNPlan(size_t plan):
 *     status = cudnnDestroyPersistentRNNPlan(<PersistentRNNPlan>plan)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1237
 * 
 * 
 * cpdef destroyPersistentRNNPlan(size_t plan):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroyPersistentRNNPlan(<PersistentRNNPlan>plan)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_143destroyPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_arg_plan); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_142destroyPersistentRNNPlan[] = "destroyPersistentRNNPlan(size_t plan)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_143destroyPersistentRNNPlan(PyObject *__pyx_self, PyObject *__pyx_arg_plan) {
  size_t __pyx_v_plan;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyPersistentRNNPlan (wrapper)", 0);
  assert(__pyx_arg_plan); {
    __pyx_v_plan = __Pyx_PyInt_As_size_t(__pyx_arg_plan); if (unlikely((__pyx_v_plan == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1237, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_142destroyPersistentRNNPlan(__pyx_self, ((size_t)__pyx_v_plan));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_142destroyPersistentRNNPlan(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_plan) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyPersistentRNNPlan", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroyPersistentRNNPlan(__pyx_v_plan, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroyPersistentRNNPlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1242
 * 
 * 
 * cpdef setRNNDescriptor_v5(             # <<<<<<<<<<<<<<
 *         size_t rnnDesc, int hiddenSize, int numLayers,
 *         size_t dropoutDesc, int inputMode, int direction, int mode,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_145setRNNDescriptor_v5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setRNNDescriptor_v5(size_t __pyx_v_rnnDesc, int __pyx_v_hiddenSize, int __pyx_v_numLayers, size_t __pyx_v_dropoutDesc, int __pyx_v_inputMode, int __pyx_v_direction, int __pyx_v_mode, int __pyx_v_dataType, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setRNNDescriptor_v5", 0);

  /* "cupy/cuda/cudnn.pyx":1246
 *         size_t dropoutDesc, int inputMode, int direction, int mode,
 *         int dataType):
 *     status = cudnnSetRNNDescriptor_v5(             # <<<<<<<<<<<<<<
 *         <RNNDescriptor>rnnDesc, hiddenSize, numLayers,
 *         <DropoutDescriptor>dropoutDesc, <RNNInputMode>inputMode,
 */
  __pyx_v_status = cudnnSetRNNDescriptor_v5(((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_hiddenSize, __pyx_v_numLayers, ((cudnnDropoutDescriptor_t)__pyx_v_dropoutDesc), ((cudnnRNNInputMode_t)__pyx_v_inputMode), ((cudnnDirectionMode_t)__pyx_v_direction), ((cudnnRNNMode_t)__pyx_v_mode), ((cudnnDataType_t)__pyx_v_dataType));

  /* "cupy/cuda/cudnn.pyx":1250
 *         <DropoutDescriptor>dropoutDesc, <RNNInputMode>inputMode,
 *         <DirectionMode>direction, <RNNMode>mode, <DataType>dataType)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1242
 * 
 * 
 * cpdef setRNNDescriptor_v5(             # <<<<<<<<<<<<<<
 *         size_t rnnDesc, int hiddenSize, int numLayers,
 *         size_t dropoutDesc, int inputMode, int direction, int mode,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setRNNDescriptor_v5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_145setRNNDescriptor_v5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_144setRNNDescriptor_v5[] = "setRNNDescriptor_v5(size_t rnnDesc, int hiddenSize, int numLayers, size_t dropoutDesc, int inputMode, int direction, int mode, int dataType)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_145setRNNDescriptor_v5(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_rnnDesc;
  int __pyx_v_hiddenSize;
  int __pyx_v_numLayers;
  size_t __pyx_v_dropoutDesc;
  int __pyx_v_inputMode;
  int __pyx_v_direction;
  int __pyx_v_mode;
  int __pyx_v_dataType;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setRNNDescriptor_v5 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rnnDesc,&__pyx_n_s_hiddenSize,&__pyx_n_s_numLayers,&__pyx_n_s_dropoutDesc,&__pyx_n_s_inputMode,&__pyx_n_s_direction,&__pyx_n_s_mode,&__pyx_n_s_dataType,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hiddenSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v5", 1, 8, 8, 1); __PYX_ERR(0, 1242, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_numLayers)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v5", 1, 8, 8, 2); __PYX_ERR(0, 1242, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dropoutDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v5", 1, 8, 8, 3); __PYX_ERR(0, 1242, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inputMode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v5", 1, 8, 8, 4); __PYX_ERR(0, 1242, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_direction)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v5", 1, 8, 8, 5); __PYX_ERR(0, 1242, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v5", 1, 8, 8, 6); __PYX_ERR(0, 1242, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v5", 1, 8, 8, 7); __PYX_ERR(0, 1242, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setRNNDescriptor_v5") < 0)) __PYX_ERR(0, 1242, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1243, __pyx_L3_error)
    __pyx_v_hiddenSize = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_hiddenSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1243, __pyx_L3_error)
    __pyx_v_numLayers = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_numLayers == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1243, __pyx_L3_error)
    __pyx_v_dropoutDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_dropoutDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1244, __pyx_L3_error)
    __pyx_v_inputMode = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_inputMode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1244, __pyx_L3_error)
    __pyx_v_direction = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_direction == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1244, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1244, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1245, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v5", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1242, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setRNNDescriptor_v5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_144setRNNDescriptor_v5(__pyx_self, __pyx_v_rnnDesc, __pyx_v_hiddenSize, __pyx_v_numLayers, __pyx_v_dropoutDesc, __pyx_v_inputMode, __pyx_v_direction, __pyx_v_mode, __pyx_v_dataType);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_144setRNNDescriptor_v5(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_rnnDesc, int __pyx_v_hiddenSize, int __pyx_v_numLayers, size_t __pyx_v_dropoutDesc, int __pyx_v_inputMode, int __pyx_v_direction, int __pyx_v_mode, int __pyx_v_dataType) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setRNNDescriptor_v5", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setRNNDescriptor_v5(__pyx_v_rnnDesc, __pyx_v_hiddenSize, __pyx_v_numLayers, __pyx_v_dropoutDesc, __pyx_v_inputMode, __pyx_v_direction, __pyx_v_mode, __pyx_v_dataType, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setRNNDescriptor_v5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1253
 * 
 * 
 * cpdef setRNNDescriptor_v6(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int hiddenSize, int numLayers,
 *         size_t dropoutDesc, int inputMode, int direction, int mode,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_147setRNNDescriptor_v6(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setRNNDescriptor_v6(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_hiddenSize, int __pyx_v_numLayers, size_t __pyx_v_dropoutDesc, int __pyx_v_inputMode, int __pyx_v_direction, int __pyx_v_mode, int __pyx_v_algo, int __pyx_v_dataType, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setRNNDescriptor_v6", 0);

  /* "cupy/cuda/cudnn.pyx":1257
 *         size_t dropoutDesc, int inputMode, int direction, int mode,
 *         int algo, int dataType):
 *     status = cudnnSetRNNDescriptor_v6(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <RNNDescriptor>rnnDesc, hiddenSize, numLayers,
 *         <DropoutDescriptor>dropoutDesc, <RNNInputMode>inputMode,
 */
  __pyx_v_status = cudnnSetRNNDescriptor_v6(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_hiddenSize, __pyx_v_numLayers, ((cudnnDropoutDescriptor_t)__pyx_v_dropoutDesc), ((cudnnRNNInputMode_t)__pyx_v_inputMode), ((cudnnDirectionMode_t)__pyx_v_direction), ((cudnnRNNMode_t)__pyx_v_mode), ((cudnnRNNAlgo_t)__pyx_v_algo), ((cudnnDataType_t)__pyx_v_dataType));

  /* "cupy/cuda/cudnn.pyx":1262
 *         <DirectionMode>direction, <RNNMode>mode, <RNNAlgo>algo,
 *         <DataType>dataType)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1253
 * 
 * 
 * cpdef setRNNDescriptor_v6(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int hiddenSize, int numLayers,
 *         size_t dropoutDesc, int inputMode, int direction, int mode,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setRNNDescriptor_v6", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_147setRNNDescriptor_v6(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_146setRNNDescriptor_v6[] = "setRNNDescriptor_v6(size_t handle, size_t rnnDesc, int hiddenSize, int numLayers, size_t dropoutDesc, int inputMode, int direction, int mode, int algo, int dataType)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_147setRNNDescriptor_v6(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_hiddenSize;
  int __pyx_v_numLayers;
  size_t __pyx_v_dropoutDesc;
  int __pyx_v_inputMode;
  int __pyx_v_direction;
  int __pyx_v_mode;
  int __pyx_v_algo;
  int __pyx_v_dataType;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setRNNDescriptor_v6 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_hiddenSize,&__pyx_n_s_numLayers,&__pyx_n_s_dropoutDesc,&__pyx_n_s_inputMode,&__pyx_n_s_direction,&__pyx_n_s_mode,&__pyx_n_s_algo,&__pyx_n_s_dataType,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 1); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hiddenSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 2); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_numLayers)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 3); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dropoutDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 4); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_inputMode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 5); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_direction)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 6); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mode)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 7); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_algo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 8); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, 9); __PYX_ERR(0, 1253, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setRNNDescriptor_v6") < 0)) __PYX_ERR(0, 1253, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
    __pyx_v_hiddenSize = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_hiddenSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
    __pyx_v_numLayers = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_numLayers == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
    __pyx_v_dropoutDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dropoutDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L3_error)
    __pyx_v_inputMode = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_inputMode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L3_error)
    __pyx_v_direction = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_direction == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L3_error)
    __pyx_v_mode = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_mode == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L3_error)
    __pyx_v_algo = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_algo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1256, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1256, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setRNNDescriptor_v6", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1253, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setRNNDescriptor_v6", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_146setRNNDescriptor_v6(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_hiddenSize, __pyx_v_numLayers, __pyx_v_dropoutDesc, __pyx_v_inputMode, __pyx_v_direction, __pyx_v_mode, __pyx_v_algo, __pyx_v_dataType);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_146setRNNDescriptor_v6(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_hiddenSize, int __pyx_v_numLayers, size_t __pyx_v_dropoutDesc, int __pyx_v_inputMode, int __pyx_v_direction, int __pyx_v_mode, int __pyx_v_algo, int __pyx_v_dataType) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setRNNDescriptor_v6", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setRNNDescriptor_v6(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_hiddenSize, __pyx_v_numLayers, __pyx_v_dropoutDesc, __pyx_v_inputMode, __pyx_v_direction, __pyx_v_mode, __pyx_v_algo, __pyx_v_dataType, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setRNNDescriptor_v6", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1265
 * 
 * 
 * cpdef getRNNWorkspaceSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc):
 *     cdef size_t sizeInBytes
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_149getRNNWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNWorkspaceSize(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_sizeInBytes;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNWorkspaceSize", 0);

  /* "cupy/cuda/cudnn.pyx":1268
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc):
 *     cdef size_t sizeInBytes
 *     status = cudnnGetRNNWorkspaceSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 *         <TensorDescriptor*>xDesc, &sizeInBytes)
 */
  __pyx_v_status = cudnnGetRNNWorkspaceSize(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_seqLength, ((cudnnTensorDescriptor_t *)__pyx_v_xDesc), (&__pyx_v_sizeInBytes));

  /* "cupy/cuda/cudnn.pyx":1271
 *         <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 *         <TensorDescriptor*>xDesc, &sizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return sizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1272
 *         <TensorDescriptor*>xDesc, &sizeInBytes)
 *     check_status(status)
 *     return sizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_sizeInBytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1265
 * 
 * 
 * cpdef getRNNWorkspaceSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc):
 *     cdef size_t sizeInBytes
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_149getRNNWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_148getRNNWorkspaceSize[] = "getRNNWorkspaceSize(size_t handle, size_t rnnDesc, int seqLength, size_t xDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_149getRNNWorkspaceSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_seqLength;
  size_t __pyx_v_xDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRNNWorkspaceSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_seqLength,&__pyx_n_s_xDesc,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNWorkspaceSize", 1, 4, 4, 1); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_seqLength)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNWorkspaceSize", 1, 4, 4, 2); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNWorkspaceSize", 1, 4, 4, 3); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getRNNWorkspaceSize") < 0)) __PYX_ERR(0, 1265, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_seqLength = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_seqLength == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getRNNWorkspaceSize", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1265, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_148getRNNWorkspaceSize(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_148getRNNWorkspaceSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNWorkspaceSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getRNNWorkspaceSize(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNWorkspaceSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1275
 * 
 * 
 * cpdef getRNNTrainingReserveSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc):
 *     cdef size_t sizeInBytes
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_151getRNNTrainingReserveSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNTrainingReserveSize(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_sizeInBytes;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNTrainingReserveSize", 0);

  /* "cupy/cuda/cudnn.pyx":1278
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc):
 *     cdef size_t sizeInBytes
 *     status = cudnnGetRNNTrainingReserveSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 *         <TensorDescriptor*>xDesc, &sizeInBytes)
 */
  __pyx_v_status = cudnnGetRNNTrainingReserveSize(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_seqLength, ((cudnnTensorDescriptor_t *)__pyx_v_xDesc), (&__pyx_v_sizeInBytes));

  /* "cupy/cuda/cudnn.pyx":1281
 *         <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 *         <TensorDescriptor*>xDesc, &sizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return sizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1282
 *         <TensorDescriptor*>xDesc, &sizeInBytes)
 *     check_status(status)
 *     return sizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_sizeInBytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1275
 * 
 * 
 * cpdef getRNNTrainingReserveSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc):
 *     cdef size_t sizeInBytes
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNTrainingReserveSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_151getRNNTrainingReserveSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_150getRNNTrainingReserveSize[] = "getRNNTrainingReserveSize(size_t handle, size_t rnnDesc, int seqLength, size_t xDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_151getRNNTrainingReserveSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_seqLength;
  size_t __pyx_v_xDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRNNTrainingReserveSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_seqLength,&__pyx_n_s_xDesc,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNTrainingReserveSize", 1, 4, 4, 1); __PYX_ERR(0, 1275, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_seqLength)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNTrainingReserveSize", 1, 4, 4, 2); __PYX_ERR(0, 1275, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNTrainingReserveSize", 1, 4, 4, 3); __PYX_ERR(0, 1275, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getRNNTrainingReserveSize") < 0)) __PYX_ERR(0, 1275, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1276, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1276, __pyx_L3_error)
    __pyx_v_seqLength = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_seqLength == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1276, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1276, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getRNNTrainingReserveSize", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1275, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNTrainingReserveSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_150getRNNTrainingReserveSize(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_150getRNNTrainingReserveSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNTrainingReserveSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getRNNTrainingReserveSize(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNTrainingReserveSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1285
 * 
 * 
 * cpdef getRNNParamsSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, size_t xDesc, int dataType):
 *     cdef size_t sizeInBytes
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_153getRNNParamsSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNParamsSize(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, size_t __pyx_v_xDesc, int __pyx_v_dataType, CYTHON_UNUSED int __pyx_skip_dispatch) {
  size_t __pyx_v_sizeInBytes;
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNParamsSize", 0);

  /* "cupy/cuda/cudnn.pyx":1288
 *         size_t handle, size_t rnnDesc, size_t xDesc, int dataType):
 *     cdef size_t sizeInBytes
 *     status = cudnnGetRNNParamsSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <RNNDescriptor>rnnDesc, <TensorDescriptor>xDesc,
 *         &sizeInBytes, <DataType>dataType)
 */
  __pyx_v_status = cudnnGetRNNParamsSize(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), (&__pyx_v_sizeInBytes), ((cudnnDataType_t)__pyx_v_dataType));

  /* "cupy/cuda/cudnn.pyx":1291
 *         <Handle>handle, <RNNDescriptor>rnnDesc, <TensorDescriptor>xDesc,
 *         &sizeInBytes, <DataType>dataType)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return sizeInBytes
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1292
 *         &sizeInBytes, <DataType>dataType)
 *     check_status(status)
 *     return sizeInBytes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_sizeInBytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1285
 * 
 * 
 * cpdef getRNNParamsSize(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, size_t xDesc, int dataType):
 *     cdef size_t sizeInBytes
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNParamsSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_153getRNNParamsSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_152getRNNParamsSize[] = "getRNNParamsSize(size_t handle, size_t rnnDesc, size_t xDesc, int dataType)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_153getRNNParamsSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  size_t __pyx_v_xDesc;
  int __pyx_v_dataType;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRNNParamsSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_xDesc,&__pyx_n_s_dataType,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNParamsSize", 1, 4, 4, 1); __PYX_ERR(0, 1285, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNParamsSize", 1, 4, 4, 2); __PYX_ERR(0, 1285, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNParamsSize", 1, 4, 4, 3); __PYX_ERR(0, 1285, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getRNNParamsSize") < 0)) __PYX_ERR(0, 1285, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1286, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1286, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1286, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1286, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getRNNParamsSize", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1285, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNParamsSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_152getRNNParamsSize(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_xDesc, __pyx_v_dataType);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_152getRNNParamsSize(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, size_t __pyx_v_xDesc, int __pyx_v_dataType) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNParamsSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getRNNParamsSize(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_xDesc, __pyx_v_dataType, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNParamsSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1295
 * 
 * 
 * cpdef getRNNLinLayerMatrixParams(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int layer, size_t xDesc, size_t wDesc,
 *         size_t w, int linLayerID, size_t linLayerMatDesc, size_t linLayerMat):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_155getRNNLinLayerMatrixParams(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNLinLayerMatrixParams(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_layer, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_w, int __pyx_v_linLayerID, size_t __pyx_v_linLayerMatDesc, size_t __pyx_v_linLayerMat, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNLinLayerMatrixParams", 0);

  /* "cupy/cuda/cudnn.pyx":1298
 *         size_t handle, size_t rnnDesc, int layer, size_t xDesc, size_t wDesc,
 *         size_t w, int linLayerID, size_t linLayerMatDesc, size_t linLayerMat):
 *     status = cudnnGetRNNLinLayerMatrixParams(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <RNNDescriptor>rnnDesc, layer,
 *         <TensorDescriptor>xDesc, <FilterDescriptor>wDesc, <void*>w,
 */
  __pyx_v_status = cudnnGetRNNLinLayerMatrixParams(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_layer, ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((void *)__pyx_v_w), __pyx_v_linLayerID, ((cudnnFilterDescriptor_t)__pyx_v_linLayerMatDesc), ((void **)__pyx_v_linLayerMat));

  /* "cupy/cuda/cudnn.pyx":1302
 *         <TensorDescriptor>xDesc, <FilterDescriptor>wDesc, <void*>w,
 *         linLayerID, <FilterDescriptor>linLayerMatDesc, <void**>linLayerMat)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1295
 * 
 * 
 * cpdef getRNNLinLayerMatrixParams(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int layer, size_t xDesc, size_t wDesc,
 *         size_t w, int linLayerID, size_t linLayerMatDesc, size_t linLayerMat):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNLinLayerMatrixParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_155getRNNLinLayerMatrixParams(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_154getRNNLinLayerMatrixParams[] = "getRNNLinLayerMatrixParams(size_t handle, size_t rnnDesc, int layer, size_t xDesc, size_t wDesc, size_t w, int linLayerID, size_t linLayerMatDesc, size_t linLayerMat)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_155getRNNLinLayerMatrixParams(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_layer;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_w;
  int __pyx_v_linLayerID;
  size_t __pyx_v_linLayerMatDesc;
  size_t __pyx_v_linLayerMat;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRNNLinLayerMatrixParams (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_layer,&__pyx_n_s_xDesc,&__pyx_n_s_wDesc,&__pyx_n_s_w,&__pyx_n_s_linLayerID,&__pyx_n_s_linLayerMatDesc,&__pyx_n_s_linLayerMat,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, 1); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_layer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, 2); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, 3); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, 4); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, 5); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_linLayerID)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, 6); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_linLayerMatDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, 7); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_linLayerMat)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, 8); __PYX_ERR(0, 1295, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getRNNLinLayerMatrixParams") < 0)) __PYX_ERR(0, 1295, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_layer = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_layer == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_w == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
    __pyx_v_linLayerID = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_linLayerID == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
    __pyx_v_linLayerMatDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_linLayerMatDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
    __pyx_v_linLayerMat = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_linLayerMat == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getRNNLinLayerMatrixParams", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1295, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNLinLayerMatrixParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_154getRNNLinLayerMatrixParams(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_layer, __pyx_v_xDesc, __pyx_v_wDesc, __pyx_v_w, __pyx_v_linLayerID, __pyx_v_linLayerMatDesc, __pyx_v_linLayerMat);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_154getRNNLinLayerMatrixParams(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_layer, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_w, int __pyx_v_linLayerID, size_t __pyx_v_linLayerMatDesc, size_t __pyx_v_linLayerMat) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNLinLayerMatrixParams", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getRNNLinLayerMatrixParams(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_layer, __pyx_v_xDesc, __pyx_v_wDesc, __pyx_v_w, __pyx_v_linLayerID, __pyx_v_linLayerMatDesc, __pyx_v_linLayerMat, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNLinLayerMatrixParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1305
 * 
 * 
 * cpdef getRNNLinLayerBiasParams(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int layer, size_t xDesc, size_t wDesc,
 *         size_t w, int linLayerID, size_t linLayerBiasDesc,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_157getRNNLinLayerBiasParams(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_getRNNLinLayerBiasParams(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_layer, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_w, int __pyx_v_linLayerID, size_t __pyx_v_linLayerBiasDesc, size_t __pyx_v_linLayerBias, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNLinLayerBiasParams", 0);

  /* "cupy/cuda/cudnn.pyx":1309
 *         size_t w, int linLayerID, size_t linLayerBiasDesc,
 *         size_t linLayerBias):
 *     status = cudnnGetRNNLinLayerBiasParams(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <RNNDescriptor>rnnDesc, layer,
 *         <TensorDescriptor>xDesc, <FilterDescriptor>wDesc, <void*>w,
 */
  __pyx_v_status = cudnnGetRNNLinLayerBiasParams(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_layer, ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((void *)__pyx_v_w), __pyx_v_linLayerID, ((cudnnFilterDescriptor_t)__pyx_v_linLayerBiasDesc), ((void **)__pyx_v_linLayerBias));

  /* "cupy/cuda/cudnn.pyx":1313
 *         <TensorDescriptor>xDesc, <FilterDescriptor>wDesc, <void*>w,
 *         linLayerID, <FilterDescriptor>linLayerBiasDesc, <void**>linLayerBias)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1305
 * 
 * 
 * cpdef getRNNLinLayerBiasParams(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int layer, size_t xDesc, size_t wDesc,
 *         size_t w, int linLayerID, size_t linLayerBiasDesc,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNLinLayerBiasParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_157getRNNLinLayerBiasParams(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_156getRNNLinLayerBiasParams[] = "getRNNLinLayerBiasParams(size_t handle, size_t rnnDesc, int layer, size_t xDesc, size_t wDesc, size_t w, int linLayerID, size_t linLayerBiasDesc, size_t linLayerBias)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_157getRNNLinLayerBiasParams(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_layer;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_w;
  int __pyx_v_linLayerID;
  size_t __pyx_v_linLayerBiasDesc;
  size_t __pyx_v_linLayerBias;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getRNNLinLayerBiasParams (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_layer,&__pyx_n_s_xDesc,&__pyx_n_s_wDesc,&__pyx_n_s_w,&__pyx_n_s_linLayerID,&__pyx_n_s_linLayerBiasDesc,&__pyx_n_s_linLayerBias,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, 1); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_layer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, 2); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, 3); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, 4); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, 5); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_linLayerID)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, 6); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_linLayerBiasDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, 7); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_linLayerBias)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, 8); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getRNNLinLayerBiasParams") < 0)) __PYX_ERR(0, 1305, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1306, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1306, __pyx_L3_error)
    __pyx_v_layer = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_layer == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1306, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1306, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1306, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_w == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1307, __pyx_L3_error)
    __pyx_v_linLayerID = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_linLayerID == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1307, __pyx_L3_error)
    __pyx_v_linLayerBiasDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_linLayerBiasDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1307, __pyx_L3_error)
    __pyx_v_linLayerBias = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_linLayerBias == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1308, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getRNNLinLayerBiasParams", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1305, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNLinLayerBiasParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_156getRNNLinLayerBiasParams(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_layer, __pyx_v_xDesc, __pyx_v_wDesc, __pyx_v_w, __pyx_v_linLayerID, __pyx_v_linLayerBiasDesc, __pyx_v_linLayerBias);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_156getRNNLinLayerBiasParams(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_layer, size_t __pyx_v_xDesc, size_t __pyx_v_wDesc, size_t __pyx_v_w, int __pyx_v_linLayerID, size_t __pyx_v_linLayerBiasDesc, size_t __pyx_v_linLayerBias) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getRNNLinLayerBiasParams", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_getRNNLinLayerBiasParams(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_layer, __pyx_v_xDesc, __pyx_v_wDesc, __pyx_v_w, __pyx_v_linLayerID, __pyx_v_linLayerBiasDesc, __pyx_v_linLayerBias, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.getRNNLinLayerBiasParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1316
 * 
 * 
 * cpdef RNNForwardInference(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc,
 *         size_t x, size_t hxDesc, size_t hx, size_t cxDesc,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_159RNNForwardInference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_RNNForwardInference(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_hyDesc, size_t __pyx_v_hy, size_t __pyx_v_cyDesc, size_t __pyx_v_cy, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("RNNForwardInference", 0);

  /* "cupy/cuda/cudnn.pyx":1322
 *         size_t y, size_t hyDesc, size_t hy, size_t cyDesc,
 *         size_t cy, size_t workspace, size_t workSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnRNNForwardInference(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1323
 *         size_t cy, size_t workspace, size_t workSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnRNNForwardInference(
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1324
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnRNNForwardInference(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 *             <TensorDescriptor*>xDesc, <void*>x,
 */
        __pyx_v_status = cudnnRNNForwardInference(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_seqLength, ((cudnnTensorDescriptor_t *)__pyx_v_xDesc), ((void *)__pyx_v_x), ((cudnnTensorDescriptor_t)__pyx_v_hxDesc), ((void *)__pyx_v_hx), ((cudnnTensorDescriptor_t)__pyx_v_cxDesc), ((void *)__pyx_v_cx), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((void *)__pyx_v_w), ((cudnnTensorDescriptor_t *)__pyx_v_yDesc), ((void *)__pyx_v_y), ((cudnnTensorDescriptor_t)__pyx_v_hyDesc), ((void *)__pyx_v_hy), ((cudnnTensorDescriptor_t)__pyx_v_cyDesc), ((void *)__pyx_v_cy), ((void *)__pyx_v_workspace), __pyx_v_workSpaceSizeInBytes);
      }

      /* "cupy/cuda/cudnn.pyx":1323
 *         size_t cy, size_t workspace, size_t workSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnRNNForwardInference(
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1334
 *             <TensorDescriptor>cyDesc, <void*>cy,
 *             <void*>workspace, workSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1316
 * 
 * 
 * cpdef RNNForwardInference(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc,
 *         size_t x, size_t hxDesc, size_t hx, size_t cxDesc,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNForwardInference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_159RNNForwardInference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_158RNNForwardInference[] = "RNNForwardInference(size_t handle, size_t rnnDesc, int seqLength, size_t xDesc, size_t x, size_t hxDesc, size_t hx, size_t cxDesc, size_t cx, size_t wDesc, size_t w, size_t yDesc, size_t y, size_t hyDesc, size_t hy, size_t cyDesc, size_t cy, size_t workspace, size_t workSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_159RNNForwardInference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_seqLength;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_hxDesc;
  size_t __pyx_v_hx;
  size_t __pyx_v_cxDesc;
  size_t __pyx_v_cx;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_w;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  size_t __pyx_v_hyDesc;
  size_t __pyx_v_hy;
  size_t __pyx_v_cyDesc;
  size_t __pyx_v_cy;
  size_t __pyx_v_workspace;
  size_t __pyx_v_workSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RNNForwardInference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_seqLength,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_hxDesc,&__pyx_n_s_hx,&__pyx_n_s_cxDesc,&__pyx_n_s_cx,&__pyx_n_s_wDesc,&__pyx_n_s_w,&__pyx_n_s_yDesc,&__pyx_n_s_y,&__pyx_n_s_hyDesc,&__pyx_n_s_hy,&__pyx_n_s_cyDesc,&__pyx_n_s_cy,&__pyx_n_s_workspace,&__pyx_n_s_workSpaceSizeInBytes,0};
    PyObject* values[19] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 1); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_seqLength)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 2); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 3); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 4); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 5); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 6); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 7); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 8); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 9); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 10); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 11); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 12); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 13); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 14); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 15); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 16); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workspace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 17); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, 18); __PYX_ERR(0, 1316, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "RNNForwardInference") < 0)) __PYX_ERR(0, 1316, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 19) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1317, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1317, __pyx_L3_error)
    __pyx_v_seqLength = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_seqLength == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1317, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1317, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1318, __pyx_L3_error)
    __pyx_v_hxDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_hxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1318, __pyx_L3_error)
    __pyx_v_hx = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_hx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1318, __pyx_L3_error)
    __pyx_v_cxDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_cxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1318, __pyx_L3_error)
    __pyx_v_cx = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_cx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1319, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1319, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_w == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1319, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1319, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
    __pyx_v_hyDesc = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_hyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
    __pyx_v_hy = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_hy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
    __pyx_v_cyDesc = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_cyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L3_error)
    __pyx_v_cy = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_cy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
    __pyx_v_workspace = __Pyx_PyInt_As_size_t(values[17]); if (unlikely((__pyx_v_workspace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[18]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("RNNForwardInference", 1, 19, 19, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1316, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNForwardInference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_158RNNForwardInference(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc, __pyx_v_x, __pyx_v_hxDesc, __pyx_v_hx, __pyx_v_cxDesc, __pyx_v_cx, __pyx_v_wDesc, __pyx_v_w, __pyx_v_yDesc, __pyx_v_y, __pyx_v_hyDesc, __pyx_v_hy, __pyx_v_cyDesc, __pyx_v_cy, __pyx_v_workspace, __pyx_v_workSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_158RNNForwardInference(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_hyDesc, size_t __pyx_v_hy, size_t __pyx_v_cyDesc, size_t __pyx_v_cy, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("RNNForwardInference", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_RNNForwardInference(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc, __pyx_v_x, __pyx_v_hxDesc, __pyx_v_hx, __pyx_v_cxDesc, __pyx_v_cx, __pyx_v_wDesc, __pyx_v_w, __pyx_v_yDesc, __pyx_v_y, __pyx_v_hyDesc, __pyx_v_hy, __pyx_v_cyDesc, __pyx_v_cy, __pyx_v_workspace, __pyx_v_workSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNForwardInference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1337
 * 
 * 
 * cpdef RNNForwardTraining(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc, size_t x,
 *         size_t hxDesc, size_t hx, size_t cxDesc, size_t cx,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_161RNNForwardTraining(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_RNNForwardTraining(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_hyDesc, size_t __pyx_v_hy, size_t __pyx_v_cyDesc, size_t __pyx_v_cy, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("RNNForwardTraining", 0);

  /* "cupy/cuda/cudnn.pyx":1344
 *         size_t workspace, size_t workSpaceSizeInBytes, size_t reserveSpace,
 *         size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnRNNForwardTraining(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1345
 *         size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnRNNForwardTraining(
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1346
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnRNNForwardTraining(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 *             <TensorDescriptor*>xDesc, <void*>x,
 */
        __pyx_v_status = cudnnRNNForwardTraining(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_seqLength, ((cudnnTensorDescriptor_t *)__pyx_v_xDesc), ((void *)__pyx_v_x), ((cudnnTensorDescriptor_t)__pyx_v_hxDesc), ((void *)__pyx_v_hx), ((cudnnTensorDescriptor_t)__pyx_v_cxDesc), ((void *)__pyx_v_cx), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((void *)__pyx_v_w), ((cudnnTensorDescriptor_t *)__pyx_v_yDesc), ((void *)__pyx_v_y), ((cudnnTensorDescriptor_t)__pyx_v_hyDesc), ((void *)__pyx_v_hy), ((cudnnTensorDescriptor_t)__pyx_v_cyDesc), ((void *)__pyx_v_cy), ((void *)__pyx_v_workspace), __pyx_v_workSpaceSizeInBytes, ((void *)__pyx_v_reserveSpace), __pyx_v_reserveSpaceSizeInBytes);
      }

      /* "cupy/cuda/cudnn.pyx":1345
 *         size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnRNNForwardTraining(
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1357
 *             <void*>workspace, workSpaceSizeInBytes,
 *             <void*>reserveSpace, reserveSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1337
 * 
 * 
 * cpdef RNNForwardTraining(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc, size_t x,
 *         size_t hxDesc, size_t hx, size_t cxDesc, size_t cx,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNForwardTraining", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_161RNNForwardTraining(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_160RNNForwardTraining[] = "RNNForwardTraining(size_t handle, size_t rnnDesc, int seqLength, size_t xDesc, size_t x, size_t hxDesc, size_t hx, size_t cxDesc, size_t cx, size_t wDesc, size_t w, size_t yDesc, size_t y, size_t hyDesc, size_t hy, size_t cyDesc, size_t cy, size_t workspace, size_t workSpaceSizeInBytes, size_t reserveSpace, size_t reserveSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_161RNNForwardTraining(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_seqLength;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_hxDesc;
  size_t __pyx_v_hx;
  size_t __pyx_v_cxDesc;
  size_t __pyx_v_cx;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_w;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  size_t __pyx_v_hyDesc;
  size_t __pyx_v_hy;
  size_t __pyx_v_cyDesc;
  size_t __pyx_v_cy;
  size_t __pyx_v_workspace;
  size_t __pyx_v_workSpaceSizeInBytes;
  size_t __pyx_v_reserveSpace;
  size_t __pyx_v_reserveSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RNNForwardTraining (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_seqLength,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_hxDesc,&__pyx_n_s_hx,&__pyx_n_s_cxDesc,&__pyx_n_s_cx,&__pyx_n_s_wDesc,&__pyx_n_s_w,&__pyx_n_s_yDesc,&__pyx_n_s_y,&__pyx_n_s_hyDesc,&__pyx_n_s_hy,&__pyx_n_s_cyDesc,&__pyx_n_s_cy,&__pyx_n_s_workspace,&__pyx_n_s_workSpaceSizeInBytes,&__pyx_n_s_reserveSpace,&__pyx_n_s_reserveSpaceSizeInBytes,0};
    PyObject* values[21] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 1); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_seqLength)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 2); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 3); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 4); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 5); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 6); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 7); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 8); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 9); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 10); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 11); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 12); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 13); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 14); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 15); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 16); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workspace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 17); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 18); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 19); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 20:
        if (likely((values[20] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, 20); __PYX_ERR(0, 1337, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "RNNForwardTraining") < 0)) __PYX_ERR(0, 1337, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 21) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
      values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_seqLength = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_seqLength == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1338, __pyx_L3_error)
    __pyx_v_hxDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_hxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1339, __pyx_L3_error)
    __pyx_v_hx = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_hx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1339, __pyx_L3_error)
    __pyx_v_cxDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_cxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1339, __pyx_L3_error)
    __pyx_v_cx = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_cx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1339, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1340, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_w == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1340, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1340, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1340, __pyx_L3_error)
    __pyx_v_hyDesc = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_hyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1341, __pyx_L3_error)
    __pyx_v_hy = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_hy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1341, __pyx_L3_error)
    __pyx_v_cyDesc = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_cyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1341, __pyx_L3_error)
    __pyx_v_cy = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_cy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1341, __pyx_L3_error)
    __pyx_v_workspace = __Pyx_PyInt_As_size_t(values[17]); if (unlikely((__pyx_v_workspace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1342, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[18]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1342, __pyx_L3_error)
    __pyx_v_reserveSpace = __Pyx_PyInt_As_size_t(values[19]); if (unlikely((__pyx_v_reserveSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1342, __pyx_L3_error)
    __pyx_v_reserveSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[20]); if (unlikely((__pyx_v_reserveSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1343, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("RNNForwardTraining", 1, 21, 21, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1337, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNForwardTraining", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_160RNNForwardTraining(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc, __pyx_v_x, __pyx_v_hxDesc, __pyx_v_hx, __pyx_v_cxDesc, __pyx_v_cx, __pyx_v_wDesc, __pyx_v_w, __pyx_v_yDesc, __pyx_v_y, __pyx_v_hyDesc, __pyx_v_hy, __pyx_v_cyDesc, __pyx_v_cy, __pyx_v_workspace, __pyx_v_workSpaceSizeInBytes, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_160RNNForwardTraining(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_hyDesc, size_t __pyx_v_hy, size_t __pyx_v_cyDesc, size_t __pyx_v_cy, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("RNNForwardTraining", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_RNNForwardTraining(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc, __pyx_v_x, __pyx_v_hxDesc, __pyx_v_hx, __pyx_v_cxDesc, __pyx_v_cx, __pyx_v_wDesc, __pyx_v_w, __pyx_v_yDesc, __pyx_v_y, __pyx_v_hyDesc, __pyx_v_hy, __pyx_v_cyDesc, __pyx_v_cy, __pyx_v_workspace, __pyx_v_workSpaceSizeInBytes, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNForwardTraining", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1360
 * 
 * 
 * cpdef RNNBackwardData(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t yDesc, size_t y,
 *         size_t dyDesc, size_t dy, size_t dhyDesc, size_t dhy,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_163RNNBackwardData(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_RNNBackwardData(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_dhyDesc, size_t __pyx_v_dhy, size_t __pyx_v_dcyDesc, size_t __pyx_v_dcy, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_dhxDesc, size_t __pyx_v_dhx, size_t __pyx_v_dcxDesc, size_t __pyx_v_dcx, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("RNNBackwardData", 0);

  /* "cupy/cuda/cudnn.pyx":1369
 *         size_t workSpaceSizeInBytes, size_t reserveSpace,
 *         size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnRNNBackwardData(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1370
 *         size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnRNNBackwardData(
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1371
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnRNNBackwardData(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 *             <TensorDescriptor*>yDesc, <void*>y,
 */
        __pyx_v_status = cudnnRNNBackwardData(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_seqLength, ((cudnnTensorDescriptor_t *)__pyx_v_yDesc), ((void *)__pyx_v_y), ((cudnnTensorDescriptor_t *)__pyx_v_dyDesc), ((void *)__pyx_v_dy), ((cudnnTensorDescriptor_t)__pyx_v_dhyDesc), ((void *)__pyx_v_dhy), ((cudnnTensorDescriptor_t)__pyx_v_dcyDesc), ((void *)__pyx_v_dcy), ((cudnnFilterDescriptor_t)__pyx_v_wDesc), ((void *)__pyx_v_w), ((cudnnTensorDescriptor_t)__pyx_v_hxDesc), ((void *)__pyx_v_hx), ((cudnnTensorDescriptor_t)__pyx_v_cxDesc), ((void *)__pyx_v_cx), ((cudnnTensorDescriptor_t *)__pyx_v_dxDesc), ((void *)__pyx_v_dx), ((cudnnTensorDescriptor_t)__pyx_v_dhxDesc), ((void *)__pyx_v_dhx), ((cudnnTensorDescriptor_t)__pyx_v_dcxDesc), ((void *)__pyx_v_dcx), ((void *)__pyx_v_workspace), __pyx_v_workSpaceSizeInBytes, ((void *)__pyx_v_reserveSpace), __pyx_v_reserveSpaceSizeInBytes);
      }

      /* "cupy/cuda/cudnn.pyx":1370
 *         size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnRNNBackwardData(
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1385
 *             <void*>workspace, workSpaceSizeInBytes,
 *             <void*>reserveSpace, reserveSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1360
 * 
 * 
 * cpdef RNNBackwardData(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t yDesc, size_t y,
 *         size_t dyDesc, size_t dy, size_t dhyDesc, size_t dhy,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNBackwardData", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_163RNNBackwardData(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_162RNNBackwardData[] = "RNNBackwardData(size_t handle, size_t rnnDesc, int seqLength, size_t yDesc, size_t y, size_t dyDesc, size_t dy, size_t dhyDesc, size_t dhy, size_t dcyDesc, size_t dcy, size_t wDesc, size_t w, size_t hxDesc, size_t hx, size_t cxDesc, size_t cx, size_t dxDesc, size_t dx, size_t dhxDesc, size_t dhx, size_t dcxDesc, size_t dcx, size_t workspace, size_t workSpaceSizeInBytes, size_t reserveSpace, size_t reserveSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_163RNNBackwardData(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_seqLength;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  size_t __pyx_v_dyDesc;
  size_t __pyx_v_dy;
  size_t __pyx_v_dhyDesc;
  size_t __pyx_v_dhy;
  size_t __pyx_v_dcyDesc;
  size_t __pyx_v_dcy;
  size_t __pyx_v_wDesc;
  size_t __pyx_v_w;
  size_t __pyx_v_hxDesc;
  size_t __pyx_v_hx;
  size_t __pyx_v_cxDesc;
  size_t __pyx_v_cx;
  size_t __pyx_v_dxDesc;
  size_t __pyx_v_dx;
  size_t __pyx_v_dhxDesc;
  size_t __pyx_v_dhx;
  size_t __pyx_v_dcxDesc;
  size_t __pyx_v_dcx;
  size_t __pyx_v_workspace;
  size_t __pyx_v_workSpaceSizeInBytes;
  size_t __pyx_v_reserveSpace;
  size_t __pyx_v_reserveSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RNNBackwardData (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_seqLength,&__pyx_n_s_yDesc,&__pyx_n_s_y,&__pyx_n_s_dyDesc,&__pyx_n_s_dy,&__pyx_n_s_dhyDesc,&__pyx_n_s_dhy,&__pyx_n_s_dcyDesc,&__pyx_n_s_dcy,&__pyx_n_s_wDesc,&__pyx_n_s_w,&__pyx_n_s_hxDesc,&__pyx_n_s_hx,&__pyx_n_s_cxDesc,&__pyx_n_s_cx,&__pyx_n_s_dxDesc,&__pyx_n_s_dx,&__pyx_n_s_dhxDesc,&__pyx_n_s_dhx,&__pyx_n_s_dcxDesc,&__pyx_n_s_dcx,&__pyx_n_s_workspace,&__pyx_n_s_workSpaceSizeInBytes,&__pyx_n_s_reserveSpace,&__pyx_n_s_reserveSpaceSizeInBytes,0};
    PyObject* values[27] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 27: values[26] = PyTuple_GET_ITEM(__pyx_args, 26);
        CYTHON_FALLTHROUGH;
        case 26: values[25] = PyTuple_GET_ITEM(__pyx_args, 25);
        CYTHON_FALLTHROUGH;
        case 25: values[24] = PyTuple_GET_ITEM(__pyx_args, 24);
        CYTHON_FALLTHROUGH;
        case 24: values[23] = PyTuple_GET_ITEM(__pyx_args, 23);
        CYTHON_FALLTHROUGH;
        case 23: values[22] = PyTuple_GET_ITEM(__pyx_args, 22);
        CYTHON_FALLTHROUGH;
        case 22: values[21] = PyTuple_GET_ITEM(__pyx_args, 21);
        CYTHON_FALLTHROUGH;
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 1); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_seqLength)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 2); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 3); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 4); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 5); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 6); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dhyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 7); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dhy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 8); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dcyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 9); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dcy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 10); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_wDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 11); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_w)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 12); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 13); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 14); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 15); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_cx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 16); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 17); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 18); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dhxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 19); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 20:
        if (likely((values[20] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dhx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 20); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 21:
        if (likely((values[21] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dcxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 21); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 22:
        if (likely((values[22] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dcx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 22); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 23:
        if (likely((values[23] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workspace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 23); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 24:
        if (likely((values[24] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 24); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 25:
        if (likely((values[25] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 25); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 26:
        if (likely((values[26] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, 26); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "RNNBackwardData") < 0)) __PYX_ERR(0, 1360, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 27) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
      values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
      values[21] = PyTuple_GET_ITEM(__pyx_args, 21);
      values[22] = PyTuple_GET_ITEM(__pyx_args, 22);
      values[23] = PyTuple_GET_ITEM(__pyx_args, 23);
      values[24] = PyTuple_GET_ITEM(__pyx_args, 24);
      values[25] = PyTuple_GET_ITEM(__pyx_args, 25);
      values[26] = PyTuple_GET_ITEM(__pyx_args, 26);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_seqLength = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_seqLength == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_dyDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_dyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_dy = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_dy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_dhyDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_dhyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_dhy = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_dhy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_dcyDesc = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_dcyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1363, __pyx_L3_error)
    __pyx_v_dcy = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_dcy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1363, __pyx_L3_error)
    __pyx_v_wDesc = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_wDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1363, __pyx_L3_error)
    __pyx_v_w = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_w == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1363, __pyx_L3_error)
    __pyx_v_hxDesc = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_hxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L3_error)
    __pyx_v_hx = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_hx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L3_error)
    __pyx_v_cxDesc = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_cxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L3_error)
    __pyx_v_cx = __Pyx_PyInt_As_size_t(values[16]); if (unlikely((__pyx_v_cx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L3_error)
    __pyx_v_dxDesc = __Pyx_PyInt_As_size_t(values[17]); if (unlikely((__pyx_v_dxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1365, __pyx_L3_error)
    __pyx_v_dx = __Pyx_PyInt_As_size_t(values[18]); if (unlikely((__pyx_v_dx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1365, __pyx_L3_error)
    __pyx_v_dhxDesc = __Pyx_PyInt_As_size_t(values[19]); if (unlikely((__pyx_v_dhxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1365, __pyx_L3_error)
    __pyx_v_dhx = __Pyx_PyInt_As_size_t(values[20]); if (unlikely((__pyx_v_dhx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1365, __pyx_L3_error)
    __pyx_v_dcxDesc = __Pyx_PyInt_As_size_t(values[21]); if (unlikely((__pyx_v_dcxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1366, __pyx_L3_error)
    __pyx_v_dcx = __Pyx_PyInt_As_size_t(values[22]); if (unlikely((__pyx_v_dcx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1366, __pyx_L3_error)
    __pyx_v_workspace = __Pyx_PyInt_As_size_t(values[23]); if (unlikely((__pyx_v_workspace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1366, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[24]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1367, __pyx_L3_error)
    __pyx_v_reserveSpace = __Pyx_PyInt_As_size_t(values[25]); if (unlikely((__pyx_v_reserveSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1367, __pyx_L3_error)
    __pyx_v_reserveSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[26]); if (unlikely((__pyx_v_reserveSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1368, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("RNNBackwardData", 1, 27, 27, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1360, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNBackwardData", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_162RNNBackwardData(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_yDesc, __pyx_v_y, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_dhyDesc, __pyx_v_dhy, __pyx_v_dcyDesc, __pyx_v_dcy, __pyx_v_wDesc, __pyx_v_w, __pyx_v_hxDesc, __pyx_v_hx, __pyx_v_cxDesc, __pyx_v_cx, __pyx_v_dxDesc, __pyx_v_dx, __pyx_v_dhxDesc, __pyx_v_dhx, __pyx_v_dcxDesc, __pyx_v_dcx, __pyx_v_workspace, __pyx_v_workSpaceSizeInBytes, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_162RNNBackwardData(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_dhyDesc, size_t __pyx_v_dhy, size_t __pyx_v_dcyDesc, size_t __pyx_v_dcy, size_t __pyx_v_wDesc, size_t __pyx_v_w, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_cxDesc, size_t __pyx_v_cx, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_dhxDesc, size_t __pyx_v_dhx, size_t __pyx_v_dcxDesc, size_t __pyx_v_dcx, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("RNNBackwardData", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_RNNBackwardData(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_yDesc, __pyx_v_y, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_dhyDesc, __pyx_v_dhy, __pyx_v_dcyDesc, __pyx_v_dcy, __pyx_v_wDesc, __pyx_v_w, __pyx_v_hxDesc, __pyx_v_hx, __pyx_v_cxDesc, __pyx_v_cx, __pyx_v_dxDesc, __pyx_v_dx, __pyx_v_dhxDesc, __pyx_v_dhx, __pyx_v_dcxDesc, __pyx_v_dcx, __pyx_v_workspace, __pyx_v_workSpaceSizeInBytes, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNBackwardData", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1388
 * 
 * 
 * cpdef RNNBackwardWeights(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc, size_t x,
 *         size_t hxDesc, size_t hx, size_t yDesc, size_t y,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_165RNNBackwardWeights(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_RNNBackwardWeights(size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_dwDesc, size_t __pyx_v_dw, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("RNNBackwardWeights", 0);

  /* "cupy/cuda/cudnn.pyx":1393
 *         size_t workspace, size_t workSpaceSizeInBytes, size_t dwDesc,
 *         size_t dw, size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnRNNBackwardWeights(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1394
 *         size_t dw, size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnRNNBackwardWeights(
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1395
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnRNNBackwardWeights(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 *             <TensorDescriptor*>xDesc, <void*>x,
 */
        __pyx_v_status = cudnnRNNBackwardWeights(((cudnnHandle_t)__pyx_v_handle), ((cudnnRNNDescriptor_t)__pyx_v_rnnDesc), __pyx_v_seqLength, ((cudnnTensorDescriptor_t *)__pyx_v_xDesc), ((void *)__pyx_v_x), ((cudnnTensorDescriptor_t)__pyx_v_hxDesc), ((void *)__pyx_v_hx), ((cudnnTensorDescriptor_t *)__pyx_v_yDesc), ((void *)__pyx_v_y), ((void *)__pyx_v_workspace), __pyx_v_workSpaceSizeInBytes, ((cudnnFilterDescriptor_t)__pyx_v_dwDesc), ((void *)__pyx_v_dw), ((void *)__pyx_v_reserveSpace), __pyx_v_reserveSpaceSizeInBytes);
      }

      /* "cupy/cuda/cudnn.pyx":1394
 *         size_t dw, size_t reserveSpace, size_t reserveSpaceSizeInBytes):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnRNNBackwardWeights(
 *             <Handle>handle, <RNNDescriptor>rnnDesc, seqLength,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1403
 *             <FilterDescriptor>dwDesc, <void*>dw,
 *             <void*>reserveSpace, reserveSpaceSizeInBytes)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1388
 * 
 * 
 * cpdef RNNBackwardWeights(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t rnnDesc, int seqLength, size_t xDesc, size_t x,
 *         size_t hxDesc, size_t hx, size_t yDesc, size_t y,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNBackwardWeights", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_165RNNBackwardWeights(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_164RNNBackwardWeights[] = "RNNBackwardWeights(size_t handle, size_t rnnDesc, int seqLength, size_t xDesc, size_t x, size_t hxDesc, size_t hx, size_t yDesc, size_t y, size_t workspace, size_t workSpaceSizeInBytes, size_t dwDesc, size_t dw, size_t reserveSpace, size_t reserveSpaceSizeInBytes)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_165RNNBackwardWeights(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_rnnDesc;
  int __pyx_v_seqLength;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_hxDesc;
  size_t __pyx_v_hx;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  size_t __pyx_v_workspace;
  size_t __pyx_v_workSpaceSizeInBytes;
  size_t __pyx_v_dwDesc;
  size_t __pyx_v_dw;
  size_t __pyx_v_reserveSpace;
  size_t __pyx_v_reserveSpaceSizeInBytes;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("RNNBackwardWeights (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_rnnDesc,&__pyx_n_s_seqLength,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_hxDesc,&__pyx_n_s_hx,&__pyx_n_s_yDesc,&__pyx_n_s_y,&__pyx_n_s_workspace,&__pyx_n_s_workSpaceSizeInBytes,&__pyx_n_s_dwDesc,&__pyx_n_s_dw,&__pyx_n_s_reserveSpace,&__pyx_n_s_reserveSpaceSizeInBytes,0};
    PyObject* values[15] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rnnDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 1); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_seqLength)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 2); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 3); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 4); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 5); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_hx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 6); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 7); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 8); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workspace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 9); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_workSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 10); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dwDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 11); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dw)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 12); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpace)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 13); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reserveSpaceSizeInBytes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, 14); __PYX_ERR(0, 1388, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "RNNBackwardWeights") < 0)) __PYX_ERR(0, 1388, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 15) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1389, __pyx_L3_error)
    __pyx_v_rnnDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_rnnDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1389, __pyx_L3_error)
    __pyx_v_seqLength = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_seqLength == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1389, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1389, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1389, __pyx_L3_error)
    __pyx_v_hxDesc = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_hxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1390, __pyx_L3_error)
    __pyx_v_hx = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_hx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1390, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1390, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1390, __pyx_L3_error)
    __pyx_v_workspace = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_workspace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L3_error)
    __pyx_v_workSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_workSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L3_error)
    __pyx_v_dwDesc = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_dwDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L3_error)
    __pyx_v_dw = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_dw == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L3_error)
    __pyx_v_reserveSpace = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_reserveSpace == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L3_error)
    __pyx_v_reserveSpaceSizeInBytes = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_reserveSpaceSizeInBytes == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("RNNBackwardWeights", 1, 15, 15, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1388, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNBackwardWeights", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_164RNNBackwardWeights(__pyx_self, __pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc, __pyx_v_x, __pyx_v_hxDesc, __pyx_v_hx, __pyx_v_yDesc, __pyx_v_y, __pyx_v_workspace, __pyx_v_workSpaceSizeInBytes, __pyx_v_dwDesc, __pyx_v_dw, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_164RNNBackwardWeights(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_rnnDesc, int __pyx_v_seqLength, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_hxDesc, size_t __pyx_v_hx, size_t __pyx_v_yDesc, size_t __pyx_v_y, size_t __pyx_v_workspace, size_t __pyx_v_workSpaceSizeInBytes, size_t __pyx_v_dwDesc, size_t __pyx_v_dw, size_t __pyx_v_reserveSpace, size_t __pyx_v_reserveSpaceSizeInBytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("RNNBackwardWeights", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_RNNBackwardWeights(__pyx_v_handle, __pyx_v_rnnDesc, __pyx_v_seqLength, __pyx_v_xDesc, __pyx_v_x, __pyx_v_hxDesc, __pyx_v_hx, __pyx_v_yDesc, __pyx_v_y, __pyx_v_workspace, __pyx_v_workSpaceSizeInBytes, __pyx_v_dwDesc, __pyx_v_dw, __pyx_v_reserveSpace, __pyx_v_reserveSpaceSizeInBytes, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.RNNBackwardWeights", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1410
 * ###############################################################################
 * 
 * cpdef size_t createSpatialTransformerDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef SpatialTransformerDescriptor stDesc
 *     status = cudnnCreateSpatialTransformerDescriptor(&stDesc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_167createSpatialTransformerDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static size_t __pyx_f_4cupy_4cuda_5cudnn_createSpatialTransformerDescriptor(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudnnSpatialTransformerDescriptor_t __pyx_v_stDesc;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createSpatialTransformerDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1412
 * cpdef size_t createSpatialTransformerDescriptor() except *:
 *     cdef SpatialTransformerDescriptor stDesc
 *     status = cudnnCreateSpatialTransformerDescriptor(&stDesc)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>stDesc
 */
  __pyx_v_status = cudnnCreateSpatialTransformerDescriptor((&__pyx_v_stDesc));

  /* "cupy/cuda/cudnn.pyx":1413
 *     cdef SpatialTransformerDescriptor stDesc
 *     status = cudnnCreateSpatialTransformerDescriptor(&stDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>stDesc
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1414
 *     status = cudnnCreateSpatialTransformerDescriptor(&stDesc)
 *     check_status(status)
 *     return <size_t>stDesc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_stDesc);
  goto __pyx_L0;

  /* "cupy/cuda/cudnn.pyx":1410
 * ###############################################################################
 * 
 * cpdef size_t createSpatialTransformerDescriptor() except *:             # <<<<<<<<<<<<<<
 *     cdef SpatialTransformerDescriptor stDesc
 *     status = cudnnCreateSpatialTransformerDescriptor(&stDesc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createSpatialTransformerDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_167createSpatialTransformerDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_166createSpatialTransformerDescriptor[] = "createSpatialTransformerDescriptor() -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_167createSpatialTransformerDescriptor(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createSpatialTransformerDescriptor (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_166createSpatialTransformerDescriptor(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_166createSpatialTransformerDescriptor(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createSpatialTransformerDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_createSpatialTransformerDescriptor(0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1410, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cudnn.createSpatialTransformerDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1417
 * 
 * 
 * cpdef destroySpatialTransformerDescriptor(size_t stDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroySpatialTransformerDescriptor(
 *         <SpatialTransformerDescriptor>stDesc)
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_169destroySpatialTransformerDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_stDesc); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_destroySpatialTransformerDescriptor(size_t __pyx_v_stDesc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroySpatialTransformerDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1418
 * 
 * cpdef destroySpatialTransformerDescriptor(size_t stDesc):
 *     status = cudnnDestroySpatialTransformerDescriptor(             # <<<<<<<<<<<<<<
 *         <SpatialTransformerDescriptor>stDesc)
 *     check_status(status)
 */
  __pyx_v_status = cudnnDestroySpatialTransformerDescriptor(((cudnnSpatialTransformerDescriptor_t)__pyx_v_stDesc));

  /* "cupy/cuda/cudnn.pyx":1420
 *     status = cudnnDestroySpatialTransformerDescriptor(
 *         <SpatialTransformerDescriptor>stDesc)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1417
 * 
 * 
 * cpdef destroySpatialTransformerDescriptor(size_t stDesc):             # <<<<<<<<<<<<<<
 *     status = cudnnDestroySpatialTransformerDescriptor(
 *         <SpatialTransformerDescriptor>stDesc)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroySpatialTransformerDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_169destroySpatialTransformerDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_stDesc); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_168destroySpatialTransformerDescriptor[] = "destroySpatialTransformerDescriptor(size_t stDesc)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_169destroySpatialTransformerDescriptor(PyObject *__pyx_self, PyObject *__pyx_arg_stDesc) {
  size_t __pyx_v_stDesc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroySpatialTransformerDescriptor (wrapper)", 0);
  assert(__pyx_arg_stDesc); {
    __pyx_v_stDesc = __Pyx_PyInt_As_size_t(__pyx_arg_stDesc); if (unlikely((__pyx_v_stDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1417, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroySpatialTransformerDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_168destroySpatialTransformerDescriptor(__pyx_self, ((size_t)__pyx_v_stDesc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_168destroySpatialTransformerDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_stDesc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroySpatialTransformerDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_destroySpatialTransformerDescriptor(__pyx_v_stDesc, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.destroySpatialTransformerDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1423
 * 
 * 
 * cpdef setSpatialTransformerDescriptor(             # <<<<<<<<<<<<<<
 *         size_t stDesc, size_t samplerType, int dataType,
 *         int nbDims, size_t dimA):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_171setSpatialTransformerDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_setSpatialTransformerDescriptor(size_t __pyx_v_stDesc, size_t __pyx_v_samplerType, int __pyx_v_dataType, int __pyx_v_nbDims, size_t __pyx_v_dimA, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setSpatialTransformerDescriptor", 0);

  /* "cupy/cuda/cudnn.pyx":1426
 *         size_t stDesc, size_t samplerType, int dataType,
 *         int nbDims, size_t dimA):
 *     status = cudnnSetSpatialTransformerNdDescriptor(             # <<<<<<<<<<<<<<
 *         <SpatialTransformerDescriptor>stDesc, <SamplerType>samplerType,
 *         <DataType>dataType, nbDims, <int*>dimA)
 */
  __pyx_v_status = cudnnSetSpatialTransformerNdDescriptor(((cudnnSpatialTransformerDescriptor_t)__pyx_v_stDesc), ((cudnnSamplerType_t)__pyx_v_samplerType), ((cudnnDataType_t)__pyx_v_dataType), __pyx_v_nbDims, ((int *)__pyx_v_dimA));

  /* "cupy/cuda/cudnn.pyx":1429
 *         <SpatialTransformerDescriptor>stDesc, <SamplerType>samplerType,
 *         <DataType>dataType, nbDims, <int*>dimA)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1423
 * 
 * 
 * cpdef setSpatialTransformerDescriptor(             # <<<<<<<<<<<<<<
 *         size_t stDesc, size_t samplerType, int dataType,
 *         int nbDims, size_t dimA):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setSpatialTransformerDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_171setSpatialTransformerDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_170setSpatialTransformerDescriptor[] = "setSpatialTransformerDescriptor(size_t stDesc, size_t samplerType, int dataType, int nbDims, size_t dimA)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_171setSpatialTransformerDescriptor(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_stDesc;
  size_t __pyx_v_samplerType;
  int __pyx_v_dataType;
  int __pyx_v_nbDims;
  size_t __pyx_v_dimA;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setSpatialTransformerDescriptor (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_stDesc,&__pyx_n_s_samplerType,&__pyx_n_s_dataType,&__pyx_n_s_nbDims,&__pyx_n_s_dimA,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stDesc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_samplerType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setSpatialTransformerDescriptor", 1, 5, 5, 1); __PYX_ERR(0, 1423, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dataType)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setSpatialTransformerDescriptor", 1, 5, 5, 2); __PYX_ERR(0, 1423, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nbDims)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setSpatialTransformerDescriptor", 1, 5, 5, 3); __PYX_ERR(0, 1423, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dimA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setSpatialTransformerDescriptor", 1, 5, 5, 4); __PYX_ERR(0, 1423, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setSpatialTransformerDescriptor") < 0)) __PYX_ERR(0, 1423, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_stDesc = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_stDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1424, __pyx_L3_error)
    __pyx_v_samplerType = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_samplerType == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1424, __pyx_L3_error)
    __pyx_v_dataType = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_dataType == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1424, __pyx_L3_error)
    __pyx_v_nbDims = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nbDims == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1425, __pyx_L3_error)
    __pyx_v_dimA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_dimA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1425, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setSpatialTransformerDescriptor", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1423, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.setSpatialTransformerDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_170setSpatialTransformerDescriptor(__pyx_self, __pyx_v_stDesc, __pyx_v_samplerType, __pyx_v_dataType, __pyx_v_nbDims, __pyx_v_dimA);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_170setSpatialTransformerDescriptor(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_stDesc, size_t __pyx_v_samplerType, int __pyx_v_dataType, int __pyx_v_nbDims, size_t __pyx_v_dimA) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setSpatialTransformerDescriptor", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setSpatialTransformerDescriptor(__pyx_v_stDesc, __pyx_v_samplerType, __pyx_v_dataType, __pyx_v_nbDims, __pyx_v_dimA, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.setSpatialTransformerDescriptor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1432
 * 
 * 
 * cpdef spatialTfGridGeneratorForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t stDesc, size_t theta, size_t grid):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_173spatialTfGridGeneratorForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_spatialTfGridGeneratorForward(size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_theta, size_t __pyx_v_grid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spatialTfGridGeneratorForward", 0);

  /* "cupy/cuda/cudnn.pyx":1434
 * cpdef spatialTfGridGeneratorForward(
 *         size_t handle, size_t stDesc, size_t theta, size_t grid):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnSpatialTfGridGeneratorForward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1435
 *         size_t handle, size_t stDesc, size_t theta, size_t grid):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSpatialTfGridGeneratorForward(
 *             <Handle>handle, <SpatialTransformerDescriptor> stDesc,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1436
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnSpatialTfGridGeneratorForward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SpatialTransformerDescriptor> stDesc,
 *             <void*>theta, <void*>grid)
 */
        __pyx_v_status = cudnnSpatialTfGridGeneratorForward(((cudnnHandle_t)__pyx_v_handle), ((cudnnSpatialTransformerDescriptor_t)__pyx_v_stDesc), ((void *)__pyx_v_theta), ((void *)__pyx_v_grid));
      }

      /* "cupy/cuda/cudnn.pyx":1435
 *         size_t handle, size_t stDesc, size_t theta, size_t grid):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSpatialTfGridGeneratorForward(
 *             <Handle>handle, <SpatialTransformerDescriptor> stDesc,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1439
 *             <Handle>handle, <SpatialTransformerDescriptor> stDesc,
 *             <void*>theta, <void*>grid)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1432
 * 
 * 
 * cpdef spatialTfGridGeneratorForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t stDesc, size_t theta, size_t grid):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfGridGeneratorForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_173spatialTfGridGeneratorForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_172spatialTfGridGeneratorForward[] = "spatialTfGridGeneratorForward(size_t handle, size_t stDesc, size_t theta, size_t grid)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_173spatialTfGridGeneratorForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_stDesc;
  size_t __pyx_v_theta;
  size_t __pyx_v_grid;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spatialTfGridGeneratorForward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_stDesc,&__pyx_n_s_theta,&__pyx_n_s_grid,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfGridGeneratorForward", 1, 4, 4, 1); __PYX_ERR(0, 1432, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_theta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfGridGeneratorForward", 1, 4, 4, 2); __PYX_ERR(0, 1432, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_grid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfGridGeneratorForward", 1, 4, 4, 3); __PYX_ERR(0, 1432, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spatialTfGridGeneratorForward") < 0)) __PYX_ERR(0, 1432, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1433, __pyx_L3_error)
    __pyx_v_stDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_stDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1433, __pyx_L3_error)
    __pyx_v_theta = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_theta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1433, __pyx_L3_error)
    __pyx_v_grid = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_grid == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1433, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spatialTfGridGeneratorForward", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1432, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfGridGeneratorForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_172spatialTfGridGeneratorForward(__pyx_self, __pyx_v_handle, __pyx_v_stDesc, __pyx_v_theta, __pyx_v_grid);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_172spatialTfGridGeneratorForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_theta, size_t __pyx_v_grid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spatialTfGridGeneratorForward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_spatialTfGridGeneratorForward(__pyx_v_handle, __pyx_v_stDesc, __pyx_v_theta, __pyx_v_grid, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfGridGeneratorForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1442
 * 
 * 
 * cpdef spatialTfGridGeneratorBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t stDesc, size_t dgrid, size_t dtheta):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_175spatialTfGridGeneratorBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_spatialTfGridGeneratorBackward(size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_dgrid, size_t __pyx_v_dtheta, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spatialTfGridGeneratorBackward", 0);

  /* "cupy/cuda/cudnn.pyx":1444
 * cpdef spatialTfGridGeneratorBackward(
 *         size_t handle, size_t stDesc, size_t dgrid, size_t dtheta):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnSpatialTfGridGeneratorBackward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1445
 *         size_t handle, size_t stDesc, size_t dgrid, size_t dtheta):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSpatialTfGridGeneratorBackward(
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1446
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnSpatialTfGridGeneratorBackward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 *             <void*>dgrid, <void*>dtheta)
 */
        __pyx_v_status = cudnnSpatialTfGridGeneratorBackward(((cudnnHandle_t)__pyx_v_handle), ((cudnnSpatialTransformerDescriptor_t)__pyx_v_stDesc), ((void *)__pyx_v_dgrid), ((void *)__pyx_v_dtheta));
      }

      /* "cupy/cuda/cudnn.pyx":1445
 *         size_t handle, size_t stDesc, size_t dgrid, size_t dtheta):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSpatialTfGridGeneratorBackward(
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1449
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 *             <void*>dgrid, <void*>dtheta)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1442
 * 
 * 
 * cpdef spatialTfGridGeneratorBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t stDesc, size_t dgrid, size_t dtheta):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfGridGeneratorBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_175spatialTfGridGeneratorBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_174spatialTfGridGeneratorBackward[] = "spatialTfGridGeneratorBackward(size_t handle, size_t stDesc, size_t dgrid, size_t dtheta)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_175spatialTfGridGeneratorBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_stDesc;
  size_t __pyx_v_dgrid;
  size_t __pyx_v_dtheta;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spatialTfGridGeneratorBackward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_stDesc,&__pyx_n_s_dgrid,&__pyx_n_s_dtheta,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfGridGeneratorBackward", 1, 4, 4, 1); __PYX_ERR(0, 1442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dgrid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfGridGeneratorBackward", 1, 4, 4, 2); __PYX_ERR(0, 1442, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dtheta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfGridGeneratorBackward", 1, 4, 4, 3); __PYX_ERR(0, 1442, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spatialTfGridGeneratorBackward") < 0)) __PYX_ERR(0, 1442, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1443, __pyx_L3_error)
    __pyx_v_stDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_stDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1443, __pyx_L3_error)
    __pyx_v_dgrid = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_dgrid == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1443, __pyx_L3_error)
    __pyx_v_dtheta = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_dtheta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1443, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spatialTfGridGeneratorBackward", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1442, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfGridGeneratorBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_174spatialTfGridGeneratorBackward(__pyx_self, __pyx_v_handle, __pyx_v_stDesc, __pyx_v_dgrid, __pyx_v_dtheta);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_174spatialTfGridGeneratorBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_dgrid, size_t __pyx_v_dtheta) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spatialTfGridGeneratorBackward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_spatialTfGridGeneratorBackward(__pyx_v_handle, __pyx_v_stDesc, __pyx_v_dgrid, __pyx_v_dtheta, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfGridGeneratorBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1452
 * 
 * 
 * cpdef spatialTfSamplerForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t stDesc, size_t alpha, size_t xDesc,
 *         size_t x, size_t grid, size_t beta, size_t yDesc, size_t y):
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_177spatialTfSamplerForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_spatialTfSamplerForward(size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_alpha, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_grid, size_t __pyx_v_beta, size_t __pyx_v_yDesc, size_t __pyx_v_y, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spatialTfSamplerForward", 0);

  /* "cupy/cuda/cudnn.pyx":1455
 *         size_t handle, size_t stDesc, size_t alpha, size_t xDesc,
 *         size_t x, size_t grid, size_t beta, size_t yDesc, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnSpatialTfSamplerForward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1456
 *         size_t x, size_t grid, size_t beta, size_t yDesc, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSpatialTfSamplerForward(
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1457
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnSpatialTfSamplerForward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 *             <void*>alpha, <TensorDescriptor>xDesc, <void*>x, <void*>grid,
 */
        __pyx_v_status = cudnnSpatialTfSamplerForward(((cudnnHandle_t)__pyx_v_handle), ((cudnnSpatialTransformerDescriptor_t)__pyx_v_stDesc), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((void *)__pyx_v_x), ((void *)__pyx_v_grid), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_yDesc), ((void *)__pyx_v_y));
      }

      /* "cupy/cuda/cudnn.pyx":1456
 *         size_t x, size_t grid, size_t beta, size_t yDesc, size_t y):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSpatialTfSamplerForward(
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1461
 *             <void*>alpha, <TensorDescriptor>xDesc, <void*>x, <void*>grid,
 *             <void*>beta, <TensorDescriptor>yDesc, <void*>y)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1452
 * 
 * 
 * cpdef spatialTfSamplerForward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t stDesc, size_t alpha, size_t xDesc,
 *         size_t x, size_t grid, size_t beta, size_t yDesc, size_t y):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfSamplerForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_177spatialTfSamplerForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_176spatialTfSamplerForward[] = "spatialTfSamplerForward(size_t handle, size_t stDesc, size_t alpha, size_t xDesc, size_t x, size_t grid, size_t beta, size_t yDesc, size_t y)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_177spatialTfSamplerForward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_stDesc;
  size_t __pyx_v_alpha;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_grid;
  size_t __pyx_v_beta;
  size_t __pyx_v_yDesc;
  size_t __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spatialTfSamplerForward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_stDesc,&__pyx_n_s_alpha,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_grid,&__pyx_n_s_beta,&__pyx_n_s_yDesc,&__pyx_n_s_y,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, 1); __PYX_ERR(0, 1452, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, 2); __PYX_ERR(0, 1452, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, 3); __PYX_ERR(0, 1452, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, 4); __PYX_ERR(0, 1452, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_grid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, 5); __PYX_ERR(0, 1452, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, 6); __PYX_ERR(0, 1452, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, 7); __PYX_ERR(0, 1452, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, 8); __PYX_ERR(0, 1452, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spatialTfSamplerForward") < 0)) __PYX_ERR(0, 1452, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1453, __pyx_L3_error)
    __pyx_v_stDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_stDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1453, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1453, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1453, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1454, __pyx_L3_error)
    __pyx_v_grid = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_grid == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1454, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1454, __pyx_L3_error)
    __pyx_v_yDesc = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_yDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1454, __pyx_L3_error)
    __pyx_v_y = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_y == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1454, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spatialTfSamplerForward", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1452, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfSamplerForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_176spatialTfSamplerForward(__pyx_self, __pyx_v_handle, __pyx_v_stDesc, __pyx_v_alpha, __pyx_v_xDesc, __pyx_v_x, __pyx_v_grid, __pyx_v_beta, __pyx_v_yDesc, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_176spatialTfSamplerForward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_alpha, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_grid, size_t __pyx_v_beta, size_t __pyx_v_yDesc, size_t __pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spatialTfSamplerForward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_spatialTfSamplerForward(__pyx_v_handle, __pyx_v_stDesc, __pyx_v_alpha, __pyx_v_xDesc, __pyx_v_x, __pyx_v_grid, __pyx_v_beta, __pyx_v_yDesc, __pyx_v_y, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfSamplerForward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cudnn.pyx":1464
 * 
 * 
 * cpdef spatialTfSamplerBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t stDesc, size_t alpha, size_t xDesc,
 *         size_t x, size_t beta, size_t dxDesc, size_t dx, size_t alphaDgrid,
 */

static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_179spatialTfSamplerBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_5cudnn_spatialTfSamplerBackward(size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_alpha, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_alphaDgrid, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_grid, size_t __pyx_v_betaDgrid, size_t __pyx_v_dgrid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spatialTfSamplerBackward", 0);

  /* "cupy/cuda/cudnn.pyx":1468
 *         size_t x, size_t beta, size_t dxDesc, size_t dx, size_t alphaDgrid,
 *         size_t dyDesc, size_t dy, size_t grid, size_t betaDgrid, size_t dgrid):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cudnnSpatialTfSamplerBackward(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1469
 *         size_t dyDesc, size_t dy, size_t grid, size_t betaDgrid, size_t dgrid):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSpatialTfSamplerBackward(
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cudnn.pyx":1470
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cudnnSpatialTfSamplerBackward(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 *             <void*>alpha, <TensorDescriptor>xDesc, <void*>x, <void*>beta,
 */
        __pyx_v_status = cudnnSpatialTfSamplerBackward(((cudnnHandle_t)__pyx_v_handle), ((cudnnSpatialTransformerDescriptor_t)__pyx_v_stDesc), ((void *)__pyx_v_alpha), ((cudnnTensorDescriptor_t)__pyx_v_xDesc), ((void *)__pyx_v_x), ((void *)__pyx_v_beta), ((cudnnTensorDescriptor_t)__pyx_v_dxDesc), ((void *)__pyx_v_dx), ((void *)__pyx_v_alphaDgrid), ((cudnnTensorDescriptor_t)__pyx_v_dyDesc), ((void *)__pyx_v_dy), ((void *)__pyx_v_grid), ((void *)__pyx_v_betaDgrid), ((void *)__pyx_v_dgrid));
      }

      /* "cupy/cuda/cudnn.pyx":1469
 *         size_t dyDesc, size_t dy, size_t grid, size_t betaDgrid, size_t dgrid):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cudnnSpatialTfSamplerBackward(
 *             <Handle>handle, <SpatialTransformerDescriptor>stDesc,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cudnn.pyx":1476
 *             <TensorDescriptor>dyDesc, <void*>dy, <void*>grid,
 *             <void*>betaDgrid, <void*>dgrid)
 *     check_status(status)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cudnn.pyx":1464
 * 
 * 
 * cpdef spatialTfSamplerBackward(             # <<<<<<<<<<<<<<
 *         size_t handle, size_t stDesc, size_t alpha, size_t xDesc,
 *         size_t x, size_t beta, size_t dxDesc, size_t dx, size_t alphaDgrid,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfSamplerBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_179spatialTfSamplerBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_5cudnn_178spatialTfSamplerBackward[] = "spatialTfSamplerBackward(size_t handle, size_t stDesc, size_t alpha, size_t xDesc, size_t x, size_t beta, size_t dxDesc, size_t dx, size_t alphaDgrid, size_t dyDesc, size_t dy, size_t grid, size_t betaDgrid, size_t dgrid)";
static PyObject *__pyx_pw_4cupy_4cuda_5cudnn_179spatialTfSamplerBackward(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_handle;
  size_t __pyx_v_stDesc;
  size_t __pyx_v_alpha;
  size_t __pyx_v_xDesc;
  size_t __pyx_v_x;
  size_t __pyx_v_beta;
  size_t __pyx_v_dxDesc;
  size_t __pyx_v_dx;
  size_t __pyx_v_alphaDgrid;
  size_t __pyx_v_dyDesc;
  size_t __pyx_v_dy;
  size_t __pyx_v_grid;
  size_t __pyx_v_betaDgrid;
  size_t __pyx_v_dgrid;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spatialTfSamplerBackward (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_stDesc,&__pyx_n_s_alpha,&__pyx_n_s_xDesc,&__pyx_n_s_x,&__pyx_n_s_beta,&__pyx_n_s_dxDesc,&__pyx_n_s_dx,&__pyx_n_s_alphaDgrid,&__pyx_n_s_dyDesc,&__pyx_n_s_dy,&__pyx_n_s_grid,&__pyx_n_s_betaDgrid,&__pyx_n_s_dgrid,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_stDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 1); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alpha)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 2); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_xDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 3); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 4); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_beta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 5); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dxDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 6); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 7); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_alphaDgrid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 8); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dyDesc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 9); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 10); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_grid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 11); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_betaDgrid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 12); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dgrid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, 13); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spatialTfSamplerBackward") < 0)) __PYX_ERR(0, 1464, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_handle == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1465, __pyx_L3_error)
    __pyx_v_stDesc = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_stDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1465, __pyx_L3_error)
    __pyx_v_alpha = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_alpha == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1465, __pyx_L3_error)
    __pyx_v_xDesc = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_xDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1465, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1466, __pyx_L3_error)
    __pyx_v_beta = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_beta == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1466, __pyx_L3_error)
    __pyx_v_dxDesc = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_dxDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1466, __pyx_L3_error)
    __pyx_v_dx = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_dx == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1466, __pyx_L3_error)
    __pyx_v_alphaDgrid = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_alphaDgrid == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1466, __pyx_L3_error)
    __pyx_v_dyDesc = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_dyDesc == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1467, __pyx_L3_error)
    __pyx_v_dy = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_dy == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1467, __pyx_L3_error)
    __pyx_v_grid = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_grid == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1467, __pyx_L3_error)
    __pyx_v_betaDgrid = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_betaDgrid == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1467, __pyx_L3_error)
    __pyx_v_dgrid = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_dgrid == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1467, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spatialTfSamplerBackward", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1464, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfSamplerBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_5cudnn_178spatialTfSamplerBackward(__pyx_self, __pyx_v_handle, __pyx_v_stDesc, __pyx_v_alpha, __pyx_v_xDesc, __pyx_v_x, __pyx_v_beta, __pyx_v_dxDesc, __pyx_v_dx, __pyx_v_alphaDgrid, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_grid, __pyx_v_betaDgrid, __pyx_v_dgrid);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_5cudnn_178spatialTfSamplerBackward(CYTHON_UNUSED PyObject *__pyx_self, size_t __pyx_v_handle, size_t __pyx_v_stDesc, size_t __pyx_v_alpha, size_t __pyx_v_xDesc, size_t __pyx_v_x, size_t __pyx_v_beta, size_t __pyx_v_dxDesc, size_t __pyx_v_dx, size_t __pyx_v_alphaDgrid, size_t __pyx_v_dyDesc, size_t __pyx_v_dy, size_t __pyx_v_grid, size_t __pyx_v_betaDgrid, size_t __pyx_v_dgrid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spatialTfSamplerBackward", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_5cudnn_spatialTfSamplerBackward(__pyx_v_handle, __pyx_v_stDesc, __pyx_v_alpha, __pyx_v_xDesc, __pyx_v_x, __pyx_v_beta, __pyx_v_dxDesc, __pyx_v_dx, __pyx_v_alphaDgrid, __pyx_v_dyDesc, __pyx_v_dy, __pyx_v_grid, __pyx_v_betaDgrid, __pyx_v_dgrid, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cudnn.spatialTfSamplerBackward", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_int")
 * cdef object __pyx_convert_vector_to_py_int(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_int(const std::vector<int>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_int", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_int")
 * cdef object __pyx_convert_vector_to_py_int(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_int")
 * cdef object __pyx_convert_vector_to_py_int(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t(const std::vector<cudnnConvolutionFwdAlgoPerf_t>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t")
 * cdef object __pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;
    __pyx_t_4 = __pyx_convert__to_py_cudnnConvolutionFwdAlgoPerf_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t")
 * cdef object __pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_cudnnConvolutionFwdAlgoPerf_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t(const std::vector<cudnnConvolutionBwdFilterAlgoPerf_t>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t")
 * cdef object __pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;
    __pyx_t_4 = __pyx_convert__to_py_cudnnConvolutionBwdFilterAlgoPerf_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t")
 * cdef object __pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_cudnnConvolutionBwdFilterAlgoPerf_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t(const std::vector<cudnnConvolutionBwdDataAlgoPerf_t>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t")
 * cdef object __pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;
    __pyx_t_4 = __pyx_convert__to_py_cudnnConvolutionBwdDataAlgoPerf_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t")
 * cdef object __pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {"check_status", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_1check_status, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_check_status},
  {"getVersion", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_3getVersion, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_2getVersion},
  {"create", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_5create, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_4create},
  {"destroy", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_7destroy, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_6destroy},
  {"setStream", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_9setStream, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_8setStream},
  {"getStream", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_11getStream, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_10getStream},
  {"createTensorDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_13createTensorDescriptor, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_12createTensorDescriptor},
  {"setTensor4dDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_15setTensor4dDescriptor, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_14setTensor4dDescriptor},
  {"setTensor4dDescriptorEx", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_17setTensor4dDescriptorEx, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_16setTensor4dDescriptorEx},
  {"getTensor4dDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_19getTensor4dDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_18getTensor4dDescriptor},
  {"setTensorNdDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_21setTensorNdDescriptor, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_20setTensorNdDescriptor},
  {"destroyTensorDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_23destroyTensorDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_22destroyTensorDescriptor},
  {"addTensor_v3", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_25addTensor_v3, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_24addTensor_v3},
  {"createFilterDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_27createFilterDescriptor, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_26createFilterDescriptor},
  {"setFilter4dDescriptor_v4", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_29setFilter4dDescriptor_v4, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_28setFilter4dDescriptor_v4},
  {"setFilterNdDescriptor_v4", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_31setFilterNdDescriptor_v4, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_30setFilterNdDescriptor_v4},
  {"getFilterNdDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_33getFilterNdDescriptor, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_32getFilterNdDescriptor},
  {"destroyFilterDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_35destroyFilterDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_34destroyFilterDescriptor},
  {"createConvolutionDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_37createConvolutionDescriptor, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_36createConvolutionDescriptor},
  {"setConvolutionMathType", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_39setConvolutionMathType, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_38setConvolutionMathType},
  {"getConvolutionMathType", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_41getConvolutionMathType, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_40getConvolutionMathType},
  {"setConvolutionGroupCount", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_43setConvolutionGroupCount, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_42setConvolutionGroupCount},
  {"getConvolutionGroupCount", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_45getConvolutionGroupCount, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_44getConvolutionGroupCount},
  {"setConvolution2dDescriptor_v4", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_47setConvolution2dDescriptor_v4, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_46setConvolution2dDescriptor_v4},
  {"setConvolution2dDescriptor_v5", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_49setConvolution2dDescriptor_v5, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_48setConvolution2dDescriptor_v5},
  {"setConvolutionNdDescriptor_v3", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_51setConvolutionNdDescriptor_v3, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_50setConvolutionNdDescriptor_v3},
  {"destroyConvolutionDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_53destroyConvolutionDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_52destroyConvolutionDescriptor},
  {"findConvolutionForwardAlgorithm", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_55findConvolutionForwardAlgorithm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_54findConvolutionForwardAlgorithm},
  {"findConvolutionForwardAlgorithmEx", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_57findConvolutionForwardAlgorithmEx, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_56findConvolutionForwardAlgorithmEx},
  {"getConvolutionForwardAlgorithm", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_59getConvolutionForwardAlgorithm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_58getConvolutionForwardAlgorithm},
  {"getConvolutionForwardWorkspaceSize", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_61getConvolutionForwardWorkspaceSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_60getConvolutionForwardWorkspaceSize},
  {"convolutionForward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_63convolutionForward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_62convolutionForward},
  {"convolutionBackwardBias", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_65convolutionBackwardBias, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_64convolutionBackwardBias},
  {"findConvolutionBackwardFilterAlgorithm", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_67findConvolutionBackwardFilterAlgorithm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_66findConvolutionBackwardFilterAlgorithm},
  {"findConvolutionBackwardFilterAlgorithmEx", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_69findConvolutionBackwardFilterAlgorithmEx, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_68findConvolutionBackwardFilterAlgorithmEx},
  {"getConvolutionBackwardFilterAlgorithm", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_71getConvolutionBackwardFilterAlgorithm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_70getConvolutionBackwardFilterAlgorithm},
  {"getConvolutionBackwardFilterWorkspaceSize", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_73getConvolutionBackwardFilterWorkspaceSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_72getConvolutionBackwardFilterWorkspaceSize},
  {"convolutionBackwardFilter_v3", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_75convolutionBackwardFilter_v3, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_74convolutionBackwardFilter_v3},
  {"findConvolutionBackwardDataAlgorithm", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_77findConvolutionBackwardDataAlgorithm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_76findConvolutionBackwardDataAlgorithm},
  {"findConvolutionBackwardDataAlgorithmEx", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_79findConvolutionBackwardDataAlgorithmEx, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_78findConvolutionBackwardDataAlgorithmEx},
  {"getConvolutionBackwardDataAlgorithm", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_81getConvolutionBackwardDataAlgorithm, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_80getConvolutionBackwardDataAlgorithm},
  {"getConvolutionBackwardDataWorkspaceSize", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_83getConvolutionBackwardDataWorkspaceSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_82getConvolutionBackwardDataWorkspaceSize},
  {"convolutionBackwardData_v3", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_85convolutionBackwardData_v3, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_84convolutionBackwardData_v3},
  {"createPoolingDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_87createPoolingDescriptor, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_86createPoolingDescriptor},
  {"setPooling2dDescriptor_v4", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_89setPooling2dDescriptor_v4, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_88setPooling2dDescriptor_v4},
  {"setPoolingNdDescriptor_v4", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_91setPoolingNdDescriptor_v4, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_90setPoolingNdDescriptor_v4},
  {"destroyPoolingDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_93destroyPoolingDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_92destroyPoolingDescriptor},
  {"poolingForward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_95poolingForward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_94poolingForward},
  {"poolingBackward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_97poolingBackward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_96poolingBackward},
  {"deriveBNTensorDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_99deriveBNTensorDescriptor, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_98deriveBNTensorDescriptor},
  {"batchNormalizationForwardTraining", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_101batchNormalizationForwardTraining, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_100batchNormalizationForwardTraining},
  {"batchNormalizationForwardInference", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_103batchNormalizationForwardInference, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_102batchNormalizationForwardInference},
  {"batchNormalizationBackward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_105batchNormalizationBackward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_104batchNormalizationBackward},
  {"createActivationDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_107createActivationDescriptor, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_106createActivationDescriptor},
  {"setActivationDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_109setActivationDescriptor, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_108setActivationDescriptor},
  {"destroyActivationDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_111destroyActivationDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_110destroyActivationDescriptor},
  {"softmaxForward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_113softmaxForward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_112softmaxForward},
  {"softmaxBackward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_115softmaxBackward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_114softmaxBackward},
  {"activationForward_v4", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_117activationForward_v4, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_116activationForward_v4},
  {"activationBackward_v4", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_119activationBackward_v4, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_118activationBackward_v4},
  {"createDropoutDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_121createDropoutDescriptor, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_120createDropoutDescriptor},
  {"destroyDropoutDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_123destroyDropoutDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_122destroyDropoutDescriptor},
  {"dropoutGetStatesSize", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_125dropoutGetStatesSize, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_124dropoutGetStatesSize},
  {"setDropoutDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_127setDropoutDescriptor, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_126setDropoutDescriptor},
  {"getDropoutReserveSpaceSize", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_129getDropoutReserveSpaceSize, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_128getDropoutReserveSpaceSize},
  {"dropoutForward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_131dropoutForward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_130dropoutForward},
  {"dropoutBackward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_133dropoutBackward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_132dropoutBackward},
  {"createRNNDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_135createRNNDescriptor, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_134createRNNDescriptor},
  {"destroyRNNDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_137destroyRNNDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_136destroyRNNDescriptor},
  {"createPersistentRNNPlan", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_139createPersistentRNNPlan, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_138createPersistentRNNPlan},
  {"setPersistentRNNPlan", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_141setPersistentRNNPlan, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_140setPersistentRNNPlan},
  {"destroyPersistentRNNPlan", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_143destroyPersistentRNNPlan, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_142destroyPersistentRNNPlan},
  {"setRNNDescriptor_v5", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_145setRNNDescriptor_v5, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_144setRNNDescriptor_v5},
  {"setRNNDescriptor_v6", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_147setRNNDescriptor_v6, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_146setRNNDescriptor_v6},
  {"getRNNWorkspaceSize", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_149getRNNWorkspaceSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_148getRNNWorkspaceSize},
  {"getRNNTrainingReserveSize", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_151getRNNTrainingReserveSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_150getRNNTrainingReserveSize},
  {"getRNNParamsSize", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_153getRNNParamsSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_152getRNNParamsSize},
  {"getRNNLinLayerMatrixParams", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_155getRNNLinLayerMatrixParams, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_154getRNNLinLayerMatrixParams},
  {"getRNNLinLayerBiasParams", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_157getRNNLinLayerBiasParams, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_156getRNNLinLayerBiasParams},
  {"RNNForwardInference", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_159RNNForwardInference, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_158RNNForwardInference},
  {"RNNForwardTraining", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_161RNNForwardTraining, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_160RNNForwardTraining},
  {"RNNBackwardData", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_163RNNBackwardData, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_162RNNBackwardData},
  {"RNNBackwardWeights", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_165RNNBackwardWeights, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_164RNNBackwardWeights},
  {"createSpatialTransformerDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_167createSpatialTransformerDescriptor, METH_NOARGS, __pyx_doc_4cupy_4cuda_5cudnn_166createSpatialTransformerDescriptor},
  {"destroySpatialTransformerDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_169destroySpatialTransformerDescriptor, METH_O, __pyx_doc_4cupy_4cuda_5cudnn_168destroySpatialTransformerDescriptor},
  {"setSpatialTransformerDescriptor", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_171setSpatialTransformerDescriptor, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_170setSpatialTransformerDescriptor},
  {"spatialTfGridGeneratorForward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_173spatialTfGridGeneratorForward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_172spatialTfGridGeneratorForward},
  {"spatialTfGridGeneratorBackward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_175spatialTfGridGeneratorBackward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_174spatialTfGridGeneratorBackward},
  {"spatialTfSamplerForward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_177spatialTfSamplerForward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_176spatialTfSamplerForward},
  {"spatialTfSamplerBackward", (PyCFunction)__pyx_pw_4cupy_4cuda_5cudnn_179spatialTfSamplerBackward, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_5cudnn_178spatialTfSamplerBackward},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "cudnn",
    __pyx_k_Thin_wrapper_of_cuDNN, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_CUDNN_STATUS_ALLOC_FAILED, __pyx_k_CUDNN_STATUS_ALLOC_FAILED, sizeof(__pyx_k_CUDNN_STATUS_ALLOC_FAILED), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_ARCH_MISMATCH, __pyx_k_CUDNN_STATUS_ARCH_MISMATCH, sizeof(__pyx_k_CUDNN_STATUS_ARCH_MISMATCH), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_BAD_PARAM, __pyx_k_CUDNN_STATUS_BAD_PARAM, sizeof(__pyx_k_CUDNN_STATUS_BAD_PARAM), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_EXECUTION_FAILED, __pyx_k_CUDNN_STATUS_EXECUTION_FAILED, sizeof(__pyx_k_CUDNN_STATUS_EXECUTION_FAILED), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_INTERNAL_ERROR, __pyx_k_CUDNN_STATUS_INTERNAL_ERROR, sizeof(__pyx_k_CUDNN_STATUS_INTERNAL_ERROR), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_INVALID_VALUE, __pyx_k_CUDNN_STATUS_INVALID_VALUE, sizeof(__pyx_k_CUDNN_STATUS_INVALID_VALUE), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_LICENSE_ERROR, __pyx_k_CUDNN_STATUS_LICENSE_ERROR, sizeof(__pyx_k_CUDNN_STATUS_LICENSE_ERROR), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_MAPPING_ERROR, __pyx_k_CUDNN_STATUS_MAPPING_ERROR, sizeof(__pyx_k_CUDNN_STATUS_MAPPING_ERROR), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_NOT_INITIALIZED, __pyx_k_CUDNN_STATUS_NOT_INITIALIZED, sizeof(__pyx_k_CUDNN_STATUS_NOT_INITIALIZED), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_NOT_SUPPORTED, __pyx_k_CUDNN_STATUS_NOT_SUPPORTED, sizeof(__pyx_k_CUDNN_STATUS_NOT_SUPPORTED), 0, 0, 1, 1},
  {&__pyx_n_s_CUDNN_STATUS_SUCCESS, __pyx_k_CUDNN_STATUS_SUCCESS, sizeof(__pyx_k_CUDNN_STATUS_SUCCESS), 0, 0, 1, 1},
  {&__pyx_n_s_CuDNNError, __pyx_k_CuDNNError, sizeof(__pyx_k_CuDNNError), 0, 0, 1, 1},
  {&__pyx_n_s_CuDNNError___init, __pyx_k_CuDNNError___init, sizeof(__pyx_k_CuDNNError___init), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_activationDesc, __pyx_k_activationDesc, sizeof(__pyx_k_activationDesc), 0, 0, 1, 1},
  {&__pyx_n_s_algo, __pyx_k_algo, sizeof(__pyx_k_algo), 0, 0, 1, 1},
  {&__pyx_n_s_algorithm, __pyx_k_algorithm, sizeof(__pyx_k_algorithm), 0, 0, 1, 1},
  {&__pyx_n_s_alpha, __pyx_k_alpha, sizeof(__pyx_k_alpha), 0, 0, 1, 1},
  {&__pyx_n_s_alphaDataDiff, __pyx_k_alphaDataDiff, sizeof(__pyx_k_alphaDataDiff), 0, 0, 1, 1},
  {&__pyx_n_s_alphaDgrid, __pyx_k_alphaDgrid, sizeof(__pyx_k_alphaDgrid), 0, 0, 1, 1},
  {&__pyx_n_s_alphaParamDiff, __pyx_k_alphaParamDiff, sizeof(__pyx_k_alphaParamDiff), 0, 0, 1, 1},
  {&__pyx_n_s_arrayLength, __pyx_k_arrayLength, sizeof(__pyx_k_arrayLength), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_bDesc, __pyx_k_bDesc, sizeof(__pyx_k_bDesc), 0, 0, 1, 1},
  {&__pyx_n_s_beta, __pyx_k_beta, sizeof(__pyx_k_beta), 0, 0, 1, 1},
  {&__pyx_n_s_betaDataDiff, __pyx_k_betaDataDiff, sizeof(__pyx_k_betaDataDiff), 0, 0, 1, 1},
  {&__pyx_n_s_betaDgrid, __pyx_k_betaDgrid, sizeof(__pyx_k_betaDgrid), 0, 0, 1, 1},
  {&__pyx_n_s_betaParamDiff, __pyx_k_betaParamDiff, sizeof(__pyx_k_betaParamDiff), 0, 0, 1, 1},
  {&__pyx_n_s_bnBias, __pyx_k_bnBias, sizeof(__pyx_k_bnBias), 0, 0, 1, 1},
  {&__pyx_n_s_bnScale, __pyx_k_bnScale, sizeof(__pyx_k_bnScale), 0, 0, 1, 1},
  {&__pyx_n_s_bnScaleBiasMeanVarDesc, __pyx_k_bnScaleBiasMeanVarDesc, sizeof(__pyx_k_bnScaleBiasMeanVarDesc), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_cStride, __pyx_k_cStride, sizeof(__pyx_k_cStride), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_computeType, __pyx_k_computeType, sizeof(__pyx_k_computeType), 0, 0, 1, 1},
  {&__pyx_n_s_convDesc, __pyx_k_convDesc, sizeof(__pyx_k_convDesc), 0, 0, 1, 1},
  {&__pyx_n_s_cupy_cuda_cudnn, __pyx_k_cupy_cuda_cudnn, sizeof(__pyx_k_cupy_cuda_cudnn), 0, 0, 1, 1},
  {&__pyx_kp_s_cupy_cuda_cudnn_pyx, __pyx_k_cupy_cuda_cudnn_pyx, sizeof(__pyx_k_cupy_cuda_cudnn_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_cx, __pyx_k_cx, sizeof(__pyx_k_cx), 0, 0, 1, 1},
  {&__pyx_n_s_cxDesc, __pyx_k_cxDesc, sizeof(__pyx_k_cxDesc), 0, 0, 1, 1},
  {&__pyx_n_s_cy, __pyx_k_cy, sizeof(__pyx_k_cy), 0, 0, 1, 1},
  {&__pyx_n_s_cyDesc, __pyx_k_cyDesc, sizeof(__pyx_k_cyDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dBnBiasResult, __pyx_k_dBnBiasResult, sizeof(__pyx_k_dBnBiasResult), 0, 0, 1, 1},
  {&__pyx_n_s_dBnScaleBiasDesc, __pyx_k_dBnScaleBiasDesc, sizeof(__pyx_k_dBnScaleBiasDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dBnScaleResult, __pyx_k_dBnScaleResult, sizeof(__pyx_k_dBnScaleResult), 0, 0, 1, 1},
  {&__pyx_n_s_dataType, __pyx_k_dataType, sizeof(__pyx_k_dataType), 0, 0, 1, 1},
  {&__pyx_n_s_dcx, __pyx_k_dcx, sizeof(__pyx_k_dcx), 0, 0, 1, 1},
  {&__pyx_n_s_dcxDesc, __pyx_k_dcxDesc, sizeof(__pyx_k_dcxDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dcy, __pyx_k_dcy, sizeof(__pyx_k_dcy), 0, 0, 1, 1},
  {&__pyx_n_s_dcyDesc, __pyx_k_dcyDesc, sizeof(__pyx_k_dcyDesc), 0, 0, 1, 1},
  {&__pyx_n_s_derivedBnDesc, __pyx_k_derivedBnDesc, sizeof(__pyx_k_derivedBnDesc), 0, 0, 1, 1},
  {&__pyx_n_s_destData, __pyx_k_destData, sizeof(__pyx_k_destData), 0, 0, 1, 1},
  {&__pyx_n_s_destDesc, __pyx_k_destDesc, sizeof(__pyx_k_destDesc), 0, 0, 1, 1},
  {&__pyx_n_s_destDiffData, __pyx_k_destDiffData, sizeof(__pyx_k_destDiffData), 0, 0, 1, 1},
  {&__pyx_n_s_destDiffDesc, __pyx_k_destDiffDesc, sizeof(__pyx_k_destDiffDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dgrid, __pyx_k_dgrid, sizeof(__pyx_k_dgrid), 0, 0, 1, 1},
  {&__pyx_n_s_dhx, __pyx_k_dhx, sizeof(__pyx_k_dhx), 0, 0, 1, 1},
  {&__pyx_n_s_dhxDesc, __pyx_k_dhxDesc, sizeof(__pyx_k_dhxDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dhy, __pyx_k_dhy, sizeof(__pyx_k_dhy), 0, 0, 1, 1},
  {&__pyx_n_s_dhyDesc, __pyx_k_dhyDesc, sizeof(__pyx_k_dhyDesc), 0, 0, 1, 1},
  {&__pyx_n_s_diffData, __pyx_k_diffData, sizeof(__pyx_k_diffData), 0, 0, 1, 1},
  {&__pyx_n_s_diffDesc, __pyx_k_diffDesc, sizeof(__pyx_k_diffDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dilationA, __pyx_k_dilationA, sizeof(__pyx_k_dilationA), 0, 0, 1, 1},
  {&__pyx_n_s_dilation_h, __pyx_k_dilation_h, sizeof(__pyx_k_dilation_h), 0, 0, 1, 1},
  {&__pyx_n_s_dilation_w, __pyx_k_dilation_w, sizeof(__pyx_k_dilation_w), 0, 0, 1, 1},
  {&__pyx_n_s_dimA, __pyx_k_dimA, sizeof(__pyx_k_dimA), 0, 0, 1, 1},
  {&__pyx_n_s_direction, __pyx_k_direction, sizeof(__pyx_k_direction), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dropout, __pyx_k_dropout, sizeof(__pyx_k_dropout), 0, 0, 1, 1},
  {&__pyx_n_s_dropoutDesc, __pyx_k_dropoutDesc, sizeof(__pyx_k_dropoutDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dstData, __pyx_k_dstData, sizeof(__pyx_k_dstData), 0, 0, 1, 1},
  {&__pyx_n_s_dstDesc, __pyx_k_dstDesc, sizeof(__pyx_k_dstDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dtheta, __pyx_k_dtheta, sizeof(__pyx_k_dtheta), 0, 0, 1, 1},
  {&__pyx_n_s_dw, __pyx_k_dw, sizeof(__pyx_k_dw), 0, 0, 1, 1},
  {&__pyx_n_s_dwDesc, __pyx_k_dwDesc, sizeof(__pyx_k_dwDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dx, __pyx_k_dx, sizeof(__pyx_k_dx), 0, 0, 1, 1},
  {&__pyx_n_s_dxData, __pyx_k_dxData, sizeof(__pyx_k_dxData), 0, 0, 1, 1},
  {&__pyx_n_s_dxDesc, __pyx_k_dxDesc, sizeof(__pyx_k_dxDesc), 0, 0, 1, 1},
  {&__pyx_n_s_dy, __pyx_k_dy, sizeof(__pyx_k_dy), 0, 0, 1, 1},
  {&__pyx_n_s_dyData, __pyx_k_dyData, sizeof(__pyx_k_dyData), 0, 0, 1, 1},
  {&__pyx_n_s_dyDesc, __pyx_k_dyDesc, sizeof(__pyx_k_dyDesc), 0, 0, 1, 1},
  {&__pyx_n_s_epsilon, __pyx_k_epsilon, sizeof(__pyx_k_epsilon), 0, 0, 1, 1},
  {&__pyx_n_s_estimatedMean, __pyx_k_estimatedMean, sizeof(__pyx_k_estimatedMean), 0, 0, 1, 1},
  {&__pyx_n_s_estimatedVariance, __pyx_k_estimatedVariance, sizeof(__pyx_k_estimatedVariance), 0, 0, 1, 1},
  {&__pyx_n_s_exponentialAverageFactor, __pyx_k_exponentialAverageFactor, sizeof(__pyx_k_exponentialAverageFactor), 0, 0, 1, 1},
  {&__pyx_n_s_filterData, __pyx_k_filterData, sizeof(__pyx_k_filterData), 0, 0, 1, 1},
  {&__pyx_n_s_filterDesc, __pyx_k_filterDesc, sizeof(__pyx_k_filterDesc), 0, 0, 1, 1},
  {&__pyx_n_s_filterDimA, __pyx_k_filterDimA, sizeof(__pyx_k_filterDimA), 0, 0, 1, 1},
  {&__pyx_n_s_filterStrideA, __pyx_k_filterStrideA, sizeof(__pyx_k_filterStrideA), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_gradData, __pyx_k_gradData, sizeof(__pyx_k_gradData), 0, 0, 1, 1},
  {&__pyx_n_s_gradDesc, __pyx_k_gradDesc, sizeof(__pyx_k_gradDesc), 0, 0, 1, 1},
  {&__pyx_n_s_grid, __pyx_k_grid, sizeof(__pyx_k_grid), 0, 0, 1, 1},
  {&__pyx_n_s_groupCount, __pyx_k_groupCount, sizeof(__pyx_k_groupCount), 0, 0, 1, 1},
  {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
  {&__pyx_n_s_hStride, __pyx_k_hStride, sizeof(__pyx_k_hStride), 0, 0, 1, 1},
  {&__pyx_n_s_handle, __pyx_k_handle, sizeof(__pyx_k_handle), 0, 0, 1, 1},
  {&__pyx_n_s_hiddenSize, __pyx_k_hiddenSize, sizeof(__pyx_k_hiddenSize), 0, 0, 1, 1},
  {&__pyx_n_s_horizontalPadding, __pyx_k_horizontalPadding, sizeof(__pyx_k_horizontalPadding), 0, 0, 1, 1},
  {&__pyx_n_s_horizontalStride, __pyx_k_horizontalStride, sizeof(__pyx_k_horizontalStride), 0, 0, 1, 1},
  {&__pyx_n_s_hx, __pyx_k_hx, sizeof(__pyx_k_hx), 0, 0, 1, 1},
  {&__pyx_n_s_hxDesc, __pyx_k_hxDesc, sizeof(__pyx_k_hxDesc), 0, 0, 1, 1},
  {&__pyx_n_s_hy, __pyx_k_hy, sizeof(__pyx_k_hy), 0, 0, 1, 1},
  {&__pyx_n_s_hyDesc, __pyx_k_hyDesc, sizeof(__pyx_k_hyDesc), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_inputMode, __pyx_k_inputMode, sizeof(__pyx_k_inputMode), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_layer, __pyx_k_layer, sizeof(__pyx_k_layer), 0, 0, 1, 1},
  {&__pyx_n_s_linLayerBias, __pyx_k_linLayerBias, sizeof(__pyx_k_linLayerBias), 0, 0, 1, 1},
  {&__pyx_n_s_linLayerBiasDesc, __pyx_k_linLayerBiasDesc, sizeof(__pyx_k_linLayerBiasDesc), 0, 0, 1, 1},
  {&__pyx_n_s_linLayerID, __pyx_k_linLayerID, sizeof(__pyx_k_linLayerID), 0, 0, 1, 1},
  {&__pyx_n_s_linLayerMat, __pyx_k_linLayerMat, sizeof(__pyx_k_linLayerMat), 0, 0, 1, 1},
  {&__pyx_n_s_linLayerMatDesc, __pyx_k_linLayerMatDesc, sizeof(__pyx_k_linLayerMatDesc), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mathType, __pyx_k_mathType, sizeof(__pyx_k_mathType), 0, 0, 1, 1},
  {&__pyx_n_s_maxpoolingNanOpt, __pyx_k_maxpoolingNanOpt, sizeof(__pyx_k_maxpoolingNanOpt), 0, 0, 1, 1},
  {&__pyx_n_s_memory, __pyx_k_memory, sizeof(__pyx_k_memory), 0, 0, 1, 1},
  {&__pyx_n_s_memoryLimitInbytes, __pyx_k_memoryLimitInbytes, sizeof(__pyx_k_memoryLimitInbytes), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_minibatch, __pyx_k_minibatch, sizeof(__pyx_k_minibatch), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_nStride, __pyx_k_nStride, sizeof(__pyx_k_nStride), 0, 0, 1, 1},
  {&__pyx_n_s_nbDims, __pyx_k_nbDims, sizeof(__pyx_k_nbDims), 0, 0, 1, 1},
  {&__pyx_n_s_nbDimsRequested, __pyx_k_nbDimsRequested, sizeof(__pyx_k_nbDimsRequested), 0, 0, 1, 1},
  {&__pyx_n_s_numLayers, __pyx_k_numLayers, sizeof(__pyx_k_numLayers), 0, 0, 1, 1},
  {&__pyx_n_s_padA, __pyx_k_padA, sizeof(__pyx_k_padA), 0, 0, 1, 1},
  {&__pyx_n_s_pad_h, __pyx_k_pad_h, sizeof(__pyx_k_pad_h), 0, 0, 1, 1},
  {&__pyx_n_s_pad_w, __pyx_k_pad_w, sizeof(__pyx_k_pad_w), 0, 0, 1, 1},
  {&__pyx_n_s_paddingA, __pyx_k_paddingA, sizeof(__pyx_k_paddingA), 0, 0, 1, 1},
  {&__pyx_n_s_plan, __pyx_k_plan, sizeof(__pyx_k_plan), 0, 0, 1, 1},
  {&__pyx_n_s_poolingDesc, __pyx_k_poolingDesc, sizeof(__pyx_k_poolingDesc), 0, 0, 1, 1},
  {&__pyx_n_s_preference, __pyx_k_preference, sizeof(__pyx_k_preference), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reluCeiling, __pyx_k_reluCeiling, sizeof(__pyx_k_reluCeiling), 0, 0, 1, 1},
  {&__pyx_n_s_reluNanOpt, __pyx_k_reluNanOpt, sizeof(__pyx_k_reluNanOpt), 0, 0, 1, 1},
  {&__pyx_n_s_requestedAlgoCount, __pyx_k_requestedAlgoCount, sizeof(__pyx_k_requestedAlgoCount), 0, 0, 1, 1},
  {&__pyx_n_s_reserveSpace, __pyx_k_reserveSpace, sizeof(__pyx_k_reserveSpace), 0, 0, 1, 1},
  {&__pyx_n_s_reserveSpaceSizeInBytes, __pyx_k_reserveSpaceSizeInBytes, sizeof(__pyx_k_reserveSpaceSizeInBytes), 0, 0, 1, 1},
  {&__pyx_n_s_resultRunningMean, __pyx_k_resultRunningMean, sizeof(__pyx_k_resultRunningMean), 0, 0, 1, 1},
  {&__pyx_n_s_resultRunningVariance, __pyx_k_resultRunningVariance, sizeof(__pyx_k_resultRunningVariance), 0, 0, 1, 1},
  {&__pyx_n_s_resultSaveInvVariance, __pyx_k_resultSaveInvVariance, sizeof(__pyx_k_resultSaveInvVariance), 0, 0, 1, 1},
  {&__pyx_n_s_resultSaveMean, __pyx_k_resultSaveMean, sizeof(__pyx_k_resultSaveMean), 0, 0, 1, 1},
  {&__pyx_n_s_rnnDesc, __pyx_k_rnnDesc, sizeof(__pyx_k_rnnDesc), 0, 0, 1, 1},
  {&__pyx_kp_s_s_s, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {&__pyx_n_s_samplerType, __pyx_k_samplerType, sizeof(__pyx_k_samplerType), 0, 0, 1, 1},
  {&__pyx_n_s_savedInvVariance, __pyx_k_savedInvVariance, sizeof(__pyx_k_savedInvVariance), 0, 0, 1, 1},
  {&__pyx_n_s_savedMean, __pyx_k_savedMean, sizeof(__pyx_k_savedMean), 0, 0, 1, 1},
  {&__pyx_n_s_seed, __pyx_k_seed, sizeof(__pyx_k_seed), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_seqLength, __pyx_k_seqLength, sizeof(__pyx_k_seqLength), 0, 0, 1, 1},
  {&__pyx_n_s_srcData, __pyx_k_srcData, sizeof(__pyx_k_srcData), 0, 0, 1, 1},
  {&__pyx_n_s_srcDesc, __pyx_k_srcDesc, sizeof(__pyx_k_srcDesc), 0, 0, 1, 1},
  {&__pyx_n_s_srcDiffData, __pyx_k_srcDiffData, sizeof(__pyx_k_srcDiffData), 0, 0, 1, 1},
  {&__pyx_n_s_srcDiffDesc, __pyx_k_srcDiffDesc, sizeof(__pyx_k_srcDiffDesc), 0, 0, 1, 1},
  {&__pyx_n_s_stDesc, __pyx_k_stDesc, sizeof(__pyx_k_stDesc), 0, 0, 1, 1},
  {&__pyx_n_s_stateSizeInBytes, __pyx_k_stateSizeInBytes, sizeof(__pyx_k_stateSizeInBytes), 0, 0, 1, 1},
  {&__pyx_n_s_states, __pyx_k_states, sizeof(__pyx_k_states), 0, 0, 1, 1},
  {&__pyx_n_s_status, __pyx_k_status, sizeof(__pyx_k_status), 0, 0, 1, 1},
  {&__pyx_n_s_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 0, 1, 1},
  {&__pyx_n_s_strideA, __pyx_k_strideA, sizeof(__pyx_k_strideA), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_tensorDesc, __pyx_k_tensorDesc, sizeof(__pyx_k_tensorDesc), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theta, __pyx_k_theta, sizeof(__pyx_k_theta), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_u, __pyx_k_u, sizeof(__pyx_k_u), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_verticalPadding, __pyx_k_verticalPadding, sizeof(__pyx_k_verticalPadding), 0, 0, 1, 1},
  {&__pyx_n_s_verticalStride, __pyx_k_verticalStride, sizeof(__pyx_k_verticalStride), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_s_wDesc, __pyx_k_wDesc, sizeof(__pyx_k_wDesc), 0, 0, 1, 1},
  {&__pyx_n_s_wStride, __pyx_k_wStride, sizeof(__pyx_k_wStride), 0, 0, 1, 1},
  {&__pyx_n_s_windowDimA, __pyx_k_windowDimA, sizeof(__pyx_k_windowDimA), 0, 0, 1, 1},
  {&__pyx_n_s_windowHeight, __pyx_k_windowHeight, sizeof(__pyx_k_windowHeight), 0, 0, 1, 1},
  {&__pyx_n_s_windowWidth, __pyx_k_windowWidth, sizeof(__pyx_k_windowWidth), 0, 0, 1, 1},
  {&__pyx_n_s_workSpace, __pyx_k_workSpace, sizeof(__pyx_k_workSpace), 0, 0, 1, 1},
  {&__pyx_n_s_workSpaceSizeInBytes, __pyx_k_workSpaceSizeInBytes, sizeof(__pyx_k_workSpaceSizeInBytes), 0, 0, 1, 1},
  {&__pyx_n_s_workspace, __pyx_k_workspace, sizeof(__pyx_k_workspace), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_xDesc, __pyx_k_xDesc, sizeof(__pyx_k_xDesc), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_yDesc, __pyx_k_yDesc, sizeof(__pyx_k_yDesc), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 446, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 451, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 61, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "cupy/cuda/cudnn.pyx":448
 * class CuDNNError(RuntimeError):
 * 
 *     def __init__(self, int status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         msg = cudnnGetErrorString(<Status>status)
 */
  __pyx_tuple_ = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_status, __pyx_n_s_msg); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_codeobj__2 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple_, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cupy_cuda_cudnn_pyx, __pyx_n_s_init, 448, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initcudnn(void); /*proto*/
PyMODINIT_FUNC initcudnn(void)
#else
PyMODINIT_FUNC PyInit_cudnn(void); /*proto*/
PyMODINIT_FUNC PyInit_cudnn(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_cudnn(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cudnn", __pyx_methods, __pyx_k_Thin_wrapper_of_cuDNN, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_cupy__cuda__cudnn) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cupy.cuda.cudnn")) {
      if (unlikely(PyDict_SetItemString(modules, "cupy.cuda.cudnn", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  __pyx_v_4cupy_4cuda_5cudnn_STATUS = ((PyObject*)Py_None); Py_INCREF(Py_None);
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("getVersion", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getVersion, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("create", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_create, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroy", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroy, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setStream", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setStream, "PyObject *(size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getStream", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getStream, "size_t (size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createTensorDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createTensorDescriptor, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setTensor4dDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setTensor4dDescriptor, "PyObject *(size_t, int, int, int, int, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setTensor4dDescriptorEx", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setTensor4dDescriptorEx, "PyObject *(size_t, int, int, int, int, int, int, int, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setTensorNdDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setTensorNdDescriptor, "PyObject *(size_t, int, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyTensorDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroyTensorDescriptor, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("addTensor_v3", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_addTensor_v3, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createFilterDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createFilterDescriptor, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setFilter4dDescriptor_v4", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setFilter4dDescriptor_v4, "PyObject *(size_t, int, int, int, int, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setFilterNdDescriptor_v4", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setFilterNdDescriptor_v4, "PyObject *(size_t, int, int, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getFilterNdDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getFilterNdDescriptor, "PyObject *(size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyFilterDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroyFilterDescriptor, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createConvolutionDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createConvolutionDescriptor, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setConvolutionMathType", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setConvolutionMathType, "PyObject *(size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getConvolutionMathType", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getConvolutionMathType, "size_t (size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setConvolutionGroupCount", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setConvolutionGroupCount, "PyObject *(size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getConvolutionGroupCount", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getConvolutionGroupCount, "int (size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setConvolution2dDescriptor_v4", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setConvolution2dDescriptor_v4, "PyObject *(size_t, int, int, int, int, int, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setConvolution2dDescriptor_v5", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setConvolution2dDescriptor_v5, "PyObject *(size_t, int, int, int, int, int, int, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setConvolutionNdDescriptor_v3", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setConvolutionNdDescriptor_v3, "PyObject *(size_t, int, size_t, size_t, size_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyConvolutionDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroyConvolutionDescriptor, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("findConvolutionForwardAlgorithm", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_findConvolutionForwardAlgorithm, "PyObject *(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("findConvolutionForwardAlgorithmEx", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_findConvolutionForwardAlgorithmEx, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getConvolutionForwardAlgorithm", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getConvolutionForwardAlgorithm, "int (size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getConvolutionForwardWorkspaceSize", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getConvolutionForwardWorkspaceSize, "size_t (size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("convolutionForward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_convolutionForward, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("convolutionBackwardBias", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardBias, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("findConvolutionBackwardFilterAlgorithm", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardFilterAlgorithm, "PyObject *(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("findConvolutionBackwardFilterAlgorithmEx", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardFilterAlgorithmEx, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getConvolutionBackwardFilterAlgorithm", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardFilterAlgorithm, "int (size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getConvolutionBackwardFilterWorkspaceSize", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardFilterWorkspaceSize, "size_t (size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("convolutionBackwardFilter_v3", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardFilter_v3, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("findConvolutionBackwardDataAlgorithm", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardDataAlgorithm, "PyObject *(size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("findConvolutionBackwardDataAlgorithmEx", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_findConvolutionBackwardDataAlgorithmEx, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getConvolutionBackwardDataAlgorithm", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardDataAlgorithm, "int (size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getConvolutionBackwardDataWorkspaceSize", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getConvolutionBackwardDataWorkspaceSize, "size_t (size_t, size_t, size_t, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("convolutionBackwardData_v3", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_convolutionBackwardData_v3, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createPoolingDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createPoolingDescriptor, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setPooling2dDescriptor_v4", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setPooling2dDescriptor_v4, "PyObject *(size_t, int, int, int, int, int, int, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setPoolingNdDescriptor_v4", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setPoolingNdDescriptor_v4, "PyObject *(size_t, int, int, int, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyPoolingDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroyPoolingDescriptor, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("poolingForward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_poolingForward, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("poolingBackward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_poolingBackward, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("deriveBNTensorDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_deriveBNTensorDescriptor, "PyObject *(size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("batchNormalizationForwardTraining", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationForwardTraining, "PyObject *(size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, double, size_t, size_t, double, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("batchNormalizationForwardInference", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationForwardInference, "PyObject *(size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("batchNormalizationBackward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_batchNormalizationBackward, "PyObject *(size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, double, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createActivationDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createActivationDescriptor, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setActivationDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setActivationDescriptor, "PyObject *(size_t, int, int, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyActivationDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroyActivationDescriptor, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("softmaxForward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_softmaxForward, "PyObject *(size_t, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("softmaxBackward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_softmaxBackward, "PyObject *(size_t, int, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("activationForward_v4", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_activationForward_v4, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("activationBackward_v4", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_activationBackward_v4, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createDropoutDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createDropoutDescriptor, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyDropoutDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroyDropoutDescriptor, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dropoutGetStatesSize", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_dropoutGetStatesSize, "size_t (size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setDropoutDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setDropoutDescriptor, "PyObject *(size_t, size_t, float, size_t, size_t, unsigned PY_LONG_LONG, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getDropoutReserveSpaceSize", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getDropoutReserveSpaceSize, "size_t (size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dropoutForward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_dropoutForward, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dropoutBackward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_dropoutBackward, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createRNNDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createRNNDescriptor, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyRNNDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroyRNNDescriptor, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createPersistentRNNPlan", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createPersistentRNNPlan, "size_t (size_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setPersistentRNNPlan", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setPersistentRNNPlan, "PyObject *(size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyPersistentRNNPlan", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroyPersistentRNNPlan, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setRNNDescriptor_v5", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setRNNDescriptor_v5, "PyObject *(size_t, int, int, size_t, int, int, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setRNNDescriptor_v6", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setRNNDescriptor_v6, "PyObject *(size_t, size_t, int, int, size_t, int, int, int, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getRNNWorkspaceSize", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getRNNWorkspaceSize, "PyObject *(size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getRNNTrainingReserveSize", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getRNNTrainingReserveSize, "PyObject *(size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getRNNParamsSize", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getRNNParamsSize, "PyObject *(size_t, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getRNNLinLayerMatrixParams", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getRNNLinLayerMatrixParams, "PyObject *(size_t, size_t, int, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getRNNLinLayerBiasParams", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_getRNNLinLayerBiasParams, "PyObject *(size_t, size_t, int, size_t, size_t, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("RNNForwardInference", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_RNNForwardInference, "PyObject *(size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("RNNForwardTraining", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_RNNForwardTraining, "PyObject *(size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("RNNBackwardData", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_RNNBackwardData, "PyObject *(size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("RNNBackwardWeights", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_RNNBackwardWeights, "PyObject *(size_t, size_t, int, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createSpatialTransformerDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_createSpatialTransformerDescriptor, "size_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroySpatialTransformerDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_destroySpatialTransformerDescriptor, "PyObject *(size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setSpatialTransformerDescriptor", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_setSpatialTransformerDescriptor, "PyObject *(size_t, size_t, int, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spatialTfGridGeneratorForward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_spatialTfGridGeneratorForward, "PyObject *(size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spatialTfGridGeneratorBackward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_spatialTfGridGeneratorBackward, "PyObject *(size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spatialTfSamplerForward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_spatialTfSamplerForward, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spatialTfSamplerBackward", (void (*)(void))__pyx_f_4cupy_4cuda_5cudnn_spatialTfSamplerBackward, "PyObject *(size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("cupy.cuda.stream"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_current_stream_ptr", (void (**)(void))&__pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr, "size_t (void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "cupy/cuda/cudnn.pyx":432
 * 
 * cdef dict STATUS = {
 *     0: 'CUDNN_STATUS_SUCCESS',             # <<<<<<<<<<<<<<
 *     1: 'CUDNN_STATUS_NOT_INITIALIZED',
 *     2: 'CUDNN_STATUS_ALLOC_FAILED',
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_0, __pyx_n_s_CUDNN_STATUS_SUCCESS) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_1, __pyx_n_s_CUDNN_STATUS_NOT_INITIALIZED) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_2, __pyx_n_s_CUDNN_STATUS_ALLOC_FAILED) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_3, __pyx_n_s_CUDNN_STATUS_BAD_PARAM) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_4, __pyx_n_s_CUDNN_STATUS_INTERNAL_ERROR) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_5, __pyx_n_s_CUDNN_STATUS_INVALID_VALUE) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_6, __pyx_n_s_CUDNN_STATUS_ARCH_MISMATCH) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_7, __pyx_n_s_CUDNN_STATUS_MAPPING_ERROR) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_8, __pyx_n_s_CUDNN_STATUS_EXECUTION_FAILED) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_9, __pyx_n_s_CUDNN_STATUS_NOT_SUPPORTED) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_10, __pyx_n_s_CUDNN_STATUS_LICENSE_ERROR) < 0) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_4cupy_4cuda_5cudnn_STATUS);
  __Pyx_DECREF_SET(__pyx_v_4cupy_4cuda_5cudnn_STATUS, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "cupy/cuda/cudnn.pyx":446
 * 
 * 
 * class CuDNNError(RuntimeError):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int status):
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_builtin_RuntimeError);
  __Pyx_GIVEREF(__pyx_builtin_RuntimeError);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_builtin_RuntimeError);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_2, __pyx_n_s_CuDNNError, __pyx_n_s_CuDNNError, (PyObject *) NULL, __pyx_n_s_cupy_cuda_cudnn, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "cupy/cuda/cudnn.pyx":448
 * class CuDNNError(RuntimeError):
 * 
 *     def __init__(self, int status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         msg = cudnnGetErrorString(<Status>status)
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4cupy_4cuda_5cudnn_10CuDNNError_1__init__, 0, __pyx_n_s_CuDNNError___init, NULL, __pyx_n_s_cupy_cuda_cudnn, __pyx_d, ((PyObject *)__pyx_codeobj__2)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_init, __pyx_t_5) < 0) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "cupy/cuda/cudnn.pyx":446
 * 
 * 
 * class CuDNNError(RuntimeError):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int status):
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_CuDNNError, __pyx_t_2, __pyx_t_4, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CuDNNError, __pyx_t_5) < 0) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cupy/cuda/cudnn.pyx":1
 * # distutils: language = c++             # <<<<<<<<<<<<<<
 * 
 * """Thin wrapper of cuDNN."""
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t")
 * cdef object __pyx_convert_vector_to_py_cudnnConvolutionBwdDataAlgoPerf_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /*--- Wrapped vars code ---*/
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ACTIVATION_CLIPPED_RELU);
    if (unlikely(!wrapped)) __PYX_ERR(2, 79, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ACTIVATION_CLIPPED_RELU", wrapped) < 0) __PYX_ERR(2, 79, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ACTIVATION_RELU);
    if (unlikely(!wrapped)) __PYX_ERR(2, 77, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ACTIVATION_RELU", wrapped) < 0) __PYX_ERR(2, 77, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ACTIVATION_SIGMOID);
    if (unlikely(!wrapped)) __PYX_ERR(2, 76, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ACTIVATION_SIGMOID", wrapped) < 0) __PYX_ERR(2, 76, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ACTIVATION_TANH);
    if (unlikely(!wrapped)) __PYX_ERR(2, 78, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ACTIVATION_TANH", wrapped) < 0) __PYX_ERR(2, 78, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_FEATURE_MAP);
    if (unlikely(!wrapped)) __PYX_ERR(2, 22, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ADD_FEATURE_MAP", wrapped) < 0) __PYX_ERR(2, 22, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_FULL_TENSOR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 25, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ADD_FULL_TENSOR", wrapped) < 0) __PYX_ERR(2, 25, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_IMAGE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 20, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ADD_IMAGE", wrapped) < 0) __PYX_ERR(2, 20, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_SAME_C);
    if (unlikely(!wrapped)) __PYX_ERR(2, 24, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ADD_SAME_C", wrapped) < 0) __PYX_ERR(2, 24, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_SAME_CHW);
    if (unlikely(!wrapped)) __PYX_ERR(2, 23, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ADD_SAME_CHW", wrapped) < 0) __PYX_ERR(2, 23, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_ADD_SAME_HW);
    if (unlikely(!wrapped)) __PYX_ERR(2, 21, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_ADD_SAME_HW", wrapped) < 0) __PYX_ERR(2, 21, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_BATCHNORM_PER_ACTIVATION);
    if (unlikely(!wrapped)) __PYX_ERR(2, 85, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_BATCHNORM_PER_ACTIVATION", wrapped) < 0) __PYX_ERR(2, 85, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_BATCHNORM_SPATIAL);
    if (unlikely(!wrapped)) __PYX_ERR(2, 86, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_BATCHNORM_SPATIAL", wrapped) < 0) __PYX_ERR(2, 86, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_BIDIRECTIONAL);
    if (unlikely(!wrapped)) __PYX_ERR(2, 94, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_BIDIRECTIONAL", wrapped) < 0) __PYX_ERR(2, 94, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION);
    if (unlikely(!wrapped)) __PYX_ERR(2, 27, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION", wrapped) < 0) __PYX_ERR(2, 27, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_0);
    if (unlikely(!wrapped)) __PYX_ERR(2, 58, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_ALGO_0", wrapped) < 0) __PYX_ERR(2, 58, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_1);
    if (unlikely(!wrapped)) __PYX_ERR(2, 59, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_ALGO_1", wrapped) < 0) __PYX_ERR(2, 59, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_FFT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 60, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_ALGO_FFT", wrapped) < 0) __PYX_ERR(2, 60, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_FFT_TILING);
    if (unlikely(!wrapped)) __PYX_ERR(2, 61, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_ALGO_FFT_TILING", wrapped) < 0) __PYX_ERR(2, 61, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_WINOGRAD);
    if (unlikely(!wrapped)) __PYX_ERR(2, 62, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_ALGO_WINOGRAD", wrapped) < 0) __PYX_ERR(2, 62, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_ALGO_WINOGRAD_NONFUSED);
    if (unlikely(!wrapped)) __PYX_ERR(2, 63, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_ALGO_WINOGRAD_NONFUSED", wrapped) < 0) __PYX_ERR(2, 63, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_NO_WORKSPACE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 54, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_NO_WORKSPACE", wrapped) < 0) __PYX_ERR(2, 54, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_PREFER_FASTEST);
    if (unlikely(!wrapped)) __PYX_ERR(2, 55, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_PREFER_FASTEST", wrapped) < 0) __PYX_ERR(2, 55, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_DATA_SPECIFY_WORKSPACE_LIMIT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 56, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_DATA_SPECIFY_WORKSPACE_LIMIT", wrapped) < 0) __PYX_ERR(2, 56, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_0);
    if (unlikely(!wrapped)) __PYX_ERR(2, 47, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_ALGO_0", wrapped) < 0) __PYX_ERR(2, 47, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_1);
    if (unlikely(!wrapped)) __PYX_ERR(2, 48, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_ALGO_1", wrapped) < 0) __PYX_ERR(2, 48, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_3);
    if (unlikely(!wrapped)) __PYX_ERR(2, 50, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_ALGO_3", wrapped) < 0) __PYX_ERR(2, 50, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_FFT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 49, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_ALGO_FFT", wrapped) < 0) __PYX_ERR(2, 49, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_WINOGRAD);
    if (unlikely(!wrapped)) __PYX_ERR(2, 51, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_ALGO_WINOGRAD", wrapped) < 0) __PYX_ERR(2, 51, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_ALGO_WINOGRAD_NONFUSED);
    if (unlikely(!wrapped)) __PYX_ERR(2, 52, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_ALGO_WINOGRAD_NONFUSED", wrapped) < 0) __PYX_ERR(2, 52, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_NO_WORKSPACE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 43, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_NO_WORKSPACE", wrapped) < 0) __PYX_ERR(2, 43, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_PREFER_FASTEST);
    if (unlikely(!wrapped)) __PYX_ERR(2, 44, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_PREFER_FASTEST", wrapped) < 0) __PYX_ERR(2, 44, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_BWD_FILTER_SPECIFY_WORKSPACE_LIMIT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 45, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_BWD_FILTER_SPECIFY_WORKSPACE_LIMIT", wrapped) < 0) __PYX_ERR(2, 45, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_DIRECT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 37, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_ALGO_DIRECT", wrapped) < 0) __PYX_ERR(2, 37, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_FFT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 38, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_ALGO_FFT", wrapped) < 0) __PYX_ERR(2, 38, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_FFT_TILING);
    if (unlikely(!wrapped)) __PYX_ERR(2, 39, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_ALGO_FFT_TILING", wrapped) < 0) __PYX_ERR(2, 39, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_GEMM);
    if (unlikely(!wrapped)) __PYX_ERR(2, 36, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_ALGO_GEMM", wrapped) < 0) __PYX_ERR(2, 36, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_GEMM);
    if (unlikely(!wrapped)) __PYX_ERR(2, 34, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_GEMM", wrapped) < 0) __PYX_ERR(2, 34, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM);
    if (unlikely(!wrapped)) __PYX_ERR(2, 35, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM", wrapped) < 0) __PYX_ERR(2, 35, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD);
    if (unlikely(!wrapped)) __PYX_ERR(2, 40, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD", wrapped) < 0) __PYX_ERR(2, 40, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD_NONFUSED);
    if (unlikely(!wrapped)) __PYX_ERR(2, 41, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD_NONFUSED", wrapped) < 0) __PYX_ERR(2, 41, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_NO_WORKSPACE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 30, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_NO_WORKSPACE", wrapped) < 0) __PYX_ERR(2, 30, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_PREFER_FASTEST);
    if (unlikely(!wrapped)) __PYX_ERR(2, 31, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_PREFER_FASTEST", wrapped) < 0) __PYX_ERR(2, 31, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CONVOLUTION_FWD_SPECIFY_WORKSPACE_LIMIT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 32, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CONVOLUTION_FWD_SPECIFY_WORKSPACE_LIMIT", wrapped) < 0) __PYX_ERR(2, 32, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_CROSS_CORRELATION);
    if (unlikely(!wrapped)) __PYX_ERR(2, 28, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_CROSS_CORRELATION", wrapped) < 0) __PYX_ERR(2, 28, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_DATA_DOUBLE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 8, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_DATA_DOUBLE", wrapped) < 0) __PYX_ERR(2, 8, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_DATA_FLOAT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 7, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_DATA_FLOAT", wrapped) < 0) __PYX_ERR(2, 7, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_DATA_HALF);
    if (unlikely(!wrapped)) __PYX_ERR(2, 9, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_DATA_HALF", wrapped) < 0) __PYX_ERR(2, 9, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_DEFAULT_MATH);
    if (unlikely(!wrapped)) __PYX_ERR(2, 11, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_DEFAULT_MATH", wrapped) < 0) __PYX_ERR(2, 11, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_DIVNORM_PRECOMPUTED_MEANS);
    if (unlikely(!wrapped)) __PYX_ERR(2, 83, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_DIVNORM_PRECOMPUTED_MEANS", wrapped) < 0) __PYX_ERR(2, 83, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_GRU);
    if (unlikely(!wrapped)) __PYX_ERR(2, 91, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_GRU", wrapped) < 0) __PYX_ERR(2, 91, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_LINEAR_INPUT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 100, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_LINEAR_INPUT", wrapped) < 0) __PYX_ERR(2, 100, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_LRN_CROSS_CHANNEL_DIM1);
    if (unlikely(!wrapped)) __PYX_ERR(2, 81, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_LRN_CROSS_CHANNEL_DIM1", wrapped) < 0) __PYX_ERR(2, 81, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_LSTM);
    if (unlikely(!wrapped)) __PYX_ERR(2, 90, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_LSTM", wrapped) < 0) __PYX_ERR(2, 90, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_NOT_PROPAGATE_NAN);
    if (unlikely(!wrapped)) __PYX_ERR(2, 14, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_NOT_PROPAGATE_NAN", wrapped) < 0) __PYX_ERR(2, 14, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_POOLING_AVERAGE_COUNT_EXCLUDE_PADDING);
    if (unlikely(!wrapped)) __PYX_ERR(2, 74, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_POOLING_AVERAGE_COUNT_EXCLUDE_PADDING", wrapped) < 0) __PYX_ERR(2, 74, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_POOLING_AVERAGE_COUNT_INCLUDE_PADDING);
    if (unlikely(!wrapped)) __PYX_ERR(2, 73, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_POOLING_AVERAGE_COUNT_INCLUDE_PADDING", wrapped) < 0) __PYX_ERR(2, 73, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_POOLING_MAX);
    if (unlikely(!wrapped)) __PYX_ERR(2, 72, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_POOLING_MAX", wrapped) < 0) __PYX_ERR(2, 72, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_PROPAGATE_NAN);
    if (unlikely(!wrapped)) __PYX_ERR(2, 15, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_PROPAGATE_NAN", wrapped) < 0) __PYX_ERR(2, 15, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_ALGO_PERSIST_DYNAMIC);
    if (unlikely(!wrapped)) __PYX_ERR(2, 98, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_RNN_ALGO_PERSIST_DYNAMIC", wrapped) < 0) __PYX_ERR(2, 98, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_ALGO_PERSIST_STATIC);
    if (unlikely(!wrapped)) __PYX_ERR(2, 97, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_RNN_ALGO_PERSIST_STATIC", wrapped) < 0) __PYX_ERR(2, 97, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_ALGO_STANDARD);
    if (unlikely(!wrapped)) __PYX_ERR(2, 96, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_RNN_ALGO_STANDARD", wrapped) < 0) __PYX_ERR(2, 96, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_RELU);
    if (unlikely(!wrapped)) __PYX_ERR(2, 88, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_RNN_RELU", wrapped) < 0) __PYX_ERR(2, 88, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_RNN_TANH);
    if (unlikely(!wrapped)) __PYX_ERR(2, 89, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_RNN_TANH", wrapped) < 0) __PYX_ERR(2, 89, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_SAMPLER_BILINEAR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 103, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_SAMPLER_BILINEAR", wrapped) < 0) __PYX_ERR(2, 103, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_SKIP_INPUT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 101, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_SKIP_INPUT", wrapped) < 0) __PYX_ERR(2, 101, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_ACCURATE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 66, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_SOFTMAX_ACCURATE", wrapped) < 0) __PYX_ERR(2, 66, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_FAST);
    if (unlikely(!wrapped)) __PYX_ERR(2, 65, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_SOFTMAX_FAST", wrapped) < 0) __PYX_ERR(2, 65, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_LOG);
    if (unlikely(!wrapped)) __PYX_ERR(2, 67, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_SOFTMAX_LOG", wrapped) < 0) __PYX_ERR(2, 67, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_MODE_CHANNEL);
    if (unlikely(!wrapped)) __PYX_ERR(2, 70, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_SOFTMAX_MODE_CHANNEL", wrapped) < 0) __PYX_ERR(2, 70, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_SOFTMAX_MODE_INSTANCE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 69, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_SOFTMAX_MODE_INSTANCE", wrapped) < 0) __PYX_ERR(2, 69, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_TENSOR_NCHW);
    if (unlikely(!wrapped)) __PYX_ERR(2, 17, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_TENSOR_NCHW", wrapped) < 0) __PYX_ERR(2, 17, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_TENSOR_NHWC);
    if (unlikely(!wrapped)) __PYX_ERR(2, 18, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_TENSOR_NHWC", wrapped) < 0) __PYX_ERR(2, 18, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_TENSOR_OP_MATH);
    if (unlikely(!wrapped)) __PYX_ERR(2, 12, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_TENSOR_OP_MATH", wrapped) < 0) __PYX_ERR(2, 12, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_5cudnn_CUDNN_UNIDIRECTIONAL);
    if (unlikely(!wrapped)) __PYX_ERR(2, 93, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUDNN_UNIDIRECTIONAL", wrapped) < 0) __PYX_ERR(2, 93, __pyx_L1_error);
  }

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init cupy.cuda.cudnn", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cupy.cuda.cudnn");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
    static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* CalculateMetaclass */
    static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* FetchCommonType */
    static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
    static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* Py3ClassCreate */
        static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
        static int __Pyx_CLineForTraceback(int c_line) {
#ifdef CYTHON_CLINE_IN_TRACEBACK
    return ((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0;
#else
    PyObject *use_cline;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = PyDict_GetItem(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *ptype, *pvalue, *ptraceback;
      PyObject *use_cline_obj;
      PyErr_Fetch(&ptype, &pvalue, &ptraceback);
      use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        use_cline = NULL;
      }
      PyErr_Restore(ptype, pvalue, ptraceback);
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    return c_line;
#endif
}

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
        #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_cudnnDataType_t(cudnnDataType_t value) {
    const cudnnDataType_t neg_one = (cudnnDataType_t) -1, const_zero = (cudnnDataType_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(cudnnDataType_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(cudnnDataType_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(cudnnDataType_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(cudnnDataType_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(cudnnDataType_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(cudnnDataType_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_cudnnTensorFormat_t(cudnnTensorFormat_t value) {
    const cudnnTensorFormat_t neg_one = (cudnnTensorFormat_t) -1, const_zero = (cudnnTensorFormat_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(cudnnTensorFormat_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(cudnnTensorFormat_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(cudnnTensorFormat_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(cudnnTensorFormat_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(cudnnTensorFormat_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(cudnnTensorFormat_t),
                                     little, !is_unsigned);
    }
}

static PyObject* __pyx_convert__to_py_cudnnConvolutionFwdAlgoPerf_t(cudnnConvolutionFwdAlgoPerf_t s) {
          PyObject* res;
          PyObject* member;
          res = PyDict_New(); if (unlikely(!res)) return NULL;
          member = __Pyx_PyInt_From_int(s.algo); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_algo, member) < 0)) goto bad;
          Py_DECREF(member);
          member = __Pyx_PyInt_From_int(s.status); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_status, member) < 0)) goto bad;
          Py_DECREF(member);
          member = PyFloat_FromDouble(s.time); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_time, member) < 0)) goto bad;
          Py_DECREF(member);
          member = __Pyx_PyInt_FromSize_t(s.memory); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_memory, member) < 0)) goto bad;
          Py_DECREF(member);
          return res;
          bad:
          Py_XDECREF(member);
          Py_DECREF(res);
          return NULL;
        }
        static PyObject* __pyx_convert__to_py_cudnnConvolutionBwdFilterAlgoPerf_t(cudnnConvolutionBwdFilterAlgoPerf_t s) {
          PyObject* res;
          PyObject* member;
          res = PyDict_New(); if (unlikely(!res)) return NULL;
          member = __Pyx_PyInt_From_int(s.algo); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_algo, member) < 0)) goto bad;
          Py_DECREF(member);
          member = __Pyx_PyInt_From_int(s.status); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_status, member) < 0)) goto bad;
          Py_DECREF(member);
          member = PyFloat_FromDouble(s.time); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_time, member) < 0)) goto bad;
          Py_DECREF(member);
          member = __Pyx_PyInt_FromSize_t(s.memory); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_memory, member) < 0)) goto bad;
          Py_DECREF(member);
          return res;
          bad:
          Py_XDECREF(member);
          Py_DECREF(res);
          return NULL;
        }
        static PyObject* __pyx_convert__to_py_cudnnConvolutionBwdDataAlgoPerf_t(cudnnConvolutionBwdDataAlgoPerf_t s) {
          PyObject* res;
          PyObject* member;
          res = PyDict_New(); if (unlikely(!res)) return NULL;
          member = __Pyx_PyInt_From_int(s.algo); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_algo, member) < 0)) goto bad;
          Py_DECREF(member);
          member = __Pyx_PyInt_From_int(s.status); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_status, member) < 0)) goto bad;
          Py_DECREF(member);
          member = PyFloat_FromDouble(s.time); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_time, member) < 0)) goto bad;
          Py_DECREF(member);
          member = __Pyx_PyInt_FromSize_t(s.memory); if (unlikely(!member)) goto bad;
          if (unlikely(PyDict_SetItem(res, __pyx_n_s_memory, member) < 0)) goto bad;
          Py_DECREF(member);
          return res;
          bad:
          Py_XDECREF(member);
          Py_DECREF(res);
          return NULL;
        }
        /* CIntFromPy */
        static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_As_unsigned_PY_LONG_LONG(PyObject *x) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned PY_LONG_LONG) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned PY_LONG_LONG) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned PY_LONG_LONG) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned PY_LONG_LONG) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned PY_LONG_LONG val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned PY_LONG_LONG) -1;
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned PY_LONG_LONG");
    return (unsigned PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned PY_LONG_LONG");
    return (unsigned PY_LONG_LONG) -1;
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
        static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
        static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
        static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
        #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* FunctionImport */
        #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
        static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            PyErr_Clear();
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
